// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "batterystatuswidget.hpp"
#include <QHBoxLayout>

// const QSize BatteryStatusWidget::iconSize = QSize(48, 48);
const QSize BatteryStatusWidget::iconSize = QSize(24, 24);

BatteryStatusWidget::BatteryStatusWidget(QWidget *parent) : QWidget{parent} {

    mPixmapCharged =
        QPixmap(":/adm2/images/battery/battery_charged.png").scaled(BatteryStatusWidget::iconSize, Qt::KeepAspectRatio);
    mPixmapCharging = QPixmap(":/adm2/images/battery/battery_charging.png")
                          .scaled(BatteryStatusWidget::iconSize, Qt::KeepAspectRatio);
    mPixmapUnknown =
        QPixmap(":/adm2/images/battery/battery_unknown.png").scaled(BatteryStatusWidget::iconSize, Qt::KeepAspectRatio);
    mPixmapOnBatteryLow =
        QPixmap(":/adm2/images/battery/battery_low.png").scaled(BatteryStatusWidget::iconSize, Qt::KeepAspectRatio);
    mPixmapOnBatteryVeryLow = QPixmap(":/adm2/images/battery/battery_very_low.png")
                                  .scaled(BatteryStatusWidget::iconSize, Qt::KeepAspectRatio);
    mPixmapOnBatteryMedium =
        QPixmap(":/adm2/images/battery/battery_medium.png").scaled(BatteryStatusWidget::iconSize, Qt::KeepAspectRatio);
    mPixmapOnBatteryHigh =
        QPixmap(":/adm2/images/battery/battery_high.png").scaled(BatteryStatusWidget::iconSize, Qt::KeepAspectRatio);
    mPixmapNoBattery =
        QPixmap(":/adm2/images/battery/battery_no.png").scaled(BatteryStatusWidget::iconSize, Qt::KeepAspectRatio);

    mLblIcon = new QLabel(this);
    mLblIcon->setPixmap(mPixmapUnknown);
    mLblInfos = new QLabel(this);

    QHBoxLayout *layout = new QHBoxLayout(this);
    layout->setContentsMargins(6, 0, 6, 0);
    layout->addWidget(mLblIcon, 0);
    layout->addWidget(mLblInfos, 0);

    mThread = new QThread;
    mWorker = new BatteryStatusWorker;
    QObject::connect(mWorker, &BatteryStatusWorker::finished, mThread, &QThread::quit);
    QObject::connect(mThread, &QThread::started, mWorker, &BatteryStatusWorker::process);
    QObject::connect(mWorker, &BatteryStatusWorker::statusUpdated, this, &BatteryStatusWidget::onStatusUpdated);

    mWorker->moveToThread(mThread);
    mThread->start();
}

BatteryStatusWidget::~BatteryStatusWidget() {
    if (mWorker->isRunning())
        mWorker->stop();
    mThread->exit();
    QObject::disconnect(mWorker, nullptr, this, nullptr);
    mWorker->deleteLater();
    mThread->deleteLater();
}

void BatteryStatusWidget::updateWidget() {
    QString tip;
    QString sp = mPercent == -1 ? "N/A" : QString("%1%").arg(mPercent);
    QString st = "N/A";
    QString infos;
    int nmins = mLifetime / 60;
    switch (mStatus) {
    case (BatteryStatusWorker::CHARGED):
        mLblIcon->setPixmap(mPixmapCharged);
        tip = tr("Charged (%1).").arg(sp);
        break;
    case (BatteryStatusWorker::CHARGING):
        mLblIcon->setPixmap(mPixmapCharging);
        tip = tr("Charging (%1).").arg(sp);
        break;
    case (BatteryStatusWorker::NO_BATTERY):
        mLblIcon->setPixmap(mPixmapNoBattery);
        tip = tr("No battery.");
        break;
    case (BatteryStatusWorker::ON_BATTERY):
        if (mPercent >= 75) {
            mLblIcon->setPixmap(mPixmapOnBatteryHigh);
        } else if (mPercent >= 50) {
            mLblIcon->setPixmap(mPixmapOnBatteryMedium);
        } else if (mPercent >= 25) {
            mLblIcon->setPixmap(mPixmapOnBatteryLow);
        } else {
            mLblIcon->setPixmap(mPixmapOnBatteryVeryLow);
        }
        if (mLifetime < 0) {
            tip = tr("Cannot estimate remaining time (%1).").arg(sp);
            st = "???min";
        } else if (nmins >= 60) {
            int a = (int)(nmins / 60), b = (int)(nmins % 60);
            tip = tr("~ %1 hours, %2 minutes (%3).").arg(a).arg(b).arg(sp);
            st = QString("~%1h%2min").arg(a).arg(b);
        } else {
            tip = tr("~ %1 minutes (%2).").arg(nmins).arg(sp);
            st = QString("~%1min").arg(nmins);
        }
        break;
    case (BatteryStatusWorker::UNKNOWN):
    default:
        mLblIcon->setPixmap(mPixmapUnknown);
        tip = tr("Could not determinate battery status.");
        st = "N/A";
        break;
    }

    if (mShowPercent && mShowLifetime)
        infos = QString("%1, %2").arg(sp, st);
    else if (mShowPercent)
        infos = QString("%1").arg(sp);
    else if (mShowLifetime)
        infos = QString("%1").arg(sp);

    // set infos
    mLblInfos->setText(infos.size() ? infos.leftJustified(infos.size() + 1) : "");
    // and lifetime (in tooltip)
    setToolTip(tip);
    mLblIcon->setToolTip(tip);
    mLblInfos->setToolTip(tip);
}

/**
 * @brief Show/Hide the percentage on the information label.
 * @param show True to show the percentage, false otherwise.
 * @param update True (default) to update informations immediatly on the label, otherwise update will occur on the next
 * battery query.
 */
void BatteryStatusWidget::showPercent(bool show, bool update) {
    mShowPercent = show;
    if (update)
        updateWidget();
}

/**
 * @brief Show/Hide the lifetime on the information label.
 * @param show True to show the lifetime, false otherwise.
 * @param update True (default) to update informations immediatly on the label, otherwise update will occur on the next
 * battery query.
 */
void BatteryStatusWidget::showLifeTime(bool show, bool update) {
    mShowLifetime = show;
    if (update)
        updateWidget();
}

/// @brief Stop polling battery status on hide.
void BatteryStatusWidget::hideEvent(QHideEvent *event) {
    mWorker->pause();
    QWidget::hideEvent(event);
}

/// @brief Start polling battery status on show.
void BatteryStatusWidget::showEvent(QShowEvent *event) {
    mWorker->resume();
    QWidget::showEvent(event);
}

/// @brief Update the widget displayed informations.
void BatteryStatusWidget::onStatusUpdated(BatteryStatusWorker::Status status, int percent, int lifetime) {
    mStatus = status;
    mPercent = percent;
    mLifetime = lifetime;
    updateWidget();
}
