// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "fitcontroller.hpp"
#include "fitworker.hpp"

/// @brief Constructor.
FitController::FitController(QObject *parent) : QObject{parent} {
    mEnabled = false;
    mMinTimeout = 0;
    mFitWorker = new FitWorker;
    mFitWorker->moveToThread(&mWorkerThread);
    QObject::connect(&mWorkerThread, &QThread::finished, mFitWorker, &QObject::deleteLater);
    QObject::connect(this, &FitController::blockToFit, mFitWorker, &FitWorker::fitBlock);
    QObject::connect(mFitWorker, &FitWorker::fitComplete, this, &FitController::fitComplete);
    mWorkerThread.start();
    mFreqTimer.start();
}

/// @brief Destructor.
FitController::~FitController() {
    mWorkerThread.quit();
    mWorkerThread.wait();
    mFitWorker = nullptr;
}

/// @brief Set a new fit configuration.
void FitController::setConfiguration(const FitConfig &config, bool fastCalculationMode) {
    waitForFinished(); // really ?
    mFitWorker->setConfiguration(config, fastCalculationMode);
}

/// @brief Wait for the current fit process finishes (if any).
void FitController::waitForFinished() {
    while (mFitWorker->isRunning()) {
        QThread::currentThread()->msleep(50);
    }
}

/**
 * @brief Set the maximum input frequency of block to send to the internal worker.
 * @param frequency  Frequency to set in Hertz.
 * @details
 * This method is intended to reduce the rate of block to send to the internal worker for fitting process.
 *
 */
void FitController::setMaximumInputFrequency(double frequency) {
    if (frequency <= 0)
        mMinTimeout = -1;
    else
        mMinTimeout = qMax(5, (int)(1000 / frequency));
    mFreqTimer.start();
}

/**
 * @brief Enable/disable fit processing.
 * @param enabled True to enable, false otherwise.
 * @details
 * When the fit processing is disabled, each call to FitController::fitBlock directly emits
 * FitController::fitComplete signal with the input block.
 */
void FitController::setFitProcess(bool enabled) {
    mEnabled = enabled;
    if (mEnabled) {
        Qt::ConnectionType connType = static_cast<Qt::ConnectionType>(Qt::QueuedConnection | Qt::UniqueConnection);
        QObject::connect(this, &FitController::blockToFit, mFitWorker, &FitWorker::fitBlock, connType);
        QObject::connect(mFitWorker, &FitWorker::fitComplete, this, &FitController::fitComplete, connType);
    } else {
        QObject::disconnect(this, &FitController::blockToFit, mFitWorker, &FitWorker::fitBlock);
        QObject::disconnect(mFitWorker, &FitWorker::fitComplete, this, &FitController::fitComplete);
    }
}

/// @brief Check whether or not a fit is being processed.
bool FitController::isRunning() const { return mFitWorker ? mFitWorker->isRunning() : false; }

/**
 * @brief Attempt to fit the given block of data in separate thread(s).
 * @param block Data block to process.
 * @param refTime Time reference of the incoming block.
 * @details
 * If fit processing is not enabled (see FitController::setFitProcess), the method simply
 * emits FitController::fitComplete signal with the input block.
 */
void FitController::fitBlock(const ABlock &block, double refTime) {
    // fit is not enabled: send the block to the next step of the processing chain.
    if (!mEnabled)
        emit fitComplete(block, QVector<FitReport>(), ExecTimes());
    // trash the block if max frequency feature is enabled and timeout has not been reached OR if the worker is running.
    if ((mMinTimeout > 0 && !mFreqTimer.hasExpired(mMinTimeout)) || mFitWorker->isRunning())
        return;
    // otherwise send the block to the worker.
    emit blockToFit(block, refTime);
    mFreqTimer.restart();
}
