// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "fitworker.hpp"

int maxDecimationFactor(int np, int no, int mp = 500, int mo = 4, int *op = nullptr, int *oo = nullptr);

/**
 * @brief Compute the maximum decimation factor.
 * @param np Total number of points in ramp
 * @param no Total number of points in offset.
 * @param mp Minimum number of points to keep in ramp (default to 500).
 * @param mo Minimum number of points to keep in offset (default to 4).
 * @param op Optional pointer to an int with, in output, the expected number of ramp points after decimation.
 * @param op Optional pointer to an int with, in output, the expected number of offset points after decimation.
 * @return The maximum decimation factor.
 * @details
 * The method computes the maximum decimation factor (i.e. step index) that can be used to reduce the number of points
 * in a ramp that such that there is at least @e mp points in the ramp after decimation and @e mo in the offset.
 * If no decimation factor can be computed, 1 is returned.
 */
int maxDecimationFactor(int np, int no, int mp, int mo, int *op, int *oo) {
    int d = 1;
    if (mp >= np || mo >= no)
        return d;
    while (d < 100) {
        if ((np / d) < mp || (no / d) < mo)
            break;
        d++;
    }
    d--;
    if (d <= 0)
        d = 1;
    if (op)
        *op = (d > 1) ? (int)(std::ceil((double)np / d)) : np;
    if (oo)
        *oo = (d > 1) ? (int)(std::ceil((double)no / d)) : no;
    return d;
}

FitWorker::FitWorker(int instrOffset, QObject *parent) : QObject{parent}, mInstrOffset(instrOffset) {
    mRunning = false;
    mReuse = false;
}

void FitWorker::setConfiguration(const FitConfig &config, bool fastCalculation) {
    QMutexLocker locker(&mDataMutex);
    mConfig = config;
    mFastCalculation = fastCalculation;
    mPreviousFitParams.clear();
}

/// @brief Check whether or not the worker is currently running (i.e. processing a block).
bool FitWorker::isRunning() const {
    QMutexLocker locker(&mRunMutex);
    return mRunning;
}

void FitWorker::fitBlock(const ABlock &inputBlock, double refTime) {

    mRunMutex.lock();
    mRunning = true;
    mRunMutex.unlock();

    ExecTimes ttimes;
    QElapsedTimer timer;
    qint64 tmp = 0, lastElapsed = 0;
    QMap<int, double> ptScales, mOffsets;
    QMap<int, QPair<double, double>> eSignals;
    QMap<int, DoubleVector> spectras, ramps;
    double P0, T0, CH2O0;

    ABlock block(inputBlock);
    FitParamsMap previousFitParams;
    mDataMutex.lock();
    FitConfig config = mConfig;
    bool fast = mFastCalculation;
    if (mFastCalculation)
        previousFitParams = mPreviousFitParams;
    mDataMutex.unlock();
    if (!config.isValid()) {
        emit fitComplete(block, QVector<FitReport>(), ttimes);
        mRunMutex.lock();
        mRunning = false;
        mRunMutex.unlock();
        return;
    }

    FCParameters mFCParams = config.fcParameters();

    // update the instrument number flag.
    block.setInstrument(block.instrument() + mInstrOffset);
    block.updateValue("iversion", config.cfgVersion());

    QVector<FitReport> reports(config.spectras().size());

    QMap<int, ScaleInfos> scaleInfosMap;
    int nbi = block.rampSize();
    for (const LParam &lp : config.lines()) {
        ScaleInfos s{nbi, lp.basePts, nbi, 1};
        if (fast) {
            int b = lp.offset.second - lp.offset.first + 1;
            s.d = maxDecimationFactor(s.nb, b, 500, 4, &(s.nf));
        }
        scaleInfosMap.insert(lp.id, s);
    }

    timer.start();
    // setup data block.
    // note: if decimInfos is not empty, vectors in spectras map may be decimated and then will not have the same
    // size as the original block.
    if (!setupBlockData(block, config, scaleInfosMap, P0, T0, CH2O0, spectras, ramps, mOffsets, eSignals)) {
        qDebug() << "cannot setup data from incoming block ";
        int i = 0;
        for (const SParam &sp : config.spectras()) {
            FitReport &report = reports[i];
            report.spectraId = sp.id;
            LParam lp = sp.line;
            report.lineId = lp.id;
            report.status = 2;
            report.reason = QString("[FAILURE] (%1) %2")
                                .arg(MP_ERR_INPUT)
                                .arg(QString::fromStdString(mpFitStatus(MP_ERR_INPUT)))
                                .toStdString();
            report.timeReference = refTime;
            i++;
        }
        ttimes.preFit = timer.nsecsElapsed();
        ttimes.global = ttimes.preFit;
        emit fitComplete(block, reports, ttimes);
        mRunMutex.lock();
        mRunning = false;
        mRunMutex.unlock();
        return;
    }
    // process fits.

    QMap<int, FitParams> fitParamsMap;
    int ispc = -1;
    for (const SParam &sp : config.spectras()) {
        ispc++;
        ExecTimes btimes;
        tmp = timer.nsecsElapsed();
        lastElapsed = tmp;
        LParam lp = sp.line;
        int sid = sp.id;
        int lid = lp.id;
        // 1. Setup the parameter for the fit.
        FitParams iFitParams;
        QStringList errors;
        ScaleInfos scaleInfos = scaleInfosMap.value(lid, {nbi, lp.basePts, nbi, 1});
        // Scale the fit parameters from CONFIG to FIT
        DoubleVector spc =
            setupFitParameters(sp, iFitParams, fitParamsMap, scaleInfos, P0, T0, CH2O0, spectras, &errors);
        // No spectra data: some errors occured in the setup:
        //  ==> register the (nullified) fitParams instance and process the next spectra.
        if (spc.empty()) {
            for (auto e : qAsConst(errors))
                qDebug() << e;
            // register the nullified FitParams in case of future dependents spectras.
            fitParamsMap.insert(sid, iFitParams);
            continue;
        }

        /* On fast mode: update initial guesses from the last previous successfull fit output parameters.
         *
         * A fit is succcessful if it has converged (from LM point of view) and if no free parameters are pegged.
         *
         * The parameters registered in the map are expected to be scaled to config ramp points: we NEED to scale them
         * to fit ramp points !
         */
        QPair<int, int> slid = qMakePair(sid, lid);
        if (fast && previousFitParams.contains(slid)) {
            FitParams &ofp = previousFitParams[slid];
            // molecules: update the value.
            for (const MParam &p : ofp.molecules())
                iFitParams.setValue(p.name, p.value);
            // for sigma-scale and base-line we set new vector and scale the values.
            iFitParams.setBaseLine(ofp.baseLine());
            iFitParams.setSigmaScale(ofp.sigmaScale());
            double scale = scaleInfos.c2f();
            iFitParams.scaleSigmaScale(scale);
            iFitParams.scaleBaseLine(scale);
        }
        // 2. Setup MPFIT solver
        // 2.1 build raw parameters data, mapping.
        std::map<std::string, int> mapping;
        double *varParams = new double[static_cast<unsigned int>(iFitParams.size())];
        mp_par_struct *parameters = buildMpFitParameters(iFitParams, mapping, varParams);
        if (parameters == nullptr) {
            delete[] varParams;
            iFitParams.nullify();
            fitParamsMap.insert(sid, iFitParams);
            continue;
        }
        // 2.2 generate struct of fixed parameters for the residual function.
        FitData fixedVars(mapping, iFitParams, spc);
        // 2.3 Prepare MPFIT configuration.
        mp_config config = {mFCParams.fTol,
            mFCParams.pTol,
            mFCParams.gTol,
            MP_MACHEP0,
            100.0,
            1e-12,
            mFCParams.maxIter,
            mFCParams.maxFEV,
            0,
            0,
            0,
            nullptr};
        // prepare fit result structure
        mp_result result;
        int np = mapping.size();
        double *xerror = new double[static_cast<unsigned int>(np)];
        for (int i = 0; i < np; i++)
            xerror[i] = 0.0;
        memset(&result, 0, sizeof(result));
        result.xerror = xerror;
        // 3. launch mpfit
        tmp = timer.nsecsElapsed();
        btimes.preFit += (tmp - lastElapsed);
        lastElapsed = tmp;
        int info =
            mpfit(&mpResFunc, spc.size(), np, varParams, parameters, &config, static_cast<void *>(&fixedVars), &result);
        result.status = info; // force status: it is not updated if the process fails before actual computations start.
        tmp = timer.nsecsElapsed();
        btimes.fit += (tmp - lastElapsed);
        lastElapsed = tmp;

        // 4. Compute the fit report.
        FitReport &report = reports[ispc];

        // 4.1 fill basic fields.
        mpFillReportFromFit(
            &report, result, iFitParams, varParams, mapping, mFCParams.nullLower, mFCParams.nullUpper, sid, lid);

        // 4.2 update extra fields.
        report.timeReference = refTime;

        report.mRampData = ramps[lid]; // original ramp -> NOT DECIMATED !
        int nr = report.mRampData.size();
        double scale = static_cast<double>(lp.basePts) / nr;
        report.mRampRange =
            std::make_pair(qMax(0, qCeil(lp.range.first / scale)), qMin(nr - 1, qCeil(lp.range.second / scale)));
        report.mSpectraRampRange =
            std::make_pair(qMax(0, qCeil(sp.range.first / scale)), qMin(nr - 1, qCeil(sp.range.second / scale)));

        report.mOrig = spc;
        DoubleVector indexes(spc.size());
        std::iota(indexes.begin(), indexes.end(), static_cast<double>(iFitParams.spectraIndexOffset()));
        report.mSigmas = polynom(indexes, report.finalParams.sigmaScale());
        report.mBaseLine = polynom(indexes, report.finalParams.baseLine());
        report.mFit.reserve(report.mSigmas.size());
        for (int i = 0; i < report.mSigmas.size(); ++i) {
            double s = report.mSigmas[i];
            double v = std::exp(-eval_abs_at(report.finalParams, s));
            report.mFit.push_back(v);
        }

        // Populate block with values
        int iori = 0, ifit = 0;
        RunningStat stats;

        int m = spc.size();
        if (m > 1) {
            double fv, ov, fvm, ovm;
            const DoubleVector &bline = report.mBaseLine;
            const DoubleVector &fit = report.mFit;
            fvm = fit[0];
            ovm = spc.at(0) / bline.at(0);
            for (int i = 1; i < m; i++) {
                fv = fit[i];
                ov = spc[i] / bline[i];
                stats.push(fv - ov);
                if (fv < fvm) {
                    ifit = i;
                    fvm = fv;
                }
                if (ov < ovm) {
                    iori = i;
                    ovm = ov;
                }
            }
        }
        QString suffix = QString("S%1_L%2").arg(sid).arg(lid);
        block.updateValue(QString("idx_min_tr_ori_%1").arg(suffix).toStdString(), iori);
        block.updateValue(QString("idx_min_tr_fit_%1").arg(suffix).toStdString(), ifit);
        block.updateValue(QString("res_mean_%1").arg(suffix).toStdString(), stats.mean());
        block.updateValue(QString("res_stddev_%1").arg(suffix).toStdString(), stats.standardDeviation());

        // updateDataBlock should scale output parameters from FIT to BLOCK.
        updateDataBlock(&block, report, scaleInfos);

        // IMPORTANT: add the final parameters to the map of fit parameters.
        // no need to care about scaling: ONLY CONCENTRATIONS ARE USED !!!
        fitParamsMap.insert(sid, report.finalParams);

        // free some memory !
        delete[] xerror;
        delete[] varParams;
        delete[] parameters;

        tmp = timer.nsecsElapsed();
        btimes.postFit += (tmp - lastElapsed);
        btimes.global = btimes.preFit + btimes.fit + btimes.postFit;
        report.times = btimes;
        ttimes += btimes;
    }
    // save previous parameters
    if (fast) {
        mDataMutex.lock();
        for (const BasicFitReport &report : qAsConst(reports)) {
            if (report.spectraId == -1 || report.lineId == -1 || !report.fitSuccess())
                continue;
            QPair<int, int> slid = qMakePair(report.spectraId, report.lineId);
            mPreviousFitParams[slid] = report.finalParams;
            auto lp = mConfig.lines()[report.lineId - 1];
            ScaleInfos scaleInfos = scaleInfosMap.value(report.lineId, {nbi, lp.basePts, nbi, 1});
            double scale = scaleInfos.f2c();
            mPreviousFitParams[slid].scaleSigmaScale(scale);
            mPreviousFitParams[slid].scaleBaseLine(scale);
        }
        mDataMutex.unlock();
    }
    ttimes.global = timer.nsecsElapsed();

    emit fitComplete(block, reports, ttimes);

    mRunMutex.lock();
    mRunning = false;
    mRunMutex.unlock();
}

/**
 * @brief Initialize fit parameters related to given data block.
 * @param block Block to process.
 * @param fitConfig Fit configuration.
 * @param scaleInfos Map with for each line identifier scales informations due to decimation.
 * @param pres Output pressure level (mbar).
 * @param temp Output temperature (K).
 * @param cH2O0 Output default H2O molar fraction (from relative humidity).
 * @param spectras Output list of pre-processed spectras in the block.
 * @param ramps Output list of original spectras (full ramps even not decimated !) in the block.
 * @param meanOffsets Output mean offsets (of each acquisition line).
 * @param eSignals Output original min/max signal value for each working ramp (after offset removal).
 * @param errors Optional pointer to a list with encountered errors.
 * @return True on success, false otherwise.
 * @note
 * If the method fails, no fit could be performed afterwards.
 *
 * In any case, the block is updated with all the expected output parameters. Such parameters are initialized to NaN
 * (except for the offsets and signal extremums).
 */
bool FitWorker::setupBlockData(ABlock &block, const FitConfig &fitConfig, const QMap<int, ScaleInfos> &scaleInfos,
    double &pres, double &temp, double &cH2O0, QMap<int, DoubleVector> &spectras, QMap<int, DoubleVector> &ramps,
    QMap<int, double> &meanOffsets, QMap<int, QPair<double, double>> &eSignals, QStringList *errors) {

    // 1. First create output parameters in block (even if the method fails !)
    for (auto &&p : fitConfig.expectedFitParameters())
        block.appendParameter(p);

    const double nan = qQNaN();

    bool ret = block.isValid();
    if (!ret)
        return ret;

    spectras.clear();
    ramps.clear();
    meanOffsets.clear();
    eSignals.clear();

    int ns = block.rampSize();
    // get data from internal block.

    pres = block.parameter(fitConfig.pressureName()).floatValue();
    temp = block.parameter(fitConfig.temperatureName()).floatValue() + C_T0; // fuckin celsius
    double hr0 = block.parameter(fitConfig.humidityName()).floatValue();

    // P and T are mandatory.
    if (dblCompare(pres, 0.0) || qIsNaN(pres) || qIsNaN(temp)) {
        if (errors)
            errors->append(
                QString("Pressure and/or temperature values are not available (from '%1' and '%2' parameters).")
                    .arg(QString::fromStdString(fitConfig.pressureName()),
                        QString::fromStdString(fitConfig.temperatureName())));
        ret = false;
    }
    // Compute H2O molar fraction from HR (no mandatory).
    if (qIsNaN(hr0)) {
        cH2O0 = nan;
        if (errors)
            errors->append(QString("Cannot compute H2O molar fraction for relative humidity ('%1' parameter).")
                               .arg(QString::fromStdString(fitConfig.humidityName())));
    } else
        cH2O0 = rh2Fraction(hr0, pres, temp);
    // get spectra data: WARNING ASSUME spectra index to be line identifier - 1.
    for (const LParam &lp : fitConfig.lines()) {
        int lid = lp.id, lii = lid - 1;
        // extract vector
        DoubleVector &ramp = ramps.insert(lid, block.ramp(lii)).value();
        if (isNullVector(ramp)) {
            if (errors)
                errors->append(QString("No spectra can be extracted for Line #%1 (using index %2).").arg(lid).arg(lii));
            spectras.insert(lid, ramp);
            continue;
        }
        ScaleInfos infos = scaleInfos.value(lid, {ns, lp.basePts, ns, 1});
        DoubleVector tmp;
        tmp.reserve(infos.nf);
        for (int i = 0; i < ns; i += infos.d)
            tmp.push_back(ramp[i]);

        double scale = infos.f2c(); // ie. npts[FIT]/npts[CONFIG]

        // compute indexes of ramps and offset (using pt scale)
        int bof = qFloor(lp.offset.first * scale);
        int eof = qFloor(lp.offset.second * scale);
        eof += 1; // we add one because we will deal with iterator and we INCLUDE the last point of offset.
        double scaling = lp.scaling;
        double offset = scaling * std::accumulate(tmp.begin() + bof, tmp.begin() + eof, 0.0) / (eof - bof);

        meanOffsets.insert(lid, offset);
        block.updateValue(QString("offset_L%1").arg(lid).toStdString(), offset);

        int blr = qCeil(lp.range.first * scale);
        int elr = qCeil(lp.range.second * scale);
        blr = qMax(0, blr);
        elr = qMin((int)tmp.size() - 1, elr);

        DoubleVector vec;
        vec.reserve(elr - blr + 1);
        double sum = 0;
        if (lp.reverse) {
            for (int i = elr; i > blr - 1; i--) {
                double v = tmp[i] * scaling - offset;
                sum += v;
                vec.push_back(v);
            }
        } else {
            for (int i = blr; i < elr + 1; i++) {
                double v = tmp[i] * scaling - offset;
                sum += v;
                vec.push_back(v);
            }
        }

        double vmin = vec.front(), vmax = vec.back();
        if (vmin > vmax)
            qSwap(vmin, vmax);
        eSignals.insert(lid, qMakePair(vmin, vmax));
        block.updateValue(QString("sigmin_L%1").arg(lid).toStdString(), vmin);
        block.updateValue(QString("sigmax_L%1").arg(lid).toStdString(), vmax);
        if (sum < lp.threshold) {
            if (errors)
                errors->append(QString("Line #%1: Spectra data is below noise threshold.").arg(lid));
            spectras.insert(lid, DoubleVector());
            continue;
        } else {
            std::for_each(vec.begin(), vec.end(), [vmax](double &v) { v /= vmax; });
            spectras.insert(lid, vec);
        }
    }

    return ret;
}

/**
 * @brief Setup fit parameters and spectra from spectra parameters.
 * @param spcParam Spectra parameters pointer.
 * @param fitParams Fit parameters set to update. The instance is first cleared.
 * @param errors Optional pointer to a list of errors encountered.
 * @return (raw) Spectra data to fit.
 * @details
 * The method always update fitParams from the spectra parameters. However the method may fail
 * on some situations. In such case, the output vector is empty.
 *
 * Sources of errors are various but come mostly from invalid data in the AMULSE data block (see
 * FileFitWorker::setupDataBlock):
 *   - Cannot get spectra from the block of raw data.
 *   - Pressure, temperature or path length are NaN.
 *   - Unresolved spectra dependency from fit configuration (should not happen).
 *   - At least a molecule initial concentration is NaN. This situation can occur if the method did not find the
 *     molecule dependency from previous fits.
 */
DoubleVector FitWorker::setupFitParameters(const SParam &spcParam, FitParams &fitParams,
    const QMap<int, FitParams> &fitParamsMap, const ScaleInfos &scaleInfos, const double &pres, const double &temp,
    const double &cH2O0, const QMap<int, DoubleVector> &spectras, QStringList *errors) {

    DoubleVector out;
    const SParam &sp = spcParam;
    LParam lp = spcParam.line;
    int lid = lp.id;
    int lMin = lp.range.first;
    bool ret = true;
    // (re)build the fitParams instance.
    fitParams.clear();
    fitParams.setPressure(CParam("PRESSURE", pres, qQNaN(), qQNaN(), false));
    fitParams.setTemperature(CParam("TEMPERATURE", temp, qQNaN(), qQNaN(), false));
    fitParams.setPathLength(lp.pathLength);
    fitParams.setSigmaScale(lp.sigmaScale);
    fitParams.setBaseLine(lp.baseLine);
    for (const MParam &m : qAsConst(sp.molecules)) {
        MParam fm(m);
        // Check special case H2O
        if (QString::fromStdString(fm.name).compare("H2O", Qt::CaseInsensitive) == 0 &&
            QString::fromStdString(fm.valueFlag).compare("hr", Qt::CaseInsensitive) == 0) {
            fm.value = cH2O0;
        }
        fitParams.setMolecule(fm); // do not check for success...
    }
    // check all concentrations.
    for (const MParam &m : fitParams.molecules()) {
        if (qIsNaN(m.value)) {
            ret = false;
            if (errors)
                errors->append(QString("S#%1: %2 initial concentration is NaN.").arg(sp.id).arg(QString::fromStdString(m.name)));
        }
    }
    if (qIsNaN(lp.pathLength.value)) {
        ret = false;
        if (errors)
            errors->append(QString("S#%1: Path length is NaN.").arg(sp.id));
    }

    // setup spectra data vector...

    // check for ramp and scale infos in the relevant map.
    if (!spectras.contains(lid)) {
        ret = false;
        if (errors)
            errors->append(QString("S#%1: Cannot get spectra data from acquisition line #%2.").arg(sp.id).arg(lid));
    }
    if (!ret) {
        fitParams.nullify();
        return out;
    }
    DoubleVector tmp = spectras.value(lp.id);
    if (isNullVector(tmp)) {
        if (errors)
            errors->append(QString("S#%1: Cannot get spectra data from acquisition line #%2.").arg(sp.id).arg(lid));
        fitParams.nullify();
        return out;
    }

    double scale = scaleInfos.f2c(); // ie. npts[FIT]/npts[CONFIG]
    // Crop the vector if needed.
    int tmpMax = tmp.size() - 1;
    int sMin = qBound(0, qCeil((sp.range.first - lMin) * scale), tmpMax);
    int sMax = qBound(0, qCeil((sp.range.second - lMin) * scale), tmpMax);
    if (lp.reverse) {
        int swap = sMin;
        sMin = tmpMax - sMax;
        sMax = tmpMax - swap;
    }
    fitParams.setSpectraIndexOffset(sMin);
    out = DoubleVector(tmp.begin() + sMin, tmp.begin() + sMax + 1);
    // scale coefficients of sigma-scale and base-line.
    scale = scaleInfos.c2f();
    fitParams.scaleSigmaScale(scale);
    fitParams.scaleBaseLine(scale);
    return out;
}

/**
 * @brief Update a data block with the parameters of a fit.
 * @param block Data block to update.
 * @param fitReport Basic fit report to process.
 * @attention
 * The method does not check for existence of the parameters within the data block to update:
 * ::setupDataFromBlock must have been called before on the block to create such parameters !
 */
void FitWorker::updateDataBlock(ABlock *block, const BasicFitReport &fitReport, const ScaleInfos &infos) {

    int niter = fitReport.iterations;
    int nfev = fitReport.evaluations;

    QString suffix = QString("S%1_L%2").arg(fitReport.spectraId).arg(fitReport.lineId);

    // fit status code
    block->updateValue(QString("fit_status_%1").arg(suffix).toStdString(), fitReport.mpStatus);
    // iterations and func eval.
    block->updateValue(QString("niter_%1").arg(suffix).toStdString(), niter);
    block->updateValue(QString("nfev_%1").arg(suffix).toStdString(), nfev);
    // chi square
    block->updateValue(QString("chi2_%1").arg(suffix).toStdString(), fitReport.chi2);

    // fit parameters
    FitParams fitParams = fitReport.finalParams;
    double scale = infos.f2b();
    fitParams.scaleSigmaScale(scale);
    fitParams.scaleBaseLine(scale);

    QString pname, pename;

    // concentrations
    const std::vector<MParam> &molecs = fitParams.molecules();

    for (int i = 0; i < molecs.size(); i++) {
        const MParam &m = molecs.at(i);
        pname = QString("%1_%2").arg(QString::fromStdString(m.name), suffix);
        pename = QString("ERR_%1").arg(pname);
        block->updateValue(pname.toStdString(), m.value);
        block->updateValue(pename.toStdString(), m.error);
    }

    // Base-line
    const std::vector<CParam> &blcoeffs = fitParams.baseLine();
    for (int i = 0; i < blcoeffs.size(); i++) {
        const CParam &p = blcoeffs.at(i);
        pname = QString("BL_C%1_%2").arg(i).arg(suffix);
        pename = QString("ERR_%1").arg(pname);
        block->updateValue(pname.toStdString(), p.value);
        block->updateValue(pename.toStdString(), p.error);
    }
    // sigma scale
    const std::vector<CParam> &sscoeffs = fitParams.sigmaScale();
    for (int i = 0; i < sscoeffs.size(); i++) {
        const CParam &p = sscoeffs.at(i);
        pname = QString("SS_C%1_%2").arg(i).arg(suffix);
        pename = QString("ERR_%1").arg(pname);
        block->updateValue(pname.toStdString(), p.value);
        block->updateValue(pename.toStdString(), p.error);
    }
    // Environment
    // [24/05/2022] --> path length only
    pname = QString("PL_%1").arg(suffix);
    pename = QString("ERR_%1").arg(pname);
    const CParam &p = fitParams.pathLength();
    block->updateValue(pname.toStdString(), p.value);
    block->updateValue(pename.toStdString(), p.error);
}
