// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/**
 * @file baseplot.cpp
 * @brief BasePlot class definitions file.
 * @author J. Burgalat
 * @date 2016
 */

#include "baseplot.hpp"

/// @brief Vertical axes flag
const QCPAxis::AxisTypes BasePlot::vertAxisType = QCPAxis::atLeft | QCPAxis::atRight;
/// @brief Horizontal axes flag
const QCPAxis::AxisTypes BasePlot::horiAxisType = QCPAxis::atBottom | QCPAxis::atTop;

/**
 * @brief Constructor
 * @param parent Parent widget.
 */
BasePlot::BasePlot(QWidget *parent) : QCustomPlot(parent) {
    mTitle = nullptr;

    // initialize members
    mEmitPosition = true;
    mZoomOrigin = QPoint();
    mZoomMode = false;

    mLogTicker = QSharedPointer<QCPAxisTickerLog>::create();
    mLinTicker = QSharedPointer<QCPAxisTicker>::create();
    mDateTicker = QSharedPointer<QCPAxisTickerDateTime>::create();

    //    mLogTicker = QSharedPointer<QCPAxisTickerLog>(new QCPAxisTickerLog);
    //    mLinTicker = QSharedPointer<QCPAxisTicker>(new QCPAxisTicker);
    //    mDateTicker = QSharedPointer<QCPAxisTickerDateTime>(new QCPAxisTickerDateTime);

    mTimeSpec = Qt::LocalTime;
    setDateTimeFormat("yyyy-MM-dd\nHH:mm:ss.zzz");

    setAxisTicker(QCPAxis::atLeft | QCPAxis::atRight | QCPAxis::atTop | QCPAxis::atBottom, BasePlot::LINEAR);

    setAutoAddPlottableToLegend(false);
    setNoAntialiasingOnDrag(true);
    setNotAntialiasedElements(QCP::AntialiasedElement::aePlottables);
    setInteractions(QCP::iRangeZoom | QCP::iRangeDrag);
    setZoomOrientations(Qt::Horizontal | Qt::Vertical);
    setDragOrientations(Qt::Horizontal | Qt::Vertical);
}

/// @brief Destructor.
BasePlot::~BasePlot() {}

/**
 * @brief Set range of the main X-axis of the plot.
 * @param range New range to set.
 */
void BasePlot::setXRange(const QCPRange &range, int axisRectIndex) {
    QCPAxisRect *axRect = axisRect(axisRectIndex);
    if (axRect) {
        QCPAxis *axis = axRect->axis(QCPAxis::atBottom);
        if (axis)
            axis->setRange(range);
    }
}

/**
 * @brief Set range of the main Y-axis of the axis rect given by index.
 * @param range New range to set.
 * @param axisRectIndex Index of the axis rect to process. Default to 0, that is the default axis rect.
 * @details
 * If the index is out of range or if the afferent axis rect does not have a left Y-axis, nothing is done.
 */
void BasePlot::setYRange(const QCPRange &range, int axisRectIndex) {
    QCPAxisRect *axRect = axisRect(axisRectIndex);
    if (axRect) {
        QCPAxis *axis = axRect->axis(QCPAxis::atLeft);
        if (axis)
            axis->setRange(range);
    }
}

/**
 * @brief Set a global title for the plot.
 * @param title Title to set.
 */
void BasePlot::setTitle(const QString &title, QFont font) {
    // special case: empty title => User do not want a title anymore.
    if (title.isEmpty() && mTitle) {
        if (plotLayout()->remove(mTitle)) {
            mTitle = nullptr;
            plotLayout()->simplify();
        } else {
            mTitle->setText(title);
        }
        return;
    }
    if (!mTitle) {
        mTitle = new QCPTextElement(this);
        plotLayout()->insertRow(0);
        plotLayout()->addElement(0, 0, mTitle);
    }
    mTitle->setText(title);
    if (font == QFont())
        font = qApp->font();
    mTitle->setFont(font);
}

/// @brief  Set date format (Qt conventions) for the DateTime ticker.
void BasePlot::setDateTimeFormat(const QString &format) { mDateTicker->setDateTimeFormat(format); }

/// @brief  Set date time specifier (Qt conventions) for the DateTime ticker.
void BasePlot::setDateSpec(const Qt::TimeSpec spec) {
    mDateTicker->setDateTimeSpec(spec);
    mTimeSpec = spec;
}

/// @brief Show/Hide grid(s) for each axis of each axis rect of the plot.
void BasePlot::setGrid(bool on) {
    for (QCPAxisRect *axRect : axisRects()) {
        for (QCPAxis *axis : axRect->axes())
            axis->grid()->setVisible(on);
    }
    replot();
}

/// @brief Get the ticker type associated to the given axis.
BasePlot::TickerType BasePlot::axisTickerType(QCPAxis *axis) {
    if (axis->ticker() == mLogTicker)
        return LOG;
    else if (axis->ticker() == mLinTicker)
        return LINEAR;
    else
        return DATE;
}

/// @brief Set main X-axis label
void BasePlot::setXLabel(QString label, int axisRectIndex) {
    QCPAxisRect *axRect = axisRect(axisRectIndex);
    if (axRect) {
        QCPAxis *axis = axRect->axis(QCPAxis::atBottom);
        if (axis)
            axis->setLabel(label);
    }
}

/// @brief Set main Y-axis label
void BasePlot::setYLabel(QString label, int axisRectIndex) {
    QCPAxisRect *axRect = axisRect(axisRectIndex);
    if (axRect) {
        QCPAxis *axis = axRect->axis(QCPAxis::atLeft);
        if (axis)
            axis->setLabel(label);
    }
}

/**
 * @brief Rescale all plot axes.
 * @param axRect A specific AxisRect to rescale.
 * @details
 * If axRect is given and valid, only the axes of this QCPAxisRect are rescaled. Otherwise the method updates the
 * range of all axes of the plot. In any case, the method does not call replot().
 */
void BasePlot::rescaleGraphs(QCPAxisRect *axRect) {
    QList<QCPAxisRect *> axRects;
    if (axRect)
        axRects << axRect;
    else
        axRects << axisRects();
    for (QCPAxisRect *axr : axRects) {
        for (QCPAxis *axis : axr->axes()) {
            axis->rescale(true);
        }
    }
}

/**
 * @brief Auto scale the X axis range.
 * @param axRect A specific AxisRect to rescale.
 * @details
 * If axRect is given and valid, only the x-axes of this QCPAxisRect are rescaled. Otherwise the method updates the
 * range of all x-axes of the plot. In any case, the method does not call replot().
 */
void BasePlot::rescaleXAxis(QCPAxisRect *axRect) {
    QList<QCPAxisRect *> axRects;
    if (axRect)
        axRects << axRect;
    else
        axRects << axisRects();
    for (QCPAxisRect *axr : axRects) {
        for (QCPAxis *axis : axr->axes(QCPAxis::atBottom | QCPAxis::atTop)) {
            axis->rescale(true);
        }
    }
}

/**
 * @brief Auto scale the Y axis range.
 * @param axRect A specific AxisRect to rescale.
 * @details
 * If axRect is given and valid, only the y-axes of this QCPAxisRect are rescaled. Otherwise the method updates the
 * range of all y-axes of the plot. In any case, the method does not call replot().
 */
void BasePlot::rescaleYAxis(QCPAxisRect *axRect) {
    QList<QCPAxisRect *> axRects;
    if (axRect)
        axRects << axRect;
    else
        axRects << axisRects();
    for (QCPAxisRect *axr : axRects) {
        for (QCPAxis *axis : axr->axes(QCPAxis::atLeft | QCPAxis::atRight)) {
            axis->rescale(true);
        }
    }
}

/**
 * @brief Set the axes orientations for the zoom feature.
 * @param orientation Orientation of the axes that can be zoomed.
 * @param index Optional index of the axis-rect to process. Default to 0, that is the default axis rect.
 */
void BasePlot::setZoomOrientations(Qt::Orientations orientation, int index) {
    QCPAxisRect *axRect = axisRect(index);
    if (axRect)
        axRect->setRangeZoom(orientation);
}

/**
 * @brief Set the axes orientations for the drag feature.
 * @param orientation Orientation of the axes that can be dragged.
 * @param index Optional index of the axis-rect to process. Default to 0, that is the default axis rect.
 */
void BasePlot::setDragOrientations(Qt::Orientations orientation, int index) {
    QCPAxisRect *axRect = axisRect(index);
    if (axRect)
        axRect->setRangeDrag(orientation);
}

/**
 * @brief Manage interactions on mouse press.
 * @param event A QMouseEvent object with the related event.
 * @details
 * The method starts to perform on specific interactions based on mouse position and button:
 *  - zoom for right-button.
 *  - drag tracer, legend or axes for left button.
 */
void BasePlot::mousePressEvent(QMouseEvent *event) {
    if (mZoomMode && event->button() == Qt::LeftButton) {
        // set zoom mode
        setSelectionRectMode(QCP::srmZoom);
        mZoomOrigin = event->pos();
    }
    QCustomPlot::mousePressEvent(event);
}

/**
 * @brief Manage interactions on mouse release.
 * @param event A QMouseEvent object with the related event.
 * @details
 * The method :
 *  - emits the block index (and the graph key value) related to the tracer position.
 *  - computes new range from zoom.
 *  - resets the plot interactions.
 */
void BasePlot::mouseReleaseEvent(QMouseEvent *event) {
    if (mZoomMode && event->button() == Qt::LeftButton) {
        QPoint pos = event->pos();
        QRect zoomRect = QRect(mZoomOrigin, pos).normalized();
        int surfZoom = zoomRect.width() * zoomRect.height();
        QCPSelectionRect *rect = selectionRect();
        if (surfZoom <= 25)
            rect->cancel();
        else {
            int surf0 = 0;
            QCPAxisRect *origRect = nullptr;
            for (QCPAxisRect *aRect : axisRects()) {
                QRect r = aRect->rect().intersected(zoomRect);
                int rsurf = r.width() * r.height();
                if (rsurf > surf0) {
                    origRect = aRect;
                    surf0 = rsurf;
                }
            }
            if (origRect) {
                QCPAxis *axis = origRect->axis(QCPAxis::atBottom);
                axis->setRange(rect->range(axis));
                axis = origRect->axis(QCPAxis::atLeft);
                axis->setRange(rect->range(axis));
            }
        }
        setSelectionRectMode(QCP::srmNone);
    }
    QCustomPlot::mouseReleaseEvent(event);
}

/// @brief compute the position in data coordinates of the mouse within the BasePlot.
void BasePlot::mouseMoveEvent(QMouseEvent *event) {
    if (mEmitPosition) {
        QCPAxisRect *axRect = axisRectAt(QPointF(event->pos()));
        if (axRect) {
            QString label;
            double xcoord = axRect->axis(QCPAxis::atBottom)->pixelToCoord(event->pos().x());
            double ycoord = axRect->axis(QCPAxis::atLeft)->pixelToCoord(event->pos().y());
            emit mousePositionChanged(xcoord, ycoord);
        }
    }
    QCustomPlot::mouseMoveEvent(event);
}

/**
 * @brief Enable/disable mouse position signals emission on mouse move.
 * @param enable True to enable signal emission, false otherwise.
 */
void BasePlot::emitPosition(bool enable) { mEmitPosition = enable; }

/// @brief Check whether or not the position  on mouse move are emitted.
bool BasePlot::positionEmitted() const { return mEmitPosition; }

/**
 * @brief Rescale all axes of the plot.
 * @details
 * Only visible plottables are taken into account.
 *
 * This slot method is convinient wrapper around QCustomPlot::rescaleAxes and QCustomPlot::replot methods.
 */
void BasePlot::autoRescale() {
    rescaleAxes(true);
    replot();
}

/**
 * @brief Rescale all x-axes of the plot to fit all visible plottables data.
 * @details
 * This slot method is a convinient wrapper around BasePlot::rescaleXAxis and QCustomPlot::replot methods.
 */
void BasePlot::autoRescaleX() {
    rescaleXAxis();
    replot();
}

/**
 * @brief Rescale all the y-axes of the plot to fit all visible plottables data.
 * @details
 * This slot method is a convinient wrapper around BasePlot::rescaleYAxis and QCustomPlot::replot methods.
 */
void BasePlot::autoRescaleY() {
    rescaleYAxis();
    replot();
}

/**
 * @brief Enable/Disable zoom mode.
 * @param enable true to enable zoom mode, false otherwise.
 */
void BasePlot::setZoomMode(bool enable) {
    mZoomMode = enable;
    if (mZoomMode) {
        setCursor(QCursor(Qt::CrossCursor));
        setInteraction(QCP::iRangeDrag, false);
        setSelectionRectMode(QCP::srmZoom);
    } else {
        setCursor(QCursor(Qt::ArrowCursor));
        setInteraction(QCP::iRangeDrag, true);
        setSelectionRectMode(QCP::srmNone);
    }
}

/**
 * @brief Set new ticker to given axis.
 * @param axis Axis to update.
 * @param type Type of ticker to set.
 */
void BasePlot::setAxisTicker(QCPAxis *axis, TickerType type) {
    switch (type) {
    case LOG:
        axis->setTicker(mLogTicker);
        axis->setScaleType(QCPAxis::stLogarithmic);
        axis->setNumberFormat("eb");
        axis->setNumberPrecision(0);
        break;
    case DATE:
        axis->setTicker(mDateTicker);
        axis->setScaleType(QCPAxis::stLinear);
        axis->setNumberFormat("gb");
        axis->setNumberPrecision(8);
        break;
    default:
        axis->setTicker(mLinTicker);
        axis->setScaleType(QCPAxis::stLinear);
        axis->setNumberFormat("gb");
        axis->setNumberPrecision(8);
        break;
    }
}

/**
 * @brief Set new ticker to all axes that matches given orientations.
 * @param axisTypes Flag with the axis types to update.
 * @param type Type of ticker to set.
 */
void BasePlot::setAxisTicker(QCPAxis::AxisTypes axisTypes, TickerType type) {
    for (QCPAxisRect *axRect : axisRects()) {
        for (QCPAxis *ax : axRect->axes(axisTypes))
            setAxisTicker(ax, type);
    }
}

/**
 * @brief Get the list of all X axes of the plot (from all axisRect)
 * @return list of QCPAxis pointers;
 */
QList<QCPAxis *> BasePlot::xAxes() {
    QList<QCPAxis *> axes;
    for (QCPAxisRect *axRect : axisRects())
        axes << axRect->axes(QCPAxis::atBottom | QCPAxis::atTop);
    return axes;
}

/**
 * @brief Get the list of all Y axes of the plot (from all axisRect)
 * @return list of QCPAxis pointers;
 */
QList<QCPAxis *> BasePlot::yAxes() {
    QList<QCPAxis *> axes;
    for (QCPAxisRect *axRect : axisRects())
        axes << axRect->axes(QCPAxis::atLeft | QCPAxis::atRight);
    return axes;
}
