// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/**
 * @file baseplot.hpp
 * @brief BasePlot widget header file.
 * @author J. Burgalat
 * @date 2016
 */

#ifndef BASEPLOT_HPP
#define BASEPLOT_HPP

#include <QObject>
#include "qcustomplot.h"

/**
 * @brief A base plot widget inherited from QCustomPlot.
 * @details
 * The widget adds some shortcut methods to QCustomPlot.
 * The widget implements the following features:
 *     - multiple axisRect (sub-plots) at initialization. In the base implementation subplots are stacked vertically.
 *     - methods to set labels and tickers type for axes on each (or all) axisRect.
 *     - possibility to enable/disable zoom,drag and axes synchronization for each subplot.
 *     - rectangle zoom mode for each axisRect.
 *     - zoom stack.
 *     - axes fixed/free ranges.
 *     - axis rect axes synchronization.
 *
 * It can be used as a base class for plot widgets.
 */
class BasePlot : public QCustomPlot {
    Q_OBJECT

  public:
    /// @brief Enumerate the kind of ticker type managed by the api.
    enum TickerType {
        LINEAR,
        LOG,
        DATE,
    };

    explicit BasePlot(QWidget *parent = nullptr);
    virtual ~BasePlot() override;

    void setTitle(const QString &title, QFont font = QFont());
    void setXLabel(QString label, int axisRectIndex = 0);
    void setYLabel(QString label, int axisRectIndex = 0);
    void setXRange(const QCPRange &range, int axisRectIndex = 0);
    void setYRange(const QCPRange &range, int axisRectIndex = 0);

    TickerType axisTickerType(QCPAxis *axis);
    void setAxisTicker(QCPAxis *axis, TickerType type);
    void setAxisTicker(QCPAxis::AxisTypes axisTypes, TickerType type);
    QList<QCPAxis *> xAxes();
    QList<QCPAxis *> yAxes();

    void setDateTimeFormat(const QString &format);
    void setDateSpec(const Qt::TimeSpec spec = Qt::LocalTime);

    void emitPosition(bool enable);
    bool positionEmitted() const;

    virtual void rescaleGraphs(QCPAxisRect *axRect = nullptr);
    virtual void rescaleXAxis(QCPAxisRect *axRect = nullptr);
    virtual void rescaleYAxis(QCPAxisRect *axRect = nullptr);

signals:
    /**
     * @brief Emitted when the mouse moves over the plot area.
     * @param x X-axis value in data coordinates of the mouse position.
     * @param y Y-axis value in data coordinates of the mouse position.
     * @note
     * The signal is emitted only if BasePlot::positionEmitted is true.
     */
    void mousePositionChanged(double x, double y);

  protected:
    virtual void mousePressEvent(QMouseEvent *event) override;
    virtual void mouseMoveEvent(QMouseEvent *event) override;
    virtual void mouseReleaseEvent(QMouseEvent *event) override;

  public slots:
    void setZoomMode(bool enable);
    virtual void autoRescale();
    virtual void autoRescaleX();
    virtual void autoRescaleY();

    void setZoomOrientations(Qt::Orientations orientation, int index=0);
    void setDragOrientations(Qt::Orientations orientation, int index=0);
    void setGrid(bool on);

  protected:
    static const QCPAxis::AxisTypes vertAxisType;
    static const QCPAxis::AxisTypes horiAxisType;

    bool mEmitPosition;                                ///< Control flag for mouse position signals.
    bool mZoomMode;                                    ///< Zoom mode.
    QPoint mZoomOrigin;                                ///< Zoom origin position.
    QSharedPointer<QCPAxisTickerLog> mLogTicker;       ///< Log ticker.
    QSharedPointer<QCPAxisTicker> mLinTicker;          ///< Linear ticker.
    QSharedPointer<QCPAxisTickerDateTime> mDateTicker; ///< Datetime ticker.
    Qt::TimeSpec mTimeSpec;                            ///< Time specifier for datetime tick labels.
    QCPTextElement *mTitle;                            ///< Plot title.
};

#endif // BASEPLOT_HPP
