// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "colorscale.hpp"

ColorScale::ColorScale(Qt::Orientation orientation, QWidget *parent) : QCustomPlot(parent) {
    setInteractions(QCP::iRangeDrag | QCP::iRangeZoom);
    plotLayout()->clear();

    mColorScale = new QCPColorScale(this);
    mColorScale->setGradient(QCPColorGradient::gpGrayscale);
    mColorScale->setDataRange(QCPRange(0, 1));

    plotLayout()->addElement(0, 0, mColorScale);
    if (orientation == Qt::Horizontal) {
        mColorScale->setType(QCPAxis::atBottom);
    }
    mColorScale->setRangeDrag(false);
    mColorScale->setRangeZoom(false);

    mColorScale->setMinimumMargins(QMargins(16, 16, 16, 16));
    // setup the timer used to postpone the data range signal emission.

    mTimer.setSingleShot(true);
    mTimer.callOnTimeout(this, &ColorScale::emitRangeChanges);

    QObject::connect(mColorScale, &QCPColorScale::dataRangeChanged, this, &ColorScale::onDataRangeChanged);
}

void ColorScale::setOrientation(Qt::Orientation orientation) {
    mColorScale->setType(orientation == Qt::Horizontal ? QCPAxis::atBottom : QCPAxis::atRight);
}

/**
 * @brief Lock/unlock uer interactions
 * @param lock True to block user interactions, false otherwise.
 * @details
 * The method is shortcut to :
 * @code
 * ColorScale::setRangeDrag(lock);
 * ColorScale::setRangeZoom(lock);
 * @endcode
 */
void ColorScale::lockUI(bool lock) {
    bool enabled = !lock;
    setRangeDrag(enabled);
    setRangeZoom(enabled);
}

/// @brief Enable/disable range drag.
/// @details
/// By default range drag is disabled.
void ColorScale::setRangeDrag(bool enable) { mColorScale->setRangeDrag(enable); }

/// @brief Enable/disable range zoom.
/// @details
/// By default range zoom is disabled.
void ColorScale::setRangeZoom(bool enable) { mColorScale->setRangeZoom(enable); }

/// @brief Set a title to the color scale.
void ColorScale::setTitle(const QString &title) {
    mColorScale->setLabel(title);
    replot();
}

/// @brief Set new color gradient.
void ColorScale::setGradient(const QCPColorGradient &gradient) {
    mColorScale->setGradient(gradient);
    replot();
}

/**
 * @brief Set new data range in the color scale.
 * @param range Range to set.
 * @note
 * The method does not emit the signal ColorScale2::rangeChanged.
 */
void ColorScale::setRange(const DataRange &range) {
    bool old = mColorScale->blockSignals(true);
    mColorScale->setDataRange(range.toQCPRange());
    mColorScale->blockSignals(old);
    replot();
}

/**
 * @overload
 * @brief Set new data range in the color scale.
 * @param range Range to set.
 * @note
 * The method does not emit the signal ColorScale2::rangeChanged.
 */
void ColorScale::setRange(const QCPRange &range) {
    bool old = mColorScale->blockSignals(true);
    mColorScale->setDataRange(range);
    mColorScale->blockSignals(old);
    replot();
}

/**
 * @brief Postpone the rangeChanged signal emission.
 * @details
 * The method starts the internal timer that will emit ColorScale2::rangeChanged 100 milliseconds later.
 *
 * This operation is performed in order to avoid excessive signal emission: the signal is intended to be connected
 * to the application ::DataManager that should update the map items color with the new range.
 */
void ColorScale::onDataRangeChanged(const QCPRange &) { mTimer.start(100); }

void ColorScale::emitRangeChanges() { emit rangeChanged(DataRange::fromQCPRange(mColorScale->dataRange())); }
