// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "fitplotwidget.hpp"
#include "fitplot.hpp"
#include "menusection.hpp"

/// @brief Constructor.
FitPlotWidget::FitPlotWidget(QWidget *parent) : QWidget(parent) {
    mPlot = new FitPlot(this);

    mRampData = mPlot->rampGraph()->data();
    mRawSpectraData = mPlot->rawSpectraGraph()->data();
    mFitSpectraData = mPlot->fitSpectraGraph()->data();
    mResidualData = mPlot->residualsGraph()->data();

    mAutoRescale = true;

    setupActions();
    QToolBar *tb = new QToolBar(this);
    tb->setMovable(false);
    tb->setFloatable(false);
    tb->addAction(mActionAutoRescale);
    tb->addSeparator();
    tb->addAction(mActionRescale);
    tb->addAction(mActionXRescale);
    tb->addAction(mActionYRescale);

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->setMenuBar(tb);
    layout->addWidget(mPlot, 2);
}

/// @brief Destructor.
FitPlotWidget::~FitPlotWidget() {}

/// @brief Get the size hint of the widget.
QSize FitPlotWidget::sizeHint() const { return QSize(600, 800); }

/// @brief Translate widgets text.
void FitPlotWidget::translateUI() {
    mPlot->rampGraph()->setName(tr("Original ramp"));
    mPlot->rawSpectraGraph()->setName(tr("Raw spectra"));
    mPlot->fitSpectraGraph()->setName(tr("Fit spectra"));
    mPlot->residualsGraph()->setName(tr("Residual"));
    mPlot->setXLabel(tr("Indexes"), 0);
    mPlot->setYLabel(tr("Voltage"), 0);
    mPlot->setXLabel(tr("wavenumbers"), 1);
    mPlot->setYLabel(tr("Transmission"), 1);
    mPlot->setXLabel(tr("wavenumbers"), 2);
    mPlot->setYLabel(tr("Residual"), 2);
    mPlot->replot();

    mActionAutoRescale->setText(tr("Auto-scale on new data"));
    mActionXRescale->setText(tr("Scale X-axis"));
    mActionYRescale->setText(tr("Scale Y-axis"));
    mActionRescale->setText(tr("Scale axes"));
}

/**
 * @brief Setup plots from given fit report.
 * @param report Fit report to display.
 */
void FitPlotWidget::setFitReport(const FitReport &report) {

    // ramp data
    mPlot->setRampInterval({report.rampInterval().first,report.rampInterval().second});
    mPlot->setSpectraInterval({report.spectraRampInterval().first,report.spectraRampInterval().second});
    {
        const DoubleVector &values = report.originalRamp();
        int npts = values.size();
        bool vis = npts > 0;
        QVector<QCPGraphData> tempData(npts, QCPGraphData(qQNaN(), qQNaN()));
        for (int i = 0; i < npts; i++) {
            tempData[i].key = static_cast<double>(i);
            tempData[i].value = values[i];
        }
        mRampData.data()->set(tempData, true);
        mPlot->setRampIntervalVisible(vis);
        mPlot->setSpectraIntervalVisible(vis);
    }

    const DoubleVector &sigmas = report.sigmas();
    int npts = sigmas.size();
    // raw spectra data
    QVector<QCPGraphData> tempData(npts, QCPGraphData(qQNaN(), qQNaN()));
    {
        const DoubleVector &values = report.rawSpectra();
        const DoubleVector &baseLine = report.baseLine();
        for (int i = 0; i < npts; i++) {
            tempData[i].key = sigmas[i];
            tempData[i].value = values[i] / baseLine[i];
        }
        mRawSpectraData.data()->set(tempData, true);
    }
    // fit spectra & residual data.
    {
        const DoubleVector &values = report.fitSpectra();
        QVector<QCPGraphData> resData(npts, QCPGraphData(qQNaN(), qQNaN()));
        for (int i = 0; i < npts; i++) {
            resData[i].key = tempData[i].key;
            resData[i].value = tempData[i].value - values[i];
            tempData[i].value = values[i];
        }
        mFitSpectraData.data()->set(tempData, true);
        mResidualData.data()->set(resData, true);
    }
    if (mAutoRescale)
        mPlot->rescaleGraphs();
    mPlot->replot();
}

/**
 * @brief Clear the plots.
 */
void FitPlotWidget::clearPlots() { mPlot->clearData(); }

/// @brief Enable/Disable auto-scale of plots on new coming data.
void FitPlotWidget::setAutoRescale(bool enabled) {
    mAutoRescale = enabled;
    mActionAutoRescale->setChecked(mAutoRescale);
}

/**
 * @brief Open a context menu with specific actions to perform on the internal plot widget.
 * @param pos Position of the mouse event (relative to the plot widget).
 * @details
 * The method generates a context menu with four actions that manage scaling features of the plot.
 */
void FitPlotWidget::onPlotContextMenuRequested(const QPoint &pos) {

    QMenu *menu = new QMenu(this);
    menu->setAttribute(Qt::WA_DeleteOnClose);
    menu->addAction(mActionAutoRescale);
    menu->addAction(new MenuSection(tr("Scaling"), menu));
    menu->addAction(mActionRescale);
    menu->addAction(mActionXRescale);
    menu->addAction(mActionYRescale);
    menu->popup(mPlot->mapToGlobal(pos));
}

void FitPlotWidget::setupActions() {
    mActionAutoRescale = new QAction(QIcon(":/plotting/auto-rescale.png"), tr("Auto-scale on new data"), this);
    QObject::connect(mActionAutoRescale, &QAction::triggered, this, &FitPlotWidget::setAutoRescale);
    mActionAutoRescale->setCheckable(true);
    mActionAutoRescale->setChecked(mAutoRescale);
    mActionXRescale = new QAction(QIcon(":/plotting/auto-x.png"), tr("Scale X-axis"), this);
    QObject::connect(mActionXRescale, &QAction::triggered, mPlot, &FitPlot::autoRescaleX);
    mActionYRescale = new QAction(QIcon(":/plotting/auto-y.png"), tr("Scale Y-axis"), this);
    QObject::connect(mActionYRescale, &QAction::triggered, mPlot, &FitPlot::autoRescaleY);
    mActionRescale = new QAction(QIcon(":/plotting/auto-both.png"), tr("Scale axes"), this);
    QObject::connect(mActionRescale, &QAction::triggered, mPlot, &FitPlot::autoRescale);
}
