// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "lineposition.hpp"
#include "qcp_helpers.hpp"

/**
 * @brief Constructor
 * @param axisRect Axis rect where to put the line position.
 * @param value Initial position of the line (in axis rect ratio coordinates).
 * @param orientation Orientation of the line: horizontal or vertical (default).
 */
LinePosition::LinePosition(QCPAxisRect *axisRect, double value, Qt::Orientation orientation)
    : QCPItemStraightLine(axisRect->parentPlot()) {
    mChannelFillLine = nullptr;
    mAntialiasedFill = true;
    mFixed = false;
    mAxisRect = axisRect;
    mOrientation = orientation;
    mDragged = false;
    setClipAxisRect(mAxisRect);
    setClipToAxisRect(true);
    point1->setType(QCPItemPosition::ptAxisRectRatio);
    point1->setAxes(mAxisRect->axis(QCPAxis::atBottom), axisRect->axis(QCPAxis::atLeft));
    point2->setType(QCPItemPosition::ptAxisRectRatio);
    point2->setAxes(mAxisRect->axis(QCPAxis::atBottom), axisRect->axis(QCPAxis::atLeft));
    if (mOrientation == Qt::Vertical) {
        point1->setCoords(value, 0);
        point2->setCoords(value, 1);
    } else {
        point1->setCoords(0, value);
        point2->setCoords(1, value);
    }
    setPen(QPen(Qt::black));
    setBrush(Qt::NoBrush);
}

/**
 * @brief Fix the position of the line.
 * @param fixed True to prevent movement from user interactions, false otherwise.
 * @details
 * By default, the LinePosition can be freely moved by user-interactions.
 * If fixed is true, the user cannot interact anymore (with the mouse) to change the position of the line.
 *
 * @note
 * The line position can still be moved programmatically using LinePosition::setPosition.
 */
void LinePosition::setFixed(bool fixed) { mFixed = fixed; }

/// @brief Check whether or not the line is fixed.
/// @sa LinePosition::setFixed
bool LinePosition::isFixed() const { return mFixed; }

/// @brief Bound/Unbound the line position to data coordinates.
void LinePosition::boundToData(bool enable) {
    mBoundToData = enable;
    QPointF p1, p2;
    if (mBoundToData) {
        p1 = Plotting::axisRectRatioToPlotCoord(mAxisRect, point1->coords());
        p2 = Plotting::axisRectRatioToPlotCoord(mAxisRect, point2->coords());
        point1->setType(QCPItemPosition::ptPlotCoords);
        point2->setType(QCPItemPosition::ptPlotCoords);
    } else {
        p1 = Plotting::plotCoordToAxisRectRatio(mAxisRect, point1->coords());
        p2 = Plotting::plotCoordToAxisRectRatio(mAxisRect, point2->coords());
        point1->setType(QCPItemPosition::ptAxisRectRatio);
        point2->setType(QCPItemPosition::ptAxisRectRatio);
    }
    point1->setCoords(p1);
    point2->setCoords(p2);
    // should I (re)compute the positions ??
    if (visible())
        parentPlot()->replot();
}

/// @brief Check whether or not the line position are expressed in data coordinates.
bool LinePosition::boundedToData() const { return mBoundToData; }

/**
 * @brief Manage interactions on mouse press.
 * @details
 * The method initializes the drag process of the line.
 */
void LinePosition::mousePressEvent(QMouseEvent *event, const QVariant &details) {
    Q_UNUSED(details)
    if (!mFixed && visible() && event->button() == Qt::LeftButton) {
        mIsDraggable = parentPlot()->interactions().testFlag(QCP::iRangeDrag);
        parentPlot()->setInteraction(QCP::iRangeDrag, false);
        mDragged = true;
    } else {
        mDragged = false;
    }
}

/**
 * @brief Manage interactions on mouse move.
 * @details
 * The method updates the position of the line.
 */
void LinePosition::mouseMoveEvent(QMouseEvent *event, const QPointF &startPos) {
    Q_UNUSED(startPos)
    if (visible() && mDragged) {
        QPointF pos = mBoundToData ? Plotting::pixelToPlotCoord(mAxisRect, event->localPos())
                                   : Plotting::pixelToAxisRectRatio(mAxisRect, event->localPos());
        move(pos);
        parentPlot()->replot();
    }
}

/**
 * @brief Manage interactions on mouse release.
 * @details
 * The method computes the final position of the LinePosition.
 */
void LinePosition::mouseReleaseEvent(QMouseEvent *event, const QPointF &startPos) {
    Q_UNUSED(startPos)
    if (mDragged) {
        mDragged = false;
        parentPlot()->setInteraction(QCP::iRangeDrag, mIsDraggable);
        QPointF pos = mBoundToData ? Plotting::pixelToPlotCoord(mAxisRect, event->localPos())
                                   : Plotting::pixelToAxisRectRatio(mAxisRect, event->localPos());
        move(pos);
        emit mouseReleasedPosition(mOrientation == Qt::Vertical ? pos.x() : pos.y());
        parentPlot()->replot();
    }
}

/**
 * @brief Get a cursor shape based the given mouse position
 * @param event Mouse event.
 * @return Cursor shape.
 */
Qt::CursorShape LinePosition::cursorShape(QMouseEvent *event) {
    if (mFixed || !visible())
        return Qt::ArrowCursor;
    bool isOver = selectTest(event->localPos(), false, nullptr) <= 0.99 * parentPlot()->selectionTolerance();
    return isOver ? (mOrientation == Qt::Vertical ? Qt::SizeHorCursor : Qt::SizeVerCursor) : Qt::ArrowCursor;
}

/// @brief Get the current position.
double LinePosition::position() const { return mOrientation == Qt::Vertical ? point1->key() : point1->value(); }

/// @brief Get the current position in term of Axis rect ratio.
double LinePosition::axisRectRatio() const {
    if (mBoundToData) {
        QPointF tmp = Plotting::plotCoordToAxisRectRatio(mAxisRect, QPointF(point1->key(), point2->value()));
        return mOrientation == Qt::Vertical ? tmp.x() : tmp.y();
    } else {
        return mOrientation == Qt::Vertical ? point1->key() : point1->value();
    }
}

/**
 * @brief set new position for this LinePosition.
 * @param value New position. It should reflect the current coordinate system (see LinePosition::boundedToData).
 * @details
 * The method does not emit LinePosition::positionChanged.
 *
 * The method does not request a replot of the parent plot.
 */
void LinePosition::setPosition(double value) {
    QPointF pt1 = mOrientation == Qt::Vertical ? QPointF(value, 0) : QPointF(0, value);
    QPointF pt2 = mOrientation == Qt::Vertical ? QPointF(value, 1) : QPointF(1, value);
    point1->setCoords(pt1);
    point2->setCoords(pt2);
}

/**
 * @brief Move the position of line.
 * @param pos Position to move to. It should reflect the current coordinate system (see LinePosition::boundedToData).
 * @note
 * The method does emit positionChanged signal.
 */
void LinePosition::move(QPointF pos) {
    double p = mOrientation == Qt::Vertical ? pos.x() : pos.y();
    setPosition(p);
    emit positionChanged(p);
}

/// @brief Get the channel fill brush.
QBrush LinePosition::brush() const { return mBrush; }

/// @brief Set the channel fill brush.
void LinePosition::setBrush(const QBrush &brush) { mBrush = brush; }

/// @brief Check whether or not fill channell is antialiased.
bool LinePosition::antialiasedFill() const { return mAntialiasedFill; }

/// @brief Enable/Disable antialiasing for fill channel.
void LinePosition::setAntialiasedFill(bool enabled) { mAntialiasedFill = enabled; }

/**
 * @brief Register a line for channel filling.
 * @param targetLine Line to register.
 * @details
 * Upon registration, the plot will draw a fill using the LinePosition::brush between this line and the registered line.
 * @attention
 * By default the LinePosition does not set a brush for channel fill, then LinePosition::setBrush should be called explicitly
 * otherwise the channel will not be visible.
 */
void LinePosition::setChannelFillLine(LinePosition *targetLine) {
    // prevent setting channel target to this graph itself:
    if (targetLine == this) {
        qDebug() << Q_FUNC_INFO << "targetLine is this line itself";
        mChannelFillLine = nullptr;
        return;
    }
    // prevent setting channel target to a line position not in the plot:
    if (targetLine && targetLine->mAxisRect != mAxisRect) {
        qDebug() << Q_FUNC_INFO << "targetLine not in same axis rect";
        mChannelFillLine = nullptr;
        return;
    }
    // prevent setting channel target to a line position that does not have the same orientation !

    mChannelFillLine = targetLine;
}

/// @brief Get the line registered for channel filling.
LinePosition *LinePosition::channelFillLine() const { return mChannelFillLine.data(); }

QLineF LinePosition::getLine() {
    QCPVector2D start(point1->pixelPosition());
    QCPVector2D end(point2->pixelPosition());
    // get visible segment of straight line inside clipRect:
    double clipPad = mainPen().widthF();
    return getRectClippedStraightLine(start, end - start, clipRect().adjusted(-clipPad, -clipPad, clipPad, clipPad));
}

void LinePosition::draw(QCPPainter *painter) {
    QCPItemStraightLine::draw(painter);
    drawFill(painter);
}

void LinePosition::drawFill(QCPPainter *painter) const {
    if (!mChannelFillLine || mBrush.style() == Qt::NoBrush || mBrush.color().alpha() == 0)
        return;
    if (mChannelFillLine->mOrientation != mOrientation) {
        qDebug() << Q_FUNC_INFO << "Registered LinePosition for channel filling does not have the same orientation !";
        return;
    }
    double clipPad = mainPen().widthF();
    QRect fillRect = clipRect().adjusted(-clipPad, -clipPad, clipPad, clipPad);
    double v1 = axisRectRatio(), v2 = mChannelFillLine->axisRectRatio();
    if (v1 > v2)
        qSwap(v1, v2);
    // out of the axis rect !
    if (v2 <= 0 || v1 >= 1)
        return;
    v1 = qBound(0.0, v1, 1.0);
    v2 = qBound(0.0, v2, 1.0);
    if (mOrientation == Qt::Vertical) {
        int w = fillRect.width();
        int dl = v1 * w, dr = (1.0 - v2) * w;
        fillRect.adjust(dl, 0, -dr, 0);
    } else {
        int h = fillRect.height();
        int dt = v1 * h, db = (1.0 - v2) * h;
        fillRect.adjust(0, dt, 0, -db);
    }

    QBrush pbrush = painter->brush();
    QPen ppen = painter->pen();
    painter->setBrush(mBrush);
    painter->setPen(Qt::NoPen);
    applyAntialiasingHint(painter, mAntialiasedFill, QCP::aeFills);
    painter->drawRect(fillRect);
    painter->setBrush(pbrush);
    painter->setPen(ppen);
}
