// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "textbox.hpp"

/// @brief Offset applied to AXisRect ratio coordinates for the predefined positions.
#define TB_OFF 0.005

const QMap<TextBox::Position, QPair<QPointF, Qt::Alignment>> TextBox::gPositions{
    {TextBox::TOP_LEFT, {QPointF(0 + TB_OFF, 0 + TB_OFF), Qt::Alignment(Qt::AlignTop | Qt::AlignLeft)}},
    {TextBox::TOP_CENTER, {QPointF(0.5, 0 + TB_OFF), Qt::Alignment(Qt::AlignTop | Qt::AlignHCenter)}},
    {TextBox::TOP_RIGHT, {QPointF(1.0 - TB_OFF, 0 + TB_OFF), Qt::Alignment(Qt::AlignTop | Qt::AlignRight)}},
    {TextBox::BOTTOM_LEFT, {QPointF(0 + TB_OFF, 1.0 - TB_OFF), Qt::Alignment(Qt::AlignBottom | Qt::AlignLeft)}},
    {TextBox::BOTTOM_CENTER, {QPointF(0.5, 1.0 - TB_OFF), Qt::Alignment(Qt::AlignBottom | Qt::AlignHCenter)}},
    {TextBox::BOTTOM_RIGHT, {QPointF(1.0 - TB_OFF, 1.0 - TB_OFF), Qt::Alignment(Qt::AlignBottom | Qt::AlignRight)}},

};

/**
 * @brief Constructor.
 * @param parentPlot Parent QCustomPlot instance.
 * @details
 * The TextBox instance is positioned on the top-right corner of the main axis rect of the plot.
 */
TextBox::TextBox(QCPAxisRect *axisRect) : QCPItemText(axisRect->parentPlot()) { setup(axisRect); }

/**
 * @brief Constructor
 * @param initialPosition Initial position to set. If the given position is not valid, the constructor will set the position to
 * TextBox::TOP_RIGHT.
 * @param parentPlot Parent QCustomPlot instance.
 */
TextBox::TextBox(TextBox::Position initialPosition, QCPAxisRect *axisRect) : QCPItemText(axisRect->parentPlot()) {
    setup(axisRect, initialPosition);
}

/// @brief Destructor.
TextBox::~TextBox() {}

/// @brief Setup the widget.
void TextBox::setup(QCPAxisRect *axisRect, TextBox::Position initialPosition) {
    setBrush(Qt::white);
    setSelectedBrush(Qt::white);
    position->setAxisRect(axisRect);
    position->setType(QCPItemPosition::ptAxisRectRatio);

    if (!TextBox::gPositions.contains(initialPosition))
        initialPosition = TOP_RIGHT;
    setPositionFlag(initialPosition);
    setPadding(QMargins(4, 2, 4, 2));
    setLayer("legend");
}

/**
 * @brief Set the position of the textbox.
 * @param pos Position to set.
 * @details
 * The method sets the position of the text-box and adjust the box alignment according to the position.
 */
void TextBox::setPositionFlag(TextBox::Position pos) {
    if (TextBox::gPositions.contains(pos)) {
        const QPair<QPointF, Qt::Alignment> p = TextBox::gPositions[pos];
        position->setCoords(p.first);
        setPositionAlignment(p.second);
        mPosFlag = pos;
    }
}

/// @brief Get the current position flag.
TextBox::Position TextBox::positionFlag() const { return mPosFlag; }

#undef TB_OFF
