// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "abstractudpsocket.hpp"
#include "globals.hpp"

/// @brief Constructor.
AbstractUDPSocket::AbstractUDPSocket(QObject *parent)
    : QObject(parent), mSocket(new QUdpSocket(this)), mListen(false), mName(""), mMinTimeout(-1), mBlockRate(200),
      mByteRate(200) {

    QObject::connect(mSocket, &QUdpSocket::stateChanged, this, &AbstractUDPSocket::socketStateChanged);
    QObject::connect(mSocket, &QUdpSocket::readyRead, this, &AbstractUDPSocket::readPendingDatagrams);
}

/// @brief Destructor.
AbstractUDPSocket::~AbstractUDPSocket() { mSocket->disconnectFromHost(); }

/**
 * @brief Connect and bind the socket to ip and port.
 * @param ip Ip address
 * @param port Port number.
 * The method always stop listening (see AbstractUDPSocket::setListening) and disconnects the socket from
 * previous host (if any).
 */
bool AbstractUDPSocket::connectAndBind(const QString &ip, quint16 port) {
    QHostAddress qip(ip);
    if (!port) {
        LOG_ERROR(
            tr("[UDP] No (valid) port number specified for %1 socket. No connection will be performed.").arg(mName));
        return false;
    }
    bool result;
    mListen = false;
    mSocket->disconnectFromHost();
    QString msg;
    if (!(result = mSocket->bind(qip, port, QUdpSocket::DontShareAddress))) {
        if (mName.isEmpty())
            msg = tr("[UDP] Cannot bind socket: %1").arg(mSocket->errorString());
        else
            msg = tr("[UDP] Cannot bind %1 socket: %2").arg(mName, mSocket->errorString());
        LOG_ERROR(msg);
    } else {
        QString msg1;
        if (mName.isEmpty()) {
            msg = tr("[UDP] Socket successfully bound to %1:%2").arg(ip).arg(port);
            msg1 = tr("[UDP] Socket starts listening to %1:%2").arg(ip).arg(port);
        } else {
            msg = tr("[UDP] Socket %1 successfully bound to %2:%3").arg(mName).arg(ip).arg(port);
            msg1 = tr("[UDP] Socket %1 starts listening to %2:%3").arg(mName).arg(ip).arg(port);
        }
        LOG_INFO(msg);
        LOG_INFO(msg1);
        mListen = true;
    }
    return result;
}

/**
 * @brief Disconnect from Host.
 * @details
 * The method disconnects the internal socket from Host and stop listening to incoming data block.
 */
void AbstractUDPSocket::disconnectSocket(bool log) {
    if (log) {
        QString msg;
        if (mName.isEmpty()) {
            msg = tr("[UDP] Socket stops listening to %1:%2").arg(ip()).arg(port());
        } else {
            msg = tr("[UDP] Socket %1 stops listening to %2:%3").arg(mName).arg(ip()).arg(port());
        }
        LOG_INFO(msg);
    }
    mListen = false;
    mSocket->disconnectFromHost();
    if (log) {
        LOG_INFO(mName.isEmpty() ? tr("[UDP] Disconnecting UDP socket")
                                 : tr("[UDP] Disconnecting %1 UDP socket").arg(mName));
    }
}

/// @brief Get the internal socket.
QUdpSocket *AbstractUDPSocket::socket() { return mSocket; }

/// @brief Get the socket state.
QAbstractSocket::SocketState AbstractUDPSocket::socketState() const { return mSocket->state(); }

/// @brief Check whether or not the internal socket is bound.
bool AbstractUDPSocket::isBound() const { return mSocket->state() == QAbstractSocket::BoundState; }

/// @brief Check whether or not the instance is listening to the incoming data.
bool AbstractUDPSocket::isListening() const { return mListen; }

/// @brief Get the current bounded local IP address.
QString AbstractUDPSocket::ip() const { return mSocket->localAddress().toString(); }

/// @brief Get the current bounded local port.
quint16 AbstractUDPSocket::port() const { return mSocket->localPort(); }

/// @brief Get the (friendly) name fo the socket.
const QString &AbstractUDPSocket::name() { return mName; }

/// @brief Set the (friendly) name of the socket.
void AbstractUDPSocket::setName(const QString &name) { mName = name; }

/**
 * @brief Enable/Disable listening for incoming data.
 * @param enable True to enable, false otherwise.
 * @return True on success, false otherwise.
 * @details
 * The method in the base class simply updates AbstractUDPSocket::mListen and resets
 * AbstractUDPSocket::mByteRate and AbstractUDPSocket::mBlockRate to zero.
 *
 * In subclasses, mListen should be used to check whether or not process data in AbstractUDPSocket::readPendingDatagrams
 * method.
 */
void AbstractUDPSocket::setListening(bool enable) {
    mListen = enable;
    if (enable)
        mFreqTimer.start();
}

/**
 * @brief Set the maximum rate of emission of AbstractUDPSocket::setMaximumInputFrequency
 * @param frequency Frequency to set in Hertz.
 * @details
 * The base class does not make any use of the frequency which is stored internally as a timeout in milliseconds.
 * Subclass can use the timeout (AbstractUDPSocket::mMinTimeout) and elapsed timer (AbstractUDPSocket::mFreqTimer)
 * in AbstractUDPSocket::readPendingDatagrams method to limit the rate of processed data using the
 * QElapsedTimer::hasExpired method for instance.
 *
 * The minimum timeout posible is 5 msec (200Hz).
 *
 * @note
 * The timer is (re)started each time AbstractUDPSocket::setListening is called with true as <i>enable</i> argument
 * value.
 *
 */
void AbstractUDPSocket::setMaximumInputFrequency(double frequency) {
    if (frequency <= 0)
        mMinTimeout = -1;
    else
        mMinTimeout = qMax(5, (int)(1000 / frequency));
}
