// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "arcgisaccountwidget.hpp"
#include "clickablelabel.hpp"
#include "globals.hpp"
#include "loginview.hpp"

#include "AuthenticationManager.h"
#include "Credential.h"
#include "CredentialCache.h"
#include "Portal.h"
#include "PortalTypes.h"
#include "PortalUser.h"
#include "QtWaitingSpinner/waitingspinnerwidget.h"
#include "SubscriptionInfo.h"

using namespace Esri::ArcGISRuntime;

QString subscriptionStateTip(SubscriptionState state);
QString userRole(PortalUserRole role);
QString subscriptionState(SubscriptionState state);

/// @brief Constructor.
ArcgisAccountWidget::ArcgisAccountWidget(QWidget *parent) : QGroupBox(parent) {
    QFormLayout *layout = new QFormLayout(this);
    mLblNameDescr = new QLabel(this);
    mLblOrganizationDescr = new QLabel(this);
    mLblRoleDescr = new QLabel(this);
    mLblStateDescr = new QLabel(this);
    mLblNameValue = new QLabel(this);
    mLblOrganizationValue = new QLabel(this);
    mLblRoleValue = new QLabel(this);
    mLblStateValue = new QLabel(this);
    mLblFetch = new ClickableLabel(this);
    mLblFetch->setHighlight(true);

    for (QLabel *lbl : {mLblNameValue, mLblOrganizationValue, mLblRoleValue, mLblStateValue})
        lbl->setAlignment(Qt::AlignRight);
    mLblFetch->setAlignment(Qt::AlignCenter);

    layout->addRow(mLblOrganizationDescr, mLblOrganizationValue);
    layout->addRow(mLblNameDescr, mLblNameValue);
    layout->addRow(mLblRoleDescr, mLblRoleValue);
    layout->addRow(mLblStateDescr, mLblStateValue);

    layout->addRow(mLblFetch);

    mSpinner = new WaitingSpinnerWidget(this);
    mSpinner->setRoundness(70.0);
    mSpinner->setMinimumTrailOpacity(15.0);
    mSpinner->setTrailFadePercentage(70.0);
    mSpinner->setNumberOfLines(12);
    mSpinner->setLineLength(10);
    mSpinner->setLineWidth(5);
    mSpinner->setInnerRadius(10);
    mSpinner->setRevolutionsPerSecond(1);

    QObject::connect(mLblFetch, &ClickableLabel::clicked, this, &ArcgisAccountWidget::signin);
}

/// @brief Display login view in order to renew license.
void ArcgisAccountWidget::signin() {
    auto am = AuthenticationManager::instance();
    am->credentialCache()->removeAndRevokeAllCredentials();
    LoginView *loginView = new LoginView;
    loginView->autoCloseOnSignIn(false);
    loginView->setAttribute(Qt::WA_DeleteOnClose, true);
    loginView->exec();
    QTimer::singleShot(250, this, &ArcgisAccountWidget::populateWidgets);
}

/// @brief Translate widget texts.
void ArcgisAccountWidget::translateUI() {
    setTitle(tr("ArcGIS account informations"));
    mLblNameDescr->setText(tr("User name"));
    mLblRoleDescr->setText(tr("User role"));
    mLblOrganizationDescr->setText(tr("Organization"));
    mLblStateDescr->setText(tr("Subscription status"));
    mLblFetch->setText(tr("ArcGIS sign-in (internet required)"));
    populateWidgets();
    setHelper(this, HelpFlag::ART_SIGNIN);
}

/**
 * @brief Populate widgets with license informations.
 */
void ArcgisAccountWidget::populateWidgets() {
    Portal *portal = new Portal(true, this);
    QObject::connect(portal, &Portal::loadStatusChanged, this, &ArcgisAccountWidget::onPortalLoadStatusChanged);
    portal->load();
    mSpinner->start();
}

void ArcgisAccountWidget::onPortalLoadStatusChanged(LoadStatus loadStatus) {
    Portal *portal = qobject_cast<Portal *>(QObject::sender());
    PortalUser *user = nullptr;
    QString oname;
    SubscriptionState state = SubscriptionState::Unknown;
    switch (loadStatus) {
    case LoadStatus::Loaded:
        mSpinner->stop();
        if (portal) {
            user = portal->portalUser();
            mLblNameValue->setText(QString("%1 (%2)").arg(user->username(), user->fullName()));

            mLblRoleValue->setText(userRole(user->role()));
            if (portal->portalInfo()) {
                oname = portal->portalInfo()->organizationName();
                state = portal->portalInfo()->subscriptionInfo()->state();
                mLblStateValue->setText(subscriptionState(state));
                mLblStateValue->setToolTip(subscriptionStateTip(state));
            } else {
                mLblStateValue->setText("N/A");
                mLblStateValue->setToolTip("");
            }
            mLblOrganizationValue->setText(oname.isEmpty() ? QString("N/A") : oname);
        }
        portal->deleteLater();
        break;
    case LoadStatus::Loading:
        break;
    case LoadStatus::FailedToLoad:
        mSpinner->stop();
        mLblNameValue->setText("N/A");
        mLblOrganizationValue->setText("N/A");
        mLblRoleValue->setText("N/A");
        mLblStateValue->setText("N/A");
        portal->deleteLater();
        break;
    case LoadStatus::NotLoaded:
        break;
    case LoadStatus::Unknown:
        break;
    default:
        break;
    }
}

QString userRole(PortalUserRole role) {
    switch (role) {
    case PortalUserRole::Publisher:
        return QCoreApplication::tr("Publisher");
    case PortalUserRole::Admin:
        return QCoreApplication::tr("Administrator");
    case PortalUserRole::User:
        return QCoreApplication::tr("User");
    case PortalUserRole::Unknown:
    default:
        return QCoreApplication::tr("Unknown");
    }
}

QString subscriptionState(SubscriptionState state) {
    switch (state) {
    case SubscriptionState::Active:
        return QCoreApplication::tr("Active");
    case SubscriptionState::Canceled:
        return QCoreApplication::tr("Canceled");
    case SubscriptionState::Deleted:
        return QCoreApplication::tr("Deleted");
    case SubscriptionState::Expired:
        return QCoreApplication::tr("Expired");
    case SubscriptionState::Restricted:
        return QCoreApplication::tr("Restricted");
    case SubscriptionState::Suspended:
        return QCoreApplication::tr("Suspended");
    case SubscriptionState::Unknown:
    default:
        return QCoreApplication::tr("Unknown");
    }
}

QString subscriptionStateTip(SubscriptionState state) {
    switch (state) {
    case SubscriptionState::Active:
        return QCoreApplication::tr("Subscription is active and allows all functionality.");
    case SubscriptionState::Canceled:
        return QCoreApplication::tr("Subscription was canceled.");
    case SubscriptionState::Deleted:
        return QCoreApplication::tr("Subscription was deleted.");
    case SubscriptionState::Expired:
        return QCoreApplication::tr("Subscription has expired.");
    case SubscriptionState::Restricted:
        return QCoreApplication::tr("Subscription has run out of credits.");
    case SubscriptionState::Suspended:
        return QCoreApplication::tr("Subscription was suspended.");
    case SubscriptionState::Unknown:
    default:
        return QCoreApplication::tr("Subscription state is unknown.");
    }
}
