// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "audpsocket.hpp"
#include <streambuf>

template <class CharT, class Traits = std::char_traits<CharT>>
class basic_memstreambuf : public std::basic_streambuf<CharT, Traits> {
  public:
    typedef std::basic_streambuf<CharT, Traits> BaseT;
    typedef typename std::basic_streambuf<CharT, Traits>::char_type char_type;

    typedef typename std::basic_streambuf<CharT, Traits>::int_type int_type;
    typedef typename std::basic_streambuf<CharT, Traits>::traits_type traits_type;

    typedef typename std::basic_streambuf<CharT, Traits>::pos_type pos_type;
    typedef typename std::basic_streambuf<CharT, Traits>::off_type off_type;

    using BaseT::eback;
    using BaseT::egptr;
    using BaseT::gbump;
    using BaseT::gptr;
    using BaseT::setg;

    // using BaseT::int_type;
    // using BaseT::traits_type;

  public:
    basic_memstreambuf() : BaseT() {}

    // non-standard
    explicit basic_memstreambuf(const std::basic_string<char_type> &s) : BaseT() {
        assert(!s.empty());
        setg(const_cast<char_type *>(s.begin()), const_cast<char_type *>(s.begin()), const_cast<char_type *>(s.end()));
    }

    // non-standard
    basic_memstreambuf(const char_type *s, std::streamsize n) : BaseT() {
        assert(s);
        assert(n > 0);
        setg(const_cast<char_type *>(s), const_cast<char_type *>(s), const_cast<char_type *>(s + n));
    }

    // non-standard
    basic_memstreambuf(const char_type *begin, const char_type *end) : BaseT() {
        assert(begin);
        assert(end);
        assert(begin < end);

        // check size
        const std::uintmax_t count = end - begin;
        if (count > static_cast<decltype(count)>(std::numeric_limits<std::streamsize>::max())) {
            throw std::invalid_argument("basic_memstreambuf too big");
        }

        setg(const_cast<char_type *>(begin), const_cast<char_type *>(begin), const_cast<char_type *>(end));
    }

  private:
    explicit basic_memstreambuf(const basic_memstreambuf &rhs) : BaseT(static_cast<BaseT &>(rhs)) {}
    basic_memstreambuf &operator=(const basic_memstreambuf &) = delete;

  protected:
    virtual std::streamsize showmanyc() override {
        const auto *ptr = gptr();
        const auto *end = egptr();

        assert(ptr <= end);

        return (ptr <= end) ? (end - ptr) : 0;
    }

    virtual int_type underflow() override {
        const auto *ptr = gptr();

        if (ptr >= egptr())
            return traits_type::eof();

        return traits_type::to_int_type(*ptr);
    }

    virtual pos_type seekoff(
        off_type off, std::ios_base::seekdir dir, std::ios_base::openmode which = std::ios_base::in) override {
        if (dir == std::ios_base::cur)
            gbump(off);
        else if (dir == std::ios_base::end)
            setg(eback(), egptr() + off, egptr());
        else if (dir == std::ios_base::beg)
            setg(eback(), eback() + off, egptr());
        return gptr() - eback();
    }

    virtual pos_type seekpos(pos_type sp, std::ios_base::openmode which) override {
        return seekoff(sp - pos_type(off_type(0)), std::ios_base::beg, which);
    }
};

typedef basic_memstreambuf<char> membuf;

/// @brief Constructor.
AUDPSocket::AUDPSocket(QObject *parent) : AbstractUDPSocket(parent), mBlockTemplate(ABlock()) {
    mName = "APODRONE";
    mNoValid = false;
}

/// @brief Destructor.
AUDPSocket::~AUDPSocket() {}

/**
 * @brief Connect and bind the UDP socket.
 * @param config UDP configuration.
 * @return True on success, false otherwise.
 * @details
 * The method connects the internal socket to the IP and port given in the input configuration.
 *
 * Additionally, it calls AUDPSocket::setListening.
 *
 * If the internal socket is already configured and bound to the parameters of the given configuration, the method
 * only checks for the listening state of the socket and set listening if needed.
 */
bool AUDPSocket::connectAndBind(const QString &ipAddr, quint16 portNum) {
    //    QUdpSocket::SocketState state = mSocket->state();
    //    //  Nothing has changed and the port is already bounded: get out.
    //    if (state == QUdpSocket::ConnectedState || state == QUdpSocket::BoundState) {
    //        bool ret = this->ip().compare(ipAddr, Qt::CaseInsensitive) == 0 && this->port() == portNum;
    //        if (ret) {
    //            if (!isListening())
    //                setListening(true);
    //            return true;
    //        }
    //    }
    // Connect to ip and port (the method will diconnect first).
    return AbstractUDPSocket::connectAndBind(ipAddr, portNum);
}

/// @brief Set a new block template.
void AUDPSocket::setBlockTemplate(const ABlock &blockTemplate) {
    mBlockTemplate = blockTemplate;
    mNoValid = false;
}

void AUDPSocket::resetStatistics() {
    mByteRate.reset();
    mBlockRate.reset();
}

void AUDPSocket::setListening(bool enable) {
    resetStatistics();
    AbstractUDPSocket::setListening(enable);
}

#include <iostream>

/**
 * @brief Read the pending datagrams from the port associated to Amulse instrument.
 * @details
 * The method reads any pending datagrams from the internal udp socket associated to Amulse. For each datagram
 * received, it attempts to convert the raw data in a ABlock instance.
 *
 * On success, the method emits AUDPSocket::receiveABlock signal.
 *
 * If the datagram header does not match the registered descriptor (see AUDPSocket::setBlockTemplate), the signal
 * AUDPSocket::invalidTemplate is emitted \b once. Until the block template is replaced, the method will silently skip
 * all other datagrams.
 */
void AUDPSocket::readPendingDatagrams() {
    static qint64 bc = 0;
    QHostAddress iAddress;
    quint16 iPort;
    QByteArray datagram;
    int btInstr = mBlockTemplate.instrument();
    int btVers = mBlockTemplate.version();
    // std::cout << mMinTimeout <<std::endl;
    if (!mListen || (mMinTimeout > 0 && !mFreqTimer.hasExpired(mMinTimeout))) {
        // std::cout << "trashing block " << (bc++) << std::endl;
        while (mSocket->hasPendingDatagrams()) {
            char a;
            mSocket->readDatagram(&a, 0);
        }
        return;
    }
    // qint64 elapsed = mFreqTimer.restart();
    while (mSocket->hasPendingDatagrams()) {
        bool float32 = false;
        int dsize = mSocket->pendingDatagramSize();
        datagram.resize(dsize);
        mSocket->readDatagram(datagram.data(), dsize, &iAddress, &iPort);
        if (datagram.isEmpty())
            continue;
        QDataStream in(&datagram, QIODevice::ReadOnly);
        in.setByteOrder(QDataStream::BigEndian);
        QString msg = tr("[UDP] Received %1 bytes from %2:%3").arg(dsize).arg(iAddress.toString()).arg(iPort);
        LOG_DEBUG(msg);

        // 1. Check instr+ver from incoming data and auto-detect floating point precision.
        int instr = 0, vers = 0;
        double dv1;
        float fv2[2];
        in.setFloatingPointPrecision(QDataStream::DoublePrecision);
        in >> dv1;
        memcpy(&fv2[0], &dv1, sizeof(double));
        instr = static_cast<int>(dv1);
        if (instr == btInstr) {
            // Let assume we are in float64 mode
            float32 = false;
            in >> dv1;
            vers = static_cast<int>(dv1);
        } else if ((instr = static_cast<int>(fv2[1])) == btInstr) {
            // note fv items indexes are reversed : WE READ BIG ENDIAN !!
            // Let assume we are in float32 mode.
            float32 = true;
            vers = static_cast<int>(fv2[0]);
        } else {
            emit udpStatus(false, dsize, mFreqTimer.elapsed());
            LOG_DEBUG("[UDP] Cannot deduce floating point precision of the incoming block.");
            continue;
        }

        if (!(btInstr == instr && btVers == vers)) {
            emit udpStatus(false, dsize, mFreqTimer.elapsed());
            LOG_DEBUG(QString("[UDP] block header (i: %1, v:%2) do not match template header (i: %3, v:%4)")
                          .arg(instr)
                          .arg(vers)
                          .arg(btInstr)
                          .arg(btVers));
            emit invalidTemplate(instr, vers, btInstr, btVers);
            continue;
        }

        // 2. parse data and emit receiveBlock.
        ABlock b(mBlockTemplate);
        bool ret = false;
        // ugly hack : read data as a vector
        QByteArray dle;
        dle.reserve(datagram.size());
        {
            QDataStream iin(&datagram, QIODevice::ReadOnly);
            in.setByteOrder(QDataStream::BigEndian);
            QDataStream oou(&dle, QIODevice::WriteOnly);
            oou.setByteOrder(QDataStream::LittleEndian);
            if (float32) {
                while (!iin.atEnd()) {
                    double v;
                    iin >> v;
                    oou << v;
                }
            } else {
                while (!iin.atEnd()) {
                    float v;
                    iin >> v;
                    oou << v;
                }
            }
        }
        membuf sbuf(dle.constData(), dle.size());
        std::istream isn(&sbuf);
        ret = b.fillFromDataStream(&isn, float32, false);

        qint64 v = mFreqTimer.elapsed();
        double vv = 1000.0 / v, vvv = vv * (dsize / 1024.0);
        if (ret) {
            mBlockRate.push(vv);
            mByteRate.push(vvv);
        }

        emit udpStatus(ret, dsize, v);
        if (ret)
            emit receiveABlock(b);
        else
            LOG_DEBUG("[UDP] failed to parse incoming block");
    }
    mFreqTimer.restart();
}

void AUDPSocket::getRates(double *block, double *bytes) {
    if (block)
        *block = mBlockRate();
    if (bytes)
        *bytes = mByteRate();
}
