// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "basemapselector.hpp"
#include "MapTypes.h"
#include "clickablelabel.hpp"

using namespace Esri::ArcGISRuntime;

/// @brief Constructor.
BasemapSelector::BasemapSelector(QWidget *parent) : QScrollArea(parent) {
    setupUI(Qt::Horizontal);
    translateUI();
}

/// @brief Constructor.
BasemapSelector::BasemapSelector(Qt::Orientation orientation, QWidget *parent) : QScrollArea(parent) {
    setupUI(orientation);
    translateUI();
}

void BasemapSelector::resetCustomItem() {
    mItems[1].path = "";
    translateCustom();
}

/**
 * @brief Select basemap (radiobutton) using the given basemap identifier.
 * @param id Basemap identifier to search.
 * @param signalNeeded True to emit BasemapSelector::selectionChanged signal. Default to false.
 */
void BasemapSelector::selectId(int id, bool signalNeeded) {
    // find button id:
    int bid = 0;
    for (int i = 0; i < mItems.size(); i++) {
        if (id == mItems.at(i).id) {
            bid = i;
            break;
        }
    }
    if (signalNeeded) {
        mButtonGroup->button(bid)->click();
    } else {
        mButtonGroup->button(bid)->setChecked(true);

        // bool old = mButtonGroup->blockSignals(true);
        //   mButtonGroup->button(bid)->click();
        // mButtonGroup->blockSignals(old);
    }
}

/// @brief Get the index of the given identifier.
/// @note
/// If the identifier is not found, 0 is returned that is the no basemap index.
int BasemapSelector::indexOfId(int id) {
    for (int i = 0; i < mItems.size(); i++) {
        if (id == mItems.at(i).id)
            return i;
    }
    return 0;
}

/// @brief Translate widgets texts.
void BasemapSelector::translateUI() {
    QVector<QString> names{tr("None"), tr("Custom"), tr("Imagery"), tr("Streets"), tr("Topographic"), tr("OSM")};
    for (int i = 0; i < names.size(); i++)
        mLabels[i]->setText(names.at(i));
    translateCustom();
    for (QWidget *w : mBasemapWidgets)
        w->adjustSize();
    mWidget->adjustSize();
}

/// @brief Translate texts related to the custom basemap.
void BasemapSelector::translateCustom() {
    QString path = mItems.at(1).path, msg1 = tr("Click on the label to select a new tile package."),
            msg2 = tr("Current tile package: %1");
    mCustomButton->setStatusTip(msg1);
    if (path.isEmpty()) {
        mLabels[1]->setStatusTip(msg2.arg(tr("None")));
        mCustomButton->setToolTip(msg1);
    } else {
        msg2 = msg2.arg(QFileInfo(path).fileName());
        mLabels[1]->setStatusTip(msg2);
        mCustomButton->setToolTip(QString("<p>%1</p><p>%2</p>").arg(msg2, msg1));
    }
}

/**
 * @brief Setup UI
 * @param orientation Layout orientation.
 * @internal
 * The method also setups the internal items:
 * - If one add a new basemap, it should also be managed in MapWidget::onBasemapSelectionChanged.
 * - One should not use vector based basemap, it requires too heavy calculations at initialization for NRT app.
 */
void BasemapSelector::setupUI(Qt::Orientation orientation) {
    QVector<QLabel *> mLabelNames;
    QVector<QPixmap> pixmaps{
        QPixmap(":/adm2/images/basemaps/none.png"),        QPixmap(":/adm2/images/basemaps/custom.png"),
        QPixmap(":/adm2/images/basemaps/imagery.jpg"),     QPixmap(":/adm2/images/basemaps/streets.jpg"),
        QPixmap(":/adm2/images/basemaps/topographic.jpg"), QPixmap(":/adm2/images/basemaps/osm.jpg")};
    QVector<QString> names{tr("None"), tr("Custom"), tr("Imagery"), tr("Streets"), tr("Topographic"), tr("OSM")};
    mItems = {
        {-2, ""},
        {-1, ""},
        {static_cast<int>(BasemapStyle::ArcGISImagery), ""},
        {static_cast<int>(BasemapStyle::ArcGISStreets), ""},
        {static_cast<int>(BasemapStyle::ArcGISStreetsRelief), ""},
        {static_cast<int>(BasemapStyle::OsmStandard), ""},
    };
    QMargins margins(2, 0, 2, 0);
    QBoxLayout::Direction direction = QBoxLayout::LeftToRight;
    Qt::Orientation sepOrient = Qt::Vertical;
    if (orientation == Qt::Vertical) {
        margins = QMargins(0, 2, 0, 2);
        sepOrient = Qt::Horizontal;
        direction = QBoxLayout::TopToBottom;
        setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
        setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    } else {
        setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
        setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
    }

    mWidget = new QWidget(this);
    mLayout = new QBoxLayout(direction, mWidget);

    mWidget->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Ignored);
    mButtonGroup = new QButtonGroup(this);

    for (int i = 0; i < mItems.size(); i++) {

        QPixmap pixmap = pixmaps.at(i).scaledToWidth(150);
        const QString &name = names.at(i);
        const BasemapInfos &item = mItems.at(i);

        QWidget *bw = new QWidget(mWidget);
        QRadioButton *radioButton = new QRadioButton(bw);
        QLabel *label;

        radioButton->setIconSize(pixmap.size());
        radioButton->setIcon(QIcon(pixmap));

        if (item.id == -1) {
            mCustomLabel = new ClickableLabel(bw);
            mCustomLabel->setHighlight(true);
            label = mCustomLabel;
            mCustomButton = radioButton;
        } else {
            label = new QLabel(bw);
        }
        label->setAlignment(Qt::AlignCenter);
        label->setText(name);

        mButtonGroup->addButton(radioButton, i);
        mLabels.push_back(label);

        QVBoxLayout *vLayout = new QVBoxLayout(bw);
        vLayout->setContentsMargins(0, 0, 0, 0);
        vLayout->addWidget(radioButton, 1);
        vLayout->addWidget(label, 0);

        mLayout->addWidget(bw, 0);
        if (item.id < 0)
            mLayout->addWidget(getSeparator(mWidget, sepOrient), 0);
        bw->adjustSize();
        mBasemapWidgets.push_back(bw);
    }

    mLayout->setContentsMargins(margins);
    mLayout->addStretch(1);
    mLastId = 0;
    mButtonGroup->button(mLastId)->setChecked(true);
    QObject::connect(mButtonGroup, &QButtonGroup::idClicked, this, &BasemapSelector::onButtonChecked);
    QObject::connect(mCustomLabel, &ClickableLabel::clicked, this, &BasemapSelector::updateCustomFilePath);
    setWidget(mWidget);
    mWidget->adjustSize();
    adjustSize();
}

/**
 * @brief Get the current filepath associated to the custom basemap.
 */
QString BasemapSelector::customFilePath() const { return mItems.at(1).path; }

/// @brief Get the selected basemap informations.
const BasemapInfos &BasemapSelector::selectedBasemap() {
    int id = mButtonGroup->checkedId();
    if (id == -1)
        id = 0;
    return mItems.at(id);
}

/// @brief Get the selected basemap identifier.
int BasemapSelector::selectedId() {
    int id = mButtonGroup->checkedId();
    if (id == -1)
        id = 0;
    return mItems.at(id).id;
}

/// @brief Manage radiobuttons clicks.
void BasemapSelector::onButtonChecked(int buttonId) {
    // really needed ?
    if (buttonId == -1)
        return;
    BasemapInfos &item = mItems[buttonId];
    // one tries to click on the custom basemap button but there is no file registered yet.
    if (item.id == -1 && item.path.isEmpty()) {
        if (!updateCustomFilePath()) {
            // failed to update the file path: re-select the last button.
            mButtonGroup->button(mLastId == buttonId ? 0 : mLastId)->click();
        } else {
            mLastId = buttonId;
            translateCustom();
            return;
        }
    } else {
        mLastId = buttonId;
        translateCustom();
        emit selectionChanged(item);
    }
}

/**
 * @brief Update the custom basemap file path.
 * @return True if a new file path has been registered, false otherwise.
 * @details
 * The method shows a file dialog to let the user choose a new tile package file. Upon success, the internal
 * file path is updated and BasemapSelector::selectionChanged is emitted if the checked radio button is associated
 * to the custom basemap.
 */
bool BasemapSelector::updateCustomFilePath() {
    BasemapInfos &item = mItems[1];
    QSettings settings;
    QString lastTpkDir = settings.value("map/tpkdir", gDataDir).toString();
    QString fileName =
        QFileDialog::getOpenFileName(this, tr("Select tile package"), lastTpkDir, tr("Tile package files (*.tpk)"));

    if (!fileName.isEmpty()) {
        QFileInfo finfo(fileName);
        settings.setValue("map/tpkdir", finfo.absolutePath());
        item.path = finfo.absoluteFilePath();
        if (mButtonGroup->checkedId() == 1) // the current checked item is the custom basemap: update !
            emit selectionChanged(item);
        return true;
    }
    return false;
}
