// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "basertplotwidget.hpp"

#include "baseplot.hpp"
#include "lineposition.hpp"
#include "menusection.hpp"
#include "qcp_helpers.hpp"

BaseRtPlotWidget::BaseRtPlotWidget(QWidget *parent) : QWidget(parent) {
    mWithXLabel = true;
    mTitlePrefix = tr("Graph");
    mName = tr("None");

    mReplayMode = false;
    mFreeMode = false;    // free mode NOT enabled by default.
    mTimeWindow = 5 * 60; // 5 minutes of time window by default.
    mTimeOffset = 0.100;  // 0.100 secs extra time offset.
    mPlot = new BasePlot(this);
    mPlot->setAutoAddPlottableToLegend(false);

    mPlot->setContextMenuPolicy(Qt::CustomContextMenu);
    mPlot->setMinimumHeight(60);
    mPlot->setAutoAddPlottableToLegend(false);
    mPlot->setDateTimeFormat("HH:mm:ss");
    mPlot->setAxisTicker(QCPAxis::atBottom | QCPAxis::atTop, BasePlot::DATE);
    mPlot->setDateSpec(Qt::UTC); // TOTAL people want UTC datetime labels !
    mPlot->setXLabel(tr("Time (UTC)"));
    // synchronize vertical axes:
    for (int i = 0; i < mPlot->axisRectCount(); i++) {
        QObject::connect(mPlot->axisRect(i)->axis(QCPAxis::atLeft),
                         static_cast<void (QCPAxis::*)(const QCPRange &)>(&QCPAxis::rangeChanged),
                         mPlot->axisRect(i)->axis(QCPAxis::atRight),
                         static_cast<void (QCPAxis::*)(const QCPRange &)>(&QCPAxis::setRange));
    }

    mLastValueBox = new TextBox(mPlot->axisRect());
    mLastValueBox->setVisible(true);
    mLastValueBox->setText("N/A");
    mLastValueBox->setPen(QPen(Qt::black));

    mReplayTracer = new ItemTracer(mPlot);
    mReplayTracer->setInterpolating(false);
    mReplayTracer->setSize(14);
    mReplayTracer->setPen(QPen(Qt::red));
    mReplayTracer->setVisible(false);

    // add lines below main plottables in the plot
    mPlot->addLayer("lines", mPlot->layer("main"), QCustomPlot::limBelow);
    mPlot->setCurrentLayer("lines");
    mUpperLine = new LinePosition(mPlot->axisRect(), 0, Qt::Horizontal);
    mLowerLine = new LinePosition(mPlot->axisRect(), 0, Qt::Horizontal);
    mThresholdLine = new LinePosition(mPlot->axisRect(), 0, Qt::Horizontal);
    mPlot->setCurrentLayer("main");

    mThLineVisible = false;
    for (LinePosition *lp : {mUpperLine, mLowerLine, mThresholdLine}) {
        lp->setVisible(false);
        lp->setFixed(true);
        lp->boundToData(true);
    }

    mToolBar = new QToolBar(tr("Graph Actions"), this);

    setupActions();

    QVBoxLayout *vLayout = new QVBoxLayout(this);
    vLayout->setContentsMargins(4, 4, 4, 4);
    vLayout->setSpacing(4);
    vLayout->setMenuBar(mToolBar);
    vLayout->addWidget(mPlot, 2);

    QObject::connect(mActionFreeMode, &QAction::triggered, this, &BaseRtPlotWidget::setFreeMode);
    QObject::connect(mActionFreeMode, &QAction::triggered, mActionRescale, &QAction::setEnabled);
    QObject::connect(mActionRescale, &QAction::triggered, mPlot, &BasePlot::autoRescale);
    QObject::connect(mActionYRescale, &QAction::triggered, this, &BaseRtPlotWidget::rescaleYAxis);
    QObject::connect(mActionXRescale, &QAction::triggered, mPlot, &BasePlot::autoRescaleX);

    setFreeMode(false);

    translateUI();

    mRgTimer.setSingleShot(true);
    QObject::connect(&mRgTimer, &QTimer::timeout, this, &BaseRtPlotWidget::emitRangeChange);
    mThTimer.setSingleShot(true);
    QObject::connect(&mThTimer, &QTimer::timeout, this, &BaseRtPlotWidget::emitThresholdChange);
}

/// @brief Set new text in the textbox.
void BaseRtPlotWidget::setTextBoxText(const QString &text) {
    mLastValueBox->setText(text);
    mPlot->replot(QCustomPlot::rpQueuedReplot);
}

/// @brief set the position of the text box.
void BaseRtPlotWidget::setTextBoxPosition(const TextBox::Position &position) {
    mLastValueBox->setPositionFlag(position);
}

/// brief Get the internal plot instance.
BasePlot *BaseRtPlotWidget::plot() const { return mPlot; }

/// @brief Create toolbar and action of the widget.
void BaseRtPlotWidget::setupActions() {
    mActionFreeMode = new QAction(QIcon::fromTheme("free-mode"), tr("Enable/disable free-mode"), this);
    mActionFreeMode->setCheckable(true);
    mActionFreeMode->setChecked(false);

    mActionYRescale = new QAction(QIcon::fromTheme("fit-height"), tr("Y-axis auto-scale"), this);
    mActionXRescale = new QAction(QIcon::fromTheme("fit-width"), tr("X-axis auto-scale"), this);
    mActionXRescale->setEnabled(false);
    mActionRescale = new QAction(QIcon::fromTheme("fit-both"), tr("Auto-scale"), this);
    mActionRescale->setEnabled(false);

    mToolBar->addAction(mActionFreeMode);
    mToolBar->addSeparator();
    mToolBar->addAction(mActionYRescale);
    mToolBar->addAction(mActionXRescale);
    mToolBar->addAction(mActionRescale);

    QWidget *spacer = new QWidget();
    spacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    mToolBar->addWidget(spacer);
    mMousePositionLabel = new QLabel(this);
    mToolBar->addWidget(mMousePositionLabel);
    mMousePositionLabel->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
    mToolBar->setContentsMargins(0, 0, 0, 0);

    mToolBar->setIconSize(TB_ICO_SIZE);
}

/**
 * @brief On mouse leave actions management.
 * @param event Associated event.
 * @details
 * The method simply clears the mouse position label text. this is a naive way to cleanup the data displayed when the
 * mouse leaves the widget.
 */
void BaseRtPlotWidget::leaveEvent(QEvent * /*event*/) { mMousePositionLabel->setText(""); }

/// @brief Translate widgets displayed text.
void BaseRtPlotWidget::translateUI() {
    mPlot->setXLabel(mWithXLabel ? tr("Time (UTC)") : "");

    mActionFreeMode->setText(mFreeMode ? tr("Disable free-mode") : tr("Enable free-mode"));
    mActionYRescale->setText(tr("Y-axis auto-scale"));
    mActionXRescale->setText(tr("X-axis auto-scale"));
    mActionRescale->setText(tr("Auto-rescale"));

    mActionFreeMode->setToolTip(tr("Enable/disable free interactions mode"));
    mActionYRescale->setToolTip(tr("Rescale Y-axis to fit whole data"));
    mActionXRescale->setToolTip(tr("Rescale X-axis to fit whole data"));
    mActionRescale->setToolTip(tr("Rescale both axes to fit whole data"));
    for (QAction *action : findChildren<QAction *>(QString(), Qt::FindDirectChildrenOnly))
        action->setStatusTip(action->toolTip());
}

/// @brief Check whether or not free mode is enabled.
bool BaseRtPlotWidget::freeMode() const { return mFreeMode; }

/// @brief Enable/disable free mode.
void BaseRtPlotWidget::setFreeMode(bool enabled) {
    mFreeMode = enabled;
    if (mFreeMode) {
        mActionFreeMode->setText(tr("Disable free-mode"));
        mPlot->setDragOrientations(Qt::Horizontal | Qt::Vertical);
        mPlot->setZoomOrientations(Qt::Horizontal | Qt::Vertical);
    } else {
        mActionFreeMode->setText(tr("Enable free-mode"));
        mPlot->setDragOrientations(Qt::Vertical);
        mPlot->setZoomOrientations(Qt::Vertical);
    }
    mActionFreeMode->setChecked(mFreeMode);
    mActionXRescale->setEnabled(mFreeMode);
    mActionRescale->setEnabled(mFreeMode);

    mPlot->replot();
}

/**
 * @brief Set the time window.
 * @param seconds Time window in seconds to set.
 * @details
 * The method sets/updates the time window to display on the X-axis of the plot in scrolling mode.
 *
 * The minimum time window is 60 seconds.
 */
void BaseRtPlotWidget::setTimeWindow(int seconds) {
    mTimeWindow = qMax(60, seconds);
    // The value will be taken into account on the next (auto) scroll.
}
/// @brief Get the current time window in seconds.
int BaseRtPlotWidget::timeWindow() const { return mTimeWindow; }

/**
 * @brief Set the extra offset to add to the rightmost value displayed by the X-axis.
 * @param milliseconds Time offset in milliseconds to set.
 *
 * The minimum offset is 50 milliseconds.
 */
void BaseRtPlotWidget::setTimeOffset(int milliseconds) {
    mTimeOffset = qMax(50, milliseconds) / 1000.0;
    // The value will be taken into account on the next (auto) scroll.
}

/// @brief Get the time offset (in milliseconds).
int BaseRtPlotWidget::timeOffset() const { return mTimeOffset * 1000; }

/**
 * @brief Show/Hide the toolbar.
 * @details
 * By default the toolbar is visible.
 *
 * Note that all actions of the toolbar are accessible from the context menu of the plot. When the toolbar is hidden,
 * only the information about the data coordinates under the mouse position are lost.
 */
void BaseRtPlotWidget::setToolbarVisible(bool visible) { mToolBar->setVisible(visible); }

/// @brief Check whether or not the toolbar is visible.
bool BaseRtPlotWidget::toolbarVisible() const { return mToolBar->isVisible(); }

/// @brief Set visibility of the X-axis label.
void BaseRtPlotWidget::setXLabelVisible(bool visible) {
    mWithXLabel = visible;
    mPlot->setXLabel(mWithXLabel ? tr("Time (UTC)") : "");
}

/// @brief Check whether or not the X-axis label is visible.
bool BaseRtPlotWidget::xLabelVisible() const { return mWithXLabel; }

/// @brief Set Y-axis label.
void BaseRtPlotWidget::setYLabel(const QString &label) { mPlot->setYLabel(label); }

/**
 * @brief Generate a new graph in the given AxisRect.
 * @return A new QCPGraph instance.
 * @details
 * The method builds a graph and sets line and marker style for the new graph according to application settings.
 */
QCPGraph *BaseRtPlotWidget::buildGraph() {

    QCPGraph *g = mPlot->addGraph();
    if (!g)
        return g;

    QSettings settings;
    settings.beginGroup("graph");
    bool wLine = settings.value("wline", true).toBool();
    double lw = settings.value("lwidth", 1.0).toDouble();
    QCPScatterStyle::ScatterShape shape = Plotting::getScatterShape(settings.value("mstyle", 0).toInt());
    double ms = settings.value("msize", 6.0).toDouble();
    settings.endGroup();

    QPen pen(Qt::black);
    pen.setWidthF(lw);
    QCPScatterStyle symbol(shape, ms);
    g->setPen(pen);
    g->selectionDecorator()->setPen(pen);
    g->setScatterStyle(symbol);
    g->setLineStyle(wLine ? QCPGraph::lsLine : QCPGraph::lsNone);
    return g;
}

/**
 * @brief Update the style of all graph using current application settings.
 */
void BaseRtPlotWidget::updateFromSettings() {
    QSettings settings;
    settings.beginGroup("graph");
    bool htb = settings.value("hide_toolbar", false).toBool();
    bool hxl = settings.value("hide_xlabel", false).toBool();
    setToolbarVisible(!htb);
    setXLabelVisible(!hxl);

    mLastValueBox->setPositionFlag(static_cast<TextBox::Position>(settings.value("box_pos", 2).toInt()));
    setTimeWindow(settings.value("wtime", 300).toInt());   // secs
    setTimeOffset(settings.value("woffset", 100).toInt()); // msecs

    bool wLine = settings.value("wline", true).toBool();
    double lw = settings.value("lwidth", 1.0).toDouble();
    QCPScatterStyle::ScatterShape shape = Plotting::getScatterShape(settings.value("mstyle", 0).toInt());
    double ms = settings.value("msize", 6.0).toDouble();
    settings.endGroup();
    QPen pen(Qt::black);
    pen.setWidthF(lw);
    QCPScatterStyle symbol(shape, ms);
    for (int i = 0; i < mPlot->graphCount(); i++) {
        QCPGraph *g = mPlot->graph(i);
        g->setPen(pen);
        g->selectionDecorator()->setPen(pen);
        g->setScatterStyle(symbol);
        g->setLineStyle(wLine ? QCPGraph::lsLine : QCPGraph::lsNone);
    }
    mPlot->replot();
}

/// @brief For ce the time (x-axis) range.
void BaseRtPlotWidget::forceTimeRange(const QCPRange &range) {
    mPlot->xAxis->setRange(range);
    mPlot->replot();
}

/**
 * @brief Scroll the x-axis range to the given time reference.
 * @param timeReference Time reference (seconds since Epoch) of the right edge of the range.
 * @details
 * The method updates the x-axis range of the plot so it displays an interval of BaseRTPlot::timeWindow that ends at \a
 * timeReference plus BaseRTPlot::timeOffset.
 *
 * Then it schedules a replot with QCustomPlot::rpQueuedReplot priority.
 *
 * @note
 * The method only works if the widget is not free mode.
 * @attention
 * The method does not check if reference is earlier than the current range upper bound.
 */
void BaseRtPlotWidget::scrollXRange(double reference) {
    if (!mFreeMode) {
        mPlot->axisRect(0)->axis(QCPAxis::atBottom)->setRange(reference + mTimeOffset, mTimeWindow, Qt::AlignRight);
        mPlot->replot(QCustomPlot::rpQueuedReplot);
    }
}

/**
 * @brief Add the user interactions actions of the plot widget to the given menu.
 * @param menu Pointer to the menu to update.
 * @details
 * If menu is null, nothing is done. Otherwise the method appends the following actions:
 *
 * - A custom separator with the key-word Interaction as title.
 * - The free-mode switch action.
 * - A custom separator with the key-word Scaling as title.
 * - The Y-axis rescale action.
 * - The X-axis rescale action.
 * - The auto rescale action.
 *
 * The two last actions may be disabled if the plot is not in free-mode.
 */
void BaseRtPlotWidget::addInteractionsActions(QMenu *menu) {
    if (menu) {
        MenuSection *section;
        if (!mReplayMode) {
            section = new MenuSection(tr("Interaction"), menu);
            menu->addAction(section);
            menu->addAction(mActionFreeMode);
        }
        section = new MenuSection(tr("Scaling"), menu);
        menu->addAction(section);
        menu->addAction(mActionYRescale);
        menu->addAction(mActionXRescale);
        menu->addAction(mActionRescale);
    }
}

/// @brief Get the mouse position label (on the toolbar).
QLabel *BaseRtPlotWidget::mousePositionLabel() const { return mMousePositionLabel; }

/**
 * @brief Refresh the plot.
 * @details
 * The base implementation on schedules a replot.
 *
 * @note
 * Such method should be called anytime a new block of data has been processed in the data manager.
 */
void BaseRtPlotWidget::refreshPlot() {
    if (!mFreeMode) {
        QCPGraph *g = mPlot->graph();
        if (g)
            g->rescaleValueAxis(true, true);
    }
    mPlot->replot(QCustomPlot::rpQueuedReplot);
}

/**
 * @brief Rescale Y-axis of the plot to fit all data.
 * @details
 * If the axis range has changed, it is updated a second time with an offset corresponding to 1% of the new range.
 *
 * This operation is only cosmetic and forces the graphs to be within the axis-rect.
 */
void BaseRtPlotWidget::rescaleYAxis() {
    QCPAxis *yaxis = mPlot->axisRect()->axis(QCPAxis::atLeft);
    QCPRange oldRange = yaxis->range();
    yaxis->rescale(true);
    QCPRange newRange = yaxis->range();
    if (oldRange != newRange) {
        double offset = newRange.size() / 100.0;
        newRange = QCPRange(newRange.lower - offset, newRange.upper + offset);
        yaxis->setRange(newRange);
        mPlot->replot(QCustomPlot::rpQueuedReplot);
    }
}

/// @brief Get the window title prefix.
QString BaseRtPlotWidget::titlePrefix() const { return mTitlePrefix; }

/// @brief Set the window title prefix.
void BaseRtPlotWidget::setTitlePrefix(const QString &titlePrefix) {
    mTitlePrefix = titlePrefix;
    setWindowTitle(QString("%1 - %2").arg(mTitlePrefix, mName));
}

/// @brief Get the widget parameter name.
QString BaseRtPlotWidget::name() const { return mName; }

/// @brief Set the parameter name of the widget.
void BaseRtPlotWidget::setName(const QString &name) {
    mName = name;
    setWindowTitle(QString("%1 - %2").arg(mTitlePrefix, mName));
}

/**
 * @brief Clear the plot.
 * @details
 * The method removes all graphs from the internal plot, clears the Y-axis label and resets the Y-axis range to [0,1].
 */
void BaseRtPlotWidget::clearPlot() {
    mReplayTracer->setGraph(nullptr);
    mReplayTracer->setVisible(false);
    mPlot->clearGraphs();
    mPlot->setYLabel("");
    mPlot->setYRange(QCPRange(0, 1));
    setName(tr("None"));
}

/**
 * @brief Add a new graph to the plot.
 * @param data Data to set in the graph.
 */
void BaseRtPlotWidget::addGraph(GraphContainerPointer data) {
    if (data.isNull())
        return;
    QCPGraph *g = buildGraph();
    g->setData(data);
    mReplayTracer->setGraph(g);
}

/**
 * @brief Remove any graphs that make use of the given data container.
 * @param data GraphDataContainer pointer to process.
 */
void BaseRtPlotWidget::removeGraph(GraphContainerPointer data) {
    if (data.isNull())
        return;
    QVector<QCPGraph *> graphs;
    for (int i = 0; i < mPlot->graphCount(); ++i) {
        QCPGraph *g = mPlot->graph(i);
        if (g->data() == data)
            graphs.push_back(g);
    }
    QCPGraph *gtracer = mReplayTracer->graph();
    for (QCPGraph *g : graphs) {
        if (gtracer == g) {
            mReplayTracer->setGraph(nullptr);
            mReplayTracer->setVisible(false);
        }
        mPlot->removeGraph(g);
    }
}

/**
 * @brief Set new flight graphs
 * @param data Vector of graph data container to set in the graphs.
 * @param item Optional pointer to a DataItem
 */
void BaseRtPlotWidget::setGraphs(GraphContainerVector data, bool replot) {
    setUpdatesEnabled(false);
    mReplayTracer->setGraph(nullptr);
    int gc = mPlot->graphCount(), dc = data.size();
    while (gc > dc) {
        mPlot->removeGraph(gc - 1);
        gc = mPlot->graphCount();
    }
    while (gc < dc) {
        buildGraph();
        gc = mPlot->graphCount();
    }
    for (int i = 0; i < dc; i++)
        mPlot->graph(i)->setData(data[i]);

    if (gc)
        mReplayTracer->setGraph(mPlot->graph());

    setUpdatesEnabled(true);
    if (!mFreeMode)
        rescaleYAxis(); // will enlarge the y-axis; if not wanted use mPlot->autoRescaleY().
    if (replot)
        mPlot->replot();
}

/// @brief Check whether or not replay mode is enabled.
bool BaseRtPlotWidget::replayMode() const { return mReplayMode; }

/**
 * @brief Enable/disable replay mode.
 * @param replayMode True to enable, false otherwise.
 * @details
 * In replay mode, the free mode is forced and the user cannot interacts with the afferent action on the widget.
 */
void BaseRtPlotWidget::setReplayMode(bool replayMode) {
    mReplayMode = replayMode;
    setFreeMode(mReplayMode);
    mActionFreeMode->setEnabled(!mReplayMode);
}

/// @brief Enable/Disable replay tracer.
void BaseRtPlotWidget::setReplayTracer(bool enable, double *key) {
    if (enable) {
        mReplayTracer->setGraph(mPlot->graph());
        mReplayTracer->setVisible(true);
        if (key)
            mReplayTracer->setGraphKey(*key);
    } else {
        mReplayTracer->setGraph(nullptr);
        mReplayTracer->setVisible(false);
    }
}

/**
 * @brief Set the replay tracer to the given time key.
 * @param key Time key to set.
 * @details
 * If \a key is negative, the tracer is hidden otherwise, it is set visible and its position will be updated to the
 * given time key upon the next draw (assuming there is a graph associated to the tracer).
 */
void BaseRtPlotWidget::setReplayKey(double key) {
    mReplayTracer->setVisible(key > 0);
    mReplayTracer->setGraphKey(key);
}

/// @brief Set color range lines visibility.
void BaseRtPlotWidget::setRgLinesVisible(bool visible) {
    mUpperLine->setVisible(visible);
    mLowerLine->setVisible(visible);
}

/// @brief Check whether or not Range lines are visible.
bool BaseRtPlotWidget::rgLinesVisible() const { return mUpperLine->visible(); }

/**
 * @brief Set new positions for color range lines.
 * @param range Data range with the new values to set for the lower and upper lines.
 * @note
 * The method does not emit BaseRtPlotWidget::colorRangeChanged.
 */
void BaseRtPlotWidget::setRgLinesValues(const DataRange &range) {
    mUpperLine->setPosition(range.upper());
    mLowerLine->setPosition(range.lower());
    mPlot->replot(QCustomPlot::rpQueuedRefresh);
}

/**
 * @brief Enable/Disable user interactions on the range lines.
 * @param enable True to enable user interactions.
 * @details
 * If enabled, the user may move the lines and setup a new data range.
 */
void BaseRtPlotWidget::setRgLinesEditable(bool enable) {
    if (enable == rgLinesEditable())
        return;
    mLowerLine->setFixed(!enable);
    mUpperLine->setFixed(!enable);
    QObject::disconnect(mUpperLine, &LinePosition::positionChanged, this, &BaseRtPlotWidget::onRgLineChanged);
    QObject::disconnect(mLowerLine, &LinePosition::positionChanged, this, &BaseRtPlotWidget::onRgLineChanged);
    if (enable) {
        QObject::disconnect(mUpperLine, &LinePosition::positionChanged, this, &BaseRtPlotWidget::onRgLineChanged);
        QObject::disconnect(mLowerLine, &LinePosition::positionChanged, this, &BaseRtPlotWidget::onRgLineChanged);
    }
}

/// @brief Check whether or not the range lines are user-editable.
bool BaseRtPlotWidget::rgLinesEditable() const { return !mUpperLine->isFixed(); }

/// @brief Postpone emission of new color range on line position changes.
void BaseRtPlotWidget::onRgLineChanged(double) { mRgTimer.start(50); }

/// @brief Compute and emit new color range.
void BaseRtPlotWidget::emitRangeChange() {
    double a = mLowerLine->position(), b = mUpperLine->position();
    if (b < a)
        qSwap(b, a);
    else if (a == b)
        b *= 1.0000001;
    emit colorRangeChanged(DataRange(a, b));
}

/// @brief Set color threshold line visibility.
void BaseRtPlotWidget::setThLineVisible(bool visible) {
    mThLineVisible = visible;
    mThresholdLine->setVisible(visible);
}

/// @brief Check whether or not threshold line is visible.
bool BaseRtPlotWidget::thLineVisible() const { return mThLineVisible; }

/**
 * @brief Set new position for the threshold line.
 * @param value New value to set.
 * @note
 * The method does not emit BaseRtPlotWidget::thresholdChanged.
 */
void BaseRtPlotWidget::setThLineValue(double value) {
    if (qIsNaN(value)) {
        mThresholdLine->setVisible(false);
    } else {
        mThresholdLine->setPosition(value);
        mThresholdLine->setVisible(mThLineVisible);
    }
    mPlot->replot(QCustomPlot::rpQueuedRefresh);
}

/**
 * @brief Enable/Disable user interactions on the threshold line.
 * @param enable True to enable user interactions.
 * @details
 * If enabled, the user may move the line and setup a new threshold.
 */
void BaseRtPlotWidget::setThLineEditable(bool enable) {
    if (enable == thLineEditable())
        return;
    mThresholdLine->setFixed(!enable);
    QObject::disconnect(mThresholdLine, &LinePosition::positionChanged, this, &BaseRtPlotWidget::thresholdChanged);
    if (enable)
        QObject::disconnect(mThresholdLine, &LinePosition::positionChanged, this, &BaseRtPlotWidget::thresholdChanged);
}

/// @brief Check whether or not the range lines are user-editable.
bool BaseRtPlotWidget::thLineEditable() const { return !mThresholdLine->isFixed(); }

/// @brief Postpone emission of new threshold value on line position changes.
void BaseRtPlotWidget::onThLineChanged(double) { mThTimer.start(50); }

/// @brief Compute and emit new color range on line position changes.
void BaseRtPlotWidget::emitThresholdChange() { emit thresholdChanged(mThresholdLine->position()); }

/// @brief Enable/disable filling between the color range lines.
void BaseRtPlotWidget::fillRgLines(bool enabled) { mLowerLine->setChannelFillLine(enabled ? mUpperLine : nullptr); }

/// @brief Set brush for color range lines filling.
void BaseRtPlotWidget::setRgLinesFillBrush(QBrush brush) { mLowerLine->setBrush(brush); }

/// @brief Set pen for threshold line.
void BaseRtPlotWidget::setThLinePen(QPen pen) { mThresholdLine->setPen(pen); }

/// @brief Set pen for color range lines.
void BaseRtPlotWidget::setRgLinesPen(QPen pen) {
    mLowerLine->setPen(pen);
    mUpperLine->setPen(pen);
}
