// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "cpumemwidget.hpp"
#include "cpumemusage.hpp"

CpuMemWidget::CpuMemWidget(QWidget *parent) : QWidget(parent) {

    mCpuVisible = true;
    mRamVisible = true;
    mRioVisible = true;
    mWioVisible = true;

    mTotalRAM = totalMemory();
    mFmt = static_cast<int>(std::floor(std::log10(mTotalRAM))) + 1;

    mLblCPUd = new QLabel("CPU", this);
    mLblRAMd = new QLabel("RAM", this);
    mLblRIOd = new QLabel("I/O(R)", this);
    mLblWIOd = new QLabel("I/O(W)", this);
    mLblCPU = new QLabel(this);
    mLblRAM = new QLabel(this);
    mLblRIO = new QLabel(this);
    mLblWIO = new QLabel(this);

    mLayout = new QFormLayout(this);
    mLayout->addRow(mLblCPUd, mLblCPU);
    mLayout->addRow(mLblRAMd, mLblRAM);
    mLayout->addRow(mLblRIOd, mLblRIO);
    mLayout->addRow(mLblWIOd, mLblWIO);

    for (QLabel *lbl : {mLblCPU, mLblRAM, mLblRIO, mLblWIO})
        lbl->setAlignment(Qt::AlignRight);

    mLblCPU->setText("N/A");
    mLblRAM->setText("N/A");
    mLblRIO->setText("N/A");
    mLblWIO->setText("N/A");

    mSysTimer = new QTimer(this);
    QObject::connect(mSysTimer, &QTimer::timeout, this, &CpuMemWidget::onTimeout);
    setTimeRate(250);
}

/// @brief Get the size hint of the widget.
QSize CpuMemWidget::sizeHint() const {
    QSize size = minimumSize();
    size.rwidth() += 20;
    return size;
}

/// @brief Stop update timer on hide.
void CpuMemWidget::hideEvent(QHideEvent *event) {
    mSysTimer->stop();
    QWidget::hideEvent(event);
}

/// @brief Start update timer on hide.
void CpuMemWidget::showEvent(QShowEvent *event) {
    mSysTimer->start(mTimeRate);
    QWidget::showEvent(event);
}

/// @brief Set a new time rate (in milliseconds) for the updates of cpu/ram informations.
void CpuMemWidget::setTimeRate(int timeRate) {
    mTimeRate = qMax(200, timeRate);
    mSysTimer->setInterval(mTimeRate);
}

/// @brief Translate widgets texts.
void CpuMemWidget::translateUI() {
    setWindowTitle(tr("System resources usage"));
    mLblCPUd->setText(tr("CPU"));
    mLblRAMd->setText(tr("RAM"));
}

/// @brief Compute CPU/RAM statistics on time out.
void CpuMemWidget::onTimeout() {
    int *mem = nullptr;
    double *cpu = nullptr, *rio = nullptr, *wio = nullptr;
    if (mRamVisible)
        mem = new int;
    if (mCpuVisible)
        cpu = new double;
    if (mRioVisible)
        rio = new double;
    if (mWioVisible)
        wio = new double;
    getStatistics(cpu, mem, rio, wio);
    if (mem) {
        mLblRAM->setText(QString("%1 MB (%2%)").arg((*mem), mFmt).arg(static_cast<int>((*mem) * 100.0 / mTotalRAM), 3));
        delete mem;
    }
    if (cpu) {
        mLblCPU->setText(QString("%1%").arg(*cpu, 5, 'f', 1));
        delete cpu;
    }
    if (wio) {
        mLblWIO->setText(QString("%1 KB/s").arg(*wio, 0, 'f', 1));
        delete wio;
    }
    if (rio) {
        mLblRIO->setText(QString("%1 KB/s").arg(*rio, 0, 'f', 1));
        delete rio;
    }
}

/// @brief Show/Hide the IO write counter label.
void CpuMemWidget::setWioVisible(bool visible) {
    mLblWIO->setVisible(visible);
    mLblWIO->setVisible(visible);
    mWioVisible = visible;
}
/// @brief Show/Hide the IO read counter label.
void CpuMemWidget::setRioVisible(bool visible) {
    mLblRIO->setVisible(visible);
    mLblRIO->setVisible(visible);
    mRioVisible = visible;
}
/// @brief Show/Hide the RAM counter label.
void CpuMemWidget::setRamVisible(bool visible) {
    mLblRAM->setVisible(visible);
    mLblRAM->setVisible(visible);
    mRamVisible = visible;
}
/// @brief Show/Hide the CPU counter label.
void CpuMemWidget::setCpuVisible(bool visible) {
    mLblCPU->setVisible(visible);
    mLblCPU->setVisible(visible);
    mCpuVisible = visible;
}

/// @brief Check whether or not the IO write counter label is visible.
bool CpuMemWidget::wioVisible() const { return mWioVisible; }

/// @brief Check whether or not the IO read counter label is visible.
bool CpuMemWidget::rioVisible() const { return mRioVisible; }

/// @brief Check whether or not the RAM counter label is visible.
bool CpuMemWidget::ramVisible() const { return mRamVisible; }

/// @brief Check whether or not the CPU counter label is visible.
bool CpuMemWidget::cpuVisible() const { return mCpuVisible; }
