// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "dataitem.hpp"

/**
 * @brief Constructor
 * @param param Parameter informations.
 * @param alias Parameter alias.
 * @param colorRange Color range to set.
 * @param threshold Data visibility threshold to set.
 * @param parent Optional QObject parent.
 */
DataItem::DataItem(const AParam &param, const QString &alias, const DataRange &colorRange, double threshold,
                   QObject *parent)
    : QObject(parent), mParam(param), mAlias(alias), mColorRange(colorRange), mThreshold(threshold),
      mSaveColorRange(colorRange), mSaveThreshold(threshold) {}

/// @brief Destructor
DataItem::~DataItem() {}

/// @brief Update the @a saved threshold with the current value.
void DataItem::updateSaveThreshold() { mSaveThreshold = mThreshold; }

/// @brief Update the @a saved color range with the current value.
void DataItem::updateSaveRange() { mSaveColorRange = mColorRange; }

/**
 * @brief Get the @a saved threshold.
 * @details
 * The @a saved threshold is part of the mechanism to restore the initial parameters of the DataItem.
 * It is first set in the constructor from the afferent argument and can only be @a updated with the
 * value of the current threshold using DataItem::updateSaveThreshold.
 */
double DataItem::saveThreshold() const { return mSaveThreshold; }

/**
 * @brief Get the @a saved color range.
 * @details
 * The @a saved color range is part of the mechanism to restore the initial parameters of the DataItem.
 * It is first set in the constructor from the afferent argument and can only be @a updated with the
 * value of the current threshold using DataItem::updateSaveRange.
 */
const DataRange &DataItem::saveColorRange() const { return mSaveColorRange; }

/**
 * @brief Check whether or not this dataitem is comparable to another.
 * @param other Other DataItem to compare to.
 * @return True if both object are comparable, false otherwise.
 * @details
 * Two DataItem instances are comparable if:
 *
 * - Their parameter name, alias and units are the same.
 * - The range and units of their ColorRange are the same.
 */
bool DataItem::comparable(const DataItem &other) const {
    // compare:
    // 1. parameter name and units and alias.
    bool cmp = mAlias.compare(other.mAlias) == 0 && mParam.name().compare(other.mParam.name()) == 0 &&
               mParam.units == other.mParam.units;
    if (!cmp)
        return cmp;
    // 2. Color range
    return mColorRange == other.mColorRange;
}

/// @brief Get the current color range (const ref).
const DataRange &DataItem::colorRange() const { return mColorRange; }

/// @brief Get the remanence threshold (const ref).
double DataItem::threshold() const { return mThreshold; }

/// @brief Get the current color range.
DataRange &DataItem::colorRange() { return mColorRange; }

/// @brief Get the remanence threshold.
double &DataItem::threshold() { return mThreshold; }

/**
 * @brief Set a new color range.
 * @param range New color range to set.
 * @return True if the range has been actually updated, false otherwise.
 * @attention
 * If range is not finite (see DataRange::isFinite) the method does not register the range:
 * Color range must have finite boundaries !
 */
bool DataItem::setColorRange(const DataRange &range) {
    bool ret = range.isFinite();
    if (ret)
        mColorRange = range;
    return ret;
}

/// @brief Set the remanence threshold.
void DataItem::setThreshold(double value) { mThreshold = value; }

/// @brief Get the name of this data item.
QString DataItem::alias() const { return mAlias; }

/// @brief Set a name for this data item.
/// @note
/// The method does emit aliasChanged signal.
void DataItem::setAlias(const QString &alias) {
    mAlias = alias;
    emit aliasChanged();
}

/// @brief Get the associated parameter (const ref).
const AParam &DataItem::parameter() const { return mParam; }

/// @brief Get the associated parameter.
AParam &DataItem::parameter() { return mParam; }

/// @brief Get the item units.
Units::Unit DataItem::units() const { return mParam.units; }

/// @brief Set a new value in the item.
void DataItem::setValue(double value) { mParam.value = value; }

/// @brief Get the last registered value.
double DataItem::value() const { return mParam.floatValue(); }

/**
 * @brief Get a string representation of the last  value including units if any.
 * @param prec Optional decimal precision (default to 3).
 */
QString DataItem::valueAndUnits(int prec) const {
    double val = value();
    if (qIsNaN(val))
        return "N/A";
    QString sym(QString::fromStdString(Units::unitToString(mParam.units)));
    return (sym.compare("N/A") == 0) ? QString::number(val, 'g', prec)
                                     : QString("%1 %2").arg(val, 0, 'g', prec).arg(sym);
}

/// @brief Get the name and units of this data item.
QString DataItem::aliasAndUnits(bool split) const {
    QString units = QString::fromStdString(Units::unitToString(mParam.units));
    if (units.compare("N/A") == 0)
        return mAlias;
    else {
        return QString("%1%2(%3)").arg(mAlias, (mAlias.size() > 16 && split) ? "\n" : " ", units);
    }
}

/**
 * @brief Check whether or not the given value is visible.
 * @details
 * A value is assumed to be visible if it is not a NaN and if it si greater or equal to the data item threshold. If the
 * latter is NaN then any value is assumed to be hidden.
 */
bool DataItem::isVisible(double value) const { return !(qIsNaN(mThreshold) || qIsNaN(value)) && value >= mThreshold; }

/**
 * @brief Set new units.
 * @param unit New unit to set.
 * @param update True (default) to update internal values.
 * @details
 * If \a update is set to true and units are consistent, the method updates the internal values of the item
 * (current value, color range and threshold).
 */
void DataItem::setUnits(Units::Unit unit, bool update) {
    if (update && !Units::inconsistent(unit, mParam.units)) {
        double v = mParam.floatValue();
        if (!qIsNaN(v))
            mParam.value = Units::convert(v, mParam.units, unit);
        mColorRange.setLower(Units::convert(mColorRange.lower(), mParam.units, unit));
        mColorRange.setUpper(Units::convert(mColorRange.upper(), mParam.units, unit));
        if (!qIsNaN(mThreshold))
            mThreshold = Units::convert(mThreshold, mParam.units, unit);
    }
    mParam.units = unit;
    // emit unitChanged();
}

/// @brief Get a XML representation of the data item.
QByteArray DataItem::toXML(bool *onMap, int *plotId) const {
    QString mapAttr, plotAttr;
    if (onMap && (*onMap))
        mapAttr = QString(" map=\"true\"");
    if (plotId && ((*plotId) > -1))
        plotAttr = QString(" plot=\"%1\"").arg((*plotId));
    QString prefix = QString("<parameter%1%2>").arg(mapAttr, plotAttr);
    QString out = QString("%1"
                          "<name>%2</name>"
                          "<alias>%3</alias>"
                          "<units>%4</units>"
                          "<min>%5</min>"
                          "<max>%6</max>"
                          "<remanence>%7</remanence>"
                          "</parameter>")
                      .arg(prefix, QString::fromStdString(mParam.name()), mAlias, QString::fromStdString(Units::unitToString(mParam.units, true)))
                      .arg(mColorRange.lower())
                      .arg(mColorRange.upper())
                      .arg(qIsNaN(mThreshold) ? "" : QString::number(mThreshold));
    return out.toLatin1();
}

/**
 * @brief Duplicate the given item.
 * @return New item.
 */
DataItem *DataItem::copyFrom(DataItem *item, QObject *parent) {
    if (!item)
        return nullptr;
    return new DataItem(item->parameter(), item->alias(), item->colorRange(), item->threshold(), parent);
}
