// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef DATAITEM_HPP
#define DATAITEM_HPP

#include "datarange.hpp"
#include "globals.hpp"

/**
 * @brief The DataItem class
 * @details
 * This class acts as a container for data points.
 *
 * Data is stored under two forms:
 *
 * - An series of graphics (simple markers) stored in an ARCGIS overlay (DataItem::graphicOverlay).
 * - a QCP graph data container with the data's timeline (DataItem::graphData).
 *
 * Data can only be added to the container (see DataItem::addData) but never be removed unless such operation
 * is performed directly on the overlay and/or the graph data container.
 *
 */
class DataItem : public QObject {
    Q_OBJECT
  public:
    explicit DataItem(const AParam &param, const QString &alias, const DataRange &colorRange, double threshold,
                      QObject *parent = nullptr);
    virtual ~DataItem();

    QString alias() const;
    void setAlias(const QString &alias);
    const AParam &parameter() const;
    AParam &parameter();
    Units::Unit units() const;
    void setValue(double value);
    double value() const;
    QString valueAndUnits(int prec = 3) const;
    QString aliasAndUnits(bool split = false) const;

    QByteArray toXML(bool *onMap = nullptr, int *plotId = nullptr) const;

    bool comparable(const DataItem &other) const;
    bool isVisible(double value) const;
    void setUnits(Units::Unit unit, bool update = true);

    const DataRange &colorRange() const;
    double threshold() const;

    DataRange &colorRange();
    double &threshold();

    bool setColorRange(const DataRange &range);
    void setThreshold(double value);

    static DataItem *copyFrom(DataItem *item, QObject *parent = nullptr);

    void updateSaveThreshold();
    void updateSaveRange();

    double saveThreshold() const;
    const DataRange &saveColorRange() const;

signals:
    void unitChanged(bool rangeUpdated);
    void aliasChanged();

  private:
    AParam mParam;         ///< Associated AParam.
    QString mAlias;        ///< Alias to display for user.
    DataRange mColorRange; ///< Range used for color scale.
    double mThreshold;     ///< Visibility threshold.

    DataRange mSaveColorRange; ///< Initial range used for color scale.
    double mSaveThreshold;     ///< Initial Visibility threshold.

  private:
    Q_DISABLE_COPY(DataItem)
};

#endif // DATAITEM_HPP
