// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "dataloader.hpp"
#include "amdataset.hpp"
#include "prm_utils.hpp"
#include "itemmapper.hpp"

/// @brief Constructor.
DataLoader::DataLoader(DPP parameters, const QVector<long> &indexes, const ItemMapper *itemMapper, const AMDataset *dataset,
                       QObject *parent)
    : QObject(parent), mParams(parameters), mValidIndexes(indexes), mItemMapper(itemMapper), mDataset(dataset),
      mIsRunning(false), mCancel(false) {}

/// @brief Destructor.
DataLoader::~DataLoader() {}

/// @brief Abort operations.
void DataLoader::cancel() {
    QMutexLocker locker(&mDataMutex);
    mCancel = true;
}

/// @brief Check whether or not the worker in the running.
bool DataLoader::isRunning() const {
    QMutexLocker locker(&mDataMutex);
    return mIsRunning;
}

/// @brief Check whether or not the loading process has been canceled.
bool DataLoader::canceled() const {
    QMutexLocker locker(&mDataMutex);
    return mCancel;
}

/// @brief Load files in the dataset.
void DataLoader::process() {
    mDataMutex.lock();
    mCancel = false;
    mIsRunning = true;
    bool cancel = mCancel;
    mDataMutex.unlock();
    QVector<AParam> params = mItemMapper->parameters();

    int ni = mValidIndexes.size(), no = ni;
    ReplayData *replayData = new ReplayData;
    QMap<AParam, QDoubleVector> &dataMap = replayData->data;
    // Common step: data extraction.
    emit stepMessage(tr("Step: Extracting parameters."));
    QVector<AParam> missings;
    for (const AParam &p : params) {
        if (mDataset->hasParameter(p)) {
            emit infoMessage(tr("Info: Extracting parameter: %1").arg(QString::fromStdString(p.name())));
            auto && tmp = mDataset->parameterAt(p, mValidIndexes);
            QDoubleVector data;
            std::copy(tmp.begin(),tmp.end(),std::back_inserter(data));
            dataMap[p] = data;
        } else {
            missings.push_back(p); // save missing parameters: we'll build nan vectors for them later.
        }
    }

    AParam timePrm = mItemMapper->parameter(ItemMapper::PCMON_TIME);
    if (!mDataset->hasParameter(timePrm))
        timePrm = mItemMapper->parameter(ItemMapper::SENSOR_TIME);

    emit stepMessage(tr("Step: Reducing dataset."));

    QMap<AParam, QDoubleVector>::iterator it = dataMap.begin();
    switch (mParams.process) {
    case DPP::TRUNCATE: {
        no = qBound(0,mParams.keepPoints,ni);
        while (it != dataMap.end()) {
            const AParam &p = it.key();
            QDoubleVector &values = it.value();
            emit stepMessage(tr("Info: Truncating parameter: %1.").arg(QString::fromStdString(p.name())));
            values.resize(no);
            values.shrink_to_fit();
            it++;
            mDataMutex.lock();
            cancel = mCancel;
            mDataMutex.unlock();
            if (cancel)
                break;
        }
        break;
    }
    case DPP::DECIMATE: {
        no = static_cast<int>(qCeil(static_cast<double>(ni) / mParams.decimationStep));
        QDoubleVector tmp;
        tmp.reserve(no);
        while (it != dataMap.end()) {
            const AParam &p = it.key();
            QDoubleVector &values = it.value();
            emit infoMessage(tr("Info: Decimating parameter: %1").arg(QString::fromStdString(p.name())));
            for (int i = 0; i < ni; i += mParams.decimationStep)
                tmp.push_back(values.at(i));
            tmp.shrink_to_fit(); // tmp size may exceed by 1.
            values = tmp;
            tmp.clear();
            it++;
            mDataMutex.lock();
            cancel = mCancel;
            mDataMutex.unlock();
            if (cancel)
                break;
        }
        no = dataMap.first().size();
        break;
    }
    case DPP::RESAMPLE: {
        std::list<QPair<int, int>> groups = timePairIndexes(dataMap[timePrm], mParams.resampleStep, dataMap[timePrm]);
        no = dataMap[timePrm].size();
        QDoubleVector tmp(no);
        while (it != dataMap.end()) {
            const AParam &p = it.key();
            QDoubleVector &values = it.value();

            if (p != timePrm) {
                emit infoMessage(tr("Info: Resampling parameter: %1").arg(QString::fromStdString(p.name())));
                int i = 0;
                for (auto it = groups.begin(); it != groups.end(); it++, i++) {
                    int first = it->first, second = it->second;
                    tmp[i] = reduceData(p.reduction, values.begin() + first, values.begin() + second);
                }
                values = tmp;
            }
            it++;
            mDataMutex.lock();
            cancel = mCancel;
            mDataMutex.unlock();
            if (cancel)
                break;
        }
        break;
    }
    default:
        break;
    }
    mDataMutex.lock();
    cancel = mCancel;
    mDataMutex.unlock();
    // check for cancellation.
    if (cancel) {
        emit infoMessage(tr("Operation canceled by the user."));
        delete replayData;
        replayData = nullptr;
    } else {
        QDoubleVector &times = dataMap[timePrm];
        no = times.size();
        replayData->frameCount = no;
        // setup missing parameters.
        emit stepMessage(tr("Step: Updating missing parameters."));
        QDoubleVector nans(no, qQNaN());
        for (const AParam &p : missings) {
            emit infoMessage(tr("Info: Adding parameter %1").arg(QString::fromStdString(p.name())));
            dataMap[p] = nans;
        }
        // final steps: positions and timeline.
        emit stepMessage("Step: Building trajectory positions and timeline.");
        emit infoMessage("Info: Building timeline.");
        QDoubleVector &lon = dataMap[mItemMapper->parameter(ItemMapper::LONGITUDE)];
        QDoubleVector &lat = dataMap[mItemMapper->parameter(ItemMapper::LATITUDE)];
        QDoubleVector &alt = dataMap[mItemMapper->parameter(ItemMapper::ALTITUDE)];
        QDoubleVector &nbs = dataMap[mItemMapper->parameter(ItemMapper::NB_SATS)];
        replayData->times.reserve(no);
        for (auto v : times)
            replayData->times.push_back(v + EPOCH1904_OFFSET);
        // get positions.
        emit infoMessage("Info: Building trajectory positions.");
        replayData->positions.reserve(no);
        replayData->gpsValidity.reserve(no);
        for (int i = 0; i < no; ++i){
            replayData->positions.push_back(Esri::ArcGISRuntime::Point(lon.at(i), lat.at(i), alt.at(i), gWGS84));
            replayData->gpsValidity.push_back(nbs.at(i) > 3);
        }
        emit infoMessage("Info: Job complete !");
        emit dataReady(replayData);
    }
    mDataMutex.lock();
    mIsRunning = false;
    mDataMutex.unlock();
    emit finished(cancel);
    return;
}
