// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef DATAMANAGER_HPP
#define DATAMANAGER_HPP

#include "datarange.hpp"
#include "flightmodel.hpp"
#include "globals.hpp"

class ScenariosModel;
class Scenario;
class ScenarioSelector;
class DataItem;
class ItemMapper;

class ScenarioWidget;
class MapWidget;
class RtPlotWidget;
class IndicatorsWidget;
class FlightWidget;
class MediaControlWidget;

namespace Esri {
namespace ArcGISRuntime {
class Map;
class Basemap;
class MapGraphicsView;
class GraphicsOverlay;
class Graphic;
class Point;
class SimpleLineSymbol;
class SimpleMarkerSymbol;
enum class SimpleMarkerSymbolStyle;
} // namespace ArcGISRuntime
} // namespace Esri

#include "SpatialReference.h"

/// @brief Defines a pair of double values.
typedef QPair<double, double> DoublePair;
/// @brief Defines a vector of double pair values.
typedef QVector<DoublePair> DoublePairVector;

typedef QPointer<ScenarioSelector> ScenarioSelectorPointer;

typedef QVector<QVector<double>> QDoubleMatrix;

/**
 * @brief The DataManager class
 * @details
 * Manage parameter and flights (or at least try to do so ^^).
 */
class DataManager : public QObject {
    Q_OBJECT

  public:
    explicit DataManager(QObject *parent = nullptr);
    virtual ~DataManager();

    // data acquisition related method.
    ItemMapper *mapper();

    void translateModelHeaders();

    void resetMapper(const ItemMapper &mapper);
    void cleanup();
    void resetData();
    bool updateData(bool *newFlightCreated = nullptr);
    void updateWidgets();

    bool processBlock(const ABlock &block, bool *newFlightCreated = nullptr);
    void setPCTimeAsReference(bool enable);
    bool pcTimeIsReference() const;
    QString computeSensorFileName(bool *noSD = nullptr) const;

    FlightModel *flightModel() const;

    bool deleteFlight(int index);
    int flightCount() const;
    int flightLength() const;
    int flightLength(int index) const;
    DoublePair dataExtrema(DataItem *item) const;
    DoublePair flightExtrema(DataItem *item, int flightIndex = -1) const;
    DoublePairVector flightsExtremas(DataItem *item, QVector<int> flightIndexes = QVector<int>()) const;
    QDateTime lastStartTime() const;

    int newFlightDelay() const;
    void setNewFlightDelay(int newFlightDelay);
    bool newFlightOnStrap() const;
    void setNewFlightOnStrap(bool newFlightOnStrap);

    // scenarios and data item related methods.
    void setScenarios(const QVector<Scenario *> &scenarios, int *selected = nullptr);

    bool saveScenarios(const QString &xmlPath);

    QVector<Scenario *> scenarios() const;
    int scenarioCount() const;

    Scenario *selectedScenario() const;
    DataItem *mapItem() const;
    DataItem *plotItem(int index) const;

    // widget registration methods.

    bool registerScenarioSelector(ScenarioSelector *widget);
    bool unregisterScenarioSelector(ScenarioSelector *widget);

    void registerMediaControlWidget(MediaControlWidget *widget);
    void registerFlightWidget(FlightWidget *widget);
    void registerScenarioWidget(ScenarioWidget *widget);
    void registerMapWidget(MapWidget *widget);
    bool registerPlotWidget(int index, RtPlotWidget *widget);
    void registerIndicatorsWidget(IndicatorsWidget *widget);

    // cosmetics methods (map items style)
    void setColorGradient(const QCPColorGradient &gradient);
    void updateFromSettings(bool applyMarkerStyle = true);
    int size(QVector<int> *flightSizes = nullptr) const;

    void setAutoScrollingInterval(int interval);
    int autoScrollingInterval() const;

    bool nowIsTimeReference() const;
    void useNowAsTimeReference(bool enable);

    void enableTimeScrolling(bool enable);
  public slots:
    void createNewFlight();

  signals:

    /**
     * @brief Emitted whenever a new flight is added.
     * @details
     * The signal is emitted whenever a new flight is added, that is when DataManager::createNewFlight has been called.
     *
     * Note that such operation can be performed when a block is processed.
     */
    void newFlightAdded();

    /// @brief Emitted whenever the internal mapper has been updated.
    void mapperUpdated();

  protected:
    virtual void timerEvent(QTimerEvent *event) override;

  private slots:
    void setAutoCenter(bool enable);
    void updateIndicators(Scenario *scenario);
    void setItemOnMap(DataItem *item);
    void setItemOnPlot(int graphIndex, DataItem *item);
    void onPlotWidgetContextMenuRequested(const QPoint &pos);
    void onMapContextMenuRequested(const QPoint &pos, bool isMap);
    void selectScenario(int index);
    void onFlightStateChanged(int index, bool checked);

    void updateColorRange(const DataRange &range);
    void updateRemanenceThreshold(double value);
    void updateMapSpatialReference(const Esri::ArcGISRuntime::SpatialReference &reference);

    void deleteFlightFromAction();
    void checkAllFlights();
    void uncheckAllFlights();

  private:
    QBasicTimer mAutoScrollTimer;
    int mAutoScrollInterval;
    double mLastTime;
    bool mNowIsReference;

    /// Last flight starting datetime.
    QDateTime mLastStartTime;

    /// Internal parameter key mapper.
    ItemMapper *mItemMapper;

    /// Scenarios.
    QVector<Scenario *> mScenarios;

    /// Flight model.
    FlightModel *mFlightModel;

    /// Last registered Amulse block.
    ABlock mLastBlock;

    /// New-Flight creation delay.
    int mNewFlightDelay;
    /// New-flight creation on strap control flag.
    bool mNewFlightOnStrap;
    /// Registered block counter.
    int mBlockCount;

    /// PC / Sensor time selector.
    bool mUsePCTime;

    QDoubleMatrix mSensorTimes;
    QDoubleMatrix mPcMonTimes;
    QVector<BoolVector> mValidGPS;
    QVector<PointVector> mPositions;
    QMap<AParam, QDoubleMatrix> mLiveData;
    QMap<AParam, DoublePairVector> mExtremas;

    Scenario *mSelectedScenario;

    // map view data
    Esri::ArcGISRuntime::SpatialReference mMapSpatialReference;
    QVector<Esri::ArcGISRuntime::GraphicsOverlay *> mOverlays;
    Esri::ArcGISRuntime::SimpleLineSymbol *mOutline;

    bool mBypassGPS;

    bool mMVShow;
    QColor mMVColor;
    Esri::ArcGISRuntime::SimpleMarkerSymbolStyle mMVStyle;
    qreal mMVSize;

    Esri::ArcGISRuntime::SimpleMarkerSymbolStyle mCFStyle;
    qreal mCFSize;

    Esri::ArcGISRuntime::SimpleMarkerSymbolStyle mPFStyle;
    qreal mPFSize;

    GraphContainerVector mMapGraphData;
    GraphContainerVector mPlotsGraphData[N_PLOTS];
    // registered widgets:

    QVector<ScenarioSelectorPointer> mScenarioSelectors;
    QPointer<ScenarioWidget> mScenarioWidget;
    QPointer<FlightWidget> mFlightWidget;
    QPointer<MapWidget> mMapWidget;
    QPointer<IndicatorsWidget> mIndicatorsWidget;
    QPointer<RtPlotWidget> mPlotWidgets[N_PLOTS];
    QPointer<MediaControlWidget> mMediaControlWidget;

    int mTotalCount;

    bool mVisUseDuration;
    int mNVisible;
    int mSecVisible;

    bool mAutoCenter;
    int mMinSat;
    QCPColorGradient mColorGradient;

  private:
    bool checkForNewFlight(const ABlock &block) const;
    void createNewFlightPvt();
    void setupInternals();
    void setupMarkerStyles();
    void setupMarkerColors();
    void setupMarkerVisibility();

    Esri::ArcGISRuntime::Graphic *getGraphicFromEnd(int interval);
    void setupWidgetsUpdatesEnabled(bool enable);
    void addParameters(QStringList names);
    // static const Esri::ArcGISRuntime::SpatialReference gWGS84;

    // Replay related:
  public:
    void setReplayMode(bool enable);
    void setReplayData(ReplayData *data);
    void setReplayFrameRate(int frameRate);

  private:
    ReplayState mReplayState;
    bool mReplayMode;
    bool mAutoRepeat;
    QTimer mReplayTimer;

    int mReplayFrameRate;
    int mFrameStep;
    int mFrameIndex;
    ReplayData *mReplayData;

    Esri::ArcGISRuntime::GraphicsOverlay *mReplayOverlay;
    GraphContainerPointer mReplayMapGraphData;
    GraphContainerPointer mReplayPlotsGraphData[N_PLOTS];

    void resizeReplayContainers();
    void forceTimeRange();
    void updateValueRange();

    int visibleIndex();
    int getReplayVisibleMinIndex(int index);

    QPair<int, int> globalToLocals(int gi);
    int localsToGlobal(int fi, int li);
  private slots:

    void stopReplay();
    void pauseReplay();
    void playReplay();
    void nextFrame();
    void prevFrame();
    void goToFrame(int index);
    void goToStart();
    void goToEnd();
    void setAutoRepeat(bool enable);
    void changeFrameStep(int step);

  signals:
    /// @brief Emited when the Replay state changed.
    void ReplayStateChanged(ReplayState state);
};

#endif // DATAMANAGER_HPP
