// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "datarange.hpp"

/**
 * @brief Initialize an infinite range.
 * @details
 * Internal boundaries are set to NaN meaning no boundaries. See DataRange::contains for more details about NaN
 * management. Moreover, units are not defined.
 */
DataRange::DataRange() : mLower(qQNaN()), mUpper(qQNaN()) {}

/// @brief Initialize a DataRange with given lower and upper boundaries.
DataRange::DataRange(double lower, double upper) : mLower(lower), mUpper(upper) { normalize(); }

/// @brief Initialize a DataRange with a QCPRange.
DataRange DataRange::fromQCPRange(const QCPRange &range) { return DataRange(range.lower, range.upper); }

/**
 * @brief Check whether or not two data range are comparable.
 * @param other Other data range to compare with.
 * @return True if both data range are equals, false otherwise.
 */
bool DataRange::operator==(const DataRange &other) const {
    if (!((qIsNaN(mLower) && qIsNaN(other.mLower)) || dblCompare(mLower, other.mLower)))
        return false;
    if (!((qIsNaN(mUpper) && qIsNaN(other.mUpper)) || dblCompare(mUpper, other.mUpper)))
        return false;
    return true;
}

/**
 * @brief Check whether or not two data range are different.
 * @param other Other data range to compare with.
 * @return True if both data range are different, false otherwise.
 * @note
 * The method is the strict opposite to DataRange::operator==.
 */
bool DataRange::operator!=(const DataRange &other) const { return !(*this == other); }

/**
 * @brief Normalize the range.
 * @details
 * The method swaps lower and upper boundaries if needed so it is internally defined in increasing order.
 */
void DataRange::normalize() {
    if (!(qIsNaN(mLower) || qIsNaN(mUpper))) {
        if (mLower > mUpper)
            qSwap(mLower, mUpper);
    }
}
/**
 * @brief Check whether value lies in the data range.
 * @param value Value to check.
 * @details
 * The method checks if the value is lower/greater than the upper/lower boundaries of the range:
 *
 * - If value is NaN then false is returned whatever the range is.
 * - If a range's boundary is set to NaN, no tests for that boundary is performed (the test results in true).
 *
 * The tests are defined to simulate (semi) open ranges. But a NaN input value is assumed to be a missing value thus
 * it cannot be contained in any kind of range.
 */
bool DataRange::contains(double value) const {
    bool nv = qIsNaN(value), nl = qIsNaN(mLower), nu = qIsNaN(mUpper);
    if (nv)
        return false;
    if (nl && nu)
        return true;
    else if (nl)
        return value <= mUpper;
    else if (nu)
        return value >= mLower;
    else
        return value >= mLower && value <= mUpper;
}

/// @brief Check whether or not the range is finite.
bool DataRange::isFinite() const { return qIsFinite(mLower) && qIsFinite(mUpper); }

/**
 * @brief Convert this data range to a QCPRange.
 * @attention
 * The output QCPRange may be invalid, QCPRange does not accept NaN/Inf as boundary.
 */
QCPRange DataRange::toQCPRange() const { return QCPRange(mLower, mUpper); }

/// @brief Get the lower boundary of the range.
double DataRange::lower() const { return mLower; }

/// @brief Get the upper boundary of the range.
double DataRange::upper() const { return mUpper; }

/**
 * @brief Set the lower boundary of the range.
 * @attention
 * The method does not normalize (i.e. swaps min and max values) if needed.
 * It is user responsibility to normalize the range if needed.
 */
void DataRange::setLower(double lower) { mLower = lower; }

/**
 * @brief Set the upper boundary of the range.
 * @attention
 * The method does not normalize (i.e. swaps min and max values) if needed.
 * It is user responsibility to normalize the range if needed.
 */
void DataRange::setUpper(double upper) { mUpper = upper; }

/**
 * @brief Set the range boundaries from a QCPRange.
 * @param range QCPRange to process.
 * @note
 * The method normalizes (i.e. swaps lower and upper) the range if needed.
 */
void DataRange::setRange(const QCPRange &range) {
    mLower = range.lower;
    mUpper = range.upper;
    normalize();
}
/**
 * @brief Set the range boundaries.
 * @param lower Lower boundary.
 * @param upper Upper boundary.
 * @note
 * The method normalizes (i.e. swaps lower and upper) the range if needed.
 */
void DataRange::setRange(double lower, double upper) {
    mLower = lower;
    mUpper = upper;
    normalize();
}
