// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "comboboxunitsdelegate.hpp"
#include "isp/units.hpp"

typedef QPair<QString, Units::Unit> UnitPair;

// Get the list of sibling units and their name.
static QVector<UnitPair> getUnitsAndStrings(Units::Unit unit) {
    auto &&tmp = Units::siblingUnits(unit);
    QVector<Units::Unit> units;
    std::copy(tmp.begin(), tmp.end(), std::back_inserter(units));
    QVector<UnitPair> out;
    out.reserve(units.size());
    for (Units::Unit u : Units::siblingUnits(unit))
        out.push_back(qMakePair(QString::fromStdString(Units::unitToString(u, true)), u));
    return out;
}

/// @brief Constructor
ComboboxUnitsDelegate::ComboboxUnitsDelegate(QObject *parent) : QStyledItemDelegate(parent) {}

/// @brief Create the combobox editor.
QWidget *ComboboxUnitsDelegate::createEditor(
    QWidget *parent, const QStyleOptionViewItem & /* option */, const QModelIndex &index) const {
    QComboBox *editor = new QComboBox(parent);
    const QAbstractItemModel *model = index.model();
    Units::Unit unit = model->data(index, Qt::EditRole).value<Units::Unit>();

    QVector<UnitPair> pus = getUnitsAndStrings(unit);
    int i = 0, idx = -1;
    for (auto pu : pus) {
        editor->addItem(pu.first, pu.second);
        if (pu.second == unit)
            idx = i;
        i++;
    }
    editor->setCurrentIndex(idx);
    return editor;
}

/// @brief Set data in the editor.
/// @note
/// The method assumes that \a index's model data() method return a Units::Unit qvariant for the edit role.
void ComboboxUnitsDelegate::setEditorData(QWidget *editor, const QModelIndex &index) const {
    const QAbstractItemModel *model = index.model();
    Units::Unit unit = model->data(index, Qt::EditRole).value<Units::Unit>();
    QComboBox *comboBox = static_cast<QComboBox *>(editor);
    comboBox->setCurrentIndex(comboBox->findData(unit));
}

/// @brief Set data in the model.
void ComboboxUnitsDelegate::setModelData(QWidget *editor, QAbstractItemModel *model, const QModelIndex &index) const {
    QComboBox *comboBox = static_cast<QComboBox *>(editor);
    if (comboBox->currentIndex() == -1)
        return;
    model->setData(index, comboBox->currentData(), Qt::EditRole);
}

/// @brief Compute the delegate size hint.
QSize ComboboxUnitsDelegate::sizeHint(const QStyleOptionViewItem &option, const QModelIndex &index) const {
    // from https://stackoverflow.com/questions/44909897/resizing-qtableview-section-with-custom-editor
    // many thanks !
    QSize hint = QStyledItemDelegate::sizeHint(option, index);
    QFontMetrics fm(option.font);
    QStyleOptionComboBox comboOption;
    comboOption.rect = option.rect;
    comboOption.state = option.state | QStyle::State_Enabled;

    const QAbstractItemModel *model = index.model();
    Units::Unit unit = model->data(index, Qt::EditRole).value<Units::Unit>();
    for (const UnitPair &pu : getUnitsAndStrings(unit)) {
        hint = hint.expandedTo(qApp->style()->sizeFromContents(
            QStyle::CT_ComboBox, &comboOption, QSize(fm.horizontalAdvance(pu.first), hint.height())));
    }
    return hint;
}
