// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "flightmodel.hpp"

/**
 * @brief Format time duration.
 * @param msecs Number of milliseconds.
 * @return A string with the duration expressed in term of minutes and seconds.
 */
QString formatDuration(qint64 msecs) {
    double v = msecs / (60.0 * 1000.0);
    double r = v - static_cast<int>(v);
    return QString("%1:%2").arg(static_cast<int>(v)).arg(static_cast<int>(r * 60), 2, 10, QChar('0'));
}

/**
 * @brief formatDatetime
 * @param datetime Datetime to format.
 * @return A string with the datetime expressed either as a date and a time or simply a time if its data is today.
 */
QString formatDatetime(const QDateTime &datetime) {
    return (datetime.date() == QDate::currentDate()) ? datetime.toString("HH:mm:ss")
                                                     : datetime.toString("HH:mm:ss (yyyy-MM-dd)");
}

/**
 * @brief Constructor.
 * @param parameters List of parameters to manage.
 * @param parent Optional QObject parent.
 * @details
 * Initialize a flight manager with the given list of parameters to process in incoming data blocks.
 */
FlightModel::FlightModel(QObject *parent) : QAbstractTableModel(parent) { translateHeaders(); }

/// @brief Destructor.
FlightModel::~FlightModel() {}

/// @brief Add a new flight.
void FlightModel::addFlight(const QDateTime &startTime, const QDateTime &endTime, int itemCount, Qt::CheckState state) {
    int nr = rowCount();
    beginInsertRows(QModelIndex(), nr, nr);
    mStartTimes.push_back(startTime);
    mEndTimes.push_back(endTime);
    mItemCount.push_back(itemCount);
    mStates.push_back(state);
    endInsertRows();
}

/**
 * @brief Remove the flight at given index.
 * @param index Index of the flight to remove.
 * @return True on success, false otherwise.
 * @note
 * The method fails only if index is out of range.
 */
bool FlightModel::removeFlight(int index) {
    if (index < 0 || index >= mStartTimes.size())
        return false;
    beginRemoveRows(QModelIndex(), index, index);
    mStartTimes.remove(index);
    mEndTimes.remove(index);
    mItemCount.remove(index);
    mStates.remove(index);
    endRemoveRows();
    return true;
}

Qt::ItemFlags FlightModel::flags(const QModelIndex &index) const {
    Qt::ItemFlags fl(Qt::ItemIsEnabled | Qt::ItemIsSelectable);
    if (!(index.isValid() && index.column()))
        fl |= Qt::ItemIsUserCheckable;
    return fl;
}

/// @brief Get the header displayed data.
QVariant FlightModel::headerData(int section, Qt::Orientation orientation, int role) const {
    if (role == Qt::DisplayRole && orientation == Qt::Horizontal && section >= 0 && section < mHeaders.size())
        return QVariant(mHeaders.at(section));
    else
        return QVariant();
}

/**
 * @brief Get the starting date time of the flight at given index.
 * @param index Index of the flight.
 * @details
 * If index is out of range, a null datetime is returned.
 */
QDateTime FlightModel::startTime(int index) const {
    return (index < 0 || index >= mStartTimes.size()) ? QDateTime() : mStartTimes[index];
}

/**
 * @brief Get the ending date time of the flight at given index.
 * @param index Index of the flight.
 * @details
 * If index is out of range, a null datetime is returned.
 */
QDateTime FlightModel::endTime(int index) const {
    return (index < 0 || index >= mEndTimes.size()) ? QDateTime() : mEndTimes[index];
}

/**
 * @brief Get the number of item that have been registered for the flight at given index.
 * @param index Index of the flight.
 * @details
 * If index is out of range, zero is returned.
 */
int FlightModel::itemCount(int index) const { return (index < 0 || index >= mItemCount.size()) ? 0 : mItemCount[index]; }

/**
 * @brief Get the check state of the flight at given index.
 * @param index Index of the flight.
 * @return The check state of the flight.
 * @details
 * If index is out of range, Qt::Unchecked is returned.
 */
Qt::CheckState FlightModel::checkState(int index) const {
    return (index < 0 || index >= mStates.size()) ? Qt::Unchecked : mStates[index];
}

/**
 * @brief Get the data of the model.
 * @param index Model index.
 * @param role Role of the data.
 * @return Data at the given index for the given role.
 * @details
 * The method cannot return data for the current flight item, use FlightManager::current instead to access the current flight.
 *
 * The method only returns data for Qt::DisplayRole, Qt::CheckStateRole and Qt::ToolTipRole.
 */
QVariant FlightModel::data(const QModelIndex &index, int role) const {
    int ir = index.row(), jc = index.column();

    if (!index.isValid() || ir >= mStartTimes.size())
        return QVariant();

    if (role == Qt::DisplayRole) {
        switch (jc) {
        case 0:
            return mStartTimes[ir].toString("HH:mm:ss");
        case 1:
            return mEndTimes[ir].toString("HH:mm:ss");
        case 2:
            return formatDuration(mStartTimes[ir].msecsTo(mEndTimes[ir]));
        default:
            return QVariant();
        }
    } else if (role == Qt::ToolTipRole) {
        if (jc == 0)
            return mStartTimes[ir].toString("yyyy-MM-dd, HH:mm:ss");
        else if (jc == 1)
            return mEndTimes[ir].toString("yyyy-MM-dd, HH:mm:ss");
        else
            return QVariant();
    } else if (role == Qt::CheckStateRole && jc == 0) {
        return mStates[ir];
    }
    return QVariant();
}

/**
 * @brief Set data in the model.
 * @param index Index to process.
 * @param value Value to set.
 * @param role Role of the data.
 * @return True on success, false otherwise.
 * @details
 * The method simply updates the checkstates of the flight at the given index.
 * It emits FlightModel::stateChanged signal with the index of the flight and its new state.
 */
bool FlightModel::setData(const QModelIndex &index, const QVariant &value, int role) {

    if (index.isValid() && index.column() == 0 && role == Qt::CheckStateRole) {
        mStates[index.row()] = static_cast<Qt::CheckState>(value.toInt());
        emit dataChanged(index, index, QVector<int>() << role);
        emit stateChanged(index.row(), mStates[index.row()] == Qt::Checked);
        return true;
    }
    return false;
}

/// @brief Check all flights.
void FlightModel::checkAll(bool signalNeeded) {
    for (int i = 0; i < mStates.size(); i++) {
        mStates[i] = Qt::Checked;
        if (signalNeeded)
            emit stateChanged(i, true);
    }
    emit dataChanged(index(0, 0), index(rowCount() - 1, 0), QVector<int>() << Qt::CheckStateRole);
}

/// @brief Uncheck all flights.
void FlightModel::uncheckAll(bool signalNeeded) {
    for (int i = 0; i < mStates.size(); i++) {
        mStates[i] = Qt::Unchecked;
        if (signalNeeded)
            emit stateChanged(i, false);
    }
    emit dataChanged(index(0, 0), index(rowCount() - 1, 0), QVector<int>() << Qt::CheckStateRole);
}

/**
 * @brief Get the number of rows under the given parent.
 * @param parent Parent model index.
 * @return The number of flights in the model.
 */
int FlightModel::rowCount(const QModelIndex &parent) const {
    Q_UNUSED(parent)
    return mStartTimes.size();
}

/**
 * @brief Get the number of column of the given parent (fixed value).
 * @param parent Parent model index.
 * @return Size of the model header (see FlightModel::g_headers).
 */
int FlightModel::columnCount(const QModelIndex &parent) const {
    Q_UNUSED(parent)
    return mHeaders.size();
}

/**
 * @brief translate model texts.
 * @details
 * Texts (Column headers) are translated for all instances of the models.
 */
void FlightModel::translateHeaders() {
    mHeaders =
        QStringList{QCoreApplication::tr("Start date"), QCoreApplication::tr("End date"), QCoreApplication::tr("Duration")};
    emit headerDataChanged(Qt::Horizontal, 0, mHeaders.size() - 1);
}

/// @brief Reset the model.
void FlightModel::resetFlights() {
    beginResetModel();
    mStartTimes.clear();
    mEndTimes.clear();
    mItemCount.clear();
    mStates.clear();
    endResetModel();
}

/**
 * @brief Get the number of flights in the model.
 * @note
 * The model only keeps track of previous flights. Then the number of flights returned by the model
 * is equal to the number of flights of its parent data manager minus one.
 */
int FlightModel::flightCount() const { return mStartTimes.size(); }

/// @brief Get the list of indexes of checked flights.
QVector<int> FlightModel::checkedFlights() const {
    QVector<int> out;
    for (int i = 0; i < mStates.size(); ++i) {
        if (mStates[i] == Qt::Checked)
            out.push_back(i);
    }
    return out;
}

/// @brief Get the list of indexes of unchecked flights.
QVector<int> FlightModel::unCheckedFlights() const {
    QVector<int> out;
    for (int i = 0; i < mStates.size(); ++i) {
        if (mStates[i] == Qt::Unchecked)
            out.push_back(i);
    }
    return out;
}
