// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "gpswidget.hpp"

#include "itemmapper.hpp"

/**
 * @brief Constructor.
 * @param parent Optional parent widget.
 */
GPSWidget::GPSWidget(QWidget *parent) : QWidget(parent) {

    mLastColor = Qt::black;
    mMinSat = 4;
    mMapper = nullptr;

    mLblNbSatDescr = new QLabel(this);
    mLblLatDescr = new QLabel(this);
    mLblLonDescr = new QLabel(this);
    mLblAltDescr = new QLabel(this);

    mLblNbSatValue = new QLabel(this);
    mLblLatValue = new QLabel(this);
    mLblLonValue = new QLabel(this);
    mLblAltValue = new QLabel(this);

    for (auto *l : {mLblNbSatValue, mLblLatValue, mLblLonValue, mLblAltValue})
        l->setAlignment(Qt::AlignVCenter | Qt::AlignRight);

    QFormLayout *layout = new QFormLayout(this);

    layout->addRow(mLblNbSatDescr, mLblNbSatValue);
    layout->addRow(mLblLatDescr, mLblLatValue);
    layout->addRow(mLblLonDescr, mLblLonValue);
    layout->addRow(mLblAltDescr, mLblAltValue);

    // setup initial value !
    mLblNbSatValue->setText("N/A");
    mLblLatValue->setText("N/A");
    mLblLonValue->setText("N/A");
    mLblAltValue->setText("N/A");

    translateUI();
}

/// @brief Get the size hint of the window.
QSize GPSWidget::sizeHint() const { return QSize(150, 60); }

/**
 * @brief Set the key mapper.
 * @details
 * The method registers the given mapper in the widget. Without a (consistent) mapper, the widget cannot display
 * anything.
 */
void GPSWidget::setMapper(ItemMapper *mapper) { mMapper = mapper; }

/**
 * @brief Update display values using mapper's last registered values.
 * @details
 * If no mapper has been set (see GPSWidget::setMapper), the method will fail to process the input block and
 * the text "N/A" is set for all labels.
 *
 * The method updates the foreground color of the labels based on the number of satellites:
 *
 * - red if there is not satellites acquired.
 * - orange if the number of satellites is less than 4.
 * - green otherwise.
 */
void GPSWidget::updateFromMapper() {
    double lon, lat, alt;
    int nsa;
    if (mMapper.isNull()) {
        reset();
        return;
    }

    nsa = mMapper->parameter(ItemMapper::NB_SATS).intValue();
    lon = mMapper->parameter(ItemMapper::LONGITUDE).floatValue();
    lat = mMapper->parameter(ItemMapper::LATITUDE).floatValue();
    alt = mMapper->parameter(ItemMapper::ALTITUDE).floatValue();

    mLblNbSatValue->setText(QString::number(nsa));
    mLblLatValue->setText(QString("%1°N").arg(lat, 10, 'f', 6)); // one can use formatLatitude from globals.hpp
    mLblLonValue->setText(QString("%1°E").arg(lon, 11, 'f', 6));
    mLblAltValue->setText(QString("%1 m").arg(alt, 7, 'f', 2));
    QColor color(nsa >= mMinSat ? Qt::black : Qt::red);
    if (color != mLastColor) {
        mLastColor = color;
        updateTextColor();
    }
}

/**
 * @brief Translate widget texts.
 * @details
 * The method translates the text of the description labels as well as the window title of the widget.
 */
void GPSWidget::translateUI() {
    setWindowTitle(tr("GPS status"));
    mLblNbSatDescr->setText(tr("N. of Satellites"));
    mLblLatDescr->setText(tr("Latitude"));
    mLblLonDescr->setText(tr("Longitude"));
    mLblAltDescr->setText(tr("Altitude"));
    mLblNbSatDescr->setToolTip(tr("Number of satellites acquired by the Adafruit GPS module."));
}

/// @brief Reset widget with a default value.
void GPSWidget::reset() {
    static const QColor black(Qt::black);
    if (mLastColor != black) {
        mLastColor = black;
        updateTextColor();
    }
    mLblNbSatValue->setText("N/A");
    mLblLatValue->setText("N/A");
    mLblLonValue->setText("N/A");
    mLblAltValue->setText("N/A");
}

/**
 * @brief Setup the minimum number of satellites.
 * @param nsat Number of satellite threshold.
 * @details
 * The method sets the minimum number of satellites threshold. Below this value, the
 * texts of the widgets are displayed in red.
 */
void GPSWidget::setMinSat(int nsat) { mMinSat = qMax(0, nsat); }

/**
 * @brief Update texts colors as a function of the last registered number of satellite.
 * @details
 * The method is responsible for applying the color indicator on the texts of the widget.
 */
void GPSWidget::updateTextColor() {
    QPalette pal = palette();
    pal.setColor(QPalette::WindowText, mLastColor);

    mLblNbSatDescr->setPalette(pal);
    mLblLatDescr->setPalette(pal);
    mLblLonDescr->setPalette(pal);
    mLblAltDescr->setPalette(pal);

    mLblNbSatValue->setPalette(pal);
    mLblLatValue->setPalette(pal);
    mLblLonValue->setPalette(pal);
    mLblAltValue->setPalette(pal);
}
