// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "indicatorseditor.hpp"
#include "delegates.hpp"
#include "dndviewstyle.hpp"
#include "indicatorchoicedialog.hpp"
#include "indicatormodel.hpp"
#include "indicators.hpp"
#include "tableview.hpp"

IndicatorsEditor::IndicatorsEditor(QWidget *parent) : QWidget(parent) {
    mEmptyModel = new IndicatorModel(this);
    mItemModel = nullptr;
    mTableView = new TableView(this);
    mTableView->setModel(mEmptyModel);

    mTableView->setSelectionBehavior(QTableView::SelectRows);
    mTableView->setSelectionMode(QTableView::SingleSelection);

    mTableView->setItemDelegateForColumn(2, new LineEditDelegate(false, mTableView));         // Name.
    mTableView->setItemDelegateForColumn(3, new ComboboxUnitsDelegate(mTableView));           // Units
    mTableView->setItemDelegateForColumn(4, new BooleanDelegate(mTableView));                 // Reverse/Down boolean flag.
    mTableView->setItemDelegateForColumn(5, new ValueLineEditDelegate(true, true, mTableView)); // Warning threshold.
    mTableView->setItemDelegateForColumn(6, new ValueLineEditDelegate(true, true, mTableView)); // Critical threshold.

    // setup column resizing policy: all column except the second (stretched) resize to its content.
    QHeaderView *header = mTableView->horizontalHeader();
    header->setSectionsMovable(false);
    header->setSectionResizeMode(QHeaderView::ResizeToContents);
    header->setSectionResizeMode(2, QHeaderView::Stretch);

    // view dnd
    mTableView->setDragEnabled(true);
    mTableView->setDropIndicatorShown(true);
    mTableView->setDefaultDropAction(Qt::MoveAction);
    mTableView->setDragDropMode(QTableView::InternalMove);
    mTableView->setDragDropOverwriteMode(false);
    DnDViewStyle *tvStyle = new DnDViewStyle;
    tvStyle->setParent(this);
    mTableView->setStyle(tvStyle);

    QToolBar *toolBar = new QToolBar(this);
    mActionAdd = new QAction(QIcon::fromTheme("add"), "", this);

    mActionDuplicate = new QAction(QIcon::fromTheme("duplicate"), "", this);
    mActionDelete = new QAction(QIcon::fromTheme("remove"), "", this);

    mActionDuplicate->setEnabled(false);
    mActionDelete->setEnabled(false);

    toolBar->addAction(mActionAdd);
    toolBar->addAction(mActionDuplicate);
    toolBar->addAction(mActionDelete);

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->addWidget(toolBar, 0);
    layout->addWidget(mTableView);

    QObject::connect(mActionAdd, &QAction::triggered, this, &IndicatorsEditor::addIndicator);
    QObject::connect(mActionDuplicate, &QAction::triggered, this, &IndicatorsEditor::duplicateIndicator);
    QObject::connect(mActionDelete, &QAction::triggered, this, &IndicatorsEditor::deleteIndicator);
    QObject::connect(mTableView->selectionModel(), &QItemSelectionModel::selectionChanged, this,
                     &IndicatorsEditor::onSelectionChanged);
    translateUI();
}

/// @brief Destructor.
IndicatorsEditor::~IndicatorsEditor() {}

/// @brief Translate widgets texts.
void IndicatorsEditor::translateUI() {
    QString msg;
    msg = tr("Add a new indicator.");
    mActionAdd->setToolTip(msg);
    mActionAdd->setStatusTip(msg);
    msg = tr("Duplicate the selected indicator.");
    mActionDuplicate->setToolTip(msg);
    mActionDuplicate->setStatusTip(msg);
    msg = tr("Remove the selected indicator.");
    mActionDelete->setToolTip(msg);
    mActionDelete->setStatusTip(msg);
}

/**
 * @brief Set a new model to edit in the view.
 * @param model Model to edit.
 * @details
 * If \a model is null, then the view is reset with an empty model and this widget is disabled.
 */
void IndicatorsEditor::setIndicatorModel(IndicatorModel *model) {
    QItemSelectionModel *ism = mTableView->selectionModel();
    if (ism)
        ism->clear();
    if (model) {
        mItemModel = model;
        mTableView->setModel(model);
        setEnabled(true);
    } else {
        mItemModel = nullptr;
        mTableView->setModel(mEmptyModel);
        setEnabled(false);
    }
    if (ism) {
        QObject::disconnect(ism, &QItemSelectionModel::selectionChanged, this, &IndicatorsEditor::onSelectionChanged);
        delete ism;
    }
    ism = mTableView->selectionModel();
    if (ism)
        QObject::connect(ism, &QItemSelectionModel::selectionChanged, this, &IndicatorsEditor::onSelectionChanged);
}

/// @brief Set new parameters
void IndicatorsEditor::setParameters(const QVector<AParam> &parameters) { mParameters = parameters; }

/**
 * @brief Add a new item.
 * @details
 * The new item is constructed from user choice (see ItemChoiceDialog).
 */
void IndicatorsEditor::addIndicator() {
    if (mParameters.isEmpty() || !mItemModel)
        return;
    bool ok;
    IndicatorChoicePair pair = IndicatorChoiceDialog::getIndicator(this, mParameters, -1, BaseIndicator::BASIC, &ok);
    AParam &param = pair.second;
    if (!(ok && param.isValid()))
        return;
    BaseIndicator *ptr = nullptr;
    switch (pair.first) {
    case BaseIndicator::BASIC:
        ptr = new BasicIndicator(param, QString::fromStdString(param.name()));
        break;
    case BaseIndicator::SWITCH:
        ptr = new SwitchIndicator(param, QString::fromStdString(param.name()));
        break;
    case BaseIndicator::RANGE:
        ptr = new RangeIndicator(param, QString::fromStdString(param.name()));
        break;
    default:
        return;
    }
    QModelIndex index = mItemModel->addIndicator(ptr);
    if (index.isValid()) {
        QModelIndex sibling = index.sibling(index.row(), 2); // index of alias.
        mTableView->setCurrentIndex(sibling);
        mTableView->scrollTo(sibling);
        mTableView->edit(sibling);
    }
}

/// @brief Duplicate the selected item.
void IndicatorsEditor::duplicateIndicator() {
    if (!mItemModel)
        return;
    QModelIndex index = mItemModel->duplicateIndicator(selectedIndex(), true);
    if (index.isValid()) {
        QModelIndex sibling = index.sibling(index.row(), 2); // index of alias.
        mTableView->setCurrentIndex(sibling);
        mTableView->scrollTo(sibling);
        mTableView->edit(sibling); // column 1: Item's alias !
    }
}

/// @brief Delete selected item.
void IndicatorsEditor::deleteIndicator() {
    if (!mItemModel)
        return;
    mItemModel->removeIndicator(selectedIndex());
    mTableView->selectionModel()->clear();
}

/// @brief Manage actions state on view selection changes.
void IndicatorsEditor::onSelectionChanged(const QItemSelection &, const QItemSelection &) {
    QModelIndex idx = selectedIndex();
    bool isValid = idx.isValid();
    mActionDuplicate->setEnabled(isValid);
    mActionDelete->setEnabled(isValid);
}

/// @brief Get the model index that is currently selected in the view.
QModelIndex IndicatorsEditor::selectedIndex() const {
    auto *m = mTableView->selectionModel();
    if (m->hasSelection()) {
        QModelIndex idx = m->selectedIndexes().first();
        return idx.sibling(idx.row(), 0);
    } else {
        return QModelIndex();
    }
}
