// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "indicatorswidget.hpp"
#include "colorindicator.hpp"
#include "indicatorwidget.hpp"
#include "itemmapper.hpp"

IndicatorsWidget::IndicatorsWidget(QWidget *parent) : QScrollArea(parent) {

    mMapper = nullptr;
    mContainer = new QWidget(this);
    mContainer->setMinimumWidth(80);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    mLayout = new QVBoxLayout(mContainer);
    mStretchItem = new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Expanding);
    mLayout->addSpacerItem(mStretchItem);
    mLayout->setContentsMargins(6, 0, 6, 0);
    mLayout->setSpacing(0);
    setWidget(mContainer);
    setFrameShape(QFrame::Shape::NoFrame);
    setMinimumHeight(32);
}

/// @brief Translate widgets texts.
void IndicatorsWidget::translateUI() { return; }

/**
 * @brief Set the key mapper.
 * @details
 * The method registers the given mapper in the widget. Without a (consistent) mapper, the widget cannot display anything.
 */
void IndicatorsWidget::setMapper(ItemMapper *mapper) { mMapper = mapper; }

/**
 * @brief Reset the UI to its default state.
 * @details
 * The method resets the visual appareance of the widget to its default state: that is set grey led for all indicators with a
 * N/A
 * value.
 */
void IndicatorsWidget::reset() {
    double nan = qQNaN();
    for (IndicatorWidget *w : mWidgets)
        w->setValue(nan);
}

/**
 * @brief Initialize indicators widgets.
 * @param indicators Indicator to visualize.
 * @attention
 * Neither the container nor its children take ownership of the given indicators. Indicators are registered in the widgets as
 * QPointer.
 */
void IndicatorsWidget::setupIndicators(const QVector<BaseIndicatorPointer> &indicators) {

    int nt = indicators.size();
    bool removed = mWidgets.size() > nt;

    setUpdatesEnabled(false);

    mLayout->removeItem(mStretchItem);

    // remove extras
    while (mWidgets.size() > nt) {
        IndicatorWidget *w = mWidgets.last();
        mLayout->removeWidget(w);
        w->deleteLater();
        mWidgets.removeLast();
    }
    if (removed)
        mLayout->invalidate();
    // add missing
    while (mWidgets.size() < nt) {
        IndicatorWidget *w = new IndicatorWidget(mContainer);
        mLayout->addWidget(w, 0, Qt::AlignTop);
        mWidgets.push_back(w);
    }

    mLayout->addSpacerItem(mStretchItem);
    // set the indicators
    for (int i = 0; i < nt; i++) {
        IndicatorWidget *w = mWidgets[i];
        w->setIndicator(indicators[i]);
    }
    mContainer->adjustSize();

    // qDebug() << mLayout->totalMinimumSize() << mLayout->totalSizeHint();

    setUpdatesEnabled(true);
    setWidgetResizable(true);
}

/// @brief Refresh the values displayed in each indicator widget.
void IndicatorsWidget::refreshValues() {
    for (IndicatorWidget *w : mWidgets)
        w->refreshValues();
}

/**
 * @brief Update indicator widgets from the registered mapper.
 * @details
 * If no mapper has been registered, the method does nothing !
 */
void IndicatorsWidget::updateFromMapper() {
    if (mMapper.isNull()) {
        qDebug() << "[Indicators widget]: No key mapper has been registered.";
        return;
    }
    for (IndicatorWidget *w : mWidgets) {
        QPointer<BaseIndicator> ind = w->indicator();
        if (!ind.isNull()) {
            AParam &p = ind->parameter();
            p.value = mMapper->parameter(QString::fromStdString(p.name()), p.units).floatValue();
        } else {
            // should I destroy the widget here ?
            // In the current algorithm, the widget is set hidden if the indicator is null which means that it has been
            // destroyed: i.e. removed from the scenario's indicator model.
        }
        w->refreshValues();
    }
}
