// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "indicatorwidget.hpp"
#include "colorindicator.hpp"

/**
 * @brief Constructor
 * @param parent Optional parent widget.
 */
IndicatorWidget::IndicatorWidget(QWidget *parent) : QWidget(parent) {
    setup();
    setIndicator(nullptr);
}

/**
 * @brief Constructor
 * @param indicator Indicator to set at initialization.
 * @param parent Optional parent widget.
 */
IndicatorWidget::IndicatorWidget(BaseIndicator *indicator, QWidget *parent) : QWidget(parent) {
    setup();
    setIndicator(indicator);
}

/// @brief Destructor.
IndicatorWidget::~IndicatorWidget() {}

/// @brief Perform initialization steps common to all constructors.
void IndicatorWidget::setup() {
    mIndicator = nullptr;
    mIndicatorLabel = new ColorIndicator(this);
    mNameLabel = new QLabel(this);
    mValueLabel = new QLabel(this);
    mNameLabel->setAlignment(Qt::AlignLeft | Qt::AlignVCenter);
    mValueLabel->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
    mLayout = new QHBoxLayout(this);
    mLayout->addWidget(mIndicatorLabel, 0, Qt::AlignLeft);
    mLayout->addWidget(mNameLabel, 0, Qt::AlignLeft);
    mLayout->addWidget(mValueLabel, 2, Qt::AlignRight);
    mLayout->setContentsMargins(0, 6, 0, 6);
    setMinimumWidth(60);
    mNameLabel->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
    mValueLabel->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
}

/**
 * @brief Set a new indicator.
 * @param indicator Pointer to the indicator to set.
 * @return The previous indicator (may be null).
 * @details
 * The method sets a new indicator. If a null pointer is given, the widget is set hidden.
 */
BaseIndicatorPointer IndicatorWidget::setIndicator(BaseIndicator *indicator) {
    BaseIndicatorPointer old = mIndicator;
    mIndicator = indicator;
    if (mIndicator) {
        mIndicatorLabel->setColor(StatusColor::Grey);
        mNameLabel->setText(mIndicator->alias());
        mValueLabel->setText("N/A");
        BaseIndicator::Type type = mIndicator->type();
        setColorTipVisible(type != BaseIndicator::BASIC);
        setValueVisible(type != BaseIndicator::SWITCH);
    }
    setVisible(mIndicator != nullptr);
    return old;
}

/// @brief Get the current indicator.
BaseIndicatorPointer IndicatorWidget::indicator() const { return mIndicator; }

/// @brief Show/Hide the indicator color tip.
void IndicatorWidget::setColorTipVisible(bool visible) { mIndicatorLabel->setVisible(visible); }

/// @brief Check whether or not the indicator color tip is visible.
bool IndicatorWidget::colorTipVisible() const { return mIndicatorLabel->isVisible(); }

/// @brief Show/Hide the indicator value.
void IndicatorWidget::setValueVisible(bool visible) {
    mValueLabel->setVisible(visible);
    mLayout->setStretch(1, visible ? 0 : 2);
    mLayout->setStretch(2, visible ? 2 : 0);
}

/// @brief Check whether or not the indicator value is visible.
bool IndicatorWidget::valueVisible() const { return mValueLabel->isVisible(); }

/// @brief Refresh displayed values.
void IndicatorWidget::refreshValues() {
    if (!mIndicator) {
        setVisible(false);
        return;
    }
    setVisible(true);
    mIndicatorLabel->setColor(mIndicator->valueStatusColor());
    mValueLabel->setText(mIndicator->valueAndUnits());
}

/**
 * @brief Update the widget with the value stored in the given block.
 * @details
 * The method is convinient wrapper to IndicatorWidget::setValue that extract the value to process directly
 * from the block using the registered indicator's parameter.
 */
void IndicatorWidget::updateFromBlock(const ABlock &block) {
    setValue(mIndicator ? block(mIndicator->parameter(), true) : qQNaN());
}

/**
 * @brief Update the widget based on the given value.
 * @param value Value to process.
 * @details
 * If the widget has no registered Indicator, then the method hides the widget and returns immediately.
 * Otherwise, the method updates the color and texts of the widget accordingly to the given value.
 *
 * If the indicator is a switch, the value is set to "ON" or "OFF" according to the status color (ON for Green).
 *
 * @attention
 * The method does not update the internal value of the indicator but uses its methods to retrieve the color and
 * text that should be associated to the given value.
 */
void IndicatorWidget::setValue(double value) {

    if (!mIndicator) {
        setVisible(false);
        return;
    }
    setVisible(true);

    StatusColor color = mIndicator->statusColor(value);
    QString text = mIndicator->valueToString(value);

    mIndicatorLabel->setColor(color);
    mValueLabel->setText(text);
}
