// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "inversionstatuswidget.hpp"

InversionStatusWidget::InversionStatusWidget(QWidget *parent) : QWidget(parent) {

    mLblIPFDescr = new QLabel(this);
    mLblIPFValue = new QLabel("0.0 Hz", this);
    QHBoxLayout *hlayout = new QHBoxLayout;
    hlayout->addWidget(mLblIPFDescr, 2, Qt::AlignLeft);
    hlayout->addWidget(mLblIPFValue, 0, Qt::AlignRight);

    mIndicatorsLayout = new QVBoxLayout;
    mStretchItem = new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Expanding);
    mIndicatorsLayout->addSpacerItem(mStretchItem);
    mIndicatorsLayout->setContentsMargins(0, 0, 0, 0);
    mIndicatorsLayout->setSpacing(0);

    QVBoxLayout *layout = new QVBoxLayout(this);

    layout->addLayout(hlayout);
    layout->addLayout(mIndicatorsLayout);

    reset();
    translateUI();
}

/// @brief Translate widgets texts.
void InversionStatusWidget::translateUI() {
    setWindowTitle(tr("Data Inversion"));
    mLblIPFDescr->setText(tr("Process frequency"));
}

void InversionStatusWidget::setInversionUpdate(bool enable) {
    if (!enable) {
        reset();
    }
    setEnabled(enable);
}

/**
 * @brief Update the inversion frequency calculation.
 * @details
 * Upon calling the method computes/updates the inversion frequency using the time elapsed between successive call of
 * the function.
 *
 * A reset timer is also (re)started with an interval of 2 seconds. On timeout, the method
 * InversionStatusWidget::resetInversionUpdate is called which reset frequency value.
 */
void InversionStatusWidget::updateInversion(double lastBPS, const QVector<FitReport> &reports) {
    // compute approx inversion frequency.
    if (mRateTimer.isValid()) {
        mInvRate.push(qMin(1000.0 / mRateTimer.restart(), lastBPS));
    } else {
        mInvRate.reset();
        mRateTimer.start();
    }
    for (int i = 0; i < mFitIdentifiers.size(); i++) {
        const QPair<int, int> &p = mFitIdentifiers.at(i);
        BaseIndicator *ind = mFitIndicators.at(i);
        auto it = std::find_if(reports.begin(), reports.end(), [p](const FitReport &report) {
            return p.first == report.spectraId && p.second == report.lineId;
        });
        ind->setValue((it == reports.end()) ? qQNaN() : it->status);
    }
    for (IndicatorWidget *w : qAsConst(mIndicatorWidgets))
        w->refreshValues();
}

/**
 * @brief Reset the widget.
 * @details
 * The method resets the inversion frequency calculations and the UI.
 */
void InversionStatusWidget::reset() {
    mInvRate.reset();
    mRateTimer.invalidate();
    mLblIPFValue->setText(QString("~%1 Hz").arg(mInvRate(), 0, 'f', 0));
    for (BaseIndicator *ind : qAsConst(mFitIndicators)) {
        ind->setValue(qQNaN());
    }
    for (IndicatorWidget *w : qAsConst(mIndicatorWidgets))
        w->refreshValues();
}

void InversionStatusWidget::refreshText() {
    double rate = mInvRate();
    mLblIPFValue->setText(QString("~%1 Hz").arg(rate, 0, 'f', (rate < 1) ? 1 : 0));
}

/**
 * @brief Setup indicators from inversion process configuration.
 * @param config Inversion configuration.
 */
void InversionStatusWidget::setConfiguration(const FitConfig &config) {
    // delete previous indicators.
    mFitIdentifiers.clear();
    qDeleteAll(mFitIndicators);
    mFitIndicators.clear();
    QVector<BaseIndicatorPointer> qptrs;

    // build indicators
    const std::vector<SParam> &spcs = config.spectras();
    qptrs.reserve(spcs.size());
    for (const SParam &sp : spcs) {
        int sid = sp.id;
        int lid = sp.line.id;
        QString name = QString::fromStdString(sp.alias).simplified(), suffix = QString("(S#%2,L#%3)").arg(sid).arg(lid);
        if (!name.contains(suffix))
            name = QString("%1 %2").arg(name, suffix);
        AParam prm(name.toStdString(), qQNaN(), Units::NOUNIT, StringList(), tr("Spectra fit process status.").toStdString());
        SwitchIndicator *ind = new SwitchIndicator(prm, name, true, this);
        mFitIdentifiers.push_back(qMakePair(sid, lid));
        mFitIndicators.push_back(ind);
        qptrs.push_back(ind);
    }

    // build indicators
    int nt = qptrs.size();
    bool removed = mIndicatorWidgets.size() > nt;

    setUpdatesEnabled(false);
    mIndicatorsLayout->removeItem(mStretchItem);
    // remove extras
    while (mIndicatorWidgets.size() > nt) {
        IndicatorWidget *w = mIndicatorWidgets.last();
        mIndicatorsLayout->removeWidget(w);
        w->deleteLater();
        mIndicatorWidgets.removeLast();
    }
    if (removed)
        mIndicatorsLayout->invalidate();
    // add missing
    while (mIndicatorWidgets.size() < nt) {
        IndicatorWidget *w = new IndicatorWidget(this);
        mIndicatorsLayout->addWidget(w, 0, Qt::AlignTop);
        mIndicatorWidgets.push_back(w);
    }
    mIndicatorsLayout->addSpacerItem(mStretchItem);
    // set the indicators
    for (int i = 0; i < nt; i++) {
        IndicatorWidget *w = mIndicatorWidgets[i];
        w->setIndicator(mFitIndicators[i]);
    }
    setUpdatesEnabled(true);
}
