// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef ITEMMODEL_HPP
#define ITEMMODEL_HPP

#include "globals.hpp"

class ItemMapper;
class DataItem;

typedef QPointer<DataItem> DataItemPointer;

class ItemModel : public QAbstractTableModel {
    Q_OBJECT

  public:
    ItemModel();

  public:
    static const int mapColumn = 6;                   ///< @brief Index of the map column in the model.
    static const int plotColumnStart = mapColumn + 1; ///< @brief Starting index of the plot columns in the model.

    explicit ItemModel(QObject *parent = nullptr);
    ItemModel(const QVector<DataItem *> &items, QObject *parent = nullptr);
    ~ItemModel();
    void translateHeaders();

    virtual Qt::ItemFlags flags(const QModelIndex &index) const override;
    virtual QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const override;
    virtual bool setData(const QModelIndex &index, const QVariant &value, int role = Qt::EditRole) override;
    virtual QVariant headerData(int section, Qt::Orientation orientation, int role = Qt::DisplayRole) const override;
    virtual int rowCount(const QModelIndex &parent = QModelIndex()) const override;
    virtual int columnCount(const QModelIndex &parent = QModelIndex()) const override;

    DataItemPointer mapItem() const;
    DataItemPointer plotItem(int index) const;

    int mapItemRowIndex() const;
    int plotItemRowIndex(int index) const;

    QVector<DataItemPointer> items() const;
    DataItemPointer item(QModelIndex index) const;

    void addItems(QVector<DataItem *> items, QVector<DataItem *> &rejected);
    QModelIndex addItem(DataItem *item);
    QModelIndex duplicateItem(DataItem *item, bool uniqueName = false);
    QModelIndex duplicateItem(QModelIndex index, bool uniqueName = false);
    bool removeItem(const QModelIndex &index);

    QMenu *plotContextMenu(int plotIndex, QMenu *parentMenu);
    QMenu *mapContextMenu(QMenu *parentMenu);

    void reOrderItem(int oldIndex, int newIndex);
    void sortItems(const QVector<int> &order);

    void normalizeRanges();
    void resetValues();
    void registerValues(const ABlock &block);
    void registerValues(const ItemMapper *mapper);

    QByteArray toXML() const;

    virtual Qt::DropActions supportedDropActions() const override;
    virtual Qt::DropActions supportedDragActions() const override;
    virtual QStringList mimeTypes() const override;
    virtual QMimeData *mimeData(const QModelIndexList &indexes) const override;
    virtual bool canDropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column,
                                 const QModelIndex &parent) const override;
    virtual bool dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column,
                              const QModelIndex &parent) override;

  signals:

    /**
     * @brief Emitted when an item has been added.
     * @param item Added item.
     */
    void itemAdded(DataItemPointer item);

    /**
     * @brief Emitted when an item is about to be removed.
     * @param item Removed item.
     * @details
     * The signal is emitted when DataItemModel::removeItem is called, right before the item is deleted by the model.
     */
    void itemAboutToBeRemoved(DataItemPointer item);

    /**
     * @brief Emitted when an item should be displayed on the map widget.
     * @param item Data item that should be displayed.
     * @details
     * The signal is emitted when the checkstate role of a model index (for the third column) has changed.
     */
    void itemOnMapRequested(DataItem *item);

    /**
     * @brief Emitted when an item should be displayed on a plot widget.
     * @param graphIndex Index of the plot widget to process.
     * @param item Data item that should be displayed.
     * @details
     * The signal is emitted when the checkstate role of a model index (for the last four columns) has changed.
     */
    void itemOnGraphRequested(int graphIndex, DataItem *item);

  private:
    QStringList mHeaders;
    QVector<DataItem *> mItems;
    DataItem *mMapItem;
    DataItem *mPlotItems[N_PLOTS];
};

class ItemROModel : public QSortFilterProxyModel {
    Q_OBJECT
  public:
    explicit ItemROModel(QObject *parent = nullptr) : QSortFilterProxyModel(parent) {}

    virtual Qt::ItemFlags flags(const QModelIndex &index) const override {
        auto *sm = sourceModel();
        if (!(sm && index.isValid()))
            return QSortFilterProxyModel::flags(index);
        QModelIndex si = mapToSource(index);
        Qt::ItemFlags f = sm->flags(si);
        f.setFlag(Qt::ItemIsSelectable, false);
        if (si.column() == 1)
            f.setFlag(Qt::ItemIsEditable, false);
        return f;
    }

  protected:
    virtual bool filterAcceptsColumn(int source_column, const QModelIndex &) const override {
        return source_column == 1 || source_column >= ItemModel::mapColumn;
    }
};

#endif // ITEMMODEL_HPP
