// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "itemseditor.hpp"
#include "dataitem.hpp"
#include "datarange.hpp"
#include "itemchoicedialog.hpp"
#include "itemmodel.hpp"

#include "delegates.hpp"
#include "dndviewstyle.hpp"
#include "tableview.hpp"

ItemsEditor::ItemsEditor(QWidget *parent)
    : QWidget(parent) {
    mEmptyModel = new ItemModel(this);
    mItemModel = nullptr;
    mTableView = new TableView(this);
    mTableView->setModel(mEmptyModel);

    mTableView->setSelectionBehavior(QTableView::SelectRows);
    mTableView->setSelectionMode(QTableView::SingleSelection);
    mTableView->setItemDelegateForColumn(1, new LineEditDelegate(false, mTableView));           // Alias.
    mTableView->setItemDelegateForColumn(2, new ComboboxUnitsDelegate(mTableView));             // Units.
    mTableView->setItemDelegateForColumn(3, new ValueLineEditDelegate(false, false, mTableView)); // Color range lower boundary.
    mTableView->setItemDelegateForColumn(4, new ValueLineEditDelegate(false, false, mTableView)); // Color range upper boundary.
    mTableView->setItemDelegateForColumn(5, new ValueLineEditDelegate(true, true, mTableView));   // Remanence threshold.
    int mapIndex = 6;
    int pltIndex = mapIndex + 1;
    mTableView->setItemDelegateForColumn(mapIndex, new BooleanDelegate(mTableView));
    for (int i = 0; i < N_PLOTS; i++) {
        mTableView->setItemDelegateForColumn(i + pltIndex, new BooleanDelegate(mTableView));
    }
    // setup column resizing policy: all column except the second (stretched) resize to its content.
    QHeaderView *header = mTableView->horizontalHeader();
    header->setSectionsMovable(false);
    header->setSectionResizeMode(QHeaderView::ResizeToContents);
    header->setSectionResizeMode(1, QHeaderView::Stretch);
    header->setHighlightSections(false);

    // view dnd
    mTableView->setDragEnabled(true);
    mTableView->setDropIndicatorShown(true);
    mTableView->setDefaultDropAction(Qt::MoveAction);
    mTableView->setDragDropMode(QTableView::InternalMove);
    mTableView->setDragDropOverwriteMode(false);
    DnDViewStyle *tvStyle = new DnDViewStyle;
    tvStyle->setParent(this);
    mTableView->setStyle(tvStyle);

    QToolBar *toolBar = new QToolBar(this);
    mActionAdd = new QAction(QIcon::fromTheme("add"), "", this);
    mActionDuplicate = new QAction(QIcon::fromTheme("duplicate"), "", this);
    mActionDelete = new QAction(QIcon::fromTheme("remove"), "", this);

    mActionDuplicate->setEnabled(false);
    mActionDelete->setEnabled(false);

    toolBar->addAction(mActionAdd);
    toolBar->addAction(mActionDuplicate);
    toolBar->addAction(mActionDelete);

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->addWidget(toolBar, 0);
    layout->addWidget(mTableView);

    QObject::connect(mActionAdd, &QAction::triggered, this, &ItemsEditor::addItem);
    QObject::connect(mActionDuplicate, &QAction::triggered, this, &ItemsEditor::duplicateItem);
    QObject::connect(mActionDelete, &QAction::triggered, this, &ItemsEditor::deleteItem);
    translateUI();
}

/// @brief Destructor.
ItemsEditor::~ItemsEditor() {}

/// @brief Translate widgets texts.
void ItemsEditor::translateUI() {
    QString msg;
    msg = tr("Add a new parameter.");
    mActionAdd->setToolTip(msg);
    mActionAdd->setStatusTip(msg);
    msg = tr("Duplicate the selected parameter.");
    mActionDuplicate->setToolTip(msg);
    mActionDuplicate->setStatusTip(msg);
    msg = tr("Remove the selected parameter.");
    mActionDelete->setToolTip(msg);
    mActionDelete->setStatusTip(msg);
}

/**
 * @brief Set a new model to edit in the view.
 * @param model Model to edit.
 * @details
 * If \a model is null, then the view is reset with an empty model and this widget is disabled.
 */
void ItemsEditor::setItemModel(ItemModel *model) {
    QItemSelectionModel *ism = mTableView->selectionModel();
    if (ism)
        ism->clear();
    if (model) {
        mItemModel = model;
        mTableView->setModel(model);
        setEnabled(true);
    } else {
        mItemModel = nullptr;
        mTableView->setModel(mEmptyModel);
        setEnabled(false);
    }
    if (ism) {
        QObject::disconnect(ism, &QItemSelectionModel::selectionChanged, this, &ItemsEditor::onSelectionChanged);
        delete ism;
    }
    ism = mTableView->selectionModel();
    if (ism)
        QObject::connect(ism, &QItemSelectionModel::selectionChanged, this, &ItemsEditor::onSelectionChanged);
}

/// @brief Set new parameters
void ItemsEditor::setParameters(const QVector<AParam> &parameters) {
    mParameters = parameters;
}

/**
 * @brief Add a new item.
 * @details
 * The new item is constructed from user choice (see ItemChoiceDialog).
 */
void ItemsEditor::addItem() {
    if (mParameters.isEmpty() || !mItemModel)
        return;
    bool ok;
    AParam param = ItemChoiceDialog::getParameter(this, mParameters, -1, &ok);
    if (!(ok && param.isValid()))
        return;
    QModelIndex index = mItemModel->addItem(new DataItem(param, QString::fromStdString(param.name()), DataRange(0, 1), qQNaN()));
    if (index.isValid()) {
        mTableView->setCurrentIndex(index);
        mTableView->scrollTo(index);
        mTableView->edit(index.sibling(index.row(), 1));
    }
}

/// @brief Duplicate the selected item.
void ItemsEditor::duplicateItem() {
    if (!mItemModel)
        return;
    QModelIndex index = mItemModel->duplicateItem(selectedIndex(), true);
    if (index.isValid()) {
        mTableView->setCurrentIndex(index);
        mTableView->scrollTo(index);
        mTableView->edit(index.sibling(index.row(), 1)); // column 1: Item's alias !
    }
}

/// @brief Delete selected item.
void ItemsEditor::deleteItem() {
    if (!mItemModel)
        return;
    if (mItemModel->removeItem(selectedIndex()))
        mTableView->selectionModel()->clear();
}

/// @brief Manage actions state on view selection changes.
void ItemsEditor::onSelectionChanged(const QItemSelection &, const QItemSelection &) {
    QModelIndex idx = selectedIndex();
    bool isValid = idx.isValid();
    mActionDuplicate->setEnabled(isValid);
    mActionDelete->setEnabled(isValid);
}

/// @brief Get the model index that is currently selected in the view.
QModelIndex ItemsEditor::selectedIndex() const {
    auto *m = mTableView->selectionModel();
    if (m->hasSelection()) {
        QModelIndex idx = m->selectedIndexes().first();
        return idx.sibling(idx.row(), 0);
    } else {
        return QModelIndex();
    }
}
