// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "layoutmanager.hpp"
#include "mainwindow.hpp"
#include "mapwidget.hpp"
#include <QApplication>
#include <QPixmap>
#include <QScreen>
#include <QWindow>

static const QByteArray noImageB64 =
    QByteArray::fromBase64("iVBORw0KGgoAAAANSUhEUgAAAoAAAAHgAQMAAAAPH06nAAAABlBMVEX///"
                           "8AAABVwtN+"
                           "AAAACXBIWXMAAA7EAAAOxAGVKw4bAAAAPUlEQVR4nO3BAQ0AAADCoPdPbQ43oAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
                           "AAAAAAAAAAAAAAAAAAADgzwCX4AAB9Dl2RwAAAABJRU5ErkJggg==");

/**
 * @brief Layout constructor.
 * @param name Name of the layout.
 * @param state Optional global layout's state as a byte array.
 * @param splitterState Optional (map widget) splitter state as a byte array.
 * @param image Optional layout screenshot.
 * @details
 *
 * All members of the Layout cannot be modified after intialization.
 */
Layout::Layout(const QString &name, const QByteArray &state, const QByteArray &splitterState, const QPixmap &image)
    : mName(name), mStateData(state), mSplitterData(splitterState), mScreenShot(image) {}

Layout::Layout(const Layout &other) {
    mName = other.mName;
    mStateData = other.mStateData;
    mSplitterData = other.mSplitterData;
    mScreenShot = other.mScreenShot;
}

/// @brief Copy assignment operator.
Layout &Layout::operator=(const Layout &other) {
    if (this == &other)
        return *this;
    mName = other.mName;
    mStateData = other.mStateData;
    mSplitterData = other.mSplitterData;
    mScreenShot = other.mScreenShot;
    return *this;
}

/// @brief Equality operator: Compare Layout names (case-sensitive).
bool Layout::operator==(const Layout &other) { return mName.compare(other.mName) == 0; }

/// @brief Less than operator: Compare Layout names (case-sensitive).
bool Layout::operator<(const Layout &other) const { return mName.compare(other.mName) < 0; }

/// @brief Set the name of the layout.
void Layout::setName(const QString &name) { mName = name; }

/// @brief Set the state of the layout.
void Layout::setState(const QByteArray &array) { mStateData = array; }

/// @brief Set the (map widget) splitter state of the layout.
void Layout::setSplitterState(const QByteArray &array) { mSplitterData = array; }

/// @brief Set the screenshot of the layout.
void Layout::setScreenShot(const QPixmap &image) { mScreenShot = image; }

/// @brief Get the name of the layout.
const QString &Layout::name() const { return mName; }

/// @brief Get the (map widget) splitter state of the layout (may be empty).
const QByteArray &Layout::splitterState() const { return mSplitterData; }

/// @brief Get the screenshot of the layout.
const QPixmap &Layout::screenShot() const { return mScreenShot; }

/// @brief Get the state of the layout (may be empty).
const QByteArray &Layout::state() const { return mStateData; }

/**
 * @brief LayoutManager constructor
 * @param window QMainWindow to manage.
 */
LayoutManager::LayoutManager(MainWindow *window, QObject *parent) : QAbstractListModel(parent), mWindow(window) {}

/**
 * @brief Set new window to manage.
 * @param window Window to manage.
 */
void LayoutManager::setWindow(MainWindow *window) { mWindow = window; }

/// @brief Get the pointer to managed MainWindow.
MainWindow *LayoutManager::window() const { return mWindow.data(); }

/// @brief Get the list of the layouts names available in the manager.
QStringList LayoutManager::layoutNames() const {
    QStringList out;
    out.reserve(mLayouts.size());
    for (const Layout &layout : qAsConst(mLayouts))
        out.append(layout.name());
    return out;
}

/// @brief Clear all layouts in the model.
void LayoutManager::clearLayouts() {
    beginResetModel();
    mLayouts.clear();
    mLayouts.shrink_to_fit();
    endResetModel();
}

void LayoutManager::addLayouts(const QVector<Layout> &layouts) {
    int nl = layouts.size();
    if (!nl)
        return;
    beginInsertRows(QModelIndex(), rowCount(), rowCount() + nl);
    mLayouts.reserve(mLayouts.size() + nl);
    for (const Layout &layout : layouts)
        mLayouts.push_back(layout);
    endInsertRows();
}

/**
 * @brief Load layout at given index.
 * @param idx (row) Index of the layout to load.
 * @return True on success, false otherwise.
 * @details
 * If @e idx is valid, the method restores the registered main window from the state stored in the layout at given
 * index.
 */
bool LayoutManager::loadLayout(int idx) {
    if (idx < 0 || idx >= mLayouts.size() || mWindow.isNull())
        return false;
    const Layout &layout = mLayouts[idx];
    mWindow->restoreState(layout.state());
    const QByteArray &splitter = layout.splitterState();
    if (splitter.size())
        mWindow->mapWidget()->setSplitterState(splitter);
    return true;
}

/**
 * @overload
 * @brief Load layout by name.
 * @param name Name of the layout to load.
 * @return True on success, false otherwise.
 * @details
 * A convinient wrapper to LayoutManager::loadLayout(LayoutManager::indexOf(name));
 */
bool LayoutManager::loadLayout(const QString &name) { return loadLayout(indexOf(name)); }

/**
 * @brief Get the (row) index of the first layout matching <i>name</i> in the model.
 * @param name Name of the layout to search.
 * @return Index of the layout or -1 if no layout matches the given name.
 */
int LayoutManager::indexOf(const QString &name) {
    auto it = std::find_if(mLayouts.cbegin(), mLayouts.cend(),
                           [=](const Layout &layout) { return name.compare(layout.name()) == 0; });
    if (it == mLayouts.cend())
        return -1;
    return std::distance(mLayouts.cbegin(), it);
}

/**
 * @brief Remove layout with given name.
 * @param name Name of the layout to remove.
 * @return True on success, false otherwise.
 * @details
 * If a layout is actually removed, the method emits LayoutManager::layoutRemoved signal.
 */
bool LayoutManager::removeLayout(const QString &name) {
    int idx = indexOf(name);
    if (idx == -1)
        return false;
    beginRemoveRows(QModelIndex(), idx, idx);
    mLayouts.removeAt(idx);
    endRemoveRows();
    emit layoutRemoved(name);
    return true;
}

bool LayoutManager::removeRows(int row, int count, const QModelIndex &parent) {
    if ((row + count) > mLayouts.size())
        return false;
    beginRemoveRows(parent, row, row + count - 1);
    mLayouts.erase(mLayouts.begin() + row, mLayouts.begin() + row + count);
    endRemoveRows();
    return true;
}

///  @brief Get the item flags for the given index.
Qt::ItemFlags LayoutManager::flags(const QModelIndex &) const {
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable | Qt::ItemIsEditable;
}

/// @brief Get data from the model.
QVariant LayoutManager::data(const QModelIndex &index, int role) const {
    if (!index.isValid())
        return QVariant();
    int row = index.row();
    if (role == Qt::DisplayRole) {
        return mLayouts[row].name();
    } else if (role == StateRole) {
        return mLayouts[row].state();
    } else if (role == SplitterRole) {
        return mLayouts[row].splitterState();
    } else if (role == ScreenshotRole) {
        return QVariant::fromValue(mLayouts[row].screenShot());
    } else
        return QVariant();
}

/**
 * @brief Set data in the model.
 * @param index Index of data to set.
 * @param value Value to set.
 * @param role Role of the data to set.
 * @return True on success, false otherwise.
 * @details
 * The method allow to update the name of a Layout if the given data can be converted in string and if
 * the role is set to Qt::EditRole.
 * No check is done on the name then layouts with name duplicates can be stored using this function.
 * On success, the method emits LayoutManager::layoutRenamed signal with the old and new names.
 */
bool LayoutManager::setData(const QModelIndex &index, const QVariant &value, int role) {
    if (!index.isValid())
        return false;
    Layout &layout = mLayouts[index.row()];
    if (role == Qt::EditRole) {
        QString oldName = layout.name(), newName = value.toString();
        layout.setName(newName);
        emit layoutRenamed(oldName, newName);
    } else if (role == StateRole) {
        layout.setState(value.toByteArray());
    } else if (role == SplitterRole) {
        layout.setSplitterState(value.toByteArray());
    } else if (role == ScreenshotRole) {
        QPixmap pixmap = value.value<QPixmap>();
        qDebug() << pixmap.isNull();
        layout.setScreenShot(pixmap);
    } else
        return false;
    emit dataChanged(index, index, {Qt::DisplayRole});
    return true;
}

/// @brief Get the total number of rows in the model.
int LayoutManager::rowCount(const QModelIndex &) const { return mLayouts.size(); }

/**
 * @brief Save window's current layout.
 * @param name Name to set for the layout to save.
 * @return Index of the layout in the model. -1 is returned if the method fails to save the layout.
 * @details
 * The method retrieves the current state of the registered window and saves it in the manager.
 * If the @b name refers to an already registered layout, it is updated.
 * The method returns the (row) index of the layout in the model.
 */
int LayoutManager::saveLayout(const QString &name) {
    if (name.isEmpty() || mWindow.isNull())
        return -1;
    int idx = indexOf(name);
    QByteArray state = mWindow->saveState();
    QPixmap img = getScreenshot();
    QByteArray splitter = mWindow->mapWidget()->splitterState();
    if (idx == -1) {
        idx = mLayouts.size();
        beginInsertRows(QModelIndex(), idx, idx);
        mLayouts.push_back(Layout(name, state, splitter, img));
        endInsertRows();
    } else {
        mLayouts[idx].setState(state);
        mLayouts[idx].setScreenShot(img);
        mLayouts[idx].setSplitterState(splitter);
    }
    return idx;
}

/**
 * @brief Get a screenshot of the main window
 * @return Screenshot as a base64 encoded string.
 */
QPixmap LayoutManager::getScreenshot() {
    if (!mWindow)
        return QPixmap::fromImage(QImage::fromData(noImageB64, "PNG")).scaledToHeight(500);
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    QPixmap pixmap(mWindow->size());
    mWindow->render(&pixmap);
    if (pixmap.height() > 500)
        pixmap = pixmap.scaledToHeight(500, Qt::SmoothTransformation);
    QApplication::restoreOverrideCursor();
    return pixmap;
}

/// @brief Save the content of the manager to file.
bool LayoutManager::saveToFile(const QString &path) {
    QSettings settings(path, QSettings::IniFormat);
    settings.sync();
    if (settings.status() != QSettings::NoError)
        return false;
    settings.clear();
    settings.beginWriteArray("layouts", mLayouts.size());
    for (int i = 0; i < mLayouts.size(); ++i) {
        settings.setArrayIndex(i);
        settings.setValue("name", mLayouts[i].name());
        settings.setValue("state", mLayouts[i].state());
        settings.setValue("splitter", mLayouts[i].splitterState());
        settings.setValue("image", mLayouts[i].screenShot());
    }
    settings.endArray();
    return true;
}

/// @brief Parse layout XML file and returns a vector of Layouts.
QVector<Layout> LayoutManager::readFromFile(const QString &path, bool *ok) {
    if (ok)
        *ok = false;
    QVector<Layout> out;
    QSettings settings(path, QSettings::IniFormat);
    settings.sync();
    if (settings.status() != QSettings::NoError)
        return out;
    if (ok)
        *ok = true;
    int nl = settings.beginReadArray("layouts");
    out.reserve(nl);
    for (int i = 0; i < nl; ++i) {
        settings.setArrayIndex(i);
        QString name = settings.value("name").toString();
        QByteArray state = settings.value("state").toByteArray();
        QByteArray splitter = settings.value("splitter").toByteArray();
        QPixmap image = settings.value("image").value<QPixmap>();
        out.push_back(Layout(name, state, splitter, image));
    }
    settings.endArray();
    return out;
}
