// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef LAYOUTMANAGER_HPP
#define LAYOUTMANAGER_HPP

#include <QPixmap>
#include <QtCore>

class MainWindow;

class Layout {
  public:
    Layout() = default;
    Layout(const QString &name, const QByteArray &state = QByteArray(), const QByteArray &geometry = QByteArray(),
           const QPixmap &image = QPixmap());
    Layout(const Layout &other);
    Layout &operator=(const Layout &other);
    bool operator==(const Layout &other);
    bool operator<(const Layout &other) const;
    const QString &name() const;
    const QByteArray &state() const;
    const QByteArray &splitterState() const;
    const QPixmap &screenShot() const;
    void setName(const QString &name);
    void setState(const QByteArray &array);
    void setSplitterState(const QByteArray &array);
    void setScreenShot(const QPixmap &image);

  private:
    QString mName;
    QByteArray mStateData;
    QByteArray mSplitterData;
    QPixmap mScreenShot;
};

/**
 * @brief The LayoutManager class
 * @details
 * This class is intended to manage the state and geometry of a QMainWindow.
 *
 * It stores a map of Layout instances which themselves store both geometry and states.
 *
 * The manager allows to save/load layouts to/from XML formatted files.
 */
class LayoutManager : public QAbstractListModel {
    Q_OBJECT
  public:
    enum LayoutRoles { StateRole = Qt::UserRole, SplitterRole, ScreenshotRole };

    explicit LayoutManager(MainWindow *window = nullptr, QObject *parent = nullptr);
    virtual ~LayoutManager() = default;
    void setWindow(MainWindow *window);
    MainWindow *window() const;
    bool saveToFile(const QString &path);
    static QVector<Layout> readFromFile(const QString &path, bool *ok = nullptr);

    QStringList layoutNames() const;

    void clearLayouts();
    void addLayouts(const QVector<Layout> &layouts);
    int saveLayout(const QString &name);

    bool removeLayout(const QString &name);

    virtual Qt::ItemFlags flags(const QModelIndex &index) const override;
    virtual QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const override;
    virtual bool setData(const QModelIndex &index, const QVariant &value, int role = Qt::EditRole) override;
    virtual int rowCount(const QModelIndex &parent = QModelIndex()) const override;

    virtual bool removeRows(int row, int count, const QModelIndex &parent = QModelIndex());

    int indexOf(const QString &name);

  public slots:
    bool loadLayout(const QString &name);
    bool loadLayout(int index);

  signals:
    void layoutRemoved(const QString &name);
    void layoutAdded(const QString &name);
    void layoutRenamed(const QString &oldName, const QString &newName);

  private:
    QPointer<MainWindow> mWindow = nullptr;
    QVector<Layout> mLayouts;
    QPixmap getScreenshot();
};

#endif // LAYOUTMANAGER_HPP
