// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "layoutnamedialog.hpp"
#include <QCompleter>
#include <QVBoxLayout>
/**
 * @brief Constructor.
 * @param initialName Initial name to display in the dialog.
 * @param availableNames List of (already) available names.
 * @param parent Parent widget.
 * @param f Window flags.
 */
LayoutNameDialog::LayoutNameDialog(const QString &initialName, const QStringList &availableNames, QWidget *parent,
                                   Qt::WindowFlags f)
    : QDialog(parent, f) {
    setWindowTitle(tr("Layout name"));
    setMinimumWidth(300);
    setMaximumWidth(600);
    mButtonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel, this);
    mLblName = new QLabel(tr("Enter a unique name"), this);
    mLdtName = new QLineEdit(initialName, this);
    QCompleter *completer = new QCompleter(availableNames, mLdtName);
    completer->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
    mLdtName->setCompleter(completer);
    mLblInfos = new QLabel(
        tr("If the name already exist (i.e. taken from the popup list), afferent layout will be updated."), this);
    mLblInfos->setWordWrap(true);
    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->addWidget(mLblName);
    layout->addWidget(mLdtName);
    layout->addWidget(mLblInfos);
    layout->addStretch(1);
    layout->addWidget(mButtonBox);

    QObject::connect(mButtonBox, &QDialogButtonBox::accepted, this, &QDialog::accept);
    QObject::connect(mButtonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);
}

/// @brief Get the size hint of the widget.
QSize LayoutNameDialog::sizeHint() const {
    QSize size = minimumSizeHint();
    int &w = size.rwidth();
    w = qBound(300, w, 800);
    return size;
}

/// @brief Get the text value.
QString LayoutNameDialog::textValue() { return mLdtName->text(); }

/**
 * @brief Get a layout name.
 * @param initialName Initial name to set in the dialog.
 * @param availableNames List of available names for completion.
 * @param ok Optional control flag with, in output, true if the OK button was clicked, false otherwise.
 * @param parent Parent widget.
 * @return Name of the layout or an empty string if user canceled the dialog.
 */
QString LayoutNameDialog::getLayoutName(const QString &initialName, const QStringList &availableNames, bool *ok,
                                        QWidget *parent) {
    LayoutNameDialog *dialog = new LayoutNameDialog(initialName, availableNames, parent);
    const int ret = dialog->exec();
    QString out;
    if (ok)
        *ok = !!ret;
    if (ret) {
        out = dialog->textValue();
    }
    dialog->deleteLater();
    return out;
}
