// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "mainwindow.hpp"
#include "audpsocket.hpp"
#include "batterystatuswidget.hpp"
#include "clockwidget.hpp"
#include "commwidget.hpp"
#include "cpumemwidget.hpp"
#include "dataloaderdialog.hpp"
#include "datamanager.hpp"
#include "filenamewidget.hpp"
#include "fitcontroller.hpp"
#include "fitplotwidget.hpp"
#include "flightwidget.hpp"
#include "gpsstatuswidget.hpp"
#include "gpswidget.hpp"
#include "indicatorswidget.hpp"
#include "inversionstatuswidget.hpp"
#include "itemmapper.hpp"
#include "layoutmanager.hpp"
#include "layoutmanagerdialog.hpp"
#include "layoutnamedialog.hpp"
#include "layoutselector.hpp"
#include "mapwidget.hpp"
#include "mediacontrolwidget.hpp"
#include "rtplotwidget.hpp"
#include "scenarioseditor.hpp"
#include "scenarioselector.hpp"
#include "scenariowidget.hpp"
#include "screenshotmanager.hpp"
#include "settings/settingsdialog.hpp"
#include "timestampwidget.hpp"
#include "udpstatuswidget.hpp"
#include "wlanwidget.hpp"
#include "xmlutils.hpp"

#include "ArcGISRuntimeEnvironment.h"

static QVector<PairOfFiles> checkAmulseFiles(const QStringList &paths, ABlock *tpl = nullptr);

/**
 * @brief Check the list of files.
 * @param paths List of file paths to check.
 * @param tpl Optional block template to check against.
 * @return Vector of pairs with file path as first item and block template as second item for each consistent files.
 * @details
 * The method attempts to read informations (see getFileInfos) from each given path and sorts them by ascending order of
 * starting date. All files that are not compatible with the API are filtered out.
 *
 * Then it filters the list according to the given block template, if any, or from the template of the first
 * sorted file.
 */
static QVector<PairOfFiles> checkAmulseFiles(const QStringList &paths, ABlock *tpl) {
    QVector<PairOfFiles> tmps;
    // get all template.
    for (const QString &path : paths) {
        ABlock b = getFileInfos(path.toStdString(), false);
        if (b) {
            tmps.push_back(std::make_pair(path.toStdString(), b));
        } else
            qDebug() << QString("Cannot extract informations from {}. Skipping file.").arg(path);
    }
    if (tmps.isEmpty())
        return tmps;
    std::sort(tmps.begin(), tmps.end(), [=](PairOfFiles &left, PairOfFiles &right) {
        return left.second.time() < right.second.time();
    });

    QMutableVectorIterator<PairOfFiles> it(tmps);

    if (!tpl && tmps.size())
        tpl = &(tmps.first().second);

    // only one item: check the template and returns appropriate vector.
    if (tmps.size() == 1) {
        if (!tmps.first().second.comparable(*tpl)) {
            tmps.clear();
        }
        return tmps;
    }
    // several items: filter out inconsistent blocks.
    while (it.hasNext()) {
        PairOfFiles &b = it.next();
        if (!b.second.comparable(*tpl)) {
            it.remove();
        }
    }
    return tmps;
}

/// @brief Constructor.
MainWindow::MainWindow(const QString &configPath, QWidget *parent, Qt::WindowFlags flags) : QMainWindow(parent, flags) {

    QSettings settings;
    mShowFit = settings.value("monitoring/show_fit", false).toBool();
    mSaveData = settings.value("storage/save", true).toBool();
    mNewFileOnNewFlight = settings.value("storage/nfonf", true).toBool();
    mScreenshotManager = new ScreenshotManager(this);
    mScreenshotManager->setFolder(QString("%1/captures").arg(gDataDir));
    mScreenshotManager->setFileName("capture");

    mNoSDWarned = false;

    mResultFileHandler = nullptr;
    mFCParams = FCParameters(20, 0, 1e-8, 1e-8, 1e-10);

    mLastDataRootDir = "";
    mInversionEnabled = false;
    mInvalidTemplateReported = false;

    mUdpSocket = new AUDPSocket(this);
    mFitController = new FitController(this);

    mDataManager = new DataManager;

    mLayoutManager = new LayoutManager(this, this);
    {
        QVector<Layout> tmp = LayoutManager::readFromFile(gLayoutsFile);
        mLayoutManager->addLayouts(tmp);
    }

    mTabWidget = nullptr;
    mFlightWidget = nullptr;
    mLayoutSelector = nullptr;
#ifndef SCN_SEL_ON_MAP
    mScenarioSelector = nullptr;
#endif
    mScenarioWidget = nullptr;
    mMapWidget = nullptr;
    for (int i = 0; i < N_PLOTS; i++)
        mPlotWidgets[i] = nullptr;
    mTimestampWidget = nullptr;
    mIndicatorsWidget = nullptr;
    mInstrumentWidget = nullptr;
    mGPSWidget = nullptr;
    mCpuMemWidget = nullptr;
    mCommWidget = nullptr;
    mInversionWidget = nullptr;
    mMediaControlWidget = nullptr;

    mClockWidget = nullptr;
    mBatteryWidget = nullptr;
    // setup the GUI: menus, actions, toolbar and widgets.
    // ==> ALWAYS AFTER MODEL AND MANAGER INIT !!
    setupUI();

    // register widgets in the data manager
#ifndef SCN_SEL_ON_MAP
    mDataManager->registerScenarioSelector(mScenarioSelector);
#endif
    mDataManager->registerFlightWidget(mFlightWidget);
    mDataManager->registerScenarioWidget(mScenarioWidget);
    mDataManager->registerMapWidget(mMapWidget);
    for (int i = 0; i < N_PLOTS; i++)
        mDataManager->registerPlotWidget(i, mPlotWidgets[i]);
    mDataManager->registerIndicatorsWidget(mIndicatorsWidget);

    mDataManager->registerMediaControlWidget(mMediaControlWidget);

    // connections
    QObject::connect(mFitController, &FitController::fitComplete, this, &MainWindow::onBlockProcessed);

    // data manager
    QObject::connect(mDataManager, &DataManager::mapperUpdated, this, &MainWindow::onMapperUpdated);
    QObject::connect(mDataManager, &DataManager::newFlightAdded, this, &MainWindow::onNewFlightCreated);
    // UDP communications.
    QObject::connect(mUdpSocket, &AUDPSocket::invalidTemplate, this, &MainWindow::reportInvalidTemplate);
    QObject::connect(mUdpSocket, &AUDPSocket::receiveABlock, this, &MainWindow::incomingBlock);
    QObject::connect(mUdpSocket, &AUDPSocket::udpStatus, mCommWidget->udpWidget(), &UDPStatusWidget::updateStatus);
    // media controls widget.
    QObject::connect(mMediaControlWidget, &MediaControlWidget::loadRequest, this, &MainWindow::loadReplayData);

    // scenario selectors.
#ifndef SCN_SEL_ON_MAP
    for (auto p : {mScenarioSelector, mScenarioWidget->selector()}) {
#else
    for (auto p : {mMapWidget->selector(), mScenarioWidget->selector()}) {
#endif
        QObject::connect(p, &ScenarioSelector::scenarioEditionRequested, this, &MainWindow::editScenarios);
        QObject::connect(p, &ScenarioSelector::scenarioSaveRequested, this, &MainWindow::saveScenarios);
    }

    // text refresh
    QObject::connect(&mRefreshTimer, &QTimer::timeout, this, &MainWindow::refreshTexts);

    // ui reset
    QObject::connect(&mResetTimer, &QTimer::timeout, this, &MainWindow::onResetTimeout);

    translateUI();

    // Optionally load the file given in arguments.
    QFileInfo infos(configPath);
    if (infos.exists() && infos.suffix().compare("adm2", Qt::CaseInsensitive) == 0)
        loadConfiguration(configPath);

    // load (some) application settings.
    mFitController->setMaximumInputFrequency(settings.value("monitoring/max_inv_freq", 5).toInt());

    mMultiThreaded = settings.value("monitoring/multithread", false).toBool();
    setMultiThreadMode(mMultiThreaded);
    // udp socket
    mUdpSocket->setMaximumInputFrequency(settings.value("monitoring/max_input_freq", 10).toInt());
    mDataManager->updateFromSettings(true); // data manager and map style.
    // auto center option is initialized outside updateFromSettings
    // otherwise each time the user modify the option it will be update in the GUI !
    bool enabled = settings.value("monitoring/follow_loc", false).toBool();
    mMapWidget->autoCenterAction()->setChecked(enabled);
    // setup min sat for gps widget
    mGPSWidget->setMinSat(settings.value("monitoring/min_sat", 4).toInt());
    // load ui settings
    loadUI();

    mDataManager->enableTimeScrolling(true);
    mNoRefresh = false;
    mNoSDWarned = !settings.value("warnings/no_sd_detected", true).toBool();
    mRefreshTimer.setInterval(666);
    mRefreshTimer.start();

    mResetInterval = qMax(1500, settings.value("monitoring/reset_delay", 1500).toInt());
    mResetTimer.setInterval(mResetInterval);
    mResetTimer.setSingleShot(true);

    // setup help button: if manual is not found disable the button.
    QString path = QString("%1/manual/index.html").arg(gResDir);
    mActionHelp->setEnabled(QFileInfo::exists(path));
}

/// @brief Destructor.
MainWindow::~MainWindow() {
    mUdpSocket->setListening(false);
    mFitController->waitForFinished();
    freeResultFileHandler();
    mDataManager->deleteLater();
}

/**
 * @brief (Un)Lock the positions of the dockwidgets.
 * @param locked True to lock positions, false otherwise.
 */
void MainWindow::lockUI(bool locked) {
    QDockWidget::DockWidgetFeatures features =
        locked ? QDockWidget::NoDockWidgetFeatures
               : (QDockWidget::DockWidgetClosable | QDockWidget::DockWidgetMovable | QDockWidget::DockWidgetFloatable);

    mActionLockUI->setIcon(QIcon::fromTheme(locked ? "lock" : "unlock"));
    const auto list = findChildren<QDockWidget *>();
    for (QDockWidget *dockWidget : list) {
        dockWidget->setFeatures(features);
    }
}

/**
 * @brief changeEvent event handler.
 * @details
 * If event is of type QEvent::LanguageChange, the method (re)translates (completely) the application, then
 * it passes the event to the base class changeEvent method (any other case).
 */
void MainWindow::changeEvent(QEvent *event) {
    if (event->type() == QEvent::LanguageChange) {
        translateUI();
    } /* else if (event->type() == QEvent::LocaleChange) {
         QString locale = QLocale::system().name();
         locale.truncate(locale.lastIndexOf('_'));
         if (gI18N)
             gI18N->setLanguage(locale);
     }*/
    QMainWindow::changeEvent(event);
}

void MainWindow::editScenarios(int current) {
    //    if (current == -1)
    //        return;
    ScenariosEditor *editor = new ScenariosEditor(this);
    editor->setParameters(mDataManager->mapper()->parameters());
    editor->setScenarios(mDataManager->scenarios(), &current);
    bool SDFLS = mNoSDWarned;
    mNoSDWarned = true;
    int ret = editor->exec();
    if (ret == QDialog::Accepted) {
        int selected = editor->selectedScenarioIndex();
        mDataManager->setScenarios(editor->scenarios(), &selected);
        // save scenarios (only if on the fly option is enabled !)
        saveScenariosOnTheFly();
    }
    mNoSDWarned = SDFLS;
    editor->deleteLater();
}

/// @brief Save scenarios (according to application settings "save on the fly").
bool MainWindow::saveScenariosOnTheFly() {
    QSettings settings;
    if (settings.value("global/save_on_fly", true).toBool())
        return saveScenarios();
    return false;
}
/**
 * @brief Save the current scenarios to disk.
 * @return True on success, false otherwise.
 * @details
 * The method checks if the current ADM configuration is valid and if the save-on-the-fly mode is enabled.
 * In such case, it saves scenarios to the path stored in the ADM configuration.
 */
bool MainWindow::saveScenarios() {
    if (!mAdmConfig.isValid())
        return false;
    QString fp = mAdmConfig.scenarioFilePath();
    bool ret = mDataManager->saveScenarios(fp);
    statusBar()->showMessage(ret ? tr("Scenarios saved to disk.") : tr("Failed to save scenarios to disk."));
    return ret;
}

/// @brief Get the size hint of the window.
QSize MainWindow::sizeHint() const {
    QSize screenSize = getActiveScreen(this)->availableSize();
    return screenSize * 0.8;
}

/**
 * @brief Center the window to the active screen.
 */
void MainWindow::centerToScreen() {
    QRect screenGeometry = getActiveScreen(this)->availableGeometry();
    int x = (screenGeometry.width() - width()) / 2;
    int y = (screenGeometry.height() - height()) / 2;
    move(x, y);
}

/// @brief Get the map widget.
MapWidget *MainWindow::mapWidget() const { return mMapWidget; }

/// @brief Setup the UI.
void MainWindow::setupUI() {

    // creates actions, menus and toolbar first
    createActions();
    mMapWidget = new MapWidget(this);

    mTabWidget = new QTabWidget(this);
    setCentralWidget(mTabWidget);
    mTabWidget->addTab(mMapWidget, tr("Map"));

    // Dock widgets:
    setDockOptions(QMainWindow::AnimatedDocks | QMainWindow::AllowNestedDocks | QMainWindow::AllowTabbedDocks |
                   QMainWindow::GroupedDragging);

    setCorner(Qt::TopLeftCorner, Qt::LeftDockWidgetArea);
    setCorner(Qt::BottomLeftCorner, Qt::LeftDockWidgetArea);
    setCorner(Qt::TopRightCorner, Qt::RightDockWidgetArea);
    setCorner(Qt::BottomRightCorner, Qt::RightDockWidgetArea);

    QDockWidget *dockWidget = nullptr;
    // Setup plots:

    // the number of plots is compilation dependent and defined as N_PLOTS.
    for (int i = 0; i < N_PLOTS; i++) {
        QString prefix = QString::number(i + 1);
        dockWidget = new QDockWidget(tr("Plot %1").arg(prefix), this);
        RtPlotWidget *plotWidget = new RtPlotWidget(i, dockWidget);
        plotWidget->enableParameterSelection(true);
        dockWidget->setObjectName(QString("dw_plot%1").arg(i + 1));
        dockWidget->setWidget(plotWidget);
        QObject::connect(plotWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
        plotWidget->setTitlePrefix(prefix);
        mPlotWidgets[i] = plotWidget;
        mPlotDockWidgets.push_back(dockWidget);
        addDockWidget(Qt::TopDockWidgetArea, dockWidget);
        QAction *action = dockWidget->toggleViewAction();
        action->setShortcut(QKeySequence(QString("Ctrl+%1").arg(prefix)));
        mViewPlotSubMenu->addAction(action);
    }

    // System resources monitor widget.
    mCpuMemWidget = new CpuMemWidget(this);
    mCpuMemWidget->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Minimum);
    dockWidget = new QDockWidget(tr("System resources usage"), this);
    dockWidget->setObjectName("dw_sysres");
    dockWidget->setWidget(mCpuMemWidget);
    QObject::connect(mCpuMemWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::LeftDockWidgetArea, dockWidget, Qt::Vertical);
    mViewMenu->addAction(dockWidget->toggleViewAction());

    // network widget.
    mCommWidget = new CommWidget(this);
    mCommWidget->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Minimum);
    dockWidget = new QDockWidget(tr("Communications"), this);
    dockWidget->setObjectName("dw_comm");
    dockWidget->setWidget(mCommWidget);
    QObject::connect(mCommWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::LeftDockWidgetArea, dockWidget, Qt::Vertical);
    mViewMenu->addAction(dockWidget->toggleViewAction());

    // inversion widget.
    dockWidget = new QDockWidget(tr("Inversion"), this);
    mInversionWidget = new InversionStatusWidget(this);
    mInversionWidget->setInversionUpdate(false);
    mInversionWidget->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Minimum);

    dockWidget->setObjectName("dw_invs");
    dockWidget->setWidget(mInversionWidget);
    QObject::connect(mInversionWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::LeftDockWidgetArea, dockWidget, Qt::Vertical);
    mViewMenu->addAction(dockWidget->toggleViewAction());

    // media control widget.
    mMediaControlWidget = new MediaControlWidget(this);
    mMediaControlWidget->setEnabled(false);
    mMediaControlWidget->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Minimum);
    dockWidget = new QDockWidget(tr("Replay controls"), this);
    dockWidget->setObjectName("dw_mcw");
    dockWidget->setWidget(mMediaControlWidget);
    QObject::connect(mMediaControlWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::BottomDockWidgetArea, dockWidget, Qt::Vertical);
    // mViewMenu->addAction(dockWidget->toggleViewAction());
    mDockMCW = dockWidget;
    mDockMCW->setVisible(false);

    // Parameter view.
    mScenarioWidget = new ScenarioWidget(this);
    mScenarioWidget->setEnabled(false);
    dockWidget = new QDockWidget(tr("Scenarios"), this);
    dockWidget->setObjectName("dw_scnview");
    dockWidget->setWidget(mScenarioWidget);
    QObject::connect(mScenarioWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::LeftDockWidgetArea, dockWidget, Qt::Vertical);
    mViewMenu->addAction(dockWidget->toggleViewAction());

    // Flight view.
    mFlightWidget = new FlightWidget(this);
    mFlightWidget->setEnabled(false);
    dockWidget = new QDockWidget(tr("Flights"), this);
    dockWidget->setObjectName("dw_flightview");
    dockWidget->setWidget(mFlightWidget);
    QObject::connect(mFlightWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::LeftDockWidgetArea, dockWidget, Qt::Vertical);
    mViewMenu->addAction(dockWidget->toggleViewAction());

    // Timestamp widget
    mTimestampWidget = new TimestampWidget(this);
    mTimestampWidget->setMapper(mDataManager->mapper());
    dockWidget = new QDockWidget(tr("Timestamps"), this);
    dockWidget->setObjectName("dw_timestamps");
    dockWidget->setWidget(mTimestampWidget);
    QObject::connect(mTimestampWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::RightDockWidgetArea, dockWidget, Qt::Vertical);
    mViewIndicatorsSubMenu->addAction(dockWidget->toggleViewAction());

    // GPS widget
    mGPSWidget = new GPSWidget(this);
    mGPSWidget->setMapper(mDataManager->mapper());
    dockWidget = new QDockWidget(tr("GPS status"), this);
    dockWidget->setObjectName("dw_gpsstatus");
    dockWidget->setWidget(mGPSWidget);
    QObject::connect(mGPSWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::RightDockWidgetArea, dockWidget, Qt::Vertical);
    mViewIndicatorsSubMenu->addAction(dockWidget->toggleViewAction());

    // File names.
    mFilenameWidget = new FilenameWidget(this);
    dockWidget = new QDockWidget(tr("Filenames"), this);
    dockWidget->setObjectName("dw_filenames");
    dockWidget->setWidget(mFilenameWidget);
    QObject::connect(mFilenameWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::RightDockWidgetArea, dockWidget, Qt::Vertical);
    mViewIndicatorsSubMenu->addAction(dockWidget->toggleViewAction());

    // Instrument Parameter indicator widget
    mInstrumentWidget = new IndicatorsWidget(this);
    mInstrumentWidget->setMapper(mDataManager->mapper());
    dockWidget = new QDockWidget(tr("Sensor indicators"), this);
    dockWidget->setObjectName("dw_indicators");
    dockWidget->setWidget(mInstrumentWidget);
    QObject::connect(mInstrumentWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::RightDockWidgetArea, dockWidget, Qt::Vertical);
    mViewIndicatorsSubMenu->addAction(dockWidget->toggleViewAction());

    // (user-defined) Parameter indicator widget
    mIndicatorsWidget = new IndicatorsWidget(this);
    mIndicatorsWidget->setMapper(mDataManager->mapper());
    dockWidget = new QDockWidget(tr("User indicators"), this);
    dockWidget->setObjectName("dw_indicators");
    dockWidget->setWidget(mIndicatorsWidget);
    QObject::connect(mIndicatorsWidget, &QWidget::windowTitleChanged, dockWidget, &QDockWidget::setWindowTitle);
    addDockWidget(Qt::RightDockWidgetArea, dockWidget, Qt::Vertical);
    mViewIndicatorsSubMenu->addAction(dockWidget->toggleViewAction());

    // setup some connections
    QHeaderView *header = mScenarioWidget->tableView()->horizontalHeader();
    mHeaderOffset = header->count() - N_PLOTS; // => first indexes are reserved for name and map check.
    // toggle dockwidget view on parameter graph column click!
    QObject::connect(header, &QHeaderView::sectionClicked, this, &MainWindow::onHeaderClick);

    statusBar();
}

/// @brief Show/Hide plot widgets associated to the header of scenrario widget table view.
void MainWindow::onHeaderClick(int index) {
    index -= mHeaderOffset;
    if (index >= 0 && index < mPlotDockWidgets.size())
        mPlotDockWidgets[index]->toggleViewAction()->trigger();
}

/**
 * @brief Attempt to load dataset from user selected files.
 */
void MainWindow::loadReplayData() {
    QStringList filepaths;
    QSettings settings;
    QString lastDir =
        settings
            .value("global/lastreplaydir", QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).first())
            .toString();
    QStringList filePaths = QFileDialog::getOpenFileNames(
        this, tr("Select files"), lastDir, tr("Results files (*.results);;Binary files (*.bin)"));
    if (filePaths.isEmpty())
        return;
    settings.setValue("global/lastreplaydir", QFileInfo(filePaths.first()).dir().absolutePath());
    QVector<PairOfFiles> infos = checkAmulseFiles(filePaths);
    filepaths.clear();
    if (infos.isEmpty())
        return;
    ABlock descr(infos.first().second);
    for (const PairOfFiles &p : qAsConst(infos))
        filepaths.append(QString::fromStdString(p.first));

    DataLoaderDialog *dialog = new DataLoaderDialog(filepaths, mDataManager->mapper(), this);
    int result = dialog->exec();
    ReplayData *data = dialog->data();
    if (result == QDialog::Rejected) {
        if (data)
            delete data;
        return;
    }
    mDataManager->setReplayData(data);
}

/// @brief Refresh frequency texts of the udp and inversion status widgets.
void MainWindow::refreshTexts() {
    if (mNoRefresh)
        return;
    double a, b;
    mUdpSocket->getRates(&a, &b);
    mCommWidget->udpWidget()->refreshText(a, b);
    mInversionWidget->refreshText();
}

/**
 * @brief Reset some parts of the UI.
 * @details
 *
 * The method is intended to reset the UI when no block have been registered since a given delay.
 *
 * By default, the method resets the inversion status widget (InversionStatusWidget class) and the udp and gps status
 * widgets of the CommWidget: The values are set to zero and the LED indicators (if any) are set back to grey which
 * represent the IDLE state of the application.
 *
 * If the <i>full reset</i> settings is enabled in the application settings then all widgets that show instant data from
 * the sensor are reset, that is:
 * - the gps widget.
 * - the timestamps widget.
 * - the instrument indicators.
 * - the user-defined indicators.
 * - file names widget.
 *
 * Map and plot are left untouched. Scenarios et flights widgets are not (never) concerned by such reset.
 *
 */
void MainWindow::onResetTimeout() {
    mNoRefresh = true;
    mCommWidget->udpWidget()->reset();
    mCommWidget->gpsWidget()->reset();
    if (mInversionEnabled)
        mInversionWidget->reset();
}

/// @brief Load UI layout settings.
void MainWindow::loadUI() {
    QSettings settings;
    QString lastLayout = settings.value("layouts/last_name", "").toString();
    int idx = mLayoutManager->indexOf(lastLayout);
    mLayoutSelector->combobox()->setCurrentIndex(idx);
    if (idx == -1 || !mLayoutManager->loadLayout(idx)) {
        restoreState(settings.value("layouts/main_win/windowstate").toByteArray());
    }
    restoreGeometry(settings.value("layouts/main_win/geometry").toByteArray());
    mActionFullScreen->setChecked(isFullScreen());
}

/// @brief Save UI layout settings.
void MainWindow::saveUI() {
    QSettings settings;
    settings.setValue("layouts/last_name", mLayoutSelector->combobox()->currentText());
    settings.setValue("layouts/main_win/geometry", saveGeometry());
    settings.setValue("layouts/main_win/windowstate", saveState());
    mMapWidget->saveUI();
}

/// @brief Translate texts.
void MainWindow::translateUI() {

    mScenarioWidget->setWindowTitle(tr("Scenarios"));
    mScenarioWidget->translateUI();
    mFlightWidget->translateUI();
    mMapWidget->translateUI();
    for (auto pw : mPlotWidgets)
        pw->translateUI();
    mGPSWidget->translateUI();
    mTimestampWidget->translateUI();
    mFilenameWidget->translateUI();
    mIndicatorsWidget->setWindowTitle(tr("User indicators"));
    mInstrumentWidget->setWindowTitle(tr("Sensor indicators"));
    mInversionWidget->translateUI();
    mCommWidget->translateUI();
    mCpuMemWidget->translateUI();
    mLayoutSelector->translateUI();
    mScenarioSelector->translateUI();

    mDataManager->translateModelHeaders();

    // menus and actions

    mFileMenu->setTitle(tr("File"));
    mViewMenu->setTitle(tr("View"));
    mViewPlotSubMenu->setTitle(tr("Plots"));
    mViewIndicatorsSubMenu->setTitle(tr("Indicators"));
    mHelpMenu->setTitle(tr("Help"));

    mActionLoad->setText(tr("Open configuration"));
    mActionAbout->setText(tr("About %1").arg(QApplication::applicationName()));
#ifdef O_HELP
    mActionHelp->setText(tr("Manual"));
#endif
    mActionScreenshot->setText(tr("Take a screenshot"));
    mActionSettings->setText(tr("Settings"));
    mActionQuit->setText(tr("Quit"));
    mActionFullScreen->setText(tr("Fullscreen"));

    // tooltips & statustip
    QString msg, msg2;
    msg = tr("Load a new configuration from file.");
    mActionLoad->setToolTip(msg);
    mActionLoad->setStatusTip(msg);
    msg = tr("Stop everything and quit the application.");
    mActionQuit->setToolTip(msg);
    mActionQuit->setStatusTip(msg);
    msg = tr("Show the application settings dialog.");
    mActionSettings->setToolTip(msg);
    mActionSettings->setStatusTip(msg);
    msg = tr("Enable/disable replay mode.");
    mActionReplay->setToolTip(msg);
    mActionReplay->setStatusTip(msg);
    msg = tr("Clear the current live data.");
    mActionClear->setStatusTip(msg);
    msg2 = tr("<p><b>WARNING:</b> The operation is irreversible !</p>");
    mActionClear->setToolTip(QString("<p>%1</p>%2").arg(msg, msg2));
#ifdef O_HELP
    if (mActionHelp->isEnabled()) {
        msg = tr("Show the manual.");
    } else {
        msg = tr("Manual is not available.");
    }
    mActionHelp->setToolTip(msg);
    mActionHelp->setStatusTip(msg);

#endif
    msg = mActionScreenshot->text();
    mActionScreenshot->setToolTip(msg);
    mActionScreenshot->setStatusTip(msg);
    msg = tr("Display the application about box.");
    mActionAbout->setToolTip(msg);
    mActionAbout->setStatusTip(msg);
    msg = tr("Enable/disable the inversion process.");
    msg2 = tr("<p>The process is activated when button is checked.</p>");
    mActionToggleInv->setStatusTip(msg);
    mActionToggleInv->setToolTip(QString("<p>%1</p>%2").arg(msg, msg2));
}

/// @brief Update UI and parameters from application settings changes.
void MainWindow::updateFromSettings() {

    QSettings settings;

    // language.
    if (gI18N)
        gI18N->setLanguage(settings.value("global/language", "en").toString());

    mNoSDWarned = !settings.value("warnings/no_sd_detected", true).toBool();

    mShowFit = settings.value("monitoring/show_fit", false).toBool();
    setupFitWidgets();
    mUdpSocket->setMaximumInputFrequency(settings.value("monitoring/max_input_freq", 5).toInt());
    mFitController->setMaximumInputFrequency(settings.value("monitoring/max_inv_freq", 5).toInt());
    mCommWidget->wlanWidget()->monitor(settings.value("monitoring/monitor_wifi", true).toBool());
    mMultiThreaded = settings.value("monitoring/multithread", false).toBool();
    setMultiThreadMode(mMultiThreaded);

    mBatteryWidget->showPercent(settings.value("global/show_battery_pct", true).toBool());
    mBatteryWidget->showLifeTime(settings.value("global/show_battery_time", true).toBool());
    mActionBattery->setVisible(settings.value("global/show_battery", true).toBool());

    mClockWidget->setUTC(settings.value("global/show_clock_utc", true).toBool());
    mClockWidget->showDate(settings.value("global/show_clock_date", false).toBool());
    mClockWidget->showSeconds(settings.value("global/show_clock_sec", true).toBool());
    mActionClock->setVisible(settings.value("global/show_clock", true).toBool());

    // All widgets registered in the data manager will receive the update.
    // (that is most of all widgets that require to be updated).
    mDataManager->updateFromSettings(true); // data manager and map style.

    // update GPS widget
    mGPSWidget->setMinSat(settings.value("monitoring/min_sat", 4).toInt());
    updateLoggerFromSettings();
}

/// @brief Create actions, menus and toolbar.
void MainWindow::createActions() {

    mActionLoad = new QAction(QIcon::fromTheme("folder-open"), tr("Open configuration"), this);
    mActionLoad->setShortcut(QKeySequence("Ctrl+O"));
    mActionToggleInv = new QAction(QIcon::fromTheme("inversion"), "", this);
    mActionToggleInv->setCheckable(true);
    mActionToggleInv->setEnabled(false);

    mActionClear = new QAction(QIcon::fromTheme("clear"), "", this);
    mActionClear->setEnabled(false);

    mActionReplay = new QAction(QIcon::fromTheme("movie"), "", this);
    mActionReplay->setCheckable(true);
    mActionReplay->setEnabled(false);

    mActionAbout = new QAction(QIcon::fromTheme("about"), "", this);
    mActionSettings = new QAction(QIcon::fromTheme("settings"), "", this);
    mActionSettings->setShortcut(QKeySequence("Ctrl+P"));

    mActionScreenshot = new QAction(QIcon::fromTheme("camera"), tr("Take a screenshot"), this);
    mActionScreenshot->setShortcut(QKeySequence("F12"));
#ifdef O_FULLSCREEN
    mActionFullScreen = new QAction(QIcon::fromTheme("full-screen"), tr("Enable/Disable fullscreen"), this);
    mActionFullScreen->setShortcut(QKeySequence("F11"));
// mActionFullScreen->setCheckable(true);
#endif
    mActionQuit = new QAction(QIcon::fromTheme("exit"), "", this);
    mActionQuit->setShortcut(QKeySequence("Ctrl+Q"));

#ifdef O_HELP
    mActionHelp = new QAction(QIcon::fromTheme("help"), "", this);
    QObject::connect(mActionHelp, &QAction::triggered, this, &MainWindow::openHelp);
#endif

    mActionLockUI = new QAction(QIcon::fromTheme("unlock"), tr("Lock/Unlock UI"), this);
    mActionLockUI->setCheckable(true);
    QObject::connect(mActionLockUI, &QAction::toggled, this, &MainWindow::lockUI);

    // connection
    QObject::connect(mActionToggleInv, &QAction::toggled, this, &MainWindow::enableInversion);
    QObject::connect(mActionClear, &QAction::triggered, this, &MainWindow::clearData);
    QObject::connect(mActionReplay, &QAction::triggered, this, &MainWindow::switchModes);
    QObject::connect(mActionLoad, &QAction::triggered, this, &MainWindow::openConfiguration);
    QObject::connect(mActionSettings, &QAction::triggered, this, &MainWindow::openSettings);
    QObject::connect(mActionAbout, &QAction::triggered, this, &MainWindow::aboutApp);
    QObject::connect(mActionScreenshot, &QAction::triggered, this, &MainWindow::captureScreen);
#ifdef O_FULLSCREEN
    QObject::connect(mActionFullScreen, &QAction::triggered, this, &MainWindow::switchFullscreen);
#endif
    QObject::connect(mActionQuit, &QAction::triggered, this, &MainWindow::close);

    // toolbar
    mToolBar = addToolBar(tr("Tool bar"));
    mToolBar->setObjectName("tools");

    // Layout selection widget
    mLayoutSelector = new LayoutSelector(mToolBar);
    mLayoutSelector->setLayoutManager(mLayoutManager);
    mLayoutSelector->layout()->setContentsMargins(0, 0, 0, 0);

#ifndef SCN_SEL_ON_MAP
    mScenarioSelector = new ScenarioSelector(mToolBar);
    mScenarioSelector->layout()->setContentsMargins(0, 0, 0, 0);
#endif
    // clock and battery widget.
    mClockWidget = new ClockWidget(mToolBar);
    mBatteryWidget = new BatteryStatusWidget(mToolBar);

    mToolBar->addAction(mActionLoad);
    mToolBar->addSeparator();
    mToolBar->addAction(mActionToggleInv);
    mToolBar->addSeparator();
    mToolBar->addAction(mActionClear);
    mToolBar->addSeparator();
    mToolBar->addAction(mActionReplay);
    mToolBar->addSeparator();
    mActionLayout = mToolBar->addWidget(mLayoutSelector);
    mToolBar->addSeparator();
#ifndef SCN_SEL_ON_MAP
    mActionScenario = mToolBar->addWidget(mScenarioSelector);
    mActionScenario->setEnabled(false);
    mToolBar->addSeparator();
#endif

    QWidget *spacer = new QWidget(this);
    spacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    mToolBar->addWidget(spacer);

    mActionClock = mToolBar->addWidget(mClockWidget);
    mActionBattery = mToolBar->addWidget(mBatteryWidget);
    mToolBar->addSeparator();
    mToolBar->addAction(mActionLockUI);
    mToolBar->addSeparator();
    mToolBar->addAction(mActionScreenshot);
    mToolBar->addSeparator();
#ifdef O_HELP
    mToolBar->addAction(mActionHelp);
#endif
    mToolBar->addAction(mActionSettings);
    mToolBar->addSeparator();
#ifdef O_FULLSCREEN
    mToolBar->addAction(mActionFullScreen);
#endif
    // menus
    QMenuBar *menubar = menuBar();
    mFileMenu = menubar->addMenu(tr("File"));
    mViewMenu = menubar->addMenu(tr("View"));
#ifdef O_FULLSCREEN
    mViewMenu->addAction(mActionFullScreen);
    mViewMenu->addAction(mActionLockUI);
    mViewMenu->addSeparator();

#endif
    mViewPlotSubMenu = mViewMenu->addMenu(tr("Plots"));
    mViewIndicatorsSubMenu = mViewMenu->addMenu(tr("Indicators"));
    mViewMenu->addSeparator();
    mHelpMenu = menubar->addMenu(tr("Help"));

    mFileMenu->addAction(mActionLoad);
    mFileMenu->addAction(mActionSettings);
    mFileMenu->addSeparator();
    mFileMenu->addAction(mActionQuit);

    mViewMenu->addAction(mToolBar->toggleViewAction());

#ifdef O_HELP
    mHelpMenu->addAction(mActionHelp);
#endif
    mHelpMenu->addAction(mActionAbout);

    // setup some settings of clock and battery widget and their associated widget.
    QSettings settings;
    mBatteryWidget->showPercent(settings.value("global/show_battery_pct", true).toBool());
    mBatteryWidget->showLifeTime(settings.value("global/show_battery_time", true).toBool());
    mActionBattery->setVisible(settings.value("global/show_battery", true).toBool());

    mClockWidget->setUTC(settings.value("global/show_clock_utc", true).toBool());
    mClockWidget->showDate(settings.value("global/show_clock_date", false).toBool());
    mClockWidget->showSeconds(settings.value("global/show_clock_sec", true).toBool());
    mActionClock->setVisible(settings.value("global/show_clock", true).toBool());
}

void MainWindow::setMultiThreadMode(bool enable) {
    // defunc <-> obsolete
    mMultiThreaded = enable;
}

/// @brief Switch to fullscreen / normal mode.
void MainWindow::switchFullscreen() {
    bool fs = isFullScreen();
    if (fs) {
        showNormal();
    } else {
        showFullScreen();
    }
    fs = isFullScreen();
    menuBar()->setHidden(fs);
    mActionFullScreen->setIcon(fs ? QIcon::fromTheme("normal-screen") : QIcon::fromTheme("full-screen"));
}

/**
 * @brief Clear the current data.
 */
void MainWindow::clearData() {
    mDataManager->resetData();
    mFilenameWidget->reset();
    mTimestampWidget->reset();
    mGPSWidget->reset();
    mCommWidget->reset();
    mInstrumentWidget->updateFromMapper();
}

/// @brief Take a picture of the screen.
void MainWindow::captureScreen() {
    if (mScreenshotManager->saveScreen()) {
        statusBar()->showMessage(tr("Screen capture saved"), 4000);
    } else {
        statusBar()->showMessage(tr("Failed to save screen capture."), 4000);
    }
}

/// @brief Display about box.
void MainWindow::aboutApp() {
    QString caption = QString("<h3>ADMin v%1</h3>").arg(QCoreApplication::applicationVersion());
    QString aboutText = QMessageBox::tr(
        "<p>ApoDrone Monitoring and INversion program</p>"
        "A <i>not so dumb</i> Qt application that monitors and computes concentration from data"
        "block &#8220;networkly&#8221; sent by APODRONE.</p>"
        "<p>Copyright (C) 2018-2022, J. Burgalat, GSMA URCA.</p>"
        "<p></p>"
        "<p>Icons:</p>"
        "<ul>"
        "<li>Window Metro and Windows 10 sets by <a href=\"https://icons8.com\">Icons8</a></li>"
        "<li>Soft Craps icons set by <a href=\"https://www.deviantart.com/hopstarter\">Hopstarter</a></li>"
        "</ul>");
    QMessageBox *msgBox = new QMessageBox(this);
    msgBox->setAttribute(Qt::WA_DeleteOnClose);
    msgBox->setWindowTitle(tr("About ADMin"));
    msgBox->setText(caption);
    msgBox->setInformativeText(aboutText);
    QIcon icon = QIcon(":/images/app.png");
    QSize size = icon.actualSize(QSize(128, 128));
    msgBox->setIconPixmap(icon.pixmap(size));
    msgBox->exec();
}

/**
 * @brief Enable/disable the inversion process.
 * @param enable True to enable the process, false otherwise.
 */
void MainWindow::enableInversion(bool enable) {
    mInversionEnabled = enable;
    mInversionWidget->setInversionUpdate(enable);
    setupFitWidgets();
    mFitController->setFitProcess(mInversionEnabled);
}

/**
 * @brief Switch between RT and replay mode.
 * @param movieMode True to switch to replay mode.
 */
void MainWindow::switchModes(bool replay) {
    if (replay) {
        QSettings settings;
        if (settings.value("warnings/replay_switch", true).toBool()) {
            QString text =
                tr("<p>You are about to switch to the replay mode.</p>"
                   "<p>The application will stop listening to the sensor.</p>"
                   "<p>Sensor monitoring will only be reactivated when the application leaves the replay mode.</p>");
            QString msg = tr("<p>Do you really wish to continue ?</p>");
            QMessageBox *msgBox = new QMessageBox;
            msgBox->setWindowTitle(tr("Information"));
            msgBox->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
            msgBox->setDefaultButton(QMessageBox::No);
            msgBox->setEscapeButton(QMessageBox::No);
            msgBox->setCheckBox(new QCheckBox(tr("Do not show again")));
            msgBox->setIcon(QMessageBox::Warning);
            msgBox->setText(text);
            msgBox->setInformativeText(msg);
            bool ok = msgBox->exec() == QMessageBox::Yes;
            settings.setValue("warnings/replay_switch", !msgBox->checkBox()->isChecked());
            delete msgBox;
            if (!ok) {
                mActionReplay->setChecked(false);
                mDockMCW->setVisible(false);
                return;
            }
        }
    }
    // clear ui.
    mTimestampWidget->reset();
    mGPSWidget->reset();
    mInstrumentWidget->reset();
    mIndicatorsWidget->reset();
    mFilenameWidget->setADMinFileName("N/A");
    mFilenameWidget->setSensorFileName("N/A");

    mUdpSocket->setListening(!replay);
    mNoRefresh = replay;
    if (replay) {
        mFitController->waitForFinished();
        if (mResultFileHandler)
            mResultFileHandler->flush();
    }
    mFlightWidget->setEnabled(!replay);
    mCommWidget->setEnabled(!replay);
    // mScenarioWidget->selector()->setInteractionMode(!replay);
    mDataManager->setReplayMode(replay);
    mMediaControlWidget->setEnabled(replay);

    mActionClear->setDisabled(replay);
    mActionLoad->setDisabled(replay);
    mActionToggleInv->setEnabled(replay ? false : mFitConfig.isValid());
    mDockMCW->setVisible(replay);
}

/// @brief Manage actions before quit and quit.
void MainWindow::closeEvent(QCloseEvent *event) {
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    mUdpSocket->disconnect();
    mFitController->waitForFinished();

    QApplication::restoreOverrideCursor();
#ifdef O_FULLSCREEN
    if (isFullScreen())
        switchFullscreen();
#endif
    saveScenariosOnTheFly();
    saveUI();
    // always save layouts
    mLayoutManager->saveToFile(gLayoutsFile);
    QMainWindow::closeEvent(event);
    QApplication::quit();
}

/// @brief Open html help file.
void MainWindow::openHelp() {
    QString path = QString("%1/manual/index.html").arg(gResDir);
    if (QFileInfo::exists(path))
        QDesktopServices::openUrl(QUrl::fromLocalFile(path));
}

/// @brief Open a new configuration.
void MainWindow::openConfiguration() {
    if (mAdmConfig.isValid()) {
        QSettings settings;
        if (settings.value("warnings/conf_change", true).toBool()) {
            QString text = tr("<p>A configuration is already loaded and running...</p>"
                              "<p>Loading a new configuration file (even the same) will reset all data.</p>");
            QString msg = tr("<p>Do you really wish to continue ?</p>");
            QMessageBox *msgBox = new QMessageBox;
            msgBox->setWindowTitle(tr("Question"));
            msgBox->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
            msgBox->setDefaultButton(QMessageBox::No);
            msgBox->setEscapeButton(QMessageBox::No);
            msgBox->setCheckBox(new QCheckBox(tr("Do not show again")));
            msgBox->setIcon(QMessageBox::Question);
            msgBox->setText(text);
            msgBox->setInformativeText(msg);
            bool ok = msgBox->exec() == QMessageBox::Yes;
            settings.setValue("warnings/conf_change", !msgBox->checkBox()->isChecked());
            delete msgBox;
            if (!ok)
                return;
        }
    }
    QSettings settings;
    QString docDir = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).at(0);
    QString cfgDir = settings.value("storage/cfgdir", QString("%1/A-ToolSuite/configurations").arg(docDir)).toString();
    QString fileName = QFileDialog::getOpenFileName(this, tr("Open ADMin file"), cfgDir, tr("ADMin files (*.adm2)"));
    if (fileName.isEmpty())
        return;
    cfgDir = QFileInfo(fileName).absoluteDir().absolutePath();
    settings.setValue("storage/cfgdir", cfgDir);
    loadConfiguration(fileName);
    return;
}

/**
 * @brief Load a new configuration.
 * @param configPath Path of the configuration file.
 * @details
 * The method attempts to load a new configuration performing the following steps:
 *
 * 1. Parse input configuration file -> on error ASK for continue.
 * 2. Update the key map (lazy mode: will update only "valid" data).
 * 3. Check keymap/descriptor consistency -> STRICT check: on error EXIT.
 * 4. Load fit configuration if any (lazy mode): if no fit file is given continue processing but disable fit
 * feature).
 *    - If no existing fit config file is given: simply disable the fit feature but continue processing.
 *    - Otherwise: check for config consistency and transitions file: on failure ASK to continue (disable the fit
 * feature in the latter case).
 * 3. Stop current processes: fit, UDP comm, data save.
 * 5. Reset the managers (fit and data) - Previous flights are no more accessible unless
 *    the list of watched parameters remains the same.
 * 6. Restart listening to incoming block.
 * 7. Update some widget with configuration informations.
 */
void MainWindow::loadConfiguration(const QString &configPath) {
    QSettings settings;
    AdmConfig admFileConfig;
    FitConfig fitFileConfig;
    bool logToDisk = settings.value("storage/log", true).toBool();

    // 1. read config and check for validity.
    QStringList errors;
    StringList serrors;
    if (!admFileConfig.parse(configPath, &errors)) {
        for (const QString &e : errors)
            LOG_ERROR(QString("[CFG] %1").arg(e));
        QString text = tr("<p>The selected configuration is not valid.</p>"
                          "%1"
                          "<p>Fix the errors and reload the configuration file.</p>")
                           .arg(logToDisk ? tr("<p>Errors encountered during parsing are available in the "
                                               "error log file.</p>")
                                          : "");
        QMessageBox *msgBox = new QMessageBox(this);
        msgBox->setWindowTitle(tr("Error"));
        msgBox->setText(text);
        if (logToDisk) {
            msgBox->setStandardButtons(QMessageBox::Ok | QMessageBox::Open);
        } else {
            msgBox->setStandardButtons(QMessageBox::Ok);
        }
        msgBox->setIcon(QMessageBox::Critical);
        int ret = msgBox->exec();
        if (ret == QMessageBox::Open)
            QDesktopServices::openUrl(QUrl::fromLocalFile(gLogFile));
        return;
    }
    errors.clear();
    serrors.clear();
    // check fit config

    QFileInfo finfos(admFileConfig.fitFilePath());
    bool hasPath = finfos.exists(), fileOk = true;

    // no fit config:
    // we have a path: parse the file and check for error.
    if (hasPath) {
        fileOk = fitFileConfig.parse(finfos.absoluteFilePath().toStdString(), &serrors);
        for (const std::string &e : serrors)
            LOG_ERROR(QString("[CFG:FIT] %1").arg(QString::fromStdString(e)));
        // check num instr and version
        if (fileOk) {
            int ai = admFileConfig.instrumentNumber(), av = admFileConfig.firmwareVersion(),
                bi = fitFileConfig.instrumentNumber(), bv = fitFileConfig.firmwareVersion();
            if (ai != bi || av != bv) {
                LOG_ERROR(QString("[CFG:FIT] Instrument number/Firmware version couple (i:%1,v:%2) is not consistent "
                                  "with the ADMin "
                                  "configuration file (i:%3,v:%4).")
                              .arg(bi)
                              .arg(bv)
                              .arg(ai)
                              .arg(av));
                fileOk = false;
            }
        }
        serrors.clear();
        if (fileOk) {
            bool tmp = true;
            for (auto &sp : fitFileConfig.spectras()) {
                tmp = tmp && sp.line.loadTransitions();
                if (!tmp)
                    break;
            }
            if (!tmp) {
                fileOk = false;
                LOG_ERROR(QString("[CFG:FIT] Cannot parse transitions file(s)"));
            }
        }
    }

    // check fit flag on output:

    if (hasPath && !fileOk && settings.value("warnings/bad_fit_cfg", true).toBool()) {
        // message box !
        QString text =
            tr("<p>The fit configuration is not valid. "
               "No fit process will occur and some parameters may not be available for monitoring.</p>"
               "%1")
                .arg(logToDisk ? tr("<p>Errors encountered during parsing are available in the error log file.</p>")
                               : "");
        QMessageBox *msgBox = new QMessageBox(this);
        msgBox->setWindowTitle(tr("Warning"));
        msgBox->setText(text);
        if (logToDisk) {
            msgBox->setStandardButtons(QMessageBox::Ok | QMessageBox::Open);
            msgBox->button(QMessageBox::Open)->setText(tr("Show log"));
        } else {
            msgBox->setStandardButtons(QMessageBox::Ok);
        }
        msgBox->setIcon(QMessageBox::Warning);
        msgBox->setCheckBox(new QCheckBox(tr("Do not show again")));

        int ret = msgBox->exec();
        settings.setValue("warnings/bad_fit_cfg", !msgBox->checkBox()->isChecked());
        msgBox->deleteLater();
        if (ret == QMessageBox::Open)
            QDesktopServices::openUrl(QUrl::fromLocalFile(gLogFile));
        fitFileConfig = FitConfig();
    }

    bool noInv = !(hasPath && fileOk);

    // Setup item mapper
    ItemMapper mapper;

    // parameters to watch.
    QVector<AParam> params;
    {
        auto &&tmp = admFileConfig.descriptor().parameters();
        for (auto &p : tmp)
            params.push_back(p);
    }
    // add fit parameters
    // here we hack the categories defined by the FitConfig::expectedFitParameters so
    // all parameters except those that belong to "concentrations" and "signal" are set to internals.
    {
        std::vector<AParam> fitParams = fitFileConfig.expectedFitParameters();
        for (AParam &p : fitParams) {
            QStringList ou;
            auto &in = p.categories;
            for (auto &v : in)
                ou.push_back(QString::fromStdString(v));
            if (!(ou.contains("concentrations") || ou.contains("signal")))
                in.push_back("internal");
            params.append(p);
        }
    }
    mapper.setParameters(params);

    // update the item mapper: keymap.
    errors.clear();
    QString keymapPath = admFileConfig.keymapFilePath();
    mapper.loadKeymap(keymapPath, &errors);
    for (const QString &e : qAsConst(errors))
        LOG_WARNING(QString("[CFG:KEYMAP] %1").arg(e));
    // now check keymap names.
    QStringList omiss, mmiss;
    QPair<bool, bool> results = mapper.checkKeymap(&mmiss, &omiss);
    for (const QString &m : qAsConst(mmiss))
        LOG_ERROR(tr("[CFG:KEYMAP] Missing/Inconsistent key %1").arg(m));
    for (const QString &m : qAsConst(omiss))
        LOG_WARNING(tr("[CFG:KEYMAP] Missing/Inconsistent key %1").arg(m));
    if (!results.first) {
        mmiss.replaceInStrings("<", "&lt;").replaceInStrings(">", "&gt;");
        QString text = tr("<p>Some mandatory key(s) cannot be found in the descriptor:</p>"
                          "%2"
                          "<p>Fix the keymap file and reload the configuration file.</p>")
                           .arg(QString("<ul><li>%1</li></ul>").arg(mmiss.join("</li><li>")));
        QMessageBox *msgBox = new QMessageBox(this);
        msgBox->setWindowTitle(tr("Error: invalid keymap"));
        msgBox->setText(text);
        msgBox->setIcon(QMessageBox::Critical);
        if (logToDisk) {
            msgBox->setStandardButtons(QMessageBox::Ok | QMessageBox::Open);
            msgBox->button(QMessageBox::Open)->setText(tr("Show log"));
        } else {
            msgBox->setStandardButtons(QMessageBox::Ok);
        }
        int ret = msgBox->exec();
        msgBox->deleteLater();
        if (ret == QMessageBox::Open)
            QDesktopServices::openUrl(QUrl::fromLocalFile(gLogFile));
        return;
    }

    // internal updates: Now everything should be fine

    mUdpSocket->setListening(false);  // stop listening while setting up everything.
    mInvalidTemplateReported = false; // reset the invalid template control flag.

    // stop current processes
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

    mFitController->waitForFinished();
    mFitController->setFitProcess(false);

    freeResultFileHandler();
    // reload data save settings
    mSaveData = settings.value("storage/save", true).toBool();
    QApplication::restoreOverrideCursor();

    mCommWidget->wlanWidget()->monitor(false);
    mCommWidget->wlanWidget()->setSSID("");
    mCommWidget->reset();

    // save scenarios (only if on the fly option is enabled !)
    saveScenariosOnTheFly();
    // reset data manager (clean up and change mapper)
    mDataManager->resetMapper(mapper);

    // save new configs
    mFitConfig = fitFileConfig;
    mAdmConfig = admFileConfig;

    mFCParams = mFitConfig.fcParameters();

    // Read scenario file.
    finfos.setFile(admFileConfig.scenarioFilePath());
    //    hasPath = finfos.exists();
    //    fileOk = true;
    errors.clear();
    int selIdx = -1;
    QVector<Scenario *> scenarios =
        loadScenariosFromXML(finfos.absoluteFilePath(), mDataManager->mapper(), &selIdx, &errors);
    for (const QString &e : qAsConst(errors))
        LOG_ERROR(QString("[CFG:SCN] %1").arg(e));

    mDataManager->setScenarios(scenarios, &selIdx); // set scenarios and optionally select selIdx
    // update instrument indicators
    finfos.setFile(admFileConfig.indicatorFilePath());
    //    hasPath = finfos.exists();
    //    fileOk = true;
    errors.clear();
    qDeleteAll(mInstrumentIndicators);
    mInstrumentIndicators.clear();
    mInstrumentIndicators = loadIndicatorsFromXML(finfos.absoluteFilePath(), mDataManager->mapper(), &errors);
    for (const QString &e : qAsConst(errors))
        LOG_ERROR(QString("[CFG:IND] %1").arg(e));
    QVector<BaseIndicatorPointer> qptrs;
    qptrs.reserve(mInstrumentIndicators.size());
    for (BaseIndicator *ind : qAsConst(mInstrumentIndicators)) {
        ind->setParent(this);
        qptrs.push_back(ind);
    }
    mInstrumentWidget->setupIndicators(qptrs);
    mTimestampWidget->setWarmupTime(mAdmConfig.laserWarmup() * 1000);
    // re-initialize file handler
    errors.clear();
    QString cname = QString("AD-i%1-v%2").arg(mAdmConfig.instrumentNumber()).arg(mAdmConfig.firmwareVersion());
    mLastDataRootDir = QString("%1/data/%2").arg(gDataDir, cname);
    QDir(mLastDataRootDir).mkpath(".");
    initResultFileHandler(mDataManager->lastStartTime());
    mFilenameWidget->setADMinFileName((mResultFileHandler && mResultFileHandler->isValid())
                                          ? QString::fromStdString(mResultFileHandler->currentFileName())
                                          : "N/A");

    // uncheck toggle inv. action (i.e. disable inversion) if we have no fit of if fit config is invalid)
    mActionToggleInv->setEnabled(!noInv);
    if (noInv) {
        mActionToggleInv->setChecked(false); // will remove any plot widget in the tab widget !
    }
    setupFitWidgets();

    bool fast = settings.value("monitoring/fast_calculations", true).toBool();
    mFitController->setConfiguration(mFitConfig, fast);
    mFitController->setFitProcess(true && mActionToggleInv->isChecked());

    mInversionWidget->setConfiguration(mFitConfig);

    // Reset UDP
    mUdpSocket->setBlockTemplate(mAdmConfig.descriptor());
    const NetworkConfig &nc = mAdmConfig.networkConfig();
    mUdpSocket->connectAndBind(nc.ipAddr, nc.port);

    mCommWidget->wlanWidget()->setSSID(nc.ssidName);
    mCommWidget->wlanWidget()->monitor(settings.value("monitoring/monitor_wifi", true).toBool());

    bool valid = mAdmConfig.isValid();
    mScenarioWidget->setEnabled(valid);

#ifndef SCN_SEL_ON_MAP
    mScenarioSelector->setEnabled(valid);
#else
    mMapWidget->selector()->setEnabled(valid);
#endif
    mFlightWidget->setEnabled(valid);
    mActionClear->setEnabled(valid);
    mActionReplay->setEnabled(valid);

    // update window title
    QString filename = QFileInfo(configPath).fileName();
    setWindowFilePath(QString("%1 (%2 - %3)").arg(QApplication::applicationName(), mAdmConfig.name(), filename));

    // check for scenarios: if no scenarios has been validated, ask user for scenario edition !
    if (scenarios.isEmpty()) {
        mmiss.replaceInStrings("<", "&lt;").replaceInStrings(">", "&gt;");
        QString text =
            tr("<p>No valid scenarios have been parsed from the configuration files. Without scenarios, no parameters "
               "can be selected in the map and plots widgets.</p>"
               "<p>Would you like to open the editor to create new scenario ?</p>");
        QMessageBox *msgBox = new QMessageBox(this);
        msgBox->setAttribute(Qt::WA_DeleteOnClose);
        msgBox->setWindowTitle(tr("Warning"));
        msgBox->setText(text);
        msgBox->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        msgBox->setIcon(QMessageBox::Warning);
        int ret = msgBox->exec();
        if (ret == QMessageBox::Yes)
            editScenarios(-1);
    }
    return;
}

/// @brief Open settings dialog.
void MainWindow::openSettings() {
    SettingsDialog *dialog = new SettingsDialog(this);
    QObject::connect(dialog, &SettingsDialog::settingsChanged, this, &MainWindow::updateFromSettings);
    bool SDFLS = mNoSDWarned;
    mNoSDWarned = true;
    dialog->exec();
    QSettings settings;
    mNoSDWarned = SDFLS || (mNoSDWarned = !settings.value("warnings/no_sd_detected", true).toBool());
    dialog->deleteLater();
}

/// @brief Manage incoming block from Amulse.
void MainWindow::incomingBlock(ABlock block) {
    mNoRefresh = false;
    double ctime = QDateTime::currentMSecsSinceEpoch() / 1000.0;
    QString pcmonKey = mDataManager->mapper()->keyName(ItemMapper::PCMON_TIME);
    if (!block.hasParameter(pcmonKey.toStdString())) {
        block.appendParameter(AParam(pcmonKey.toStdString(),
            ctime - EPOCH1904_OFFSET,
            Units::unitFromString("seconds"),
            StringList{"internal"},
            std::string("Acquisition PC System time (seconds since 01/01/1904)")));
    } else {
        block.updateValue(pcmonKey.toStdString(), ctime - EPOCH1904_OFFSET);
    }

    // All widgets that required an INPUT block can be updated here.

    // checking for SD card detection flag.
    QString sdPrmName = mDataManager->mapper()->keyName(ItemMapper::SD_DETECT);
    if (sdPrmName.size() && block.hasParameter(sdPrmName.toStdString())) {
        bool hasSD =
            block.parameter(mDataManager->mapper()->keyName(ItemMapper::SD_DETECT).toStdString()).intValue() != 0;
        checkSDCard(hasSD);
    }

    int nsats = block.parameter(mDataManager->mapper()->keyName(ItemMapper::NB_SATS).toStdString()).intValue();
    mCommWidget->gpsWidget()->updateStatus(nsats);

    mFitController->fitBlock(block, ctime);
}

/**
 * @brief Perform UI updates on block processed.
 * @param block Output (fit) processed block.
 * @param reports Vector of fit reports.
 * @param times Fit execution times.
 */
void MainWindow::onBlockProcessed(const ABlock &block, const QVector<FitReport> &reports, const ExecTimes &times) {
    if (mInversionEnabled) {
        LOG_INFO(QString("Block fit exec times: %1").arg(QString::fromStdString(times.toString())));

        mInversionWidget->updateInversion(mUdpSocket->blockRate(), reports);

        int i = 0;
        bool logSummary = gLogAppender && gLogAppender->detailsLevel() == Logger::LogLevel::Debug;
        for (const FitReport &r : reports) {
            i++;
            if (r.fitSuccess()) {
                LOG_DEBUG(QString("Fit #%1 (SUCCESS): %2").arg(i).arg(QString::fromStdString(r.reason)));
            } else {
                LOG_DEBUG(QString("Fit #%1 (FAIL): %2").arg(i).arg(QString::fromStdString(r.reason)));
            }
            if (logSummary)
                LOG_DEBUG(QString::fromStdString(r.summary()));
        }
        if (mShowFit) {
            for (int i = 0; i < mFitWidgets.size(); i++) {
                if (i >= reports.size()) {
                    mFitWidgets[i]->clearPlots();
                } else {
                    mFitWidgets[i]->setFitReport(reports[i]);
                }
            }
        }
    } else if (mShowFit) {
        for (FitPlotWidget *fw : qAsConst(mFitWidgets))
            fw->clearPlots();
    }

    bool newFlight;

    // add data to the manager.
    mDataManager->processBlock(block, &newFlight);

    // save data.
    int iver = block.parameter("iversion").intValue();
    // IMPORTANT: we must check for iversion because
    // we check for iversion because it's the easiest way to know if the block has been processed.
    if (mResultFileHandler && mSaveData && iver)
        mResultFileHandler->registerBlock(block);

    // All widgets that required an OUTPUT block can be updated here.
    mFilenameWidget->setADMinFileName((mResultFileHandler && mResultFileHandler->isValid())
                                          ? QString::fromStdString(mResultFileHandler->currentFileName())
                                          : "N/A");
    bool nosd = true;
    QString &&name = mDataManager->computeSensorFileName(&nosd);
    mFilenameWidget->setSensorFileName(name, nosd);

    // once everything is done
    mResetTimer.start();
}

/// @brief Report an invalid template from the UDP socket receiver.
void MainWindow::reportInvalidTemplate(int instr, int vers) {
    if (mInvalidTemplateReported)
        return;
    QMessageBox::critical(this,
        tr("Error"),
        tr("<h4 style='white-space:pre'>Invalid descriptor</h4>"
           "<p>The descriptor of the configuration (i:%1,v:%2) does not match the incoming data block (i:%3,v:%4). "
           "The application will not be able to parse AMULSE data.</p>")
            .arg(mAdmConfig.instrumentNumber())
            .arg(mAdmConfig.firmwareVersion())
            .arg(instr)
            .arg(vers),
        QMessageBox::Ok);
    // the warning message will be display only once !
    mInvalidTemplateReported = true;
}

/**
 * @brief Update some widgets when new values have been registered in the data manager's mapper.
 */
void MainWindow::onMapperUpdated() {
    mTimestampWidget->updateFromMapper();
    mGPSWidget->updateFromMapper();
    mInstrumentWidget->updateFromMapper();
    mFilenameWidget->setSensorFileName(mDataManager->computeSensorFileName());
}

/**
 * @brief Manage new output file creation on new flight.
 */
void MainWindow::onNewFlightCreated() {
    if (mSaveData && mNewFileOnNewFlight) {
        freeResultFileHandler();
        initResultFileHandler(mDataManager->lastStartTime());
        mFilenameWidget->setADMinFileName((mResultFileHandler && mResultFileHandler->isValid())
                                              ? QString::fromStdString(mResultFileHandler->currentFileName())
                                              : "N/A");
    }
}

/**
 * @brief Delete the internal file handler.
 * @details
 * the method properly closes the internal file handler, cleans up its associated file hierarchy and finally deletes
 * and nullify it.
 */
void MainWindow::freeResultFileHandler() {
    mScreenshotManager->setFolder(QString("%1/captures").arg(gDataDir));
    if (mResultFileHandler) {
        mResultFileHandler->close();
        try {
            mResultFileHandler->cleanup();
        } catch (...) {
            // not a big deal anyway...
        }
        delete mResultFileHandler;
        mResultFileHandler = nullptr;
    }
}

/**
 * @brief Setup the fit widgets in the central tab widget.
 * @param n Number of plot widgets that should be set in the tab widget.
 */
void MainWindow::setupFitWidgets() {
    mTabWidget->setUpdatesEnabled(false);
    const std::vector<SParam> &spectras = mFitConfig.spectras();
    int nt = 1;
    if (mInversionEnabled && mShowFit)
        nt += spectras.size();
    int nw = mTabWidget->count();
    // remove widgets until we match the correct number.
    while (nw > nt) {
        int iw = nw - 1;
        QWidget *w = mTabWidget->widget(iw);
        mTabWidget->removeTab(iw);
        mFitWidgets.takeLast();
        w->deleteLater();
        nw = iw;
    }
    // add widgets until we match the correct number.
    while (nw < nt) {
        FitPlotWidget *fw = new FitPlotWidget(this);
        mFitWidgets.push_back(fw);
        mTabWidget->addTab(fw, "");
        nw = mTabWidget->count();
    }
    mTabWidget->setUpdatesEnabled(true);
    for (int i = 1, j = 0; i < mTabWidget->count(); i++, j++) {
        const SParam &sp = spectras.at(j);
        int sid = sp.id, lid = sp.line.id;
        QString title = QString("%1 (S#%2, L#%3)")
                            .arg(sp.alias.empty() ? QString("Fit") : QString::fromStdString(sp.alias))
                            .arg(sid)
                            .arg(lid);
        mTabWidget->setTabText(i, title);
        mFitWidgets[j]->clearPlots();
    }
}

/**
 * @brief Initialize the file handler.
 * @param startTime Starting datetime.
 * @return True if the file handler has been successfully initialized, false otherwise.
 * @details
 * The method initializes the internal file handler using startTime for output file name format.
 *
 * If the the file handler has been already initialized, the method does nothing and returns false.
 *
 * Files are saved in a sub-directory of MainWindow::mLastDataDirectory which is set to the starting date time with
 * the format: yyyy-MM-dd_hh.mm.ss.
 *
 * The file name is formatted as: <yyyyMMddhhmmss>_ADMIN_<X>.results
 *
 * Where:
 *   - <yyyyMMddhhmmss> is the starting date time of the flight (as given in argument of the method).
 *   - <X> is a counter related to the rolling file appender.
 */
bool MainWindow::initResultFileHandler(QDateTime startTime) {
    if (mResultFileHandler)
        return false;
    if (startTime.isNull())
        startTime = QDateTime::currentDateTimeUtc();
    QString dirPath = QString("%1/%2").arg(mLastDataRootDir, startTime.toString("yyyy-MM-dd_hh.mm.ss"));
    QString filePrefix = QString("%1_ADMIN").arg(startTime.toString("yyyyMMddhhmmss"));

    QString fname = QString("%1/%2.results").arg(dirPath, filePrefix);
    if (QDir(dirPath).mkpath(".")) {
        mScreenshotManager->setFolder(dirPath);
    } else {
        mScreenshotManager->setFolder(QString("%1/captures").arg(gDataDir));
    }
    mResultFileHandler = new BinFileHandler(fname.toStdString(), true, 200 * 1024 * 1024);

    return mResultFileHandler->isValid();
}

/**
 * @brief Check for SD card status
 * @param sdDetected
 */
void MainWindow::checkSDCard(bool sdDetected) {
    if (mNoSDWarned || sdDetected)
        return;
    QSettings settings;
    if (settings.value("warnings/no_sd_detected", true).toBool()) {
        QString text = tr("<p>No SD card has been detected by the sensor.</p>"
                          "<p>Please insert a compatible SD card and <b>restart</b> the sensor.</p>"
                          "<p><i>If you continue, data will not be saved onboard.</i></p>");
        QMessageBox *msgBox = new QMessageBox(QMessageBox::Warning, tr("Attention"), text, QMessageBox::Ok, this);
        msgBox->setDefaultButton(QMessageBox::Ok);
        msgBox->setEscapeButton(QMessageBox::Ok);
        msgBox->setCheckBox(new QCheckBox(tr("Do not show again")));
        QObject::connect(msgBox, &QMessageBox::finished, this, [=](int) {
            QSettings settings;
            bool shouldWarn = !msgBox->checkBox()->isChecked();
            settings.setValue("warnings/no_sd_detected", shouldWarn);
            msgBox->deleteLater();
            mNoSDWarned = true;
            if (shouldWarn)
                QTimer::singleShot(30000, this, [=]() { mNoSDWarned = false; });
        });
        msgBox->setFocus();
        msgBox->exec();
    }
    mNoSDWarned = true;
}
