// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "mapitemeditor.hpp"
#include "dataitem.hpp"
#include "valuelineedit.hpp"

MapItemEditor::MapItemEditor(QWidget *parent) : QWidget(parent) {
    mCurrentItem = nullptr;

    mLblTitle = new QLabel(this);

    mLblValueMin = new QLabel(this);
    mLdtValueMin = new ValueLineEdit(false, false, this);
    mLdtValueMin->setAnimation(true);

    mLblValueMax = new QLabel(this);
    mLdtValueMax = new ValueLineEdit(false, false, this);
    mLdtValueMax->setAnimation(true);

    mBtnApply1 = new QToolButton(this);
    mBtnApply1->setIcon(QIcon::fromTheme("apply"));
    mBtnApply1->setToolButtonStyle(Qt::ToolButtonIconOnly);

    mSeparator = new QFrame(this);
    mSeparator->setFrameShape(QFrame::VLine);
    mSeparator->setFrameShadow(QFrame::Sunken);
    mSeparator->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);

    mLblThreshold = new QLabel(this);
    mLdtThreshold = new ValueLineEdit(true, true, this);

    mBtnApply2 = new QToolButton(this);
    mBtnApply2->setIcon(QIcon::fromTheme("apply"));
    mBtnApply2->setToolButtonStyle(Qt::ToolButtonIconOnly);

    for (QWidget *w : {mLdtValueMin, mLdtValueMax, mLdtThreshold}) {
        w->setMinimumWidth(40);
    }

    QHBoxLayout *layout = new QHBoxLayout(this);
    layout->addStretch(1);
    layout->addWidget(mLblTitle, 0, Qt::AlignLeft);
    layout->addWidget(mLblValueMin, 0, Qt::AlignLeft);
    layout->addWidget(mLdtValueMin, 1, Qt::AlignLeft);
    layout->addWidget(mLblValueMax, 0, Qt::AlignLeft);
    layout->addWidget(mLdtValueMax, 1, Qt::AlignLeft);
    layout->addWidget(mBtnApply1, 0, Qt::AlignLeft);
    layout->addWidget(mSeparator, 0, Qt::AlignRight);
    layout->addWidget(mLblThreshold, 0, Qt::AlignRight);
    layout->addWidget(mLdtThreshold, 1, Qt::AlignRight);
    layout->addWidget(mBtnApply2, 0, Qt::AlignRight);

    QObject::connect(mBtnApply1, &QToolButton::clicked, this, &MapItemEditor::onRABClicked);
    QObject::connect(mBtnApply2, &QToolButton::clicked, this, &MapItemEditor::onTABClicked);

    translateUI();

    updateFromItem(nullptr);
}

/// @brief Check color range line edit values and emit new data range if needed.
void MapItemEditor::onRABClicked() {
    bool okMin, okMax;
    double min, max;
    min = mLdtValueMin->text().toDouble(&okMin);
    max = mLdtValueMax->text().toDouble(&okMax);
    if (!(okMin && okMax)) {
        // should highlight the line edit in the gui ?!
        return;
    }
    emit colorRangeChanged(DataRange(min, max));
}

/// @brief Check Remanence threshold line edit value and emit new data range if needed.
void MapItemEditor::onTABClicked() {
    QString sval = mLdtThreshold->text();
    emit thresholdChanged(sval.isEmpty() ? qQNaN() : sval.toDouble());
}

/**
 * @brief Set item to edit.
 * @details
 * The method associates the given item to the editor and replaces the values displayed by those of the given item.
 *
 * Call the method with a null pointer to disassociate an item of the editor.
 */
void MapItemEditor::updateFromItem(DataItem *item) {

    DataRange colorRange;
    double threshold;
    mCurrentItem = item;
    if (mCurrentItem) {
        colorRange = mCurrentItem->colorRange();
        threshold = mCurrentItem->threshold();
    } else {
        colorRange = DataRange();
        threshold = qQNaN();
    }

    mLdtValueMin->setText(QString::number(colorRange.lower()));
    mLdtValueMax->setText(QString::number(colorRange.upper()));
    mLdtThreshold->setText(qIsNaN(threshold) ? "" : QString::number(threshold));
    mLblTitle->setText(QString("%1,").arg(mCurrentItem ? mCurrentItem->aliasAndUnits() : tr("No parameter")));
    // set enabled.
    setEnabled(mCurrentItem != nullptr);
}

/// @brief Translate widgets texts.
void MapItemEditor::translateUI() {
    mLblTitle->setText(QString("%1,").arg(mCurrentItem ? mCurrentItem->aliasAndUnits() : tr("No parameter")));
    mLblValueMin->setText(tr("Min."));
    mLblValueMax->setText(tr("Max."));
    mLblThreshold->setText(tr("Rem. %1").arg(QChar(0x2265)));

    mLdtValueMin->setStatusTip(tr("Enter a minimum value for the color range."));
    mLdtValueMax->setStatusTip(tr("Enter a maximum value for the color range."));
    mLdtThreshold->setStatusTip(tr("Enter a remanence threshold for the data."));
}

/// @brief Lock/Unlock user interactions with editable widgets.
void MapItemEditor::lockUI(bool locked) {
    bool enabled = !locked;
    mLdtValueMin->setEnabled(enabled);
    mLdtValueMax->setEnabled(enabled);
    mBtnApply1->setEnabled(enabled);
    mLdtThreshold->setEnabled(enabled);
    mBtnApply2->setEnabled(enabled);
}

/**
 * @brief Set new color range.
 * @param range Range to set.
 * @details
 * The method sets a new color range and updates the widgets values consequently.
 * The method does @b not emit ItemEditor::colorRangeChanged signal.
 */
void MapItemEditor::setColorRange(const DataRange &range) {
    mLdtValueMin->setText(QString::number(range.lower()));
    mLdtValueMax->setText(QString::number(range.upper()));
}

/**
 * @overload
 * @brief Set new color range.
 * @param range Range to set.
 * @details
 * The method sets a new color range and updates the widgets values consequently.
 * The method does @b not emit ItemEditor::colorRangeChanged signal.
 */
void MapItemEditor::setColorRange(const QCPRange &range) {
    mLdtValueMin->setText(QString::number(range.lower));
    mLdtValueMax->setText(QString::number(range.upper));
}

/**
 * @brief Set new visibility data range.
 * @param range Range to set.
 * @details
 * The method sets a new visibility data range and updates the widgets values consequently.
 * The method does @b not emit ItemEditor::thresholdChanged signal.
 */
void MapItemEditor::setThreshold(double value) { mLdtThreshold->setText(qIsNaN(value) ? "" : QString::number(value)); }

/**
 * @brief Check boundaries of the color range line edits.
 * @details
 * The method checks whether or not the current extrema set in the line edits are ordered and swaps them if needed.
 */
void MapItemEditor::checkLineEditValues() {
    QLineEdit *ldt = qobject_cast<QLineEdit *>(QObject::sender());
    if (!ldt || ldt != mLdtValueMin || ldt != mLdtValueMax)
        return;
    QString slower = mLdtValueMin->text(), supper = mLdtValueMax->text();
    double lower = slower.toDouble(), upper = supper.toDouble();
    if (lower > upper) {
        mLdtValueMin->setText(supper);
        mLdtValueMax->setText(slower);
    }
}
