// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "mapwidget.hpp"
#include "basemapselector.hpp"
#include "baseplot.hpp"
#include "basertplotwidget.hpp"
#include "colorscale.hpp"
#include "dataitem.hpp"
#include "mapitemeditor.hpp"
#include "menusection.hpp"
#include "scenarioselector.hpp"

// arcgis
#include "Basemap.h"
#include "Geometry.h"
#include "GeometryEngine.h"
#include "Map.h"

//#include "GeometryBuilder.h"
#include "ArcGISTiledLayer.h"
#include "EnvelopeBuilder.h"
#include "LatitudeLongitudeGrid.h"
#include "SimpleLineSymbol.h"
using namespace Esri::ArcGISRuntime;

/// @brief Get a string representation of the given status.
QString loadStatusToString(LoadStatus status) {
    switch (status) {
    case LoadStatus::Loaded:
        return QCoreApplication::tr("Loaded !");
    case LoadStatus::Loading:
        return QCoreApplication::tr("Loading !");
    case LoadStatus::FailedToLoad:
        return QCoreApplication::tr("Failed to load !");
    case LoadStatus::NotLoaded:
        return QCoreApplication::tr("Not loaded !");
    case LoadStatus::Unknown:
    default:
        return QCoreApplication::tr("Unknown !");
    }
}

/// @brief Constructor.
MapWidget::MapWidget(QWidget *parent) : QWidget(parent) {
    mZoomFactor = 2;
    mCurrentItem = nullptr;
    mEnableParamSelection = true;

    mSplitter = new QSplitter(Qt::Vertical, this);
    QPalette pal = mSplitter->palette();
    pal.setColor(QPalette::Window, Qt::white);
    mSplitter->setAutoFillBackground(true);
    mSplitter->setPalette(pal);

    QWidget *topWidget = new QWidget(mSplitter);

    QSettings settings;
    // color scale orientation
    bool horiz = settings.value("map/cs_horiz", false).toBool();
    Qt::Orientation csOrient = horiz ? Qt::Horizontal : Qt::Vertical;
    // basemap selector orientation.
    horiz = settings.value("map/bs_horiz", false).toBool();
    Qt::Orientation bsOrient = horiz ? Qt::Horizontal : Qt::Vertical;

    mToolBar = new QToolBar(topWidget);
    mItemEditor = new MapItemEditor(this);
    if (mItemEditor->layout())
        mItemEditor->layout()->setContentsMargins(0, 0, 0, 0);

    mBasemapsWidget = new BasemapSelector(bsOrient, this);
    mBasemapsWidget->setBackgroundRole(QPalette::Base);
    mBasemapsWidget->setFrameStyle(QFrame::NoFrame);

    // arcgis map view
    mSpatialReference = SpatialReference::wgs84();
    mMapView = new MapGraphicsView(topWidget);

    mMap = new Map(mSpatialReference, this);
    mMapView->setMap(mMap);

    mLastViewpoint = mMapView->currentViewpoint(ViewpointType::BoundingGeometry);
    setupGrid();

    // Setup the color scale.
    mColorScale = new ColorScale(csOrient, topWidget);

    // Top-widget layout
    QBoxLayout::Direction direction;
    QBoxLayout *innerLayout;
    Qt::Alignment csAlign;
    QSizePolicy csSizePolicy;
    if (csOrient == Qt::Vertical) {
        direction = QBoxLayout::LeftToRight;
        csSizePolicy = QSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Expanding);
        csAlign = Qt::AlignRight;
    } else {
        direction = QBoxLayout::TopToBottom;
        csSizePolicy = QSizePolicy(QSizePolicy::Expanding, QSizePolicy::MinimumExpanding);
        csAlign = Qt::AlignBottom;
    }
    mColorScale->setSizePolicy(csSizePolicy);
    innerLayout = new QBoxLayout(direction);
    innerLayout->setContentsMargins(0, 0, 0, 0);
    innerLayout->setSpacing(0);
    innerLayout->addWidget(mMapView, 2);
    innerLayout->addWidget(mColorScale, 0, csAlign);

    direction = bsOrient == Qt::Horizontal ? QBoxLayout::TopToBottom : QBoxLayout::LeftToRight;
    QBoxLayout *outerLayout = new QBoxLayout(direction, topWidget);
    outerLayout->setMenuBar(mToolBar);
    outerLayout->addWidget(mBasemapsWidget, 0);
    outerLayout->addLayout(innerLayout, 1);

    mBasemapsWidget->setVisible(false);

    mBaseRtPlotWidget = new BaseRtPlotWidget(mSplitter);

    mSplitter->addWidget(topWidget);
    mSplitter->addWidget(mBaseRtPlotWidget);

    mSplitter->setStretchFactor(0, 3);
    mSplitter->setStretchFactor(1, 1);

    QVBoxLayout *vLayout = new QVBoxLayout(this);
    vLayout->setContentsMargins(0, 0, 0, 0);
    vLayout->addWidget(mSplitter);

    setupActions();

    QObject::connect(mBasemapsWidget, &BasemapSelector::selectionChanged, this, &MapWidget::onBasemapSelectionChanged);

    QObject::connect(mItemEditor, &MapItemEditor::colorRangeChanged, this, &MapWidget::onIERangeChanged);
    QObject::connect(mItemEditor, &MapItemEditor::thresholdChanged, this, &MapWidget::onIEThresholdChanged);

    QObject::connect(mColorScale, &ColorScale::rangeChanged, this, &MapWidget::onCScaleRangeChanged);

    // plot connections.
    QObject::connect(mBaseRtPlotWidget, &BaseRtPlotWidget::thresholdChanged, this, &MapWidget::onBPThresholdChanged);
    QObject::connect(mBaseRtPlotWidget, &BaseRtPlotWidget::colorRangeChanged, this, &MapWidget::onBPRangeChanged);

    QObject::connect(mBaseRtPlotWidget->plot(), &BasePlot::mousePositionChanged, this, &MapWidget::setPositionText);

    mBaseRtPlotWidget->plot()->setContextMenuPolicy(Qt::CustomContextMenu);
    mMapView->setContextMenuPolicy(Qt::CustomContextMenu);

    QObject::connect(mBaseRtPlotWidget->plot(), &BasePlot::customContextMenuRequested, this, [=](const QPoint &pos) {
        emit this->contextMenuRequested(mBaseRtPlotWidget->plot()->mapTo(this, pos), false);
    });

    QObject::connect(mMapView, &MapGraphicsView::customContextMenuRequested, this,
                     [=](const QPoint &pos) { emit this->contextMenuRequested(mMapView->mapToParent(pos), true); });

    loadUI();
}

/**
 * @brief Enable/Disable parameter selection from context menu.
 * @param enable True to enable, false to disable.
 */
void MapWidget::enableParameterSelection(bool enable) { mEnableParamSelection = enable; }

/// @brief Check whether or not parameter selection from context menu is enabled.
bool MapWidget::parameterSelectionEnabled() const { return mEnableParamSelection; }

/**
 * @brief Update threshold range line of the RT plot when Item editor afferent value has changed.
 * @param value New value to set.
 */
void MapWidget::onIEThresholdChanged(double value) {
    mBaseRtPlotWidget->setThLineValue(value);
    emit thresholdChanged(value);
}

/**
 * @brief Update item editor threshold value when plot threshold line has been changed.
 * @param value New value to set.
 */
void MapWidget::onBPThresholdChanged(double value) {
    mItemEditor->setThreshold(value);
    emit thresholdChanged(value);
}

/**
 * @brief Update item editor and plot range lines when the color scale is updated.
 * @param range New color range to set.
 */
void MapWidget::onCScaleRangeChanged(const DataRange &range) {
    mItemEditor->setColorRange(range);
    mBaseRtPlotWidget->setRgLinesValues(range);
    emit colorRangeChanged(range);
}

/**
 * @brief Update color scale and plot range lines when item editor is updated.
 * @param range New color range to set.
 * @note
 * The method also emits MapWidget::colorRangeChanged.
 */
void MapWidget::onIERangeChanged(const DataRange &range) {
    mColorScale->setRange(range);
    mBaseRtPlotWidget->setRgLinesValues(range);
    emit colorRangeChanged(range);
}

/**
 * @brief Update color scale and item editor when plot range lines is updated.
 * @param range New color range to set.
 * @note
 * The method also emits MapWidget::colorRangeChanged.
 */
void MapWidget::onBPRangeChanged(const DataRange &range) {
    mColorScale->setRange(range);
    mItemEditor->setColorRange(range);
    emit colorRangeChanged(range);
}

/// @brief Clear all graphs and overlays from their parent widgets.
void MapWidget::clearGraphsAndOverlays() {
    mBaseRtPlotWidget->clearPlot();
    mMapView->graphicsOverlays()->clear();
}

/// @brief Add an overlay to the map.
void MapWidget::addOverlay(GraphicsOverlay *overlay) {
    auto *model = mMapView->graphicsOverlays();
    if (!model->contains(overlay))
        model->append(overlay);
}

/// @brief Remove overlay from the map.
void MapWidget::removeOverlay(GraphicsOverlay *overlay) {
    auto *model = mMapView->graphicsOverlays();
    while (model->contains(overlay))
        model->removeOne(overlay);
}

/// @brief Add a new graph to the plot widget.
void MapWidget::addGraph(GraphContainerPointer data) { mBaseRtPlotWidget->addGraph(data); }

/**
 * @brief Remove graphs from the plot widget that make use of the given data container.
 * @sa BaseRtPlotWidget::removeGraph
 */
void MapWidget::removeGraph(GraphContainerPointer data) { mBaseRtPlotWidget->removeGraph(data); }

/**
 * @brief Update the map widget from the given data item.
 * @param item Data item to process.
 * @param overlays Map overlays that should be set on the map.
 * @param graphContainers Data containers that should be drawn on the plot.
 *
 * The method registers the given item as the current item of the widget.
 *
 * If item is null, overlays and graphContainers are quietly ignored: the widget should not display any data.
 *
 * Otherwise the method updates the overlays of the map with the given one and setup plot's graph for the given graph
 * data containers.
 */
void MapWidget::updateFromItem(DataItem *item, QVector<GraphicsOverlay *> overlays,
                               GraphContainerVector graphContainers) {

    DataRange colorRange, visibleRange;
    QString wTitle;
    bool hasItem = item != nullptr;

    if (hasItem) {
        colorRange = item->colorRange();
        mBaseRtPlotWidget->setYLabel(item->aliasAndUnits(true));
        mBaseRtPlotWidget->setGraphs(graphContainers);
        auto *model = mMapView->graphicsOverlays();
        model->clear();
        for (auto *o : overlays)
            model->append(o);
    } else {
        colorRange = DataRange(0, 1);
        clearGraphsAndOverlays();
    }

    mCurrentItem = item;
    mColorScale->setRangeDrag(hasItem);
    mColorScale->setRangeZoom(hasItem);
    mColorScale->setRange(colorRange.toQCPRange());
    mItemEditor->updateFromItem(mCurrentItem);

    if (mCurrentItem) {
        mMAReset->setEnabled(true);
        mBaseRtPlotWidget->setRgLinesVisible(mShowRange);
        mBaseRtPlotWidget->setThLineVisible(mShowThreshold);
        mBaseRtPlotWidget->setRgLinesValues(mCurrentItem->colorRange());
        mBaseRtPlotWidget->setThLineValue(mCurrentItem->threshold());
    } else {
        mMAReset->setEnabled(false);
        mBaseRtPlotWidget->setRgLinesVisible(false);
        mBaseRtPlotWidget->setThLineVisible(false);
    }

    refreshPlot();
}

/// @brief Scroll the X-axis range of the plot widget.
void MapWidget::scrollXRange(double reference) { mBaseRtPlotWidget->scrollXRange(reference); }

/**
 * @brief Refresh the plot.
 * @details
 * The method refreshes the plot, it:
 *
 * 1. Updates the last displayed in the textbox.
 * 2. Schedules a (queued) replot.
 */
void MapWidget::refreshPlot() {
    // setTextBoxText requests a (queued) replot !
    mBaseRtPlotWidget->setTextBoxText(mCurrentItem ? mCurrentItem->valueAndUnits() : "N/A");
    mBaseRtPlotWidget->refreshPlot();
}

/// @brief Set new gradient for the color scale.
void MapWidget::setColorGradient(const QCPColorGradient &gradient) { mColorScale->setGradient(gradient); }

/**
 * @brief Update the mouse position label.
 * @param x X-axis coordinate value
 * @param y Y-axis coordinate value
 * @details
 * The method formats the text to be set in the mouse-position label (right part of the toolbar).
 */
void MapWidget::setPositionText(double x, double y) {
    QString label;
    if (mCurrentItem) {
        QString sym(QString::fromStdString(Units::unitToString(mCurrentItem->parameter().units)));
        if (sym.compare("N/A") == 0)
            label = QString("%1, %2")
                        .arg(QDateTime::fromMSecsSinceEpoch(x * 1000, Qt::UTC).toString("HH:mm:ss.zzz (UTC)"))
                        .arg(y, 0, 'g', 3);
        else
            label = QString("%1, %2 %3")
                        .arg(QDateTime::fromMSecsSinceEpoch(x * 1000, Qt::UTC).toString("HH:mm:ss.zzz (UTC)"))
                        .arg(y, 0, 'g', 3)
                        .arg(sym);
    } else {
        label = QString("%1, %2")
                    .arg(QDateTime::fromMSecsSinceEpoch(x * 1000, Qt::UTC).toString("HH:mm:ss.zzz (UTC)"))
                    .arg(y, 12);
    }
    mBaseRtPlotWidget->mousePositionLabel()->setText(label);
}

#ifdef SCN_SEL_ON_MAP

/// @brief Get the scenario selector.
ScenarioSelector *MapWidget::selector() const { return mScenarioSelector; }

/// @brief Get the action associated to the internal scenario selector.
QAction *MapWidget::selectorAction() const { return mActionScenario; }

#endif

/// @brief Save the splitter state.
void MapWidget::saveUI() {
    QSettings settings;
    settings.setValue("layouts/map_widget/splitter", mSplitter->saveState());
}

/// @brief Load/restore the splitter state.
void MapWidget::loadUI() {
    QSettings settings;
    QByteArray array = settings.value("layouts/map_widget/splitter").toByteArray();
    if (array.size())
        mSplitter->restoreState(array);
}

/// @brief Load/restore the splitter state.
void MapWidget::translateUI() {

    mMACenter->setText(tr("Auto-center"));
    mMARescale->setText(tr("Rescale"));
    mMAZoomIn->setText(tr("Zoom in"));
    mMAZoomOut->setText(tr("Zoom out"));
    mMALock->setText(mMALock->isChecked() ? tr("Unlock UI") : tr("Lock UI"));
    mMAReset->setText(tr("Reset"));

    mMACenter->setToolTip(tr("Enable/Disable auto-centering on new position."));
    mMARescale->setToolTip(tr("Rescale the map to fit the currently displayed positions."));
    mMAZoomIn->setToolTip(tr("Zoom in."));
    mMAZoomOut->setToolTip(tr("Zoom out."));
    mMABasemaps->setToolTip(tr("Select basemap."));
    mMALock->setToolTip(tr("Enable/Disable user interactions with the range and threshold."));
    mMAReset->setToolTip(tr("Reset range and threshold to their last saved values."));
    for (QAction *action : {mMACenter, mMARescale, mMAZoomIn, mMAZoomOut, mMABasemaps, mMALock, mMAReset})
        action->setStatusTip(action->toolTip());
    mItemEditor->translateUI();
#ifdef SCN_SEL_ON_MAP
    mScenarioSelector->translateUI();
#endif
    mBasemapsWidget->translateUI();
    mBaseRtPlotWidget->translateUI();
}

/// @brief Get the current splitter state.
QByteArray MapWidget::splitterState() const { return mSplitter->saveState(); }

/// @brief Set splitter's state.
void MapWidget::setSplitterState(const QByteArray &state) { mSplitter->restoreState(state); }

/// @brief Get the internal plot widget.
QPointer<BaseRtPlotWidget> MapWidget::plotWidget() const { return mBaseRtPlotWidget; }

/// @brief Get the internal mapView.
QPointer<MapGraphicsView> MapWidget::mapView() const { return mMapView; }

/// @brief Get the size hint of the widget (800x600).
QSize MapWidget::sizeHint() const { return QSize(800, 600); }

/**
 * @brief Add the user interactions actions of the mapview widget to the given menu.
 * @param menu Pointer to the menu to update.
 * @details
 * If menu is null, nothing is done. Otherwise the method appends the following actions:
 *
 * - A custom separator with the key-word Interaction as title.
 * - The auto-center switch action.
 * - The "expand to fit data" action.
 * - The Zoom-in action.
 * - The Zoom-out action.
 */
void MapWidget::addMapInteractionsActions(QMenu *menu) {
    if (menu) {
        menu->addAction(new MenuSection(tr("Interactions"), menu));
        menu->addAction(mMACenter);
        menu->addAction(mMARescale);
        menu->addAction(mMAZoomIn);
        menu->addAction(mMAZoomOut);
    }
}

/// @brief Setup actions.s
void MapWidget::setupActions() {

    mMACenter = new QAction(QIcon::fromTheme("location"), "", this);
    mMACenter->setCheckable(true);
    mMARescale = new QAction(QIcon::fromTheme("expand"), "", this);
    mMAZoomIn = new QAction(QIcon::fromTheme("zoom-in"), "", this);
    mMAZoomOut = new QAction(QIcon::fromTheme("zoom-out"), "", this);
    mMABasemaps = new QAction(QIcon::fromTheme("layers"), "", this);
    mMABasemaps->setCheckable(true);

    mMALock = new QAction(QIcon::fromTheme("unlock"), "", this);
    mMALock->setCheckable(true);
    mMAReset = new QAction(QIcon::fromTheme("reset"), "", this);
    mMAReset->setEnabled(false);

    QObject::connect(mMARescale, &QAction::triggered, this, &MapWidget::rescaleMap);
    QObject::connect(mMAZoomIn, &QAction::triggered, this, &MapWidget::zoomIn);
    QObject::connect(mMAZoomOut, &QAction::triggered, this, &MapWidget::zoomOut);

    QObject::connect(mMABasemaps, &QAction::triggered, mBasemapsWidget, &BasemapSelector::setVisible);

    QObject::connect(mMAReset, &QAction::triggered, this, &MapWidget::onResetClicked);
    QObject::connect(mMALock, &QAction::toggled, this, &MapWidget::onLockToggled);

    QObject::connect(mMAZoomOut, &QAction::triggered, this, &MapWidget::zoomOut);
    QObject::connect(mMAZoomOut, &QAction::triggered, this, &MapWidget::zoomOut);

    mToolBar->addAction(mMABasemaps);
    mToolBar->addSeparator();
    mToolBar->addAction(mMACenter);
    mToolBar->addAction(mMARescale);
    mToolBar->addSeparator();
    mToolBar->addAction(mMAZoomIn);
    mToolBar->addAction(mMAZoomOut);
    mToolBar->addSeparator();
#ifdef SCN_SEL_ON_MAP
    mScenarioSelector = new ScenarioSelector(mToolBar);
    mScenarioSelector->layout()->setContentsMargins(0, 0, 0, 0);
    mActionScenario = mToolBar->addWidget(mScenarioSelector);
    mToolBar->addSeparator();
#endif
    QWidget *spacer = new QWidget();
    spacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    mToolBar->addWidget(spacer);
    mToolBar->addWidget(mItemEditor);
    mToolBar->addSeparator();
    mToolBar->addAction(mMAReset);
    mToolBar->addSeparator();
    mToolBar->addAction(mMALock);
}

/// @brief (Un)Block user interactions with the editor and color scale.
void MapWidget::onLockToggled(bool checked) {
    mMALock->setIcon(QIcon::fromTheme(checked ? "lock" : "unlock"));
    mMALock->setText(checked ? tr("Unlock UI") : tr("Lock UI"));
    mMAReset->setEnabled(!checked);
    mItemEditor->lockUI(checked);
    mColorScale->lockUI(checked);
    /*
     // WE DO NOT WANT to modify threshold and color range from the plot, it's too risky for the moment.
    mBaseRtPlotWidget->setRgLinesEditable(!checked);
    mBaseRtPlotWidget->setThLineEditable(!checked);
    */
}

/**
 * @brief Reset the threshold of the current item to their last saved values (from the configuration).
 */
void MapWidget::onResetClicked() {
    qDebug() << "I should reset the thresholds to their last saved values, for the current item !";
    if (!mCurrentItem)
        return;
    // change color range.
    const DataRange &a = mCurrentItem->saveColorRange();
    mItemEditor->setColorRange(a);
    mColorScale->setRange(a);
    mBaseRtPlotWidget->setRgLinesValues(a);

    emit colorRangeChanged(a);
    // change threshold.
    double b = mCurrentItem->saveThreshold();
    mBaseRtPlotWidget->setThLineValue(b);
    mItemEditor->setThreshold(b);
    emit thresholdChanged(b);
}

/// @brief Center the map at the given location.
void MapWidget::centerMapOn(double latitude, double longitude) {
    Point pos(longitude, latitude, SpatialReference::wgs84());
    mMapView->setViewpoint(pos.extent());
}

/// @brief Get the toolbar's action related to map auto center.
QAction *MapWidget::autoCenterAction() { return mMACenter; }

/**
 * @brief Center the map on the position of the last added graphic.
 */
void MapWidget::centerOnLastPosition() {
    auto *model = mMapView->graphicsOverlays();
    if (model->size()) {
        auto *graphics = model->last()->graphics();
        if (graphics->size()) {
            mMapView->setViewpoint(graphics->last()->geometry().extent());
        }
    }
}

/**
 * @brief Rescale the the map so that all graphics overlays are shown.
 * @note
 * If no item are displayed on the map, the method does nothing.
 */
void MapWidget::rescaleMap() {
    SpatialReference sref = mMapView->spatialReference(); // SpatialReference::wgs84();
    auto *model = mMapView->graphicsOverlays();
    int no = model->size();
    if (no) {
        EnvelopeBuilder envBuilder(GeometryEngine::project(model->first()->extent(), sref));
        if (no > 1) {
            auto it = (++model->begin());
            while (it != model->end()) {
                envBuilder.unionWithEnvelope(GeometryEngine::project((*it)->extent(), sref));
                it++;
            }
        }
        if (qIsNaN(envBuilder.width()) || qIsNaN(envBuilder.height()) || !envBuilder.center().isValid()) {
            return;
        }
        Envelope env(envBuilder.center(), envBuilder.width() + 1e-5, envBuilder.height() + 1e-5);
        mMapView->setViewpoint(env);
    }
}

/// @brief Show/Hide the map grid.
void MapWidget::setGridVisible(bool visible) {
    bool reg = mMapGrid == mMapView->grid();
    if (visible) {
        if (!reg)
            mMapView->setGrid(mMapGrid);
    } else {
        if (reg)
            mMapView->setGrid(nullptr);
    }
}

/// @brief Setup the map grid.
void MapWidget::setupGrid() {
    mMapGrid = new LatitudeLongitudeGrid(this);
    mMapGrid->setLabelPosition(GridLabelPosition::TopRight);
    mMapGrid->setLabelFormat(LatitudeLongitudeGridLabelFormat::DecimalDegrees);
    int gridLevels = mMapGrid->levelCount();
    for (int level = 0; level < gridLevels; level++) {
        const float width = 1 + level;
        SimpleLineSymbol *lineSym = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor(Qt::white), width, this);
        mMapGrid->setLineSymbol(level, lineSym);
        constexpr float size = 14.0f;
        TextSymbol *textSym =
            new TextSymbol("text", QColor(Qt::black), size, HorizontalAlignment::Left, VerticalAlignment::Bottom, this);
        textSym->setHaloColor(QColor("white"));
        textSym->setHaloWidth(2.0f + level);
        mMapGrid->setTextSymbol(level, textSym);
    }
}

/// @brief Get the spatial reference of the internal ARCGIS map.
Esri::ArcGISRuntime::SpatialReference MapWidget::spatialReference() const { return mSpatialReference; }

/// @brief Get the zoom factor used by the actions ZoomIn and ZoomOut.
double MapWidget::zoomFactor() const { return mZoomFactor; }

/**
 * @brief Set zoom factor.
 * @param zoomFactor New zoom factor.
 * @details
 * The zoom factor is used to Zoom actions.
 * The current scale is multiplied (zoom out) or divided (zoom in) by the zoom factor to obtain the new map scale.
 * @note
 * The zoom factor cannot be less than 1.01.
 */
void MapWidget::setZoomFactor(double zoomFactor) { mZoomFactor = qMin(1.01, zoomFactor); }

/**
 * @brief Update some properties from application settings.
 * @details
 * The method updates the threshold and color range lines settings of the widget's plot.
 */
void MapWidget::updateFromSettings() {
    QSettings settings;

    // setup grid visibility.
    setGridVisible(settings.value("map/m_showgrid", false).toBool());

    // update plot widget settings.
    mBaseRtPlotWidget->updateFromSettings();

    // get color range and threshold lines settings.
    settings.beginGroup("graph");
    mShowRange = settings.value("show_rgl", true).toBool();
    QString rgc = settings.value("rgl_color", "#0000ff").toString();
    double rgw = settings.value("rgl_width", 1.0).toDouble();
    bool rgf = settings.value("rgl_fill", true).toBool();
    QString rgfc = settings.value("rgl_fill_color", "#0000ff").toString();
    double rgfa = settings.value("rgl_fill_alpha", 0.25).toDouble();
    bool rgEditable = settings.value("rgl_editable", false).toBool();
    mShowThreshold = settings.value("show_thl", true).toBool();
    QString thc = settings.value("thl_color", "#ff0000").toString();
    double thw = settings.value("thl_width", 1.0).toDouble();
    bool thEditable = settings.value("thl_editable", false).toBool();
    settings.endGroup();

    mBaseRtPlotWidget->setRgLinesVisible(mShowRange);
    mBaseRtPlotWidget->fillRgLines(rgf);
    mBaseRtPlotWidget->setRgLinesFillBrush(buildBrush(rgfc, rgfa));
    mBaseRtPlotWidget->setRgLinesPen(buildPen(rgc, rgw));
    mBaseRtPlotWidget->setRgLinesEditable(rgEditable);

    mBaseRtPlotWidget->setThLineVisible(mShowThreshold);
    mBaseRtPlotWidget->setThLinePen(buildPen(thc, thw));
    mBaseRtPlotWidget->setThLineEditable(thEditable);

    if (mCurrentItem) {
        mBaseRtPlotWidget->setRgLinesValues(mCurrentItem->colorRange());
        mBaseRtPlotWidget->setThLineValue(mCurrentItem->threshold());
    } else {
        mBaseRtPlotWidget->setRgLinesVisible(false);
        mBaseRtPlotWidget->setThLineVisible(false);
    }
}

void MapWidget::zoom(double factor) {
    double cscale = mMapView->mapScale();
    if (!qIsNaN(cscale))
        mMapView->setViewpointScale(cscale * factor);
}

/// @brief Zoom In using MapWidget::zoomFactor()
void MapWidget::zoomIn() { zoom(1.0 / mZoomFactor); }

/// @brief Zoom Out using MapWidget::zoomFactor()
void MapWidget::zoomOut() { zoom(mZoomFactor); }

/// @brief Set new basemap on basemap view selection changes.
void MapWidget::onBasemapSelectionChanged(const BasemapInfos &selected) {
    mLastViewpoint = mMapView->currentViewpoint(ViewpointType::BoundingGeometry);
    // we should setup new base map.
    int id = selected.id;
    Basemap *basemap = nullptr;

    if (id == -1) {
        if (!QFileInfo::exists(selected.path)) {
            // file path is not valid: reset immediately to no basemap.
            mBasemapsWidget->resetCustomItem();
            mBasemapsWidget->selectId(-2, true);
            return;
        }
        QUrl url(QUrl::fromLocalFile(selected.path));
        ArcGISTiledLayer *tiledLayer = new ArcGISTiledLayer(url);
        basemap = new Basemap(tiledLayer);
        tiledLayer->setParent(basemap);

    } else if (id >= 0) {
        BasemapStyle bt = static_cast<BasemapStyle>(id);
        switch (bt) {
        case BasemapStyle::OsmStandard:
            basemap = new Basemap(BasemapStyle::OsmStandard);
            break;
        case BasemapStyle::ArcGISImagery:
            basemap = new Basemap(BasemapStyle::ArcGISImagery);
            break;
        case BasemapStyle::ArcGISStreets:
            basemap = new Basemap(BasemapStyle::ArcGISStreets);
            break;
        case BasemapStyle::ArcGISTopographic:
            basemap = new Basemap(BasemapStyle::ArcGISTopographic);
            break;
        default:
            Q_ASSERT_X(false, "MapWidget::onBasemapSelectionChanged", "Basemap type not managed");
        }
    }
    // set the basemap to the Map.
    Map *map = nullptr;
    static const SpatialReference wgs84 = SpatialReference::wgs84();
    if (basemap) {
        map = new Map(basemap, this);
        basemap->setParent(map);
        QObject::connect(basemap, &Basemap::loadStatusChanged, this, &MapWidget::loadStatusChanged);
        //       basemap->load();
    } else {
        map = new Map(wgs84, this);
    }
    mMapView->setMap(map); // setting a new map will load the basemap layer !
    if (!basemap) {
        auto geom = mLastViewpoint.targetGeometry();
        if (geom.spatialReference() != wgs84) {
            geom = GeometryEngine::project(geom, wgs84);
            mLastViewpoint = Viewpoint(geom);
        }
        mMapView->setViewpoint(mLastViewpoint);
    }
    mMap->basemap()->deleteLater();
    mMap->deleteLater();
    mMap = map;
    SpatialReference mapRef = mMap->spatialReference();
    if (mapRef != mSpatialReference) {
        mSpatialReference = mapRef;
        emit spatialReferenceChanged(mSpatialReference);
    }
}

void MapWidget::loadStatusChanged(LoadStatus status) {
    if (!(status == LoadStatus::FailedToLoad || status == LoadStatus::Loaded))
        return;

    // get infos for logging purpose:
    BasemapInfos infos = mBasemapsWidget->selectedBasemap();
    QString type;
    if (infos.id == -1) {
        type = tr("File (%1)").arg(infos.path);
    } else {
        BasemapStyle id = static_cast<BasemapStyle>(infos.id);
        switch (id) {
        case BasemapStyle::OsmStandard:
            type = "OSM";
            break;
        case BasemapStyle::ArcGISImagery:
            type = QCoreApplication::tr("Imagery");
            break;
        case BasemapStyle::ArcGISStreets:
            type = QCoreApplication::tr("Streets");
            break;
        case BasemapStyle::ArcGISTopographic:
            type = QCoreApplication::tr("Topographic");
            break;
        default:
            type = QCoreApplication::tr("Unknown");
        }
    }

    if (status == LoadStatus::FailedToLoad) {
        Basemap *basemap = qobject_cast<Basemap *>(QObject::sender());
        qDebug() << basemap->loadError().message() << basemap->loadError().additionalMessage();
        LOG_INFO(tr("Loading tile layer: %1: FAILED").arg(type));
        QSettings settings;
        if (settings.value("warnings/map_load_fail", true).toBool()) {
            QString text = tr("<p>Cannot load the tile layer: %1</p>"
                              "<p>The application will reset the map with no basemap.</p>")
                               .arg(type);
            QMessageBox *msgBox =
                new QMessageBox(QMessageBox::Critical, tr("Failed to load"), text, QMessageBox::Ok, this);
            msgBox->setDefaultButton(QMessageBox::Ok);
            msgBox->setEscapeButton(QMessageBox::Ok);
            msgBox->setCheckBox(new QCheckBox(tr("Do not show again")));
            msgBox->exec();
            settings.setValue("warnings/map_load_fail", !msgBox->checkBox()->isChecked());
            delete msgBox;
        }
        mBasemapsWidget->selectId(-2, true);
        return;
    }

    LOG_INFO(tr("Loading tile layer: %1: SUCCESS").arg(type));
    Basemap *basemap = qobject_cast<Basemap *>(QObject::sender());
    if (!basemap)
        return;
    auto *m = basemap->baseLayers();
    ArcGISTiledLayer *tl = nullptr;
    int n = m->size();
    if (m && n) {
        for (int i = 0; i < n; i++) {
            tl = qobject_cast<ArcGISTiledLayer *>(m->at(i));
            if (tl)
                break;
        }
    }
    if (!tl)
        return;
    auto *ca = tl ? tl->tileCache() : nullptr;
    if (ca) {
        Envelope env = ca->fullExtent();
        try {
            auto lastGeom = mLastViewpoint.targetGeometry();
            auto newGeom = env;
            if (lastGeom.spatialReference() != newGeom.spatialReference()) {
                lastGeom = GeometryEngine::project(lastGeom, newGeom.spatialReference());
            }
            if (!GeometryEngine::contains(newGeom, lastGeom))
                mLastViewpoint = Viewpoint(env);
        } catch (const std::exception &ex) {
            qDebug() << "Nasty exception occurs while trying to set the viewpoint after loading a tile layer...";
            qDebug() << ex.what();
        }
    }
    QObject::disconnect(QObject::sender());
    mMapView->setViewpoint(mLastViewpoint);
    return;
}
