// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "mediacontrolwidget.hpp"

/// @brief Constructor.
MediaControlWidget::MediaControlWidget(QWidget *parent) : QWidget(parent) {
    mState = ReplayState::STOPPED;

    mLoadButton = new QToolButton(this);
    mLoadButton->setIcon(QIcon::fromTheme("folder-open"));
    mStopButton = new QToolButton(this);
    mStopButton->setIcon(QIcon::fromTheme("stop"));
    mStopButton->setEnabled(false);
    mPlayButton = new QToolButton(this);
    mPlayButton->setIcon(QIcon::fromTheme("play"));
    mGoToStartButton = new QToolButton(this);
    mGoToStartButton->setIcon(QIcon::fromTheme("skip-to-start"));
    mGoToEndButton = new QToolButton(this);
    mGoToEndButton->setIcon(QIcon::fromTheme("skip-to-end"));
    mRepeatButton = new QToolButton(this);
    mRepeatButton->setIcon(QIcon::fromTheme("repeat"));
    mRepeatButton->setCheckable(true);
    mRepeatButton->setChecked(false);

    mSlider = new QSlider(Qt::Horizontal, this);
    mSpinBox = new QSpinBox(this);
    mCbxRate = new QComboBox(this);

    mCbxRate->addItem("1:1", QVariant(1));
    mCbxRate->addItem("1:5", QVariant(5));
    mCbxRate->addItem("1:10", QVariant(10));
    mCbxRate->addItem("1:20", QVariant(20));
    mCbxRate->setCurrentIndex(0);

    QVBoxLayout *layout = new QVBoxLayout(this);
    QHBoxLayout *hlayout = new QHBoxLayout;

    hlayout->addWidget(mSlider, 2);
    hlayout->addWidget(mSpinBox, 0);
    layout->addLayout(hlayout);

    hlayout = new QHBoxLayout;
    hlayout->addWidget(mLoadButton);
    hlayout->addStretch(0);
    hlayout->addWidget(mStopButton);
    hlayout->addWidget(mGoToStartButton);
    hlayout->addWidget(mPlayButton);
    hlayout->addWidget(mGoToEndButton);
    hlayout->addStretch(0);
    hlayout->addWidget(mRepeatButton, 0);
    hlayout->addWidget(mCbxRate, 0);
    layout->addLayout(hlayout);

    layout->addStretch(0);

    QObject::connect(mSpinBox, qOverload<int>(&QSpinBox::valueChanged), mSlider, &QSlider::setValue);
    QObject::connect(mSlider, &QSlider::valueChanged, mSpinBox, &QSpinBox::setValue);
    QObject::connect(mSlider, &QSlider::valueChanged, this, &MediaControlWidget::frameIndexChanged);
    QObject::connect(mPlayButton, &QToolButton::clicked, this, &MediaControlWidget::playClicked);
    QObject::connect(mStopButton, &QToolButton::clicked, this, &MediaControlWidget::stop);
    QObject::connect(mLoadButton, &QToolButton::clicked, this, &MediaControlWidget::loadRequest);
    QObject::connect(mGoToStartButton, &QToolButton::clicked, this, &MediaControlWidget::goToStart);
    QObject::connect(mGoToEndButton, &QToolButton::clicked, this, &MediaControlWidget::goToEnd);
    QObject::connect(mRepeatButton, &QToolButton::toggled, this, &MediaControlWidget::autoRepeat);
    QObject::connect(mCbxRate, qOverload<int>(&QComboBox::currentIndexChanged), this, &MediaControlWidget::updateFrameStep);

    mSlider->setRange(0, 0);

    setupControls(1);
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Minimum);
    // setFixedHeight(this->layout()->totalMinimumSize().height(););
}

/// @brief Get the play rate.
int MediaControlWidget::frameStep() const { return mCbxRate->itemData(mCbxRate->currentIndex()).toInt(); }

QSize MediaControlWidget::sizeHint() const { return layout()->totalMinimumSize(); }

/// @brief Set new values for the slider.
void MediaControlWidget::setFrameIndex(int value) {
    if (value == mSlider->value())
        return;
    mSlider->setValue(value);
}

void MediaControlWidget::setupControls(int frameCount) {
    bool old1 = mSlider->blockSignals(true), old2 = mSpinBox->blockSignals(true);
    mSlider->setRange(0, frameCount - 1);
    mSpinBox->setRange(0, frameCount - 1);
    mSlider->setValue(0);
    mSpinBox->setValue(0);
    mSlider->blockSignals(old1);
    mSpinBox->blockSignals(old2);
}

/**
 * @brief Reset the slider and spinbox to zero.
 * @details
 * The method blocks the signal of both the slider and spinbox and resets their display value to zero.
 */
void MediaControlWidget::resetControls() {
    bool old1 = mSlider->blockSignals(true), old2 = mSpinBox->blockSignals(true);
    mSlider->setValue(0);
    mSpinBox->setValue(0);
    mSlider->blockSignals(old1);
    mSpinBox->blockSignals(old2);
}

/// @brief Emit \a play or \a pause signals based on the current movie state.
void MediaControlWidget::playClicked() {
    switch (mState) {
    case ReplayState::STOPPED:
    case ReplayState::PAUSED:
        emit play();
        break;
    case ReplayState::PLAYING:
        emit pause();
        break;
    }
}

/// @brief Update buttons on movie state changes.
void MediaControlWidget::setState(ReplayState state) {
    if (state != mState) {
        mState = state;
        switch (state) {
        case ReplayState::STOPPED:
            mStopButton->setEnabled(false);
            mPlayButton->setIcon(QIcon::fromTheme("play"));
            resetControls(); // really ?!
            break;
        case ReplayState::PLAYING:
            mStopButton->setEnabled(true);

            mPlayButton->setIcon(QIcon::fromTheme("pause"));
            break;
        case ReplayState::PAUSED:
            mStopButton->setEnabled(true);
            mPlayButton->setIcon(QIcon::fromTheme("play"));
            break;
        }
    }
}

/// @brief Update the play rate.
void MediaControlWidget::updateFrameStep(int /*index*/) { emit frameStepChanged(frameStep()); }
