// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "movingavg.hpp"

#include <algorithm>
#include <cmath>

MovingAverage::MovingAverage(size_t nSamples)
    : mSamples(std::max((size_t)1, nSamples)), mTotalSamples{std::max((size_t)1, nSamples)}, mNumSamples{0}, mTotal{0} {

}

/**
 * @brief Set new number of samples to account for the mobile average.
 * @param n New (maximum) number of samples to account for.
 * @details
 * Unless the total number has not changed, the method also discards the previous calculations.
 */
void MovingAverage::setTotalSamples(size_t n) {
    if (n == mTotalSamples)
        return;
    mSamples.resize(n);
    if (n < mTotalSamples)
        mSamples.shrink_to_fit();
    mTotalSamples = n;
    reset();
}

#include <iostream>

/// @brief Add a new sample to the moving average.
MovingAverage &MovingAverage::push(double sample) {
    if (!std::isfinite(sample))
        return *this;
    if (mNumSamples < mTotalSamples)
        mSamples[mNumSamples++] = sample;
    else {
        double &oldest = mSamples[mNumSamples++ % mTotalSamples];
        mTotal -= oldest;
        oldest = sample;
    }
    mTotal += sample;
    return *this;
}

/// @brief Reset the mobile average.
void MovingAverage::reset() {
    mTotal = 0;
    mNumSamples = 0;
}

/// @brief Get the current average value.
double MovingAverage::operator()() const {
    return mNumSamples ? static_cast<double>(mTotal / std::min(mNumSamples, mTotalSamples)) : 0.0;
}

ExponentialMovingAverage::ExponentialMovingAverage(double smoothing) : alpha(smoothing), last(0.0), calls(0) {}

double ExponentialMovingAverage::operator()(double sample) {
    double beta = 1.0 - alpha;
    if (!std::isnan(sample)) {
        last = alpha * sample + beta * last;
        calls++;
    }
    return calls >0 ? (last / (1.0 - std::pow(beta,calls))) : last;
}

void ExponentialMovingAverage::setSmoothing(double smoothing) { alpha = smoothing; }

/// @brief Reset the mobile average.
void ExponentialMovingAverage::reset() {
    last = 0.0;
    calls = 0;
}
