// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "rtplotwidget.hpp"

#include "baseplot.hpp"
#include "dataitem.hpp"
#include "datamanager.hpp"
#include "menusection.hpp"
#include "textbox.hpp"

RtPlotWidget::RtPlotWidget(int plotIndex, QWidget *parent) : BaseRtPlotWidget(parent) {
    mIndex = plotIndex;

    mEnableParamSelection = true;
    mCurrentItem = nullptr;

    QObject::connect(mPlot, &BasePlot::mousePositionChanged, this, &RtPlotWidget::setPositionText);
    mPlot->setContextMenuPolicy(Qt::CustomContextMenu);
}

/**
 * @brief Update the plot widget from the given data item.
 * @param item Data item to process.
 * @param graphContainers Data containers that should be drawn on the plot.
 *
 * The method registers the given item as the current item of the widget.
 *
 * If item is null, graphContainers are quietly ignored: the widget should not display any data.
 *
 * Otherwise the method updates plot's graphs in order to display the given graph data containers.
 */
void RtPlotWidget::updateFromItem(DataItem *item, GraphContainerVector graphContainers) {
    if (item) {
        setGraphs(graphContainers, false);
        mPlot->setYLabel(item->aliasAndUnits(true));
        if (!freeMode())
            mPlot->autoRescaleY();
        setName(item->alias());
    } else {
        clearPlot();
        setName(tr("None"));
    }
    mCurrentItem = item;
    if (mCurrentItem) {
        setThLineVisible(mShowThreshold);
        setThLineValue(mCurrentItem->threshold());
    } else {
        setThLineVisible(false);
    }
    refreshPlot();
}

/**
 * @brief Enable/Disable parameter selection from context menu.
 * @param enable True to enable, false to disable.
 */
void RtPlotWidget::enableParameterSelection(bool enable) { mEnableParamSelection = enable; }

/// @brief Check whether or not parameter selection from context menu is enabled.
bool RtPlotWidget::parameterSelectionEnabled() const { return mEnableParamSelection; }

/**
 * @brief Refresh the plot.
 * @details
 * The method refreshes the plot, it:
 *
 * 1. Updates the last displayed in the textbox.
 * 2. Schedules a (queued) replot.
 *
 * @note
 * Such method should be called anytime a new block of data has been processed in the data manager.
 */
void RtPlotWidget::refreshPlot() {
    mLastValueBox->setText(mCurrentItem ? mCurrentItem->valueAndUnits() : "N/A");
    BaseRtPlotWidget::refreshPlot();
}

void RtPlotWidget::updateFromSettings() {
    BaseRtPlotWidget::updateFromSettings();

    // get threshold line settings.
    QSettings settings;
    settings.beginGroup("graph");
    mShowThreshold = settings.value("show_thl", true).toBool();
    QString thc = settings.value("thl_color", "#ff0000").toString();
    double thw = settings.value("thl_width", 1.0).toDouble();
    settings.endGroup();

    setRgLinesVisible(false);
    setThLineEditable(false);
    setThLineVisible(mShowThreshold);
    setThLinePen(buildPen(thc, thw));
    if (mCurrentItem) {
        setThLineValue(mCurrentItem->threshold());
    } else {
        setThLineVisible(false);
    }
}

/**
 * @brief Update the mouse position label.
 * @param x X-axis coordinate value
 * @param y Y-axis coordinate value
 * @details
 * The method formats the text to be set in the mouse-position label (right part of the toolbar).
 */
void RtPlotWidget::setPositionText(double x, double y) {
    QString label;
    if (mCurrentItem) {
        QString sym(QString::fromStdString(Units::unitToString(mCurrentItem->parameter().units)));
        if (sym.compare("N/A") == 0)
            label = QString("%1, %2")
                        .arg(QDateTime::fromMSecsSinceEpoch(x * 1000, Qt::UTC).toString("HH:mm:ss.zzz (UTC)"))
                        .arg(y, 0, 'g', 3);
        else
            label = QString("%1, %2 %3")
                        .arg(QDateTime::fromMSecsSinceEpoch(x * 1000, Qt::UTC).toString("HH:mm:ss.zzz (UTC)"))
                        .arg(y, 0, 'g', 3)
                        .arg(sym);
    } else {
        label = QString("%1, %2")
                    .arg(QDateTime::fromMSecsSinceEpoch(x * 1000, Qt::UTC).toString("HH:mm:ss.zzz (UTC)"))
                    .arg(y, 12);
    }
    mMousePositionLabel->setText(label);
}

int RtPlotWidget::index() const { return mIndex; }

DataItem *RtPlotWidget::currentItem() { return mCurrentItem; }
