// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "scenario.hpp"
#include "dataitem.hpp"
#include "indicators.hpp"
#include "itemmapper.hpp"

/// @brief Constructor.
Scenario::Scenario(QObject *parent) : QObject(parent) {
    mItemsModel = new ItemModel(this);
    mIndicatorsModel = new IndicatorModel(this);
}

/**
 * @brief Full constructor.
 * @param name Name of the scenario.
 * @param comments Comments/Description of the scenario.
 * @param layout Name of the preferred layout to load when Scenario is selected.
 * @param items List of data item to manage. The scenario takes ownership of the items.
 * @param indicators List of indicators to manage. The scenario takes ownership of the indicators.
 * @param parent Optional parent object.
 */
Scenario::Scenario(const QString &name, const QString &comments, const QVector<DataItem *> &items,
                   const QVector<BaseIndicator *> &indicators, QObject *parent)
    : QObject(parent), mName(name), mComments(comments) {
    mItemsModel = new ItemModel(items, this);
    mIndicatorsModel = new IndicatorModel(indicators, this);
}

/// @brief Destructor.
Scenario::~Scenario() {}

/**
 * @brief Create a new scenario with the same content as the given one.
 * @param scenario Scenario to copy.
 * @param parent Optional QObject parent to set to the new scenario. Default to nullptr.
 * @return Pointer to the new Scenario.
 */
Scenario *Scenario::copyFrom(Scenario *scenario, QObject *parent) {
    if (!scenario)
        return nullptr;
    QVector<DataItem *> items;
    auto *im = scenario->itemsModel();
    for (DataItem *item : im->items()) {
        items.push_back(DataItem::copyFrom(item));
    }
    QVector<BaseIndicator *> indicators;
    for (BaseIndicator *item : scenario->mIndicatorsModel->indicators()) {
        BaseIndicator *copy = BaseIndicator::copyFrom(item);
        if (copy)
            indicators.push_back(copy);
    }
    Scenario *scn =
        new Scenario(scenario->name(), scenario->comments(),  items, indicators, parent);
    auto *om = scn->itemsModel();
    int ridx = im->mapItemRowIndex();
    if (ridx != -1)
        om->setData(om->index(ridx, ItemModel::mapColumn), true);
    for (int i = 0; i < N_PLOTS; ++i) {
        ridx = im->plotItemRowIndex(i);
        if (ridx != -1)
            om->setData(om->index(ridx, ItemModel::plotColumnStart + i), true);
    }
    return scn;
}

/// @brief Get the scenario name.
QString Scenario::name() const { return mName; }

/// @brief Set a name for this scenario.
void Scenario::setName(const QString &name) { mName = name; }

/// @brief Get the scenario comments.
QString Scenario::comments() const { return mComments; }

/// @brief Set comments for this scenario.
void Scenario::setComments(const QString &comments) { mComments = comments; }

/// @brief Get the number of items stored in the scenario.
int Scenario::itemCount() const { return mItemsModel->rowCount(); }

/// @brief Get the number of indicators stored in the scenario.
int Scenario::indicatorCount() const { return mIndicatorsModel->rowCount(); }

/// @brief Get a XML representation of the scenario.
QByteArray Scenario::toXML() const {
    QString content = QString("<scenario>"
                              "<name>%1</name>"
                              "<description>%2</description>"
                              "%3"
                              "%4"
                              "</scenario>")
                          .arg(mName, QString("\n%1\n").arg(mComments.trimmed()), mItemsModel->toXML(),
                               mIndicatorsModel->toXML());
    return content.toLatin1();
}

/// @brief Get the internal data items model.
ItemModel *Scenario::itemsModel() const { return mItemsModel; }

/// @brief Get the internal indicators model.
IndicatorModel *Scenario::indicatorsModel() const { return mIndicatorsModel; }

/// @brief Reset the values of the internal models.
void Scenario::resetValues() {
    mItemsModel->resetValues();
    mIndicatorsModel->resetValues();
}

/**
 * @brief Register new values in scenario items and indicators from the given block of data.
 * @param block Block of data to process.
 */
void Scenario::registerValues(const ABlock &block) {
    mItemsModel->registerValues(block);
    mIndicatorsModel->registerValues(block);
}

/**
 * @brief Register new values in scenario items and indicators from the given item mapper.
 * @param ItemMapper Item mapper to search.
 */
void Scenario::registerValues(const ItemMapper *mapper) {
    Q_ASSERT_X(mapper, "Scenario::registerValues", "Input mapper cannot be null !");
    mItemsModel->registerValues(mapper);
    mIndicatorsModel->registerValues(mapper);
}
