// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "scenarioeditor.hpp"
#include "dndviewstyle.hpp"
#include "lineedit.hpp"
#include "scenario.hpp"
#include "scenariosmodel.hpp"

/// @brief Constructor.
ScenarioEditor::ScenarioEditor(QWidget *parent) : QWidget(parent) {

    mScenarioModel = new ScenariosModel(this);

    QToolBar *toolBar = new QToolBar(this);
    mActionAdd = new QAction(QIcon::fromTheme("add"), "", this);
    mActionDuplicate = new QAction(QIcon::fromTheme("duplicate"), "", this);
    mActionDuplicate->setEnabled(false);
    mActionDelete = new QAction(QIcon::fromTheme("remove"), "", this);
    mActionDelete->setEnabled(false);
    toolBar->addAction(mActionAdd);
    toolBar->addAction(mActionDuplicate);
    toolBar->addAction(mActionDelete);

    mListView = new QListView(this);
    mListView->setSelectionBehavior(QListView::SelectRows);
    mListView->setSelectionMode(QListView::SingleSelection);
    mListView->setModel(mScenarioModel);

    // view dnd
    mListView->setDragEnabled(true);
    mListView->setDropIndicatorShown(true);
    mListView->setDefaultDropAction(Qt::MoveAction);
    mListView->setDragDropMode(QTableView::InternalMove);
    mListView->setDragDropOverwriteMode(false);
    DnDViewStyle *tvStyle = new DnDViewStyle;
    tvStyle->setParent(this);
    mListView->setStyle(tvStyle);

    mLdtName = new LineEdit(this);
    mLdtName->setValidator(new NotEmptyValidator(this));
    mLdtName->setEnabled(false);
    mTxeDescription = new QTextEdit(this);
    mTxeDescription->setMinimumHeight(80);
    mTxeDescription->setEnabled(false);

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->setMenuBar(toolBar);
    //    layout->addWidget(toolBar, 0);
    layout->addWidget(mListView, 3);
    layout->addWidget(getSeparator(this));
    layout->addWidget(new QLabel(tr("Scenario Name"), this), 0);
    layout->addWidget(mLdtName, 0);
    layout->addWidget(new QLabel(tr("Description"), this), 0);
    layout->addWidget(mTxeDescription, 1);

    QObject::connect(mActionAdd, &QAction::triggered, this, &ScenarioEditor::addScenario);
    QObject::connect(mActionDuplicate, &QAction::triggered, this, &ScenarioEditor::duplicateScenario);
    QObject::connect(mActionDelete, &QAction::triggered, this, &ScenarioEditor::deleteScenario);

    QObject::connect(mLdtName, &QLineEdit::editingFinished, this, &ScenarioEditor::onEditingFinished);
    QObject::connect(mTxeDescription, &QTextEdit::textChanged, this, &ScenarioEditor::onEditingFinished);
    QObject::connect(mListView->selectionModel(), &QItemSelectionModel::selectionChanged, this,
                     &ScenarioEditor::onSelectionChanged);

    translateUI();
}

/// @brief Destructor.
ScenarioEditor::~ScenarioEditor() {}

///@brief Translate widgets texts.
void ScenarioEditor::translateUI() {
    mLdtName->setPlaceholderText(tr("Enter a scenario name."));
    mTxeDescription->setPlaceholderText(tr("Enter a description about the scenario."));
    QString msg;
    msg = tr("Add a new scenario.");
    mActionAdd->setToolTip(msg);
    mActionAdd->setStatusTip(msg);
    msg = tr("Duplicate the selected scenario.");
    mActionDuplicate->setToolTip(msg);
    mActionDuplicate->setStatusTip(msg);
    msg = tr("Remove the selected scenario.");
    mActionDelete->setToolTip(msg);
    mActionDelete->setStatusTip(msg);
}

/// @brief Get the selected scenario index.
int ScenarioEditor::selectedScenarioIndex() const { return selectedIndex().row(); }

/**
 * @brief Get the selected scenario.
 * @return Pointer to the selected scenario (from the view) or a null pointer if no scenario is selected.
 */
QPointer<Scenario> ScenarioEditor::selectedScenario() const { return mScenarioModel->scenario(selectedIndex()); }

/// @brief Get the model index that is currently selected in the listview.
QModelIndex ScenarioEditor::selectedIndex() const {
    auto *m = mListView->selectionModel();
    return m->hasSelection() ? m->selectedIndexes().first() : QModelIndex();
}

/**
 * @brief Setup new scenarios to edit.
 * @param scenarios Scenarios to edit.
 * @details
 * The method creates a copy of each given scenarios and sets in the internal scenarios model.
 */
void ScenarioEditor::setupScenarios(const QVector<Scenario *> scenarios, int *selectIndex) {
    int rowIdx = selectIndex ? (*selectIndex) : -1;
    if (rowIdx >= scenarios.size())
        rowIdx = -1;
    QVector<Scenario *> tmp;
    for (Scenario *scn : scenarios) {
        if (scn)
            tmp.push_back(Scenario::copyFrom(scn));
    }
    mScenarioModel->setScenarios(tmp);
    auto *m = mListView->selectionModel();
    QModelIndex idx = mScenarioModel->index(rowIdx, 0);
    if (idx.isValid()) {
        m->select(idx, QItemSelectionModel::Select);
    } else {
        m->clear();
    }
}

/**
 * @brief Get copies of the scenarios.
 * @return Vector of scenarios.
 * @details
 * The method creates a copy of each scenarios stored in the widget's scenario model and returns a vector.
 * @attention
 * It is user responsibility to delete the pointers when no more used.
 */
QVector<Scenario *> ScenarioEditor::scenarios() const {
    QVector<Scenario *> tmp;
    for (Scenario *scn : mScenarioModel->scenarios()) {
        if (scn)
            tmp.push_back(Scenario::copyFrom(scn));
    }
    return tmp;
}

/// @brief Get the number of scenarios stored in the internal model.
int ScenarioEditor::scenariosCount() const { return mScenarioModel->rowCount(); }

/// @brief Get the number of parameters stored in the scenario at given index.
int ScenarioEditor::itemsCount(int index) const {
    if (index < 0 || index >= mScenarioModel->rowCount())
        return 0;
    return mScenarioModel->scenario(mScenarioModel->index(index, 0))->itemsModel()->rowCount();
}

// SLOTS

/**
 * @brief Update scenario name or description.
 * @details
 * This slot managed name and description  edition of the currently selected scenario.
 */
void ScenarioEditor::onEditingFinished() {
    auto list = mListView->selectionModel()->selectedIndexes();
    if (list.isEmpty())
        return;
    int rowIndex = list.first().row();
    int colIndex = -1;
    QLineEdit *ldt = qobject_cast<QLineEdit *>(QObject::sender());
    QTextEdit *txe = qobject_cast<QTextEdit *>(QObject::sender());
    QString content;
    if (ldt) {
        colIndex = 0;
        content = ldt->text();
    } else if (txe) {
        colIndex = 1;
        content = txe->toPlainText();
    } else {
        return;
    }
    mScenarioModel->setData(mScenarioModel->index(rowIndex, colIndex), content);
}

/**
 * @brief Manage user selection changes in the view.
 * @details
 * The slot manages actions to perform when user has selected (or deselected) a scenario in the view:
 *
 * - Name and description in the informations groupbox are updated.
 * - ScenarioEditor::selectedScenarioChanged is emitted.
 */
void ScenarioEditor::onSelectionChanged(const QItemSelection & /*selected*/, const QItemSelection & /*deselected*/) {
    QString name, description;
    QModelIndex idx = selectedIndex();
    bool isValid = idx.isValid();
    if (isValid) {
        auto p = mScenarioModel->scenario(idx);
        name = p->name();
        description = p->comments();
    }
    mLdtName->setText(name);
    mTxeDescription->setPlainText(description);

    mLdtName->setEnabled(isValid);
    mTxeDescription->setEnabled(isValid);
    mActionDuplicate->setEnabled(isValid);
    mActionDelete->setEnabled(isValid);

    emit selectedScenarioChanged(mScenarioModel->scenario(idx));
}

/// @brief Add a new scenario.
void ScenarioEditor::addScenario() {
    Scenario *scn = new Scenario;
    scn->setName(tr("New scenario"));
    QModelIndex index = mScenarioModel->addScenario(scn);
    if (index.isValid()) {
        mListView->setCurrentIndex(index);
        mListView->scrollTo(index);
    } else {
        delete scn;
    }
}

/// @brief Duplicate the selected scenario.
void ScenarioEditor::duplicateScenario() {
    QModelIndex index = mScenarioModel->duplicateScenario(selectedScenario(), true);
    if (index.isValid()) {
        mListView->setCurrentIndex(index);
        mListView->scrollTo(index);
    }
}

/// @brief Delete the selected scenario.
void ScenarioEditor::deleteScenario() {
    QModelIndex idx = selectedIndex();
    mScenarioModel->removeScenario(idx);
}
