// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "scenarioseditor.hpp"

#include "scenariosmodel.hpp"
#include "scenario.hpp"
#include "scenarioeditor.hpp"
#include "itemseditor.hpp"
#include "indicatorseditor.hpp"

/// @brief Constructor.
ScenariosEditor::ScenariosEditor(QWidget *parent) : QDialog(parent) {

    mSplitter = new QSplitter(this);
    mScenarioEditor = new ScenarioEditor(mSplitter);
    QWidget *rightWidget = new QWidget(mSplitter);
    mTabWidget = new QTabWidget(rightWidget);
    mTabWidget->setMovable(false);
    mItemsEditor = new ItemsEditor(mTabWidget);
    mIndicatorsEditor = new IndicatorsEditor(mTabWidget);
    mTabWidget->addTab(mItemsEditor, tr("Parameters"));
    mTabWidget->addTab(mIndicatorsEditor, tr("Indicators"));

    QVBoxLayout *layout = new QVBoxLayout(rightWidget);
    layout->addWidget(mTabWidget);
    QMargins margins;
    margins = layout->contentsMargins();
    margins.setTop(0);
    margins.setBottom(0);
    layout->setContentsMargins(margins);

    mSplitter->addWidget(mScenarioEditor);
    mSplitter->addWidget(rightWidget);
    mSplitter->setStretchFactor(0, 1);
    mSplitter->setStretchFactor(1, 3);

    mButtonBox = new QDialogButtonBox(this);
    mButtonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    mButtonBox->button(QDialogButtonBox::Ok)->setAutoDefault(false);
    mButtonBox->button(QDialogButtonBox::Ok)->setDefault(false);

    // update editors layout.

    for (auto *w : {
             static_cast<QWidget *>(mScenarioEditor), static_cast<QWidget *>(mItemsEditor),
             static_cast<QWidget *>(mIndicatorsEditor),
         }) {
        margins = w->layout()->contentsMargins();
        margins.setTop(0);
        //    margins.setBottom(0);
        if (w == mScenarioEditor)
            margins.setLeft(0);
        w->layout()->setContentsMargins(margins);
    }

    layout = new QVBoxLayout(this);
    layout->addWidget(mSplitter, 2);
    layout->addWidget(mButtonBox, 0);

    QObject::connect(mScenarioEditor, &ScenarioEditor::selectedScenarioChanged, this, &ScenariosEditor::onScenarioChanged);
    QObject::connect(mButtonBox, &QDialogButtonBox::accepted, this, &ScenariosEditor::sanityCheck);
    QObject::connect(mButtonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);

    translateUI();
}

/// @brief Destructor.
ScenariosEditor::~ScenariosEditor() {}

void ScenariosEditor::keyPressEvent(QKeyEvent *e) {
    auto k = e->key();
    if (k == Qt::Key_Return || k == Qt::Key_Enter)
        return;
    QDialog::keyPressEvent(e);
}

/// @brief Get he size hint for this widget.
QSize ScenariosEditor::sizeHint() const {
    QSize sz(1024, 768);
    if (parentWidget())
        sz = parentWidget()->size() * 0.85;
    return sz;
}

/// @brief Translate widgets texts.
void ScenariosEditor::translateUI() {
    setWindowTitle(tr("Scenario editor"));
    mScenarioEditor->translateUI();
    mItemsEditor->translateUI();
    mIndicatorsEditor->translateUI();
    mTabWidget->setTabText(0, tr("Parameters"));
    mTabWidget->setTabText(1, tr("Indicators"));
}

/**
 * @brief Check scenarios health before accept the dialog.
 * @details
 * The method is responsible to check that there is still some scenario and that scenario
 * contains at least an item.
 *
 * If either conditions are not fullfilled, a warning message box is raise asking the user if she is still OK to continue...
 */
void ScenariosEditor::sanityCheck() {
    QSettings settings;
    bool warn = settings.value("warnings/empty_scn", true).toBool();
    if (!warn) {
        QDialog::accept();
        return;
    }
    // check for empty scenario and/or no more scenario.
    int ns = mScenarioEditor->scenariosCount();
    bool noScenarios = ns == 0;
    bool emptyScenarios = false;
    int ne = 0;
    for (int i = 0; i < ns; ++i) {
        if (mScenarioEditor->itemsCount(i) == 0) {
            ne++;
            emptyScenarios = true;
        }
    }

    if (warn && (noScenarios || emptyScenarios)) {
        QString text;
        QString msg = tr("<p>Do you really wish to continue ?</p>");
        if (noScenarios) {
            text = tr("<p>All scenarios have been removed...</p>"
                      "<p>If you accept the dialog, all scenarios will be removed from the configuration. You will not able to "
                      "monitor parameters on the map until a scenario is added again to the configuration.</p>");
        } else {
            if (ne == ns) {
                text = tr("<p>All scenarios are empty.</p>"
                          "<p>If you accept the dialog, scenarios will be replaced. You will not able to monitor parameters on "
                          "the map until a parameter is added to one of the scenarios.</p>");
            } else {
                text = tr("<p>Some scenarios do not store any parameters.</p>"
                          "<p>If you accept the dialog, scenarios will be replaced. You will not able to monitor parameters on "
                          "the map until from empty scenarios.</p>");
            }
        }
        QMessageBox *msgBox = new QMessageBox;
        msgBox->setWindowTitle(tr("Warning"));
        msgBox->setStandardButtons(QMessageBox::Yes | QMessageBox::No);
        msgBox->setDefaultButton(QMessageBox::No);
        msgBox->setEscapeButton(QMessageBox::No);
        msgBox->setCheckBox(new QCheckBox(tr("Do not show again")));
        msgBox->setIcon(QMessageBox::Warning);
        msgBox->setText(text);
        msgBox->setInformativeText(msg);
        bool ok = msgBox->exec() == QMessageBox::Yes;
        settings.setValue("warnings/empty_scn", !msgBox->checkBox()->isChecked());
        delete msgBox;
        if (!ok)
            return;
    }
    QDialog::accept();
    return;
}

/// @brief Set the list of parameters for items and indicators creation.
void ScenariosEditor::setParameters(const QVector<AParam> &params) {
    mItemsEditor->setParameters(params);
    mIndicatorsEditor->setParameters(params);
}

/**
 * @brief Setup new scenarios to edit.
 * @param scenarios Scenarios to edit.
 * @param selectIndex Option pointer to an int with the index of the scenario to select.
 * @details
 * The method creates a copy of each given scenarios and sets in the internal scenarios model.
 */
void ScenariosEditor::setScenarios(const QVector<Scenario *> scenarios, int *selectIndex) {
    // setup null models for parameters and indicators editors.
    // prevent the editors to be enabled in case there is no scenarios:
    // In such case, the selectedScenarioChanged is not emitted because fuckin Qt does not emit
    // selectionChanged for the afferent view on clear method (whereas the doc states that it should !).
    mItemsEditor->setItemModel(nullptr);
    mIndicatorsEditor->setIndicatorModel(nullptr);
    mScenarioEditor->setupScenarios(scenarios, selectIndex);
}


/**
 * @brief Get copies of the scenarios.
 * @return Vector of scenarios.
 * @details
 * The method returns a copy of each scenarios stored in the widget's scenario model and returns a vector.
 * @note
 * It is user responsibility to delete the pointers when no more used.
 */
QVector<Scenario *> ScenariosEditor::scenarios() const { return mScenarioEditor->scenarios(); }

/// @brief Get the selected scenario in the editor.
QPointer<Scenario> ScenariosEditor::selectedScenario() const { return mScenarioEditor->selectedScenario(); }

/// @brief Get the selected scenario in the editor.
int ScenariosEditor::selectedScenarioIndex() const { return mScenarioEditor->selectedScenarioIndex(); }

/// @brief Setup editors models on scenarios changes.
void ScenariosEditor::onScenarioChanged(QPointer<Scenario> scenario) {
    if (scenario) {
        mItemsEditor->setItemModel(scenario->itemsModel());
        mIndicatorsEditor->setIndicatorModel(scenario->indicatorsModel());
    } else {
        mItemsEditor->setItemModel(nullptr);
        mIndicatorsEditor->setIndicatorModel(nullptr);
    }
}
