// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "scenarioselector.hpp"
#include "scenario.hpp"

ScenarioSelector::ScenarioSelector(QWidget *parent) : QWidget{parent} {
    mLabel = new QLabel(this);
    mCbxScenario = new QComboBox(this);
    mCbxScenario->setMinimumWidth(70);
    mCbxScenario->setSizeAdjustPolicy(QComboBox::AdjustToContents);

    mBtnEdit = new QPushButton(QIcon::fromTheme("edit"), "", this);
    mBtnSave = new QPushButton(QIcon::fromTheme("save"), "", this);

    mBtnEdit->setStyleSheet("border : none");
    mBtnSave->setStyleSheet("border : none");

    mBtnEdit->setEnabled(true);
    mBtnSave->setEnabled(true);

    QHBoxLayout *hlayout = new QHBoxLayout(this);
    hlayout->addWidget(mLabel, 0);
    hlayout->addWidget(mCbxScenario, 0);
    hlayout->addWidget(mBtnEdit, 0);
    hlayout->addWidget(mBtnSave, 0);

    QObject::connect(mCbxScenario, qOverload<int>(&QComboBox::currentIndexChanged), this,
                     &ScenarioSelector::currentIndexChanged);
    QObject::connect(mBtnSave, &QToolButton::clicked, this, &ScenarioSelector::scenarioSaveRequested);
    QObject::connect(mBtnEdit, &QToolButton::clicked, this,
                     [=]() { emit scenarioEditionRequested(mCbxScenario->currentIndex()); });
    translateUI();
}

/// @brief Get the internal scenario combobox.
QComboBox *ScenarioSelector::combobox() const { return mCbxScenario; }

/**
 * @brief Set new scenarios.
 * @param scenarios Scenarios to process.
 * @param selectedIndex Optional current index to set in the combobox after scenarios have been processed. Default to
 * -1.
 * @details
 * The method sets new combobox items using the given scenarios.
 * Then it blocks combobox signals and sets the current index of the combobox to selectedIndex.
 */
void ScenarioSelector::setup(const QVector<Scenario *> &scenarios, int selectedIndex) {
    QSignalBlocker blocker(mCbxScenario);
    mCbxScenario->clear();
    int ns = scenarios.size();
    if (!ns)
        return;
    if (selectedIndex < 0 || selectedIndex >= ns)
        selectedIndex = -1;

    for (int i = 0; i < ns; i++) {
        Scenario *scn = scenarios.at(i);
        QString name = scn->name();
        mCbxScenario->addItem(name);
        QString com = scn->comments();
        if (!com.isEmpty())
            mCbxScenario->setItemData(i, QString("<h4 style='white-space:pre'>%1</h4><p>%2</p>").arg(name, com),
                                      Qt::ToolTipRole);
    }
    mCbxScenario->setCurrentIndex(selectedIndex);
}

/**
 * @brief Enable/Disable buttons.
 * @param enable True to enable buttons, false otherwise.
 */
void ScenarioSelector::setInteractionMode(bool enable) {
    mBtnEdit->setEnabled(enable);
    mBtnSave->setEnabled(enable);
}

/// @brief Check whether or not the user should only interact with the combobox.
void ScenarioSelector::setSelectOnly(bool enable) {
    mBtnEdit->setVisible(!enable);
    mBtnSave->setVisible(!enable);
}

/// @brief Translate widgets texts.
void ScenarioSelector::translateUI() {
    mLabel->setText(tr("Scenario"));
    mBtnSave->setToolTip(tr("Save current scenario."));
    mBtnEdit->setToolTip(tr("Open scenario editor."));
    mBtnSave->setStatusTip(tr("Save current scenario."));
    mBtnEdit->setStatusTip(tr("Open scenario editor."));
}

void ScenarioSelector::setCurrentIndex(int index, bool blockSignal) {
    if (blockSignal) {
        QSignalBlocker blocker(mCbxScenario);
        mCbxScenario->setCurrentIndex(index);
    } else {
        mCbxScenario->setCurrentIndex(index);
    }
}
