// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "generalpage.hpp"
#include "arcgisaccountwidget.hpp"
#include "globals.hpp"

/// @brief Constructor
GeneralPage::GeneralPage(QWidget *parent) : BasePage(parent) {
    mPageTitle = tr("General");
    mPageIcon = getIcon("home");
    buildUI();
    translateUI();
}

/// @brief Initialize widgets of the General group box.
void GeneralPage::buildUI() {
    mLblLanguage = new QLabel(this);
    mCbxLanguage = new QComboBox(this);
    mLblAppFont = new QLabel(this);

    QWidget *wdtAppFont = new QWidget(this);
    QHBoxLayout *hLayout = new QHBoxLayout(wdtAppFont);
    hLayout->setContentsMargins(0, 0, 0, 0);
    mLdtAppFont = new QLineEdit(wdtAppFont);
    mLdtAppFont->setReadOnly(true);
    mBtnAppFont = new QToolButton(wdtAppFont);
    mBtnAppFont->setIcon(QIcon::fromTheme("font"));
    hLayout->addWidget(mLdtAppFont);
    hLayout->addWidget(mBtnAppFont);

    mChbSaveOnFly = new QCheckBox(this);

    QVBoxLayout *layout = nullptr;
    mGbxShowBattery = new QGroupBox(this);
    mGbxShowBattery->setCheckable(true);
    mGbxShowBattery->setFlat(true);
    mChbShowBatteryPercent = new QCheckBox(this);
    mChbShowBatteryTime = new QCheckBox(this);

    layout = new QVBoxLayout(mGbxShowBattery);
    layout->addWidget(mChbShowBatteryPercent);
    layout->addWidget(mChbShowBatteryTime);

    mGbxShowClock = new QGroupBox(this);
    mGbxShowClock->setCheckable(true);
    mGbxShowClock->setFlat(true);
    mChbShowClockUTC = new QCheckBox(this);
    mChbShowClockDate = new QCheckBox(this);
    mChbShowClockSecs = new QCheckBox(this);

    layout = new QVBoxLayout(mGbxShowClock);
    layout->addWidget(mChbShowClockDate);
    layout->addWidget(mChbShowClockSecs);
    layout->addWidget(mChbShowClockUTC);

    mArcgisAccountWidget = new ArcgisAccountWidget(this);

    QFormLayout *formLayout = new QFormLayout(this);
    formLayout->addRow(mLblLanguage, mCbxLanguage);
    formLayout->addRow(mLblAppFont, wdtAppFont);
    formLayout->addRow(mChbSaveOnFly);
    formLayout->addRow(mGbxShowBattery);
    formLayout->addRow(mGbxShowClock);

    formLayout->addRow(getSeparator(this));
    formLayout->addRow(mArcgisAccountWidget);

    QObject::connect(mBtnAppFont, &QToolButton::clicked, this, &GeneralPage::chooseAppFont);
}

/// @brief Load general settings from configuration file.
void GeneralPage::loadSettings() {
    QSettings settings;
    settings.beginGroup("global");
    mCbxLanguage->clear();
    int ifb = 0;
    if (gI18N) {
        for (QLocale::Language l : gI18N->getLanguageList()) {
            if (l == QLocale::C)
                ifb = mCbxLanguage->count();
            mCbxLanguage->addItem(I18N::languageName(l), I18N::languageCode(l));
        }
    } else {
        mCbxLanguage->addItem(I18N::languageName(QLocale::C), I18N::languageCode(QLocale::C));
    }
    QString lang = settings.value("language", I18N::languageCode(QLocale::system().language())).toString();

    //    if (gI18N) {
    //                mCbxLanguage->addItem("default (english)", QLocale(QLocale::C).name());
    //                for (const QString &lang : gI18N->getLanguageList()) {
    //                    mCbxLanguage->addItem(QLocale::languageToString(QLocale(lang).language()), lang);
    //                }
    //    } else {
    //          mCbxLanguage->addItem("default (english)", QLocale(QLocale::C).name());
    //    }
    //    QString lang = settings.value("language", QLocale(QLocale::C).name()).toString();

    int idx = mCbxLanguage->findData(lang);
    if (idx == -1)
        idx = ifb;
    mCbxLanguage->setCurrentIndex(idx);

    mAppFont = settings.value("font", QApplication::font()).value<QFont>();
    mLdtAppFont->setText(fontToString(mAppFont));
    mChbSaveOnFly->setChecked(settings.value("save_on_fly", true).toBool());

    mGbxShowBattery->setChecked(settings.value("show_battery", true).toBool());
    mChbShowBatteryPercent->setChecked(settings.value("show_battery_pct", true).toBool());
    mChbShowBatteryTime->setChecked(settings.value("show_battery_time", true).toBool());

    mGbxShowClock->setChecked(settings.value("show_clock", true).toBool());
    mChbShowClockUTC->setChecked(settings.value("show_clock_utc", true).toBool());
    mChbShowClockDate->setChecked(settings.value("show_clock_date", false).toBool());
    mChbShowClockSecs->setChecked(settings.value("show_clock_sec", true).toBool());

    settings.endGroup();
}

/// @brief Save general settings in configuration file.
void GeneralPage::saveSettings() {
    QSettings settings;
    settings.beginGroup("global");
    settings.setValue("language", mCbxLanguage->currentData().toString());
    settings.setValue("font", mAppFont);
    settings.setValue("save_on_fly", mChbSaveOnFly->isChecked());
    settings.setValue("show_battery", mGbxShowBattery->isChecked());
    settings.setValue("show_battery_pct", mChbShowBatteryPercent->isChecked());
    settings.setValue("show_battery_time", mChbShowBatteryTime->isChecked());
    settings.setValue("show_clock", mGbxShowClock->isChecked());
    settings.setValue("show_clock_utc", mChbShowClockUTC->isChecked());
    settings.setValue("show_clock_date", mChbShowClockDate->isChecked());
    settings.setValue("show_clock_sec", mChbShowClockSecs->isChecked());
    settings.endGroup();
}

/// @brief Translates widgets of the general groupbox.
void GeneralPage::translateUI() {
    mPageTitle = tr("General");
    mLblLanguage->setText(tr("Language"));
    mLblAppFont->setText((tr("Font")));
    mChbSaveOnFly->setText(tr("Save configuration on the fly."));
    mGbxShowBattery->setTitle(tr("Show computer battery status"));
    mChbShowBatteryPercent->setText(tr("Show percentage."));
    mChbShowBatteryTime->setText(tr("Show remaining time."));
    mGbxShowClock->setTitle(tr("Show system clock"));
    mChbShowClockDate->setText(tr("Show date."));
    mChbShowClockSecs->setText(tr("Show seconds."));
    mChbShowClockUTC->setText(tr("Show UTC time."));
    mArcgisAccountWidget->translateUI();

    setHelper(mLblLanguage, HelpFlag::LANG);
    setHelper(mCbxLanguage, HelpFlag::LANG);
    setHelper(mLblAppFont, HelpFlag::FONT);
    setHelper(mBtnAppFont, HelpFlag::FONT);
    setHelper(mChbSaveOnFly, HelpFlag::SAVE_ON_FLY);
    setHelper(mGbxShowBattery, HelpFlag::SHOW_BATTERY);
    setHelper(mChbShowBatteryPercent, HelpFlag::SHOW_BATTERY);
    setHelper(mChbShowBatteryTime, HelpFlag::SHOW_BATTERY);

    setHelper(mGbxShowClock, HelpFlag::SHOW_CLOCK);
    setHelper(mChbShowClockDate, HelpFlag::SHOW_CLOCK);
    setHelper(mChbShowClockSecs, HelpFlag::SHOW_CLOCK);
    setHelper(mChbShowClockUTC, HelpFlag::SHOW_CLOCK);

    setHelper(mArcgisAccountWidget, HelpFlag::ART_SIGNIN);
}

/// @brief Select a font for the application.
void GeneralPage::chooseAppFont() {
    bool ok;
    QFont font = QFontDialog::getFont(&ok, mAppFont, this, tr("Select a font"));
    if (ok) {
        mAppFont = font;
        mLdtAppFont->setText(fontToString(mAppFont));
    }
}

/// @brief Get a (human readable) string representation of a font
QString GeneralPage::fontToString(const QFont &font) {
    QString out;
    out = QString("%1, %2pt").arg(font.family()).arg(font.pointSize());
    if (font.bold())
        out += ", bold";
    if (font.italic())
        out += ", italic";
    if (font.underline())
        out += ", underline";
    if (font.overline())
        out += ", overline";
    return out;
}
