// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "interfacepage.hpp"
#include "colorchooserlabel.hpp"

void fillStyleCombobox(QComboBox *cbx) {
    int idx = cbx->currentIndex();
    cbx->clear();
    cbx->addItem(QCoreApplication::tr("Circle"), 0);
    cbx->addItem(QCoreApplication::tr("Cross (+)"), 1);
    cbx->addItem(QCoreApplication::tr("Diamond"), 2);
    cbx->addItem(QCoreApplication::tr("Square"), 3);
    cbx->addItem(QCoreApplication::tr("Triangle"), 4);
    cbx->addItem(QCoreApplication::tr("Diag. cross (X)"), 5);
    cbx->setCurrentIndex(idx);
}

class MapUIWidget : public QWidget {
    Q_OBJECT
  public:
    explicit MapUIWidget(QWidget *parent = nullptr);
    void loadSettings();
    void saveSettings();
    void translateUI();

  private:
    QCheckBox *mChbMapGrid;
    QCheckBox *mChbHoriCB;
    QCheckBox *mChbHideMV;

    QGroupBox *mGbxMissingValue;

    QLabel *mLblMVColor;
    ColorChooserLabel *mLblMVColorChooser;
    QLabel *mLblMVStyle;
    QComboBox *mCbxMVStyle;
    QDoubleSpinBox *mSpbMVSize;

    QGroupBox *mGbxCurrentFlight;
    QLabel *mLblCFStyle;
    QComboBox *mCbxCFStyle;
    QDoubleSpinBox *mSpbCFSize;

    QGroupBox *mGbxPreviousFlights;
    QLabel *mLblPFStyle;
    QComboBox *mCbxPFStyle;
    QDoubleSpinBox *mSpbPFSize;

  private:
    void buildUI();
};

/**
 * @brief Constructor
 * @param parent Parent widget.
 * @details
 * The constructor only builds up the UI.
 * Additionally, MapStyleWidget::loadSettings() and MapStyleWidget::translateUI() should be called to
 * populate the widgets with values and set texts.
 */
MapUIWidget::MapUIWidget(QWidget *parent) : QWidget(parent) { buildUI(); }

/// @brief Initialize widgets of the plot group box.
void MapUIWidget::buildUI() {

    mChbMapGrid = new QCheckBox(this);
    mChbHoriCB = new QCheckBox(this);
    mChbHideMV = new QCheckBox(this);

    // missing value style
    mGbxMissingValue = new QGroupBox(this);

    mLblMVColor = new QLabel(mGbxMissingValue);
    mLblMVColorChooser = new ColorChooserLabel(mGbxMissingValue);
    mLblMVColorChooser->setHighlight(true);
    mLblMVStyle = new QLabel(mGbxMissingValue);
    mCbxMVStyle = new QComboBox(mGbxMissingValue);
    fillStyleCombobox(mCbxMVStyle);
    mSpbMVSize = new QDoubleSpinBox(mGbxMissingValue);
    mSpbMVSize->setSuffix(" px");
    mSpbMVSize->setRange(4.0, 20.0);
    mSpbMVSize->setSingleStep(0.5);

    QFormLayout *flayout = new QFormLayout(mGbxMissingValue);
    flayout->addRow(mLblMVColor, mLblMVColorChooser);
    QHBoxLayout *hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxMVStyle);
    hlayout->addWidget(mSpbMVSize);
    flayout->addRow(mLblMVStyle, hlayout);

    // current flight style
    mGbxCurrentFlight = new QGroupBox(this);
    mLblCFStyle = new QLabel(mGbxCurrentFlight);
    mCbxCFStyle = new QComboBox(mGbxCurrentFlight);
    fillStyleCombobox(mCbxCFStyle);
    mSpbCFSize = new QDoubleSpinBox(mGbxCurrentFlight);
    mSpbCFSize->setSuffix(" px");
    mSpbCFSize->setRange(4.0, 20.0);
    mSpbCFSize->setSingleStep(0.5);
    flayout = new QFormLayout(mGbxCurrentFlight);
    hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxCFStyle);
    hlayout->addWidget(mSpbCFSize);
    flayout->addRow(mLblCFStyle, hlayout);

    // previous flights style
    mGbxPreviousFlights = new QGroupBox(this);
    mLblPFStyle = new QLabel(mGbxPreviousFlights);
    mCbxPFStyle = new QComboBox(mGbxPreviousFlights);
    fillStyleCombobox(mCbxPFStyle);
    mSpbPFSize = new QDoubleSpinBox(mGbxPreviousFlights);
    mSpbPFSize->setSuffix(" px");
    mSpbPFSize->setRange(4.0, 20.0);
    mSpbPFSize->setSingleStep(0.5);

    flayout = new QFormLayout(mGbxPreviousFlights);
    hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxPFStyle);
    hlayout->addWidget(mSpbPFSize);
    flayout->addRow(mLblPFStyle, hlayout);

    QVBoxLayout *vlayout = new QVBoxLayout(this);

    vlayout->addWidget(mChbMapGrid);
    vlayout->addWidget(mChbHoriCB);
    vlayout->addWidget(mChbHideMV);

    vlayout->addWidget(mGbxMissingValue);
    vlayout->addWidget(mGbxCurrentFlight);
    vlayout->addWidget(mGbxPreviousFlights);
    vlayout->addStretch(1);

    QObject::connect(mChbHideMV, &QCheckBox::toggled, mGbxMissingValue, &QWidget::setDisabled);
}

/// @brief Load plot settings from configuration file.
void MapUIWidget::loadSettings() {
    QSettings settings;

    settings.beginGroup("map");

    mChbMapGrid->setChecked(settings.value("m_showgrid", false).toBool());
    mChbHoriCB->setChecked(settings.value("cs_horiz", false).toBool());
    mChbHideMV->setChecked(settings.value("m_hidemv", false).toBool());

    mLblMVColorChooser->setColor(QColor(settings.value("m_mvcolor", "#000000").toString()));
    mCbxMVStyle->setCurrentIndex(mCbxCFStyle->findData(settings.value("m_mvmstyle", 0).toInt()));
    mSpbMVSize->setValue(settings.value("m_mvsize", 10.0).toDouble());

    mCbxCFStyle->setCurrentIndex(mCbxCFStyle->findData(settings.value("m_cmstyle", 0).toInt()));
    mSpbCFSize->setValue(settings.value("m_cmsize", 10.0).toDouble());

    mCbxPFStyle->setCurrentIndex(mCbxPFStyle->findData(settings.value("m_omstyle", 0).toInt()));
    mSpbPFSize->setValue(settings.value("m_omsize", 10.0).toDouble());

    settings.endGroup();
}

/// @brief Save plot settings in configuration file.
void MapUIWidget::saveSettings() {
    QSettings settings;
    settings.beginGroup("map");
    settings.setValue("cs_horiz", mChbHoriCB->isChecked());
    settings.setValue("m_showgrid", mChbMapGrid->isChecked());
    settings.setValue("m_hidemv", mChbHideMV->isChecked());

    settings.setValue("m_mvmstyle", mCbxMVStyle->currentData().toInt());
    settings.setValue("m_mvcolor", mLblMVColorChooser->color().name());
    settings.setValue("m_mvmsize", mSpbMVSize->value());

    settings.setValue("m_cmstyle", mCbxCFStyle->currentData().toInt());
    settings.setValue("m_cmsize", mSpbCFSize->value());

    settings.setValue("m_omstyle", mCbxPFStyle->currentData().toInt());
    settings.setValue("m_omsize", mSpbPFSize->value());
    settings.endGroup();
}

/// @brief Translates widgets of the plot groupbox.
void MapUIWidget::translateUI() {
    // map
    mChbMapGrid->setText(tr("Show map grid."));
    mChbHoriCB->setText(tr("Display the colorbar horizontally."));
    mChbHideMV->setText(tr("Hide missing value point on the map."));
    mGbxMissingValue->setTitle(tr("Missing value point"));
    mLblMVColor->setText(tr("Color"));
    mLblMVStyle->setText(tr("Symbol"));
    mGbxCurrentFlight->setTitle(tr("Current flight"));
    mLblCFStyle->setText(tr("Symbol"));
    mGbxPreviousFlights->setTitle(tr("Previous flights"));
    mLblPFStyle->setText(tr("Symbol"));

    setHelper(mGbxCurrentFlight, HelpFlag::CF_STYLE);
    setHelper(mGbxPreviousFlights, HelpFlag::PF_STYLE);
    setHelper(mGbxMissingValue, HelpFlag::MV_STYLE);
    setHelper(mChbHoriCB, HelpFlag::CS_ORIENT);
    setHelper(mChbHideMV, HelpFlag::HIDE_MV);
}

class GraphUIWidget : public QWidget {
    Q_OBJECT
  public:
    explicit GraphUIWidget(QWidget *parent = nullptr);
    void loadSettings();
    void saveSettings();
    void translateUI();
  private slots:
    void updatePlotWidgets(int id, bool checked);

  private:
    QLabel *mLblASPeriod;
    QSpinBox *mSpbASPeriod;
    QLabel *mLblTimeWindow;
    QSpinBox *mSpbTimeWindow;
    QLabel *mLblTimeOffset;
    QSpinBox *mSpbTimeOffset;

    QLabel *mLblTextBox;
    QComboBox *mCbxTextBox;

    QButtonGroup *mButtonGroup;
    QLabel *mLblPlotType;
    QRadioButton *mBtnPLine;
    QRadioButton *mBtnPScatter;
    QLabel *mLblPLineSize;
    QLabel *mLblPSymbol;
    QDoubleSpinBox *mSpbPLineSize;
    QComboBox *mCbxPSymbol;
    QDoubleSpinBox *mSpbPSymbolSize;
    int mLastLMS;
    int mLastSMS;

  private:
    void buildUI();
};
#include "qcp_helpers.hpp"

/**
 * @brief Constructor
 * @param parent Parent widget.
 * @details
 * The constructor only builds up the UI.
 * Additionally, GraphStyleWidget::loadSettings() and GraphStyleWidget::translateUI() should be called to
 * populate the widgets with values and set texts.
 */
GraphUIWidget::GraphUIWidget(QWidget *parent) : QWidget(parent) { buildUI(); }

/// @brief Initialize widgets of the plot group box.
void GraphUIWidget::buildUI() {

    mButtonGroup = new QButtonGroup(this);
    QFormLayout *flayout = new QFormLayout(this);

    mLblASPeriod = new QLabel(this);
    mSpbASPeriod = new QSpinBox(this);
    mSpbASPeriod->setRange(40, 2000);
    mSpbASPeriod->setSingleStep(10);
    mSpbASPeriod->setSuffix(tr(" ms"));
    mSpbASPeriod->setAccelerated(true);
    flayout->addRow(mLblASPeriod, mSpbASPeriod);
    mLblTimeWindow = new QLabel(this);
    mSpbTimeWindow = new QSpinBox(this);
    mSpbTimeWindow->setRange(1, 30);
    mSpbTimeWindow->setSuffix(tr(" min"));
    mSpbTimeWindow->setAccelerated(true);
    flayout->addRow(mLblTimeWindow, mSpbTimeWindow);
    mLblTimeOffset = new QLabel(this);
    mSpbTimeOffset = new QSpinBox(this);
    mSpbTimeOffset->setRange(50, 60000);
    mSpbTimeOffset->setSuffix(tr(" ms"));
    mSpbTimeOffset->setSingleStep(10);
    mSpbTimeOffset->setStepType(QSpinBox::AdaptiveDecimalStepType);
    mSpbTimeOffset->setAccelerated(true);
    flayout->addRow(mLblTimeOffset, mSpbTimeOffset);
    mLblTextBox = new QLabel(this);
    mCbxTextBox = new QComboBox(this);
    mCbxTextBox->addItems({
        tr("Top left"), tr("Top center"), tr("Top right"), tr("Bottom left"), tr("Bottom center"), tr("Bottom right"),
    });
    flayout->addRow(mLblTextBox, mCbxTextBox);

    flayout->addRow(getSeparator(this));

    mLblPlotType = new QLabel(this);
    mBtnPLine = new QRadioButton(this);
    mBtnPScatter = new QRadioButton(this);
    mButtonGroup->addButton(mBtnPLine, 0);
    mButtonGroup->addButton(mBtnPScatter, 1);
    QHBoxLayout *hlayout = new QHBoxLayout;
    hlayout->addWidget(mBtnPLine);
    hlayout->addWidget(mBtnPScatter);
    flayout->addRow(mLblPlotType, hlayout);
    hlayout = new QHBoxLayout;
    mLblPLineSize = new QLabel(this);
    mSpbPLineSize = new QDoubleSpinBox(this);
    mSpbPLineSize->setRange(0.5, 5.0);
    mSpbPLineSize->setValue(1.0);
    mSpbPLineSize->setSingleStep(0.1);
    mSpbPLineSize->setSuffix(" px");
    mSpbPLineSize->setAccelerated(true);
    flayout->addRow(mLblPLineSize, mSpbPLineSize);
    hlayout = new QHBoxLayout;
    mLblPSymbol = new QLabel(this);
    mCbxPSymbol = new QComboBox(this);
    Plotting::fillMarkerStyleComboBox(mCbxPSymbol);
    mSpbPSymbolSize = new QDoubleSpinBox(this);
    mSpbPSymbolSize->setRange(4.0, 24.0);
    mSpbPSymbolSize->setValue(6.0);
    mSpbPSymbolSize->setSingleStep(1.0);
    mSpbPSymbolSize->setSuffix(" px");
    mSpbPSymbolSize->setAccelerated(true);
    hlayout->addWidget(mCbxPSymbol);
    hlayout->addWidget(mSpbPSymbolSize);
    flayout->addRow(mLblPSymbol, hlayout);
    QObject::connect(mCbxPSymbol, QOverload<int>::of(&QComboBox::currentIndexChanged), [=](int) {
        if (mBtnPLine->isChecked()) {
            mLastLMS = mCbxPSymbol->currentData().toInt();
        } else {
            mLastSMS = mCbxPSymbol->currentData().toInt();
        }
    });
    QObject::connect(mButtonGroup, QOverload<int, bool>::of(&QButtonGroup::buttonToggled), this,
                     &GraphUIWidget::updatePlotWidgets);
}

/**
 * @brief Update graph line connection style as a function of the checked radio button.
 * @param id Identifier of the button
 * @param checked Check state.
 * @details
 * If the linestyle button checkstate has changed, the method enables/disables the line dash pattern and width widgets.
 *
 * If the scatterstyle button checkstate has changed, the method update the scatter style combobox selected value if
 * it is set to none.
 */
void GraphUIWidget::updatePlotWidgets(int id, bool checked) {

    // connection by line selected.
    if (id == 0) {
        mSpbPLineSize->setEnabled(checked);
    }
    if (id == 1) {
        int firstData = mCbxPSymbol->itemData(0).toInt();
        bool old = mCbxPSymbol->blockSignals(true);
        if (checked) {
            if (firstData == 0)
                mCbxPSymbol->removeItem(0);
        } else {
            if (firstData != 0)
                mCbxPSymbol->insertItem(0, Plotting::getScatterIcon(0), Plotting::getScatterName(0), 0);
        }
        mCbxPSymbol->blockSignals(old);
    }
    if (mBtnPLine->isChecked())
        mCbxPSymbol->setCurrentIndex(mCbxPSymbol->findData(mLastLMS));
    if (mBtnPScatter->isChecked())
        mCbxPSymbol->setCurrentIndex(mCbxPSymbol->findData(mLastSMS));
}

/// @brief Load plot settings from configuration file.
void GraphUIWidget::loadSettings() {
    QSettings settings;

    settings.beginGroup("monitoring");
    mSpbASPeriod->setValue(settings.value("asinterval", 200).toInt());
    mSpbTimeWindow->setValue(settings.value("wtime", 300).toInt() / 60);
    mSpbTimeOffset->setValue(settings.value("woffset", 100).toInt());
    settings.endGroup();

    settings.beginGroup("graph");
    mCbxTextBox->setCurrentIndex(settings.value("box_pos", 2).toInt());
    mBtnPLine->setChecked(settings.value("wline", true).toBool());
    mSpbPLineSize->setValue(settings.value("lwidth", 1.0).toDouble());
    int ms = settings.value("mstyle", 0).toInt();
    mCbxPSymbol->setCurrentIndex(qMax(0, mCbxPSymbol->findData(ms)));
    mSpbPSymbolSize->setValue(settings.value("msize", 6.0).toDouble());
    settings.endGroup();
    if (mBtnPLine->isChecked()) {
        mLastLMS = ms;
        mLastSMS = 1;
    } else {
        mLastLMS = 0;
        mLastSMS = ms;
    }
}

/// @brief Save plot settings in configuration file.
void GraphUIWidget::saveSettings() {
    QSettings settings;

    settings.beginGroup("graph");
    settings.setValue("now_as_ref", true); // not editable yet !
    settings.setValue("asinterval", qBound(40, mSpbASPeriod->value(), 2000));
    settings.setValue("wtime", qBound(60, mSpbTimeWindow->value() * 60, 1800));
    settings.setValue("woffset", qBound(50, mSpbTimeOffset->value(), 5000));
    settings.setValue("box_pos", mCbxTextBox->currentIndex());

    settings.setValue("wline", mBtnPLine->isChecked());
    settings.setValue("lwidth", mSpbPLineSize->value());
    settings.setValue("mstyle", mCbxPSymbol->currentData().toInt());
    settings.setValue("msize", mSpbPSymbolSize->value());
    settings.endGroup();
}

/// @brief Translates widgets of the plot groupbox.
void GraphUIWidget::translateUI() {

    mLblASPeriod->setText(tr("Auto-scrolling"));
    mLblTimeWindow->setText(tr("Time window"));
    mLblTimeOffset->setText(tr("Time offset"));

    mLblTextBox->setText(tr("Last value box position"));
    int idx = mCbxTextBox->currentIndex();
    QStringList list{
        tr("Top left"), tr("Top center"), tr("Top right"), tr("Bottom left"), tr("Bottom center"), tr("Bottom right"),
    };
    mCbxTextBox->clear();
    mCbxTextBox->addItems(list);
    mCbxTextBox->setCurrentIndex(idx);

    setHelper(mLblASPeriod, HelpFlag::ASPERIOD);
    setHelper(mSpbASPeriod, HelpFlag::ASPERIOD);
    setHelper(mLblTimeWindow, HelpFlag::WTIME);
    setHelper(mSpbTimeWindow, HelpFlag::WTIME);
    setHelper(mLblTimeOffset, HelpFlag::WOFFSET);
    setHelper(mSpbTimeOffset, HelpFlag::WOFFSET);

    // graph
    Plotting::reloadNames();
    Plotting::fillMarkerStyleComboBox(mCbxPSymbol, mBtnPScatter->isChecked());
    mBtnPLine->setText(tr("line"));
    mBtnPScatter->setText(tr("scatter"));
    mLblPlotType->setText(tr("Graph type"));
    mLblPLineSize->setText(tr("Line width"));
    mLblPSymbol->setText(tr("Marker style"));
}

InterfacePage::InterfacePage(QWidget *parent) : BasePage(parent) {
    mPageTitle = tr("Interface");
    mPageIcon = getIcon("interface");
    buildUI();
    translateUI();
}

/// @brief Initialize widgets of the plot group box.
void InterfacePage::buildUI() {

    mTabWidget = new QTabWidget(this);
    mTabWidget->setMovable(false);

    mPlotWidget = new GraphUIWidget(mTabWidget);
    mMapWidget = new MapUIWidget(mTabWidget);

    mTabWidget->addTab(mMapWidget, tr("Map"));
    mTabWidget->addTab(mPlotWidget, tr("Graph"));

    QVBoxLayout *layout = new QVBoxLayout;
    layout->setContentsMargins(0, 0, 0, 0);
    layout->addWidget(mTabWidget, 2);
    setLayout(layout);
}

void InterfacePage::loadSettings() {
    mPlotWidget->loadSettings();
    mMapWidget->loadSettings();
}

void InterfacePage::saveSettings() {
    mPlotWidget->saveSettings();
    mMapWidget->saveSettings();
}

void InterfacePage::translateUI() {
    mPageTitle = tr("Interface");

    mPlotWidget->translateUI();
    mMapWidget->translateUI();

    mTabWidget->setTabText(0, tr("Map"));
    mTabWidget->setTabText(1, tr("Graph"));
}

/**
 * @brief Save the current tab widget index.
 */
void InterfacePage::saveUI() {
    QSettings settings;
    settings.setValue("layouts/cfg_diag/ui_tab", mTabWidget->currentIndex());
}

/**
 * @brief Load the current tab widget index.
 */
void InterfacePage::loadUI() {
    QSettings settings;
    bool tabIndex = settings.value("layouts/cfg_diag/ui_tab", 0).toInt();
    mTabWidget->setCurrentIndex(tabIndex);
}
