// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "settingsdialog.hpp"

#include "basepage.hpp"
#include "generalpage.hpp"
#include "graphpage.hpp"
#include "mappage.hpp"
#include "monitoringpage.hpp"
#include "storagepage.hpp"
#include "warningspage.hpp"

const int pageIconSize = 24;

/// @brief Constructor.
PageModel::PageModel(QObject *parent) : QAbstractListModel(parent) {
    QPixmap empty(pageIconSize, pageIconSize);
    empty.fill(Qt::transparent);
    mEmptyIcon = QIcon(empty);
}

///  @brief Destructor.
PageModel::~PageModel() {}

/// @brief Get the number of rows in the model.
int PageModel::rowCount(const QModelIndex &parent) const { return parent.isValid() ? 0 : mPages.size(); }

/// @brief Get the data of the model.
QVariant PageModel::data(const QModelIndex &index, int role) const {
    switch (role) {
    case Qt::DisplayRole:
        return mPages.at(index.row())->pageTitle();
    case Qt::DecorationRole: {
        QIcon icon = mPages.at(index.row())->pageIcon();
        if (icon.isNull())
            icon = mEmptyIcon;
        return icon;
    }
    }
    return QVariant();
}

/**
 * @brief Register pages in the model.
 * @param pages Page widget to register.
 * @details
 * The model does not take ownership of the widgets.
 */
void PageModel::setPages(const QVector<BasePage *> &pages) {
    beginResetModel();
    qDeleteAll(mPages);
    mPages = pages;
    endResetModel();
}

/**
 * @brief Get the page at given index.
 * @param index Index of the page.
 * @return the afferent page or a null pointer if the index is out of range.
 */
BasePage *PageModel::page(int index) { return (index < 0 || index > mPages.size()) ? nullptr : mPages.at(index); }

/**
 * @overload
 * @brief Get the page at given index.
 * @param index Index of the page.
 * @return the afferent page or a null pointer if the index is not valid.
 */
BasePage *PageModel::page(QModelIndex index) { return index.isValid() ? mPages.at(index.row()) : nullptr; }

/// @brief Get the pages registered in the model (const ref).
const QVector<BasePage *> &PageModel::pages() const { return mPages; }

/* ------------------------------------------------------------------------- */

class PageListViewDelegate : public QStyledItemDelegate {
  public:
    explicit PageListViewDelegate(QObject *parent) : QStyledItemDelegate(parent) {}

    QSize sizeHint(const QStyleOptionViewItem &option, const QModelIndex &index) const override {
        QSize size = QStyledItemDelegate::sizeHint(option, index);
        size.setHeight(qMax(size.height(), 32));
        return size;
    }
};

/**
 * Special version of a QListView that has the width of the first column as
 * minimum size.
 */
class PageListView : public QListView {
  public:
    PageListView() {
        setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Expanding);
        setItemDelegate(new PageListViewDelegate(this));
        setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    }

    QSize sizeHint() const final {
        int width = sizeHintForColumn(0) + frameWidth() * 2 + 5;
        width += verticalScrollBar()->sizeHint().width();
        return QSize(width, 100);
    }

    // QListView installs a event filter on its scrollbars
    bool eventFilter(QObject *obj, QEvent *event) final {
        if (obj == verticalScrollBar() && (event->type() == QEvent::Show || event->type() == QEvent::Hide))
            updateGeometry();
        return QListView::eventFilter(obj, event);
    }
};

/* ------------------------------------------------------------------------- */

/// @brief Constructor.
SettingsDialog::SettingsDialog(QWidget *parent) : QDialog(parent) {
    QSettings settings;
    mHasUISetting = settings.contains("layouts/cfg_diag/geometry");
    setupUI();
    loadSettings();
    for (BasePage *page : mPages)
        page->loadUI();
    translateUI();
    // load UI geometry if any.
    QByteArray array = settings.value("layouts/cfg_diag/geometry").toByteArray();
    if (array.size())
        restoreGeometry(array);
    showPage(settings.value("layouts/cfg_diag/last_page", 0).toInt());
}

/// @brief Setup the widgets.
void SettingsDialog::setupUI() {

    // setup labels
    mHeaderLabel = new QLabel;
    mCategoryLabel = new QLabel;
    QFont headerLabelFont = mHeaderLabel->font();
    headerLabelFont.setBold(true);
    const int pointSize = headerLabelFont.pointSize();
    if (pointSize > 0)
        headerLabelFont.setPointSize(pointSize + 2);
    mHeaderLabel->setFont(headerLabelFont);
    mCategoryLabel->setFont(headerLabelFont);

    // listview
    mPageList = new PageListView;
    mPageList->setIconSize(QSize(pageIconSize, pageIconSize));
    mPageList->setModel(&mModel);
    mPageList->setSelectionMode(QAbstractItemView::SingleSelection);
    mPageList->setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);

    // stack layout
    mStackedLayout = new QStackedLayout;
    mStackedLayout->setContentsMargins(0, 0, 0, 0);
    // set a first empty widget: if no page is selected in the model we show this one.
    mStackedLayout->addWidget(new QWidget(this));

    // build pages
    mPages = {new GeneralPage(this), new WarningsPage(this), new MonitoringPage(this),
              new MapPage(this),     new GraphPage(this),    new StoragePage(this)};
    mModel.setPages(mPages);
    for (BasePage *p : mPages)
        mStackedLayout->addWidget(p);

    mButtonBox = new QDialogButtonBox(this);
    mButtonBox->setOrientation(Qt::Horizontal);
    mButtonBox->setStandardButtons(QDialogButtonBox::RestoreDefaults | QDialogButtonBox::Apply | QDialogButtonBox::Ok |
                                   QDialogButtonBox::Cancel);
    mButtonBox->button(QDialogButtonBox::Ok)->setDefault(true);

    QHBoxLayout *headerHLayout = new QHBoxLayout;
    const int leftMargin = QApplication::style()->pixelMetric(QStyle::PM_LayoutLeftMargin);
    headerHLayout->addSpacerItem(new QSpacerItem(leftMargin, 0, QSizePolicy::Fixed, QSizePolicy::Ignored));
    headerHLayout->addWidget(mHeaderLabel);
    QGridLayout *gLayout = new QGridLayout;
    gLayout->addWidget(mCategoryLabel, 0, 0, 1, 1);
    gLayout->addLayout(headerHLayout, 0, 1, 1, 1);
    gLayout->addWidget(mPageList, 1, 0, 1, 1);
    gLayout->addLayout(mStackedLayout, 1, 1, 1, 1);
    gLayout->addWidget(mButtonBox, 2, 0, 1, 2);
    gLayout->setColumnStretch(1, 4);
    setLayout(gLayout);
    gLayout->setSizeConstraint(QLayout::SetMinimumSize);

    // connections
    QObject::connect(mButtonBox, &QDialogButtonBox::clicked, this, &SettingsDialog::manageButtonBox);
    QObject::connect(mPageList->selectionModel(), &QItemSelectionModel::currentRowChanged, this,
                     &SettingsDialog::currentChanged);

    mPageList->setFocus();
    setWindowTitle(tr("Options"));
}

/// @brief Show the page at given index.
void SettingsDialog::showPage(int pageId) {
    if (pageId < 0 || pageId >= mPages.size())
        pageId = 0;
    mPageList->setCurrentIndex(mModel.index(pageId));
}

/// Change visible widget on the stack layout.
void SettingsDialog::currentChanged(const QModelIndex &current) {

    BasePage *page = mModel.page(current);
    if (page) {
        mStackedLayout->setCurrentWidget(page);
        mHeaderLabel->setText(page->pageTitle());
        mVisitedPages.insert(page);
    } else {
        mStackedLayout->setCurrentIndex(0);
        mHeaderLabel->clear();
    }
}

/// @brief Translate widgets text.
void SettingsDialog::translateUI() {
    setWindowTitle(tr("Application Settings"));
    mCategoryLabel->setText(tr("Categories"));
    mButtonBox->button(QDialogButtonBox::RestoreDefaults)->setText(tr("Restore Defaults"));
    mButtonBox->button(QDialogButtonBox::Apply)->setText(tr("Apply"));
    mButtonBox->button(QDialogButtonBox::Ok)->setText(tr("Ok"));
    mButtonBox->button(QDialogButtonBox::Cancel)->setText(tr("Cancel"));
    for (BasePage *page : mPages)
        page->translateUI();
}

/// @brief Load settings of all registered pages.
void SettingsDialog::loadSettings() {
    for (BasePage *page : mPages)
        page->loadSettings();
}

/// @brief Save settings of all visited pages.
void SettingsDialog::saveSettings() {
    for (BasePage *page : mVisitedPages)
        page->saveSettings();
}

/**
 * @brief Manage click on buttonbox's buttons.
 * @param btn Button that has been clicked
 * @details The method applies changes in the settings if either Apply or Ok
 * buttons have been clicked or discards changes if Cancel button has been clicked.
 */
void SettingsDialog::manageButtonBox(QAbstractButton *btn) {
    switch (mButtonBox->buttonRole(btn)) {
    case (QDialogButtonBox::ApplyRole):
        saveSettings();
        emit settingsChanged();
        break;
    case (QDialogButtonBox::AcceptRole):
        saveSettings();
        emit settingsChanged();
        done(QDialog::Accepted);
        break;
    case (QDialogButtonBox::RejectRole):
        done(QDialog::Rejected);
        break;
    case (QDialogButtonBox::ResetRole):
        // Should we show a MessageBox here ?
        setFactoryDefaults();
        loadSettings();
        emit settingsChanged();
        break;
    default:
        break;
    }
}

/**
 * @brief Close the dialog with the given exit status.
 */
void SettingsDialog::done(int val) {
    QSettings settings;
    // save UI for each pages.
    for (BasePage *p : mPages)
        p->saveUI();
    // save last visited page
    settings.setValue("layouts/cfg_diag/last_page", mStackedLayout->currentIndex() - 1);
    // save UI geometry.
    settings.setValue("layouts/cfg_diag/geometry", saveGeometry());
    QDialog::done(val);
}

QSize SettingsDialog::sizeHint() const { return mHasUISetting ? QWidget::sizeHint() : minimumSize(); }

/**
 * @brief changeEvent event handler.
 * @details
 * The method retranslates the dialog and its widgets on  QEvent::LanguageChange event.
 */
void SettingsDialog::changeEvent(QEvent *event) {
    if (event->type() == QEvent::LanguageChange)
        translateUI();
    QDialog::changeEvent(event);
}
