// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "stylepage.hpp"
#include "colorchooserlabel.hpp"
#include "qcp_helpers.hpp"

/**
 * @brief Constructor
 * @param parent Parent widget.
 * @details
 * The constructor only builds up the UI.
 * Additionally, GraphStyleWidget::loadSettings() and GraphStyleWidget::translateUI() should be called to
 * populate the widgets with values and set texts.
 */
GraphUIWidget::GraphUIWidget(QWidget *parent) : QWidget(parent) { buildUI(); }

/// @brief Initialize widgets of the plot group box.
void GraphUIWidget::buildUI() {
    mButtonGroup = new QButtonGroup(this);
    QFormLayout *flayout = new QFormLayout(this);
    mLblPlotType = new QLabel(this);
    mBtnPLine = new QRadioButton(this);
    mBtnPScatter = new QRadioButton(this);
    mButtonGroup->addButton(mBtnPLine, 0);
    mButtonGroup->addButton(mBtnPScatter, 1);
    QHBoxLayout *hlayout = new QHBoxLayout;
    hlayout->addWidget(mBtnPLine);
    hlayout->addWidget(mBtnPScatter);
    flayout->addRow(mLblPlotType, hlayout);
    hlayout = new QHBoxLayout;
    mLblPLineSize = new QLabel(this);
    mSpbPLineSize = new QDoubleSpinBox(this);
    mSpbPLineSize->setRange(0.5, 5.0);
    mSpbPLineSize->setValue(1.0);
    mSpbPLineSize->setSingleStep(0.1);
    mSpbPLineSize->setSuffix(" px");
    mSpbPLineSize->setAccelerated(true);
    flayout->addRow(mLblPLineSize, mSpbPLineSize);
    hlayout = new QHBoxLayout;
    mLblPSymbol = new QLabel(this);
    mCbxPSymbol = new QComboBox(this);
    Plotting::fillMarkerStyleComboBox(mCbxPSymbol);
    mSpbPSymbolSize = new QDoubleSpinBox(this);
    mSpbPSymbolSize->setRange(4.0, 24.0);
    mSpbPSymbolSize->setValue(6.0);
    mSpbPSymbolSize->setSingleStep(1.0);
    mSpbPSymbolSize->setSuffix(" px");
    mSpbPSymbolSize->setAccelerated(true);
    hlayout->addWidget(mCbxPSymbol);
    hlayout->addWidget(mSpbPSymbolSize);
    flayout->addRow(mLblPSymbol, hlayout);
    QObject::connect(mCbxPSymbol, QOverload<int>::of(&QComboBox::currentIndexChanged), [=](int) {
        if (mBtnPLine->isChecked()) {
            mLastLMS = mCbxPSymbol->currentData().toInt();
        } else {
            mLastSMS = mCbxPSymbol->currentData().toInt();
        }
    });
    QObject::connect(mButtonGroup, QOverload<int, bool>::of(&QButtonGroup::buttonToggled), this,
                     &GraphUIWidget::updatePlotWidgets);
}

/**
 * @brief Update graph line connection style as a function of the checked radio button.
 * @param id Identifier of the button
 * @param checked Check state.
 * @details
 * If the linestyle button checkstate has changed, the method enables/disables the line dash pattern and width widgets.
 *
 * If the scatterstyle button checkstate has changed, the method update the scatter style combobox selected value if
 * it is set to none.
 */
void GraphUIWidget::updatePlotWidgets(int id, bool checked) {

    // connection by line selected.
    if (id == 0) {
        mSpbPLineSize->setEnabled(checked);
    }
    if (id == 1) {
        int firstData = mCbxPSymbol->itemData(0).toInt();
        bool old = mCbxPSymbol->blockSignals(true);
        if (checked) {
            if (firstData == 0)
                mCbxPSymbol->removeItem(0);
        } else {
            if (firstData != 0)
                mCbxPSymbol->insertItem(0, Plotting::getScatterIcon(0), Plotting::getScatterName(0), 0);
        }
        mCbxPSymbol->blockSignals(old);
    }
    if (mBtnPLine->isChecked())
        mCbxPSymbol->setCurrentIndex(mCbxPSymbol->findData(mLastLMS));
    if (mBtnPScatter->isChecked())
        mCbxPSymbol->setCurrentIndex(mCbxPSymbol->findData(mLastSMS));
}

/// @brief Load plot settings from configuration file.
void GraphUIWidget::loadSettings() {
    QSettings settings;
    settings.beginGroup("graph");
    mBtnPLine->setChecked(settings.value("wline", true).toBool());
    mSpbPLineSize->setValue(settings.value("lwidth", 1.0).toDouble());
    int ms = settings.value("mstyle", 0).toInt();
    mCbxPSymbol->setCurrentIndex(qMax(0, mCbxPSymbol->findData(ms)));
    mSpbPSymbolSize->setValue(settings.value("msize", 6.0).toDouble());
    settings.endGroup();
    if (mBtnPLine->isChecked()) {
        mLastLMS = ms;
        mLastSMS = 1;
    } else {
        mLastLMS = 0;
        mLastSMS = ms;
    }
}

/// @brief Save plot settings in configuration file.
void GraphUIWidget::saveSettings() {
    QSettings settings;
    settings.beginGroup("graph");
    settings.setValue("wline", mBtnPLine->isChecked());
    settings.setValue("lwidth", mSpbPLineSize->value());
    settings.setValue("mstyle", mCbxPSymbol->currentData().toInt());
    settings.setValue("msize", mSpbPSymbolSize->value());
    settings.endGroup();
}

/// @brief Translates widgets of the plot groupbox.
void GraphUIWidget::translateUI() {
    // graph
    Plotting::reloadNames();
    Plotting::fillMarkerStyleComboBox(mCbxPSymbol, mBtnPScatter->isChecked());
    mBtnPLine->setText(tr("line"));
    mBtnPScatter->setText(tr("scatter"));
    mLblPlotType->setText(tr("Graph type"));
    mLblPLineSize->setText(tr("Line width"));
    mLblPSymbol->setText(tr("Marker style"));
}

void fillStyleCombobox(QComboBox *cbx) {
    int idx = cbx->currentIndex();
    cbx->clear();
    cbx->addItem(QCoreApplication::tr("Circle"), 0);
    cbx->addItem(QCoreApplication::tr("Cross (+)"), 1);
    cbx->addItem(QCoreApplication::tr("Diamond"), 2);
    cbx->addItem(QCoreApplication::tr("Square"), 3);
    cbx->addItem(QCoreApplication::tr("Triangle"), 4);
    cbx->addItem(QCoreApplication::tr("Diag. cross (X)"), 5);
    cbx->setCurrentIndex(idx);
}

/**
 * @brief Constructor
 * @param parent Parent widget.
 * @details
 * The constructor only builds up the UI.
 * Additionally, MapStyleWidget::loadSettings() and MapStyleWidget::translateUI() should be called to
 * populate the widgets with values and set texts.
 */
MapUIWidget::MapUIWidget(QWidget *parent) : QWidget(parent) { buildUI(); }

/// @brief Initialize widgets of the plot group box.
void MapUIWidget::buildUI() {

    mChbMapGrid = new QCheckBox(this);
    mChbHoriCB = new QCheckBox(this);
    mChbHideMV = new QCheckBox(this);

    // missing value style
    mGbxMissingValue = new QGroupBox(this);

    mLblMVColor = new QLabel(mGbxMissingValue);
    mLblMVColorChooser = new ColorChooserLabel(mGbxMissingValue);
    mLblMVColorChooser->setHighlight(true);
    mLblMVStyle = new QLabel(mGbxMissingValue);
    mCbxMVStyle = new QComboBox(mGbxMissingValue);
    fillStyleCombobox(mCbxMVStyle);
    mSpbMVSize = new QDoubleSpinBox(mGbxMissingValue);
    mSpbMVSize->setSuffix(" px");
    mSpbMVSize->setRange(4.0, 20.0);
    mSpbMVSize->setSingleStep(0.5);

    QFormLayout *flayout = new QFormLayout(mGbxMissingValue);
    flayout->addRow(mLblMVColor, mLblMVColorChooser);
    QHBoxLayout *hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxMVStyle);
    hlayout->addWidget(mSpbMVSize);
    flayout->addRow(mLblMVStyle, hlayout);

    // current flight style
    mGbxCurrentFlight = new QGroupBox(this);
    mLblCFStyle = new QLabel(mGbxCurrentFlight);
    mCbxCFStyle = new QComboBox(mGbxCurrentFlight);
    fillStyleCombobox(mCbxCFStyle);
    mSpbCFSize = new QDoubleSpinBox(mGbxCurrentFlight);
    mSpbCFSize->setSuffix(" px");
    mSpbCFSize->setRange(4.0, 20.0);
    mSpbCFSize->setSingleStep(0.5);
    flayout = new QFormLayout(mGbxCurrentFlight);
    hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxCFStyle);
    hlayout->addWidget(mSpbCFSize);
    flayout->addRow(mLblCFStyle, hlayout);

    // previous flights style
    mGbxPreviousFlights = new QGroupBox(this);
    mLblPFStyle = new QLabel(mGbxPreviousFlights);
    mCbxPFStyle = new QComboBox(mGbxPreviousFlights);
    fillStyleCombobox(mCbxPFStyle);
    mSpbPFSize = new QDoubleSpinBox(mGbxPreviousFlights);
    mSpbPFSize->setSuffix(" px");
    mSpbPFSize->setRange(4.0, 20.0);
    mSpbPFSize->setSingleStep(0.5);

    flayout = new QFormLayout(mGbxPreviousFlights);
    hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxPFStyle);
    hlayout->addWidget(mSpbPFSize);
    flayout->addRow(mLblPFStyle, hlayout);

    QVBoxLayout *vlayout = new QVBoxLayout(this);
    vlayout->addWidget(mChbMapGrid);
    vlayout->addWidget(mChbHoriCB);
    vlayout->addWidget(mChbHideMV);

    vlayout->addWidget(mGbxMissingValue);
    vlayout->addWidget(mGbxCurrentFlight);
    vlayout->addWidget(mGbxPreviousFlights);
    vlayout->addStretch(1);

    QObject::connect(mChbHideMV, &QCheckBox::toggled, mGbxMissingValue, &QWidget::setDisabled);
}

/// @brief Load plot settings from configuration file.
void MapUIWidget::loadSettings() {
    QSettings settings;
    settings.beginGroup("map");

    mChbMapGrid->setChecked(settings.value("m_showgrid", false).toBool());
    mChbHoriCB->setChecked(settings.value("cs_horiz", false).toBool());
    mChbHideMV->setChecked(settings.value("m_hidemv", false).toBool());

    mLblMVColorChooser->setColor(QColor(settings.value("m_mvcolor", "#000000").toString()));
    mCbxMVStyle->setCurrentIndex(mCbxCFStyle->findData(settings.value("m_mvmstyle", 0).toInt()));
    mSpbMVSize->setValue(settings.value("m_mvsize", 10.0).toDouble());

    mCbxCFStyle->setCurrentIndex(mCbxCFStyle->findData(settings.value("m_cmstyle", 0).toInt()));
    mSpbCFSize->setValue(settings.value("m_cmsize", 10.0).toDouble());

    mCbxPFStyle->setCurrentIndex(mCbxPFStyle->findData(settings.value("m_omstyle", 0).toInt()));
    mSpbPFSize->setValue(settings.value("m_omsize", 10.0).toDouble());

    settings.endGroup();
}

/// @brief Save plot settings in configuration file.
void MapUIWidget::saveSettings() {
    QSettings settings;
    settings.beginGroup("map");
    settings.setValue("cs_horiz", mChbHoriCB->isChecked());
    settings.setValue("m_showgrid", mChbMapGrid->isChecked());
    settings.setValue("m_hidemv", mChbHideMV->isChecked());

    settings.setValue("m_mvmstyle", mCbxMVStyle->currentData().toInt());
    settings.setValue("m_mvcolor", mLblMVColorChooser->color().name());
    settings.setValue("m_mvmsize", mSpbMVSize->value());

    settings.setValue("m_cmstyle", mCbxCFStyle->currentData().toInt());
    settings.setValue("m_cmsize", mSpbCFSize->value());

    settings.setValue("m_omstyle", mCbxPFStyle->currentData().toInt());
    settings.setValue("m_omsize", mSpbPFSize->value());
    settings.endGroup();
}

/// @brief Translates widgets of the plot groupbox.
void MapUIWidget::translateUI() {
    // map
    mChbMapGrid->setText(tr("Show map grid."));
    mChbHoriCB->setText(tr("Display the colorbar horizontally."));
    mChbHideMV->setText(tr("Hide missing value point on the map."));
    mGbxMissingValue->setTitle(tr("Missing value point"));
    mLblMVColor->setText(tr("Color"));
    mLblMVStyle->setText(tr("Symbol"));
    mGbxCurrentFlight->setTitle(tr("Current flight"));
    mLblCFStyle->setText(tr("Symbol"));
    mGbxPreviousFlights->setTitle(tr("Previous flights"));
    mLblPFStyle->setText(tr("Symbol"));

    setHelper(mGbxCurrentFlight, HelpFlag::CF_STYLE);
    setHelper(mGbxPreviousFlights, HelpFlag::PF_STYLE);
    setHelper(mGbxMissingValue, HelpFlag::MV_STYLE);
    setHelper(mChbHoriCB, HelpFlag::CS_ORIENT);
    setHelper(mChbHideMV, HelpFlag::HIDE_MV);
}

StylePage::StylePage(QWidget *parent) : BasePage(parent) {
    mPageTitle = tr("Style");
    mPageIcon = getIcon("style");
    buildUI();
    translateUI();
}

/// @brief Initialize widgets of the plot group box.
void StylePage::buildUI() {

    mTabWidget = new QTabWidget(this);
    mTabWidget->setMovable(false);

    mPlotWidget = new GraphUIWidget(mTabWidget);
    mMapWidget = new MapUIWidget(mTabWidget);

    mTabWidget->addTab(mPlotWidget, tr("Graph"));
    mTabWidget->addTab(mMapWidget, tr("Map"));

    QVBoxLayout *layout = new QVBoxLayout;
    layout->setContentsMargins(0, 0, 0, 0);
    layout->addWidget(mTabWidget);
    setLayout(layout);
}

void StylePage::loadSettings() {
    mPlotWidget->loadSettings();
    mMapWidget->loadSettings();
}

void StylePage::saveSettings() {

    mPlotWidget->saveSettings();
    mMapWidget->saveSettings();
}

void StylePage::translateUI() {
    mPageTitle = tr("Style");
    mPlotWidget->translateUI();
    mMapWidget->translateUI();

    mTabWidget->setTabText(0, tr("Graph"));
    mTabWidget->setTabText(1, tr("Map"));
}

/**
 * @brief Save the current tab widget index.
 */
void StylePage::saveUI() {
    QSettings settings;
    settings.setValue("layouts/cfg_diag/ui_tab", mTabWidget->currentIndex());
}

/**
 * @brief Load the current tab widget index.
 */
void StylePage::loadUI() {
    QSettings settings;
    bool tabIndex = settings.value("layouts/cfg_diag/ui_tab", 0).toInt();
    mTabWidget->setCurrentIndex(tabIndex);
}
