// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "timestampwidget.hpp"
#include "colorindicator.hpp"
#include "itemmapper.hpp"
#include "isp/times.hpp"

using namespace ISP;

/**
 * @brief Build a datetime from the given couple of date and time.
 * @param adaDate (Adafruit GPS) date as a double.
 * @param adaTime (Adafruit GPS) time as a double.
 * @details
 * The initial formats are ddmmyy for date and hhmmss for time, but due to double conversion, any
 * leading zero disappear from the value.
 *
 * Thus the method rebuilds the original string formats and then parses the results into QDate and QTime objects.
 *
 * @internal
 * A pain in the ass, while the datetime string retrieved from GPS messages could have been converted directly to a
 * NI timestamp (secs from 1904/01/01, like h_system parameter).
 */
QDateTime buildGPSDatetime(double adaDate, double adaTime) {
    QTime time = QTime::fromString(QString("%1").arg(adaTime, 10, 'f', 3, QChar('0')), "hhmmss.zzz");
    QDate date = QDate::fromString(QString("%1").arg(static_cast<int>(adaDate), 6, 10, QChar('0')), "ddMMyy");
    if (date.year() < 2000)
        date = date.addYears(100);
    return QDateTime(date, time, Qt::UTC);
}

/**
 * @brief Constructor.
 * @param parent Optional parent widget.
 */
TimestampWidget::TimestampWidget(QWidget *parent) : QWidget(parent) {
    mWarmupTime = -1;
    mMapper = nullptr;
    mHSysColor = Qt::black;
    mLayout = new QFormLayout(this);

    buildTimeStampLabels();

    QFrame *line = new QFrame(this);
    line->setFrameShape(QFrame::HLine);
    line->setFrameShadow(QFrame::Sunken);
    mLayout->addRow(line);

    buildAdditionalLabels();

    translateUI();
}

/**
 * @brief Set new warmup time.
 * @param milliseconds Warmup time in milliseconds.
 */
void TimestampWidget::setWarmupTime(int milliseconds) { mWarmupTime = milliseconds; }

/// @brief Get the size hint of the window.
QSize TimestampWidget::sizeHint() const { return QSize(150, 60); }

void TimestampWidget::buildTimeStampLabels() {
    mLblHSysDescr = new QLabel(this);
    mLblHGpsDescr = new QLabel(this);
    mLblHAcqDescr = new QLabel(this);

    mLblHSysValue = new QLabel(this);
    mLblHGpsValue = new QLabel(this);
    mLblHAcqValue = new QLabel(this);

    for (auto *l : {mLblHGpsValue, mLblHSysValue, mLblHAcqValue}) {
        l->setAlignment(Qt::AlignVCenter | Qt::AlignRight);
        l->setText("--:--:--.---");
    }
    mLayout->addRow(mLblHGpsDescr, mLblHGpsValue);
    mLayout->addRow(mLblHSysDescr, mLblHSysValue);
    mLayout->addRow(mLblHAcqDescr, mLblHAcqValue);
}

void TimestampWidget::buildAdditionalLabels() {

    mLblAETDescr = new QLabel(this);
    mLblAETValue = new QLabel(this);
    mLblAETValue->setAlignment(Qt::AlignVCenter | Qt::AlignRight);
    mLblAETValue->setText("???");

    QHBoxLayout *layout = new QHBoxLayout;
    mLblWarmupDescr = new QLabel(this);
    mWarmupIndicator = new ColorIndicator(this);
    mWarmupIndicator->setColor(StatusColor::Grey);
    layout->addWidget(mLblWarmupDescr, 2, Qt::AlignLeft);
    layout->addWidget(mWarmupIndicator, 0, Qt::AlignRight);

    mLayout->addRow(mLblAETDescr, mLblAETValue);
    mLayout->addRow(layout);
}

/**
 * @brief Set the key mapper.
 * @details
 * The method registers the given mapper in the widget. Without a (consistent) mapper, the widget cannot display
 * anything.
 */
void TimestampWidget::setMapper(ItemMapper *mapper) { mMapper = mapper; }

/**
 * @brief Update display timestamps using mapper's last registered values.
 * @details
 * The method extracts parameters from the blocks and computes the times (in UTC): GPS time, sensor time and the
 * monitoring PC time (at the time of the reception of the UDP packet).
 *
 * If no key mapper has been set (see TimestampWidget::setKeyMapper), the method will fail to process the input block
 * and the pattern "--:--:--.---" is set for all labels.
 *
 * If the block content is not consistent with the key mapper, the same pattern will be set for each time that cannot be
 * computed.
 *
 * @todo
 * Add h sys/gps Synchronization timestamp.
 */
void TimestampWidget::updateFromMapper() {
    static const QString missing = "--:--:--.---";
    QString gps, sys, pcm, etm;
    gps = sys = pcm = missing;
    etm = "N/A";
    StatusColor color = StatusColor::Grey;
    QColor syncColor = mHSysColor;
    if (mMapper.isNull()) {
        qDebug() << "[Timestamp widget]: No key mapper has been registered.";
    } else {
        double a, b, c, d, e;
        AParam f;
        // clock sync:
        a = mMapper->value(ItemMapper::CLOCK_SYNC);
        if (!(qIsNaN(a) || a > 0.0))
            syncColor = Qt::red;
        // timestamps
        a = mMapper->parameter(ItemMapper::GPS_DATE).floatValue();
        b = mMapper->parameter(ItemMapper::GPS_TIME).floatValue();
        c = mMapper->parameter(ItemMapper::SENSOR_TIME).floatValue();
        d = mMapper->parameter(ItemMapper::PCMON_TIME).floatValue();

        if (!(qIsNaN(a) || qIsNaN(b)))
            gps = buildGPSDatetime(a, b).toString("hh:mm:ss.zzz");
        if (!qIsNaN(c))
            sys = QDateTime::fromMSecsSinceEpoch((c + EPOCH1904_OFFSET) * 1000, Qt::UTC).toString("hh:mm:ss.zzz");
        if (!qIsNaN(d))
            pcm = QDateTime::fromMSecsSinceEpoch((d + EPOCH1904_OFFSET) * 1000, Qt::UTC).toString("hh:mm:ss.zzz");

        // build elapsed time (since acquisition start <=> acq strap status ON)
        e = mMapper->parameter(ItemMapper::A_NUMBER).floatValue();
        f = mMapper->parameter(ItemMapper::A_PERIOD);
        double etime = Units::convert(f.floatValue(), f.units, Units::MSEC) * e; // elapsed time in msec
        if (qIsNaN(etime)) {
            etm = "N/A";
        } else {
            etm = QString::fromStdString(msecsToTimeFormat(etime, "%m min. %ss"));
            if (mWarmupTime > 0)
                color = etime >= mWarmupTime ? StatusColor::Green : StatusColor::Red;
        }
    }
    mLblHSysValue->setText(sys);
    mLblHGpsValue->setText(gps);
    mLblHAcqValue->setText(pcm);
    mLblAETValue->setText(etm);
    mWarmupIndicator->setColor(color);

    if (syncColor != mHSysColor) {
        mHSysColor = syncColor;
        updateHSysColor();
    }
}

/**
 * @brief Update h-system timestamp labels color.
 * @details
 * The method updates the color of the h-system timestamp labels according to the current internal saved color
 * (mHSysColor). The color value itself is updated in TimestampWidget::reset and TimestampWidget::updateFromMapper (as
 * a function of the CLOCK_SYNC key value of the item mapper).
 */
void TimestampWidget::updateHSysColor() {
    QPalette pal;
    for (QLabel *lbl : {mLblHSysDescr, mLblHSysValue}) {
        pal = lbl->palette();
        pal.setColor(QPalette::WindowText, mHSysColor);
        lbl->setPalette(pal);
    }
}

/// @brief Reset the data display in the widget.
void TimestampWidget::reset() {
    static const QColor black(Qt::black);
    if (mHSysColor != black) {
        mHSysColor = black;
        updateHSysColor();
    }
     for (auto *l : {mLblHGpsValue, mLblHSysValue, mLblHAcqValue})
        l->setText("--:--:--.---");
    mLblAETValue->setText("N/A");
    mWarmupIndicator->setColor(StatusColor::Grey);
}

/**
 * @brief Translate widget texts.
 * @details
 * The method translates the text of the description labels as well as the window title of the widget.
 */
void TimestampWidget::translateUI() {
    setWindowTitle(tr("Timestamps and Timers"));
    mLblHSysDescr->setText(tr("Sensor system"));
    mLblHGpsDescr->setText(tr("GPS module"));
    mLblHAcqDescr->setText(tr("Monitoring PC"));

    mLblAETDescr->setText(tr("Acq. elapsed time"));
    mLblWarmupDescr->setText(tr("Lasers warmup"));
}
