// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/**
 * @file SciLineEdit.cpp
 * @brief Science double editor class definitions file.
 * @author J. Burgalat
 * @date 2016
 */

#include "valuelineedit.hpp"
#include "borderanimation.hpp"
#include "validators.hpp"
#include <QFocusEvent>
#include <QMouseEvent>
#include <QRegularExpression>
#include <QtGlobal>

/// @brief NaN string representation.
const QString ValueLineEdit::mNaN = QString::number(qQNaN());

/**
 * @brief Constructor.
 * @param prec Decimals precision.
 * @param parent Parent widget.
 * @details Minimum and maximum values are set to NaN, that is no specific range is set.
 */
ValueLineEdit::ValueLineEdit(bool allowEmpty, bool allowNaN, QWidget *parent) : QLineEdit(parent) {
    mNoAnim = true;
    mJustFocused = false;
    mAllowEmpty = allowEmpty;
    mAllowNaN = allowNaN;
    mLast = mAllowEmpty ? "" : (mAllowNaN ? mNaN : "0");
    setValidator(new DoubleValidator(this, mAllowEmpty, mAllowNaN));
    setText(mLast);
}

/**
 * @brief Enable/disable border animation on wrong input.
 * @param enabled <i>true</i> to anable, <i>false</i> otherwise.
 * @details If <i>enabled</i>, the border of the widget is animated (blink in red color for few milliseconds) when the
 * focus leaves the widget and the widget text is empty.
 */
void ValueLineEdit::setAnimation(bool enabled) { mNoAnim = !enabled; }

/// @brief Select all text if the widget has got the focus recently.
void ValueLineEdit::mousePressEvent(QMouseEvent *event) {
    QLineEdit::mousePressEvent(event);
    if (mJustFocused) {
        selectAll();
        mJustFocused = false;
    }
}

/// @brief Save last entry on focus in and select all text.
void ValueLineEdit::focusInEvent(QFocusEvent *event) {
    Q_UNUSED(event)
    mJustFocused = true;
    selectAll();
    QString ctxt = text();
    mLast = ctxt;
    if (mLast.simplified().isEmpty() && !mAllowEmpty)
        mLast = mAllowNaN ? "nan" : "0";
}

/**
 * @brief Check data on focus out.
 * @param event Focus event.
 * @details
 * The method checks whether or not the QLineEdit is currently empty. In the former
 * case, the widget is animated with a blinking red border and its text is set to
 * the last saved value.
 */
void ValueLineEdit::focusOutEvent(QFocusEvent *event) {
    QString ctxt = text().simplified();
    int pos = ctxt.size();
    // bad are not good cases

    bool badCases = !((mAllowEmpty && ctxt.isEmpty()) || (ctxt.toDouble() == qQNaN() && !mAllowNaN) ||
                      (validator()->validate(ctxt, pos) == QValidator::Acceptable));
    // qDebug() << mAllowEmpty << mAllowNaN << ctxt.isEmpty() << "=>" << badCases;
    if (badCases) {
        if (mNoAnim) {
            setText(mLast);
        } else {
            BorderAnimation *anim = new BorderAnimation(this);
            anim->start(QAbstractAnimation::DeleteWhenStopped);
            QObject::connect(anim, &QAbstractAnimation::finished, this, [=]() { setText(mLast); });
        }
        //  qDebug() << "BAD CASE";
    }
    QLineEdit::focusOutEvent(event);
}
