// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "xmlutils.hpp"
#include "dataitem.hpp"
#include "itemmapper.hpp"
#include "messagehandler.hpp"

#include "indicators.hpp"
#include "scenario.hpp"

#include <QXmlSchema>
#include <QXmlSchemaValidator>
#include <QtXml>

#include "cfg_versions.hpp"

typedef QPair<bool, int> BIPair;

bool validateContent(const QByteArray &xmlContent, const QByteArray &xmlSchema, QStringList *errors);

Scenario *parseScenario(QDomElement elt, const ItemMapper *mapper = nullptr, QStringList *errors = nullptr);
QVector<DataItem *> parseItems(QDomNode parentNode, const ItemMapper *mapper = nullptr,
    QVector<BIPair> *attributes = nullptr, QStringList *errors = nullptr);
QVector<BaseIndicator *> parseIndicators(
    QDomNode parentNode, const ItemMapper *mapper = nullptr, QStringList *errors = nullptr);
BasicIndicator *parseBasicIndicator(QDomElement elt, const ItemMapper *mapper = nullptr, QStringList *errors = nullptr);
SwitchIndicator *parseSwitchIndicator(
    QDomElement elt, const ItemMapper *mapper = nullptr, QStringList *errors = nullptr);
RangeIndicator *parseRangeIndicator(QDomElement elt, const ItemMapper *mapper = nullptr, QStringList *errors = nullptr);

// public methods.

/**
 * @brief Read scenarios from XML file.
 * @param path Path of the XML file.
 * @param errors Optional pointer to a QStringList with in output, the errors messages.
 * @return A vector of Scenario instances.
 */
QVector<Scenario *> loadScenariosFromXML(
    const QString &xmlPath, const ItemMapper *mapper, int *selected, QStringList *errors) {
    QVector<Scenario *> out;
    QFile file(xmlPath);
    // cannot open file.
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QCoreApplication::tr("Cannot open %1").arg(xmlPath));
        return out;
    }

    QByteArray content = file.readAll();
    file.close();
    file.setFileName(":/adm2/xsd/scenarios.xsd");
    file.open(QIODevice::ReadOnly | QIODevice::Text);
    QByteArray schema = file.readAll();
    file.close();
    if (!validateContent(content, schema, errors)) {
        if (errors)
            errors->append(QCoreApplication::tr("%1 is not a valid scenario file (syntax error).").arg(xmlPath));
        return out;
    }
    QDomDocument document;
    document.setContent(content);
    // QDomNode
    QDomElement root = document.firstChildElement("scenarios");
    if (root.isNull()) {
        if (errors)
            errors->append(QCoreApplication::tr("No <scenarios> node found."));
        return out;
    }
    QDomNodeList nodes = root.childNodes();
    for (int i = 0; i < nodes.count(); i++) {
        QDomElement element = nodes.at(i).toElement();
        if (element.isNull())
            continue;
        QString tagName = element.tagName();
        if (tagName.compare("scenario") == 0) {
            Scenario *scn = parseScenario(element, mapper, errors);
            if (scn)
                out.push_back(scn);
        }
    }
    int i = 1;
    for (Scenario *scn : out) {
        if (scn->name().isEmpty())
            scn->setName(Scenario::tr("Scenario #%1").arg(i));
        i++;
    }

    bool ok;
    int last = root.attribute("last", "-1").toInt(&ok);
    if (!ok || last < 0 || last >= out.size())
        last = -1;
    if (selected)
        *selected = last;
    return out;
}

/**
 * @brief Save list of scenarios to XML file.
 * @param xmlPath Path of the XML file to write.
 * @param scenarios List of scenarios to save.
 * @param selected Optional pointer to the index of the selected scenario in the list (see method documentation for more
 *details.
 * @return True on success, false otherwise.
 * @details
 * The method saves the given list of scenarios in a XML file located at xmlPath. The method fails if the location is
 *not writable.
 *
 * If selected is given as one of the scenario of the list, then it is flagged as the last used scenario in the XML
 *file. When such file is loaded in the application, the scenario will be selected upon loading.
 */
bool writeScenariosToXML(
    const QString &xmlPath, const QVector<Scenario *> scenarios, int *selected, QStringList *errors) {
    QVector<Scenario *> out;
    QFile file(xmlPath);
    // cnanot open file.
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QCoreApplication::tr("Cannot open %1 for writing").arg(xmlPath));
        return false;
    }
    int idx = -1;
    if (selected)
        idx = *selected;
    if (idx < 0 || idx >= scenarios.size())
        idx = -1;
    QStringList scns;
    for (Scenario *scn : scenarios)
        scns.append(QString(scn->toXML()));
    QString content =
        QString("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?><scenarios last=\"%1\" version=\"%2\">%3</scenarios>")
            .arg(idx)
            .arg(SCN_XML_CUR_VER)
            .arg(scns.join(""));

    QDomDocument doc;
    doc.setContent(content);
    QTextStream stream(&file);
    doc.save(stream, 2);
    return true;
}

/**
 * @brief Read parameter indicators from XML file.
 * @param path Path of the XML file.
 * @param errors Optional pointer to a QStringList with in output, the errors messages.
 * @return A vector of BaseIndicator instances.
 */
QVector<BaseIndicator *> loadIndicatorsFromXML(const QString &xmlPath, const ItemMapper *mapper, QStringList *errors) {
    QVector<BaseIndicator *> out;
    QFile file(xmlPath);
    // cnanot open file.
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QCoreApplication::tr("Cannot open %1").arg(xmlPath));
        return out;
    }
    QByteArray content = file.readAll();
    file.close();
    file.setFileName(":/adm2/xsd/indicators.xsd");
    file.open(QIODevice::ReadOnly | QIODevice::Text);
    QByteArray schema = file.readAll();
    file.close();
    if (!validateContent(content, schema, errors)) {
        if (errors)
            errors->append(QCoreApplication::tr("XML file is not valid"));
        return out;
    }
    QDomDocument document;
    document.setContent(content);
    QDomNode root = document.firstChildElement("indicators");

    out = parseIndicators(root, mapper, errors);

    return out;
}

// private methods.

/**
 * @brief Validate XML content with XML schema.
 * @param xmlContent Content to validate.
 * @param xmlSchema Schema to use for validation.
 * @param errors Option pointer to a list of strings with in output the errors messages.
 * @return True if the content is validated against the schema, false otherwise.
 */
bool validateContent(const QByteArray &xmlContent, const QByteArray &xmlSchema, QStringList *errors) {
    const QByteArray &instanceData = xmlContent;
    const QByteArray &schemaData = xmlSchema;

    MessageHandler messageHandler;
    QXmlSchema schema;
    schema.setMessageHandler(&messageHandler);
    bool errorOccurred = false;
    if (!schemaData.isEmpty())
        schema.load(schemaData);
    if (!schema.isValid()) {
        errorOccurred = true;
        if (errors)
            errors->append(messageHandler.statusMessage());
    } else {
        QXmlSchemaValidator validator(schema);
        if (!validator.validate(instanceData)) {
            errorOccurred = true;
            if (errors)
                errors->append(messageHandler.statusMessage());
        }
    }
    return !errorOccurred;
}

/**
 * @brief Build a scenario from the given dom element.
 * @param elt Element to parse.
 * @param mapper Optional mapper used to check and update underlying parameters.
 * @param errors Optional pointer to a QStringList with in output, the errors messages.
 * @return A pointer to a Scenario on success, a null pointer otherwise.
 */
Scenario *parseScenario(QDomElement elt, const ItemMapper *mapper, QStringList *errors) {
    if (elt.tagName().compare("scenario") != 0)
        return nullptr;

    QString name, comments, layout;
    QVector<BaseIndicator *> indicators;
    QVector<DataItem *> parameters;
    QVector<BIPair> attributes;
    QDomNode n = elt.firstChild();
    while (!n.isNull()) {
        if (n.isElement()) {
            QDomElement e = n.toElement();
            QString tagName = e.tagName();
            if (tagName.compare("name") == 0) {
                name = e.text().trimmed();
            } else if (tagName.compare("description") == 0) {
                comments = e.text().trimmed();
            } else if (tagName.compare("parameters") == 0) {
                parameters = parseItems(e, mapper, &attributes, errors);
            } else if (tagName.compare("indicators") == 0) {
                indicators = parseIndicators(e, mapper, errors);
            }
        }
        n = n.nextSibling();
    }
    Scenario *scn = new Scenario(name, comments, parameters, indicators);
    // setup the selected map and plot items in the scenario from the attributes data.
    ItemModel *model = scn->itemsModel();
    for (int i = 0; i < attributes.size(); i++) {
        const BIPair &attrs = attributes.at(i);
        if (attrs.first) {
            model->setData(model->index(i, ItemModel::mapColumn), true);
        }
        if (attrs.second > -1) {
            int col = attrs.second + ItemModel::plotColumnStart;
            model->setData(model->index(i, col), true);
        }
    }
    return scn;
}

/**
 * @brief Build a list of data item from the given parent node.
 * @param parentNode Parent node to parse (\<parameters>).
 * @param mapper Optional mapper used to check and update underlying parameters.
 * @param attributes Optional pointer to a vector with in output the map and plot index attributes of the items.
 * @param errors Optional pointer to a QStringList with in output, the errors messages.
 * @return A vector of DataItem pointers.
 */
QVector<DataItem *> parseItems(
    QDomNode parentNode, const ItemMapper *mapper, QVector<BIPair> *attributes, QStringList *errors) {

    static const QStringList true_flags = {"1", "yes", "y", "true", "t"};

    QVector<DataItem *> out;
    QVector<BIPair> attrs;
    QDomElement parentElt = parentNode.toElement();
    if (parentElt.isNull()) {
        if (errors)
            errors->append(QCoreApplication::tr("No <parameters> node found."));

        return out;
    }

    // check version
    int cfgver = parentNode.toElement().attribute("version", "0").toInt();
    if (cfgver < PRM_XML_MIN_VER) {
        if (errors)
            errors->append(
                QCoreApplication::tr("<parameters> node version too old (must be >=%1).").arg(PRM_XML_MIN_VER));
        return out;
    }

    bool okConv;

    QDomNodeList nodes = parentElt.elementsByTagName("parameter");
    for (int i = 0; i < nodes.count(); i++) {
        QDomElement prmElt = nodes.at(i).toElement();
        // parse element.
        bool autoUnit = true;
        double cvmin = qQNaN(), cvmax = qQNaN(), rvmin = qQNaN();
        QString tmp, pname, palias;
        Units::Unit unit = Units::NOUNIT;
        QDomNode n = prmElt.firstChild();
        while (!n.isNull()) {
            if (n.isElement()) {
                QDomElement e = n.toElement();
                QString tagName = e.tagName();
                if (tagName.compare("name") == 0) {
                    pname = e.text().trimmed();
                } else if (tagName.compare("alias") == 0) {
                    palias = e.text().trimmed();
                } else if (tagName.compare("units") == 0) {
                    tmp = e.text().trimmed();
                    if (tmp.size()) {
                        autoUnit = false;
                        unit = Units::unitFromString(tmp.toStdString());
                        if (unit == Units::UKN && errors) {
                            errors->push_back(
                                QCoreApplication::tr("Cannot find matching units for <units> tag: \"%1\"").arg(tmp));
                        }
                    }
                } else if (tagName.compare("min") == 0) {
                    tmp = e.text().trimmed();
                    cvmin = tmp.toDouble(&okConv);
                    if (!okConv) {
                        cvmin = qQNaN();
                        if (errors)
                            errors->push_back(QCoreApplication::tr("Cannot convert <min> value: \"%1\"").arg(tmp));
                    }
                } else if (tagName.compare("max") == 0) {
                    tmp = e.text().trimmed();
                    cvmax = tmp.toDouble(&okConv);
                    if (!okConv) {
                        cvmax = qQNaN();
                        if (errors)
                            errors->push_back(QCoreApplication::tr("Cannot convert <max> value: \"%1\"").arg(tmp));
                    }
                } else if (tagName.compare("remanence") == 0) {
                    tmp = e.text().trimmed();
                    rvmin = tmp.toDouble(&okConv);
                    if (tmp.isEmpty() || !okConv) {
                        rvmin = qQNaN();
                        if (errors && !okConv && !tmp.isEmpty())
                            errors->push_back(
                                QCoreApplication::tr("Cannot convert <remanence> value: \"%1\"").arg(tmp));
                    }
                }
            }
            n = n.nextSibling();
        }
        if (palias.isEmpty()) {
            palias = pname.toUpper().replace("_", " ");
        }
        // check
        if (pname.isEmpty() || qIsNaN(cvmin) || qIsNaN(cvmax)) {
            if (errors)
                errors->push_back(QCoreApplication::tr("Too much errors: skipping <parameter> element"));
            continue;
        }

        AParam prm;
        if (mapper)
            prm = mapper->parameter(pname);
        if (!prm.isValid())
            prm = AParam(pname.toStdString(), qQNaN(), unit);

        // autoUnit == true if <units> is not given or filled.
        // then we assume that units are not relevant or is simply the same as the parameter found in the
        // optional param mapper (if no mapper or no param found then units are simply irrelevant).
        if (!autoUnit) {
            // check unit consistency
            if (Units::inconsistent(unit, prm.units) && errors) {
                errors->push_back(
                    QCoreApplication::tr("Parameter \"%1\" units not consistent with block descriptor: %2 / %3.")
                        .arg(QString::fromStdString(prm.name()),
                            QString::fromStdString(Units::unitToString(unit)),
                            QString::fromStdString(Units::unitToString(prm.units))));
            }
            // force param unit to be the same as the parsed one.
            prm.units = unit;
        }

        // parse <parameter> element attributes:
        // get attributes
        bool onMap = true_flags.contains(prmElt.attribute("map", "false").trimmed());
        int plotIndex = prmElt.attribute("plot", "-1").trimmed().toInt(&okConv);
        if (!okConv)
            plotIndex = -1;
        attrs.push_back(qMakePair(onMap, plotIndex));
        out.push_back(new DataItem(prm, palias, DataRange(cvmin, cvmax), rvmin));
    }
    if (attributes)
        *attributes = attrs;
    return out;
}

/**
 * @brief Parse a list of indicators (switch or range) from an \<indicators> node.
 * @param parentNode Parent node to process.
 * @param mapper Optional mapper used to check and update underlying parameters.
 * @param errors Optional pointer to a QStringList with in output, the errors messages.
 * @return Vector of BaseIndicators pointers.
 * @details
 * The method parses all indicators that are defined within parentNode and returns them as a vector
 * of BaseIndicators pointers.
 */
QVector<BaseIndicator *> parseIndicators(QDomNode parentNode, const ItemMapper *mapper, QStringList *errors) {
    QVector<BaseIndicator *> out;
    if (parentNode.isNull()) {
        if (errors)
            errors->append(QCoreApplication::tr("No <indicators> node found."));
        return out;
    }

    // check version
    int cfgver = parentNode.toElement().attribute("version", "0").toInt();
    if (cfgver < IND_XML_MIN_VER) {
        if (errors)
            errors->append(
                QCoreApplication::tr("<indicators> node version too old (must be >=%1).").arg(IND_XML_MIN_VER));
        return out;
    }

    QDomNodeList nodes = parentNode.childNodes();
    for (int i = 0; i < nodes.count(); i++) {
        QDomElement element = nodes.at(i).toElement();
        if (element.isNull())
            continue;
        QString tagName = element.tagName();
        if (tagName.compare("range") == 0) {
            RangeIndicator *ind = parseRangeIndicator(element, mapper, errors);
            if (ind)
                out.push_back(ind);
        } else if (tagName.compare("switch") == 0) {
            SwitchIndicator *ind = parseSwitchIndicator(element, mapper, errors);
            if (ind)
                out.push_back(ind);
        } else if (tagName.compare("basic") == 0) {
            BasicIndicator *ind = parseBasicIndicator(element, mapper, errors);
            if (ind)
                out.push_back(ind);
        }
    }
    return out;
}

/**
 * @brief Parse a single switch indicator from the given dom element.
 * @param elt Dom element to parse.
 * @param mapper Optional parameter mapper instance used to check and update the indicator underlying parameter.
 * @param errors Optional pointer to a QStringList with in output, the errors messages.
 * @return A switch indicator or a null pointer if the element could not be parsed successfully.
 * @details
 * The method parses a single switch node. It may fail if the given node is not a switch node or if its subnode "name"
 * (which is related to the underlying parameter of the indicator) does have a value. In such case a null pointer is
 *returned.
 *
 * If given, mapper is used to retrieve the underlying parameter, otherwise the parameter is built from scratch using
 * the subnode name value.
 */
BasicIndicator *parseBasicIndicator(QDomElement elt, const ItemMapper *mapper, QStringList *errors) {
    if (elt.tagName().compare("basic") != 0)
        return nullptr;
    QString tmp, pname, palias;
    bool autoUnit = true;
    Units::Unit unit = Units::NOUNIT;
    QDomNode n = elt.firstChild();
    while (!n.isNull()) {
        if (n.isElement()) {
            QDomElement e = n.toElement();
            QString tagName = e.tagName();
            if (tagName.compare("name") == 0) {
                pname = e.text().trimmed();
            } else if (tagName.compare("alias") == 0) {
                palias = e.text().trimmed();
            } else if (tagName.compare("units") == 0) {
                tmp = e.text().trimmed();
                if (!tmp.isEmpty()) {
                    autoUnit = false;
                    unit = Units::unitFromString(tmp.toStdString());
                    if (unit == Units::UKN && errors) {
                        errors->push_back(
                            QCoreApplication::tr("Cannot find matching units for <units> tag: \"%1\"").arg(tmp));
                    }
                }
            }
        }
        n = n.nextSibling();
    }
    if (palias.isEmpty())
        palias = pname.toUpper().replace("_", " ");

    // check
    if (pname.isEmpty()) {
        if (errors)
            errors->push_back(QCoreApplication::tr("<name> element is empty."));
        return nullptr;
    }
    AParam prm;
    if (mapper)
        prm = mapper->parameter(pname);
    if (!prm.isValid())
        prm = AParam(pname.toStdString(), qQNaN(), unit);

    // autoUnit == true if <units> is not given or filled.
    // then we assume that units are not relevant or is simply the same as the parameter found in the
    // optional param mapper (if no mapper or no param found then units are simply irrelevant).

    if (!autoUnit) {
        // check unit consistency
        if (Units::inconsistent(unit, prm.units) && errors) {
            errors->push_back(
                QCoreApplication::tr("Indicator \"%1\" units not consistent with block descriptor: %2 / %3.")
                    .arg(QString::fromStdString(prm.name()),
                        QString::fromStdString(Units::unitToString(unit)),
                        QString::fromStdString(Units::unitToString(prm.units))));
        }
        // force param unit to be the same as the parsed one.
        prm.units = unit;
    }

    return new BasicIndicator(prm, palias);
}

/**
 * @brief Parse a single switch indicator from the given dom element.
 * @param elt Dom element to parse.
 * @param mapper Optional parameter mapper instance used to check and update the indicator underlying parameter.
 * @param errors Optional pointer to a QStringList with in output, the errors messages.
 * @return A switch indicator or a null pointer if the element could not be parsed successfully.
 * @details
 * The method parses a single switch node. It may fail if the given node is not a switch node or if its subnode "name"
 * (which is related to the underlying parameter of the indicator) does have a value. In such case a null pointer is
 *returned.
 *
 * If given, mapper is used to retrieve the underlying parameter, otherwise the parameter is built from scratch using
 * the subnode name value.
 */
SwitchIndicator *parseSwitchIndicator(QDomElement elt, const ItemMapper *mapper, QStringList *errors) {
    if (elt.tagName().compare("switch") != 0)
        return nullptr;
    QString tmp, pname, palias;
    bool reverse = false;
    if (elt.attribute("reverse", "no").compare("yes", Qt::CaseInsensitive) == 0)
        reverse = true;
    QDomNode n = elt.firstChild();
    while (!n.isNull()) {
        if (n.isElement()) {
            QDomElement e = n.toElement();
            QString tagName = e.tagName();
            if (tagName.compare("name") == 0) {
                pname = e.text().trimmed();
            } else if (tagName.compare("alias") == 0) {
                palias = e.text().trimmed();
            }
        }
        n = n.nextSibling();
    }
    if (palias.isEmpty())
        palias = pname.toUpper().replace("_", " ");

    // check
    if (pname.isEmpty()) {
        if (errors)
            errors->push_back(QCoreApplication::tr("<name> element is empty."));
        return nullptr;
    }
    AParam prm;
    if (mapper)
        prm = mapper->parameter(pname);
    if (!prm.isValid())
        prm = AParam(pname.toStdString(), qQNaN(), Units::NOUNIT);

    return new SwitchIndicator(prm, palias, reverse);
}

/**
 * @brief Parse a single range indicator from the given dom element.
 * @param elt Dom element to parse.
 * @param mapper Optional parameter mapper instance used to check and update the indicator underlying parameter.
 * @param errors Optional pointer to a QStringList with in output, the errors messages.
 * @return A range indicator or a null pointer if the element could not be parsed successfully.
 * @details
 * The method parses a single "indicator" node. It may fail if the given node is not a "indicator" node or if its
 *subnode "name" (which is related to the underlying parameter of the indicator) does have a value. In such case a null
 *pointer is returned.
 *
 * If given, mapper is used to retrieve the underlying parameter, otherwise the parameter is built from scratch using
 * the subnode name value.
 */
RangeIndicator *parseRangeIndicator(QDomElement elt, const ItemMapper *mapper, QStringList *errors) {
    if (elt.tagName().compare("range") != 0)
        return nullptr;
    bool ok, down = false, autoUnit = true;
    double warn, crit;
    QString tmp, pname, palias;
    Units::Unit unit = Units::NOUNIT;
    warn = crit = qQNaN();
    if (elt.attribute("down", "no").compare("yes", Qt::CaseInsensitive) == 0)
        down = true;
    QDomNode n = elt.firstChild();
    while (!n.isNull()) {
        if (n.isElement()) {
            QDomElement e = n.toElement();
            QString tagName = e.tagName();
            if (tagName.compare("name") == 0) {
                pname = e.text().trimmed();
            } else if (tagName.compare("alias") == 0) {
                palias = e.text().trimmed();
            } else if (tagName.compare("units") == 0) {
                tmp = e.text().trimmed();
                if (!tmp.isEmpty()) {
                    autoUnit = false;
                    unit = Units::unitFromString(tmp.toStdString());
                    if (unit == Units::UKN && errors) {
                        errors->push_back(
                            QCoreApplication::tr("Cannot find matching units for <units> tag: \"%1\"").arg(tmp));
                    }
                }
            } else if (tagName.compare("warning") == 0) {
                warn = e.text().trimmed().toDouble(&ok);
                if (!ok)
                    warn = qQNaN();

            } else if (tagName.compare("critical") == 0) {
                crit = e.text().trimmed().toDouble(&ok);
                if (!ok)
                    crit = qQNaN();
            }
        }
        n = n.nextSibling();
    }
    if (palias.isEmpty()) {
        palias = pname.toUpper().replace("_", " ");
    }
    // check
    if (pname.isEmpty()) {
        if (errors)
            errors->push_back(QCoreApplication::tr("<name> element is empty."));
        return nullptr;
    }
    AParam prm;
    if (mapper)
        prm = mapper->parameter(pname);
    if (!prm.isValid())
        prm = AParam(pname.toStdString(), qQNaN(), unit);

    // autoUnit == true if <units> is not given or filled.
    // then we assume that units are not relevant or is simply the same as the parameter found in the
    // optional param mapper (if no mapper or no param found then units are simply irrelevant).

    if (!autoUnit) {
        // check unit consistency
        if (Units::inconsistent(unit, prm.units) && errors) {
            errors->push_back(
                QCoreApplication::tr("Indicator \"%1\" units not consistent with block descriptor: %2 / %3.")
                    .arg(QString::fromStdString(prm.name()),
                        QString::fromStdString(Units::unitToString(unit)),
                        QString::fromStdString(Units::unitToString(prm.units))));
        }
        // force param unit to be the same as the parsed one.
        prm.units = unit;
    }

    return new RangeIndicator(prm, palias, warn, crit, down);
}
