// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "wlaninterface.hpp"

/// @brief Default (invalid) constructor.
WlanInterface::WlanInterface() { connected = false; }

WlanInterface::WlanInterface(GUID iguid) { updateGuidInfos(iguid); }

/**
 * @brief Full constructor.
 * @param client WLAN Handle client.
 * @param guid WLAN interface GUID.
 * @details
 * This constructor is convienient wrapper to:
 * @code{.cpp}
 * WlanIface wif;
 * wif.updateGuidInfos(guid);
 * wif.updateWlanInfos(client);
 * @endcode
 */
WlanInterface::WlanInterface(GUID inputGuid, HANDLE handle) {
    updateGuidInfos(inputGuid);
    updateWlanInfos(handle);
}

/// @brief Check whether or not the instance is valid.
bool WlanInterface::valid() const { return !invalidGUID(guid); }

/**
 * @brief Update informations about the interface using the given GUID.
 * @param iguid GUID to process.
 * @return True on success, false otherwise.
 * @details
 * The method updates the name, alias, luid and guid of the instance with the values derived from the input GUID.
 * On failure (that isn win api methods fail), the instance is left unmodified.
 */
bool WlanInterface::updateGuidInfos(GUID inputGuid) {
    NET_LUID tmpLuid;
    const size_t sz = IF_MAX_STRING_SIZE + 1;
    WCHAR tmp[sz];
    QString tmpName, tmpAlias;
    if (ConvertInterfaceGuidToLuid(&inputGuid, &tmpLuid) != ERROR_SUCCESS)
        return false;
    if (ConvertInterfaceLuidToNameW(&tmpLuid, tmp, sz) != ERROR_SUCCESS)
        return false;
    tmpName = QString::fromWCharArray(&tmp[0]);
    if (ConvertInterfaceLuidToAlias(&tmpLuid, tmp, sz) != ERROR_SUCCESS)
        return false;
    tmpAlias = QString::fromWCharArray(&tmp[0]);
    guid = inputGuid;
    luid = tmpLuid;
    name = tmpName;
    alias = tmpAlias;
    return true;
}

/**
 * @brief Update current network informations.
 * @param client Handle of the wlan api client to use for query.
 * @return False if the query failed, true on success.
 * @details
 * If the query fails, the method aborts and the network informations are not updated.
 */
bool WlanInterface::updateWlanInfos(HANDLE client) {
    bool ret = false;
    if (!(client && valid()))
        return ret;
    DWORD dwResult;
    PWLAN_CONNECTION_ATTRIBUTES pConnectInfo = nullptr;
    DWORD connectInfoSize = sizeof(WLAN_CONNECTION_ATTRIBUTES);
    WLAN_OPCODE_VALUE_TYPE opCode = wlan_opcode_value_type_invalid;
    dwResult = WlanQueryInterface(client, &guid, wlan_intf_opcode_current_connection, nullptr, &connectInfoSize,
                                  (PVOID *)&pConnectInfo, &opCode);

    // if something wrong we just assume the interface is not connected to a network and exit without
    // updating the current network.
    if (dwResult != ERROR_SUCCESS) {
        connected = false;
        return ret;
    }
    QString currentProfile;
    connected = pConnectInfo->isState == wlan_interface_state_connected &&
                pConnectInfo->wlanAssociationAttributes.dot11BssType == dot11_BSS_type_infrastructure;
    if (connected) {
        network.profileName = QString::fromWCharArray(&pConnectInfo->strProfileName[0]).trimmed();
        network.signalStrength = static_cast<int>(pConnectInfo->wlanAssociationAttributes.wlanSignalQuality);
        const DOT11_SSID &dot11Ssid = pConnectInfo->wlanAssociationAttributes.dot11Ssid;
        network.SSID = QString(reinterpret_cast<const char *>(&dot11Ssid.ucSSID[0])).trimmed();
    } else {
        network = WlanInfos();
    }
    if (pConnectInfo != nullptr)
        WlanFreeMemory(pConnectInfo);
    return true;
}
