// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "wlanwatcher.hpp"

const DWORD WlanWatcher::dwClientVersion = 2;

/// @brief constructor.
WlanWatcher::WlanWatcher(QObject *parent) : QObject(parent), mHClient(nullptr) {

    DWORD dwCurVersion = 0;
    DWORD dwRetVal = WlanOpenHandle(WlanWatcher::dwClientVersion, nullptr, &dwCurVersion, &mHClient);
    if (dwRetVal == ERROR_SUCCESS) {
        dwRetVal = WlanRegisterNotification(mHClient, WLAN_NOTIFICATION_SOURCE_ACM, TRUE,
                                            &WlanWatcher::wlanNotification, this, nullptr, nullptr);
        // get interfaces
        resetIfaceInfos();
    }

    QObject::connect(&mTimer, &QTimer::timeout, this, &WlanWatcher::updateStates);
}

/// @brief Destructor.
WlanWatcher::~WlanWatcher() {
    if (mHClient) {
        WlanRegisterNotification(mHClient, WLAN_NOTIFICATION_SOURCE_NONE, TRUE, &WlanWatcher::wlanNotification, this,
                                 nullptr, nullptr);
        WlanCloseHandle(mHClient, nullptr);
    }
}

/// @brief Check whether or not the instance is valid.
bool WlanWatcher::isValid() const { return mHClient != nullptr; }

/// @brief Check whether or not the monitored SSID is available from at least one interface.
bool WlanWatcher::isAvailable() const { return mAvailable; }

/// @brief Check whether or not at least one interface is connected to the monitored SSID.
bool WlanWatcher::isConnected() const { return mConnected; }

/**
 * @brief Start monitoring registered SSID.
 * @param refreshTimeout Optional refresh timeout. Default to 1000 msecs.
 * @details
 * The method starts the monitoring process of the registered SSID.
 *
 * If no SSID has been registered, WlanWatcher::disconnected signal is emitted and the
 * connection and availibity statuses of the watcher are set to False.
 *
 * Otherwise, the method initializes the availability and connection statuses as well as the
 * signal strength of the SSID.
 */
void WlanWatcher::start(int refreshTimeout) {
    WlanRegisterNotification(mHClient, WLAN_NOTIFICATION_SOURCE_NONE, TRUE, &WlanWatcher::wlanNotification, this,
                             nullptr, nullptr);
    if (mSSID.isEmpty()) {
        mAvailable = false;
        mConnected = false;
        emit disconnected();
    } else {
        bool conn = false;
        int sigst = 101;
        // update interfaces and check for ssid
        QMap<GUID, WlanInterface>::iterator it = mInterfaces.begin();
        for (it = mInterfaces.begin(); it != mInterfaces.end(); it++) {
            WlanInterface &iface = it.value();
            iface.updateWlanInfos(mHClient);
            if (iface.connected && iface.network.SSID.compare(mSSID) == 0) {
                sigst = sigst > 100 ? iface.network.signalStrength : qMin(sigst, iface.network.signalStrength);
                conn = true;
            }
        }
        if (conn) {
            mAvailable = true;
            mConnected = true;
            emit connected(sigst);
        } else {
            mAvailable = checkAvailability(mSSID);
        }
    }
    mTimer.start(refreshTimeout);
}

/**
 *  @brief Stop monitoring registered SSID.
 *  @warning
 *  The method does not emit any signals.
 *
 */
void WlanWatcher::stop() {
    WlanRegisterNotification(mHClient, WLAN_NOTIFICATION_SOURCE_NONE, TRUE, &WlanWatcher::wlanNotification, this,
                             nullptr, nullptr);
    mTimer.stop();
}

/**
 * @brief Sett the SSID to monitor.
 * @param ssid SSID to monitor.
 * @details
 * The methods simply sets the SSID to monitor.
 * Unless it has already been called, WlanWatcher::start must be called to start monitoring the network.
 */
void WlanWatcher::monitorSSID(const QString &ssid) { mSSID = ssid; }

/**
 * @brief check whether or not the given ssid is available.
 * @param ssid SSID to search.
 * @return True if the ssid is reachable from (at least) one of the registered interface.
 */
bool WlanWatcher::checkAvailability(const QString &ssid) const {

    if (mHClient == nullptr)
        return false;

    DWORD dwResult;
    for (GUID guid : mInterfaces.keys()) {
        PWLAN_AVAILABLE_NETWORK_LIST pBssList = nullptr;
        PWLAN_AVAILABLE_NETWORK pBssEntry = nullptr;
        dwResult = WlanGetAvailableNetworkList(mHClient, &guid, 0, nullptr, &pBssList);
        if (dwResult == ERROR_SUCCESS) {
            for (DWORD j = 0; j < pBssList->dwNumberOfItems; j++) {
                pBssEntry = &pBssList->Network[j];
                // 1. BSS type
                if (!(pBssEntry->dot11BssType == dot11_BSS_type_infrastructure && pBssEntry->dot11Ssid.uSSIDLength))
                    continue;
                QString cssid(reinterpret_cast<const char *>(&pBssEntry->dot11Ssid.ucSSID[0]));
                if (cssid.trimmed().compare(ssid) == 0) {
                    WlanFreeMemory(pBssList);
                    return true;
                }
            }
        }
        if (pBssList != nullptr) {
            WlanFreeMemory(pBssList);
            pBssList = nullptr;
        }
    }
    return false;
}

/**
 * @brief Get the list of all available networks from all registered interfaces.
 * @details
 * Unlike WlanWatcher::availableSSIDs, the method can return SSID duplicates. In such cases, the networks are seen by
 * different interfaces and can be distinguished using their GUID member.
 */
QVector<WlanInfos> WlanWatcher::availableNetworks() const {
    QVector<WlanInfos> networks;
    if (mHClient == nullptr)
        return networks;
    DWORD dwResult;
    for (GUID guid : mInterfaces.keys()) {
        PWLAN_AVAILABLE_NETWORK_LIST pBssList = nullptr;
        PWLAN_AVAILABLE_NETWORK pBssEntry = nullptr;
        dwResult = WlanGetAvailableNetworkList(mHClient, &guid, 0, nullptr, &pBssList);
        if (dwResult == ERROR_SUCCESS) {
            for (DWORD j = 0; j < pBssList->dwNumberOfItems; j++) {
                pBssEntry = &pBssList->Network[j];
                // 1. BSS type
                if (pBssEntry->dot11BssType != dot11_BSS_type_infrastructure)
                    continue;
                if (pBssEntry->dot11Ssid.uSSIDLength > 0) {
                    WlanInfos infos;
                    infos.SSID = QString(reinterpret_cast<const char *>(&pBssEntry->dot11Ssid.ucSSID[0])).trimmed();
                    infos.profileName = QString::fromWCharArray(&pBssEntry->strProfileName[0]).trimmed();
                    infos.signalStrength = pBssEntry->wlanSignalQuality;
                    infos.guid = guid;
                    networks.push_back(infos);
                }
            }
        }
        if (pBssList != nullptr) {
            WlanFreeMemory(pBssList);
            pBssList = nullptr;
        }
    }
    return networks;
}

/**
 * @brief Get all available SSID seen by all registered interfaces.
 * @return List of ssid. Duplicates are removed.
 */
QStringList WlanWatcher::availableSSIDs() const {
    QStringList ssids;
    if (mHClient == nullptr)
        return ssids;
    DWORD dwResult;
    for (GUID guid : mInterfaces.keys()) {
        PWLAN_AVAILABLE_NETWORK_LIST pBssList = nullptr;
        PWLAN_AVAILABLE_NETWORK pBssEntry = nullptr;
        dwResult = WlanGetAvailableNetworkList(mHClient, &guid, 0, nullptr, &pBssList);
        if (dwResult == ERROR_SUCCESS) {
            for (DWORD j = 0; j < pBssList->dwNumberOfItems; j++) {
                pBssEntry = &pBssList->Network[j];
                // 1. BSS type
                if (pBssEntry->dot11BssType != dot11_BSS_type_infrastructure)
                    continue;
                QString ssid;
                if (pBssEntry->dot11Ssid.uSSIDLength > 0)
                    ssids.push_back(QString(reinterpret_cast<const char *>(&pBssEntry->dot11Ssid.ucSSID[0])).trimmed());
            }
        }
        if (pBssList != nullptr) {
            WlanFreeMemory(pBssList);
            pBssList = nullptr;
        }
    }
    ssids.removeDuplicates();
    return ssids;
}

/// @brief Get the current monitored SSID.
const QString &WlanWatcher::monitoredSSID() const { return mSSID; }

/**
 * @brief Enumerate WLAN interfaces.
 * @details
 * The method resets the internal map of WLAN interfaces.
 */
void WlanWatcher::resetIfaceInfos() {
    mInterfaces.clear();
    if (mHClient == nullptr)
        return;
    PWLAN_INTERFACE_INFO_LIST pIfList = nullptr;
    DWORD dwRetVal = WlanEnumInterfaces(mHClient, nullptr, &pIfList);
    if (dwRetVal == ERROR_SUCCESS) {
        for (unsigned int i = 0; i < pIfList->dwNumberOfItems; i++) {
            GUID guid = (&pIfList->InterfaceInfo[i])->InterfaceGuid;
            mInterfaces[guid] = WlanInterface(guid, mHClient);
        }
    } /*else {
        qDebug() << "Failed to enumerate interfaces...";
    }*/
    if (pIfList != nullptr)
        WlanFreeMemory(pIfList);
}

/// @brief Register the last available notification.
void WlanWatcher::setLastNotification(Notification notification, GUID guid) {
    // qDebug() << QUuid(guid).toString() << ":" << WlanWatcher::notificationToString(notification);
    mLastNotification = notification;
    if (invalidGUID(guid))
        return;
    WlanInterface iface;
    if (notification == INTERFACE_ARRIVAL) {
        // add new interface.
        mInterfaces[guid] = WlanInterface(guid, mHClient);
        iface = mInterfaces[guid];
    } else if (notification == INTERFACE_REMOVAL) {
        iface = mInterfaces.take(guid);
    } else if (notification == DISCONNECTED) {
        mInterfaces[guid].connected = false;
        iface = mInterfaces[guid];
    } else if (notification == CONNECTION_COMPLETE) {
        // update status of the network.
        mInterfaces[guid].updateWlanInfos(mHClient);
        iface = mInterfaces[guid];
    } else if (notification != SCAN_COMPLETE) {
        return;
    }
    updateStates();
    emit interfaceChange(iface);
}
/**
 * @brief Update informations about current connection state of each interfaces.
 * @details
 * The method updates the informations about the current connection state for each interface registered in the watcher.
 * Additionally it updates the connected and available internal flags if the watcher is currently monitoring an SSID.
 */
void WlanWatcher::updateStates() {
    if (mHClient == nullptr)
        return;
    if (!mSSID.isEmpty()) {
        // Retrieve new informations for each WLAN interfaces.
        QMap<GUID, WlanInterface>::iterator it = mInterfaces.begin();
        for (it = mInterfaces.begin(); it != mInterfaces.end(); it++)
            it.value().updateWlanInfos(mHClient);
        // check for SSID
        bool curC = false;
        int sigst = 101;
        for (it = mInterfaces.begin(); it != mInterfaces.end(); it++) {
            WlanInterface &iface = it.value();
            // qDebug() << iface.connected << iface.network.SSID << mSSID;
            if (iface.connected && iface.network.SSID.compare(mSSID) == 0) {
                sigst = sigst > 100 ? iface.network.signalStrength : qMin(sigst, iface.network.signalStrength);
                curC = true;
            }
        }
        if (curC && !mConnected) {
            mAvailable = true;
            mConnected = true;
            emit connected(sigst);
        } else if (curC && mConnected) {
            emit signalChanged(sigst);
        }
        if (mConnected && !curC) {
            mConnected = false;
            emit disconnected();
        }
    }
}

/// @brief Manage notification from wlan handle.
VOID WINAPI WlanWatcher::wlanNotification(PWLAN_NOTIFICATION_DATA wlanNotifData, PVOID context) {
    WlanWatcher *pThis = (WlanWatcher *)context;
    if (wlanNotifData->NotificationSource != WLAN_NOTIFICATION_SOURCE_ACM)
        return;
    Notification notification = Notification(wlanNotifData->NotificationCode);
    pThis->setLastNotification(notification, wlanNotifData->InterfaceGuid);
}

/**
 * @brief Get a string representation of the given WLAN notification.
 * @param notif Notification to process.
 * @return String representation (very minimalist).
 */
QString WlanWatcher::notificationToString(Notification notification) {
    switch (notification) {
    case Notification::START:
        return "START";
    case Notification::AUTOCONF_ENABLED:
        return "AUTOCONF ENABLED";
    case Notification::AUTOCONF_DISABLED:
        return "AUTOCONF DISABLED";
    case Notification::BACKGOUND_SCAN_ENABLED:
        return "BACKGOUND SCAN ENABLED";
    case Notification::BACKGOUND_SCAN_DISABLED:
        return "BACKGOUND SCAN DISABLED";
    case Notification::BSS_TYPE_CHANGE:
        return "BSS_TYPE_CHANGE";
    case Notification::POWER_SETTINGS_CHANGE:
        return "POWER SETTINGS CHANGE";
    case Notification::SCAN_COMPLETE:
        return "SCAN COMPLETE";
    case Notification::SCAN_FAIL:
        return "SCAN FAIL";
    case Notification::CONNECTION_START:
        return "CONNECTION START";
    case Notification::CONNECTION_COMPLETE:
        return "CONNECTION COMPLETE";
    case Notification::CONNECTION_ATTEMPT_FAIL:
        return "CONNECTION ATTEMPT FAIL";
    case Notification::FILTER_LIST_CHANGE:
        return "FILTER LIST CHANGE";
    case Notification::INTERFACE_ARRIVAL:
        return "INTERFACE ARRIVAL";
    case Notification::INTERFACE_REMOVAL:
        return "INTERFACE REMOVAL";
    case Notification::PROFILE_CHANGE:
        return "PROFILE CHANGE";
    case Notification::PROFILES_EXHAUSTED:
        return "PROFILES EXHAUSTED";
    case Notification::NETWORK_NOT_AVAILABLE:
        return "NETWORK NOT AVAILABLE";
    case Notification::NETWORK_AVAILABLE:
        return "NETWORK AVAILABLE";
    case Notification::DISCONNECTING:
        return "DISCONNECTING";
    case Notification::DISCONNECTED:
        return "DISCONNECTED";
    case Notification::ADHOC_NETWORK_STATE_CHANGE:
        return "ADHOC NETWORK STATE CHANGE";
    case Notification::PROFILE_UNBLOCKED:
        return "PROFILE UNBLOCKED";
    case Notification::SCREEN_POWER_CHANGE:
        return "SCREEN POWER CHANGE";
    case Notification::PROFILE_BLOCKED:
        return "PROFILE BLOCKED";
    case Notification::SCAN_LIST_REFRESH:
        return "SCAN LIST REFRESH";
    case Notification::END:
        return "END";
    default:
        return "UNKNOWN";
    }
}

/// @brief Get the number of interfaces registered in the wacher.
int WlanWatcher::interfacesCount() const { return mInterfaces.size(); }
