// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/**
 * @file globals.cpp
 * @brief Global variables and methods definitions file.
 * @author J. Burgalat
 * @date 2016-2018
 */

#include "globals.hpp"

SignalAppender *gSignalAppender = nullptr;
FileAppender *gFileAppender = nullptr;
I18N *gI18N = nullptr;

QString gDescrDir;
QString gLogFile;
QString gResDir;

/// @brief Initialize global variables of the application.
/// @param dir Application resources directory.
void iniGlobals() {
    QSettings settings;

// Initialize translations
#ifdef Q_OS_WIN
    gResDir = QString("%1/data/afc").arg(QApplication::applicationDirPath());
#elif defined(Q_OS_MACOS)
    // assume the app is bundled
    gResDir = QString("%1/../Resources").arg(QApplication::applicationDirPath());
#else
    gResDir = QString("%1/share/AFC").arg(QFileInfo(QApplication::applicationDirPath()).absolutePath());
#endif
    QString lang = settings.value("global/language", QLocale::system().name()).toString();
    gI18N = new I18N(QString("%1/%2").arg(gResDir, "translations"), lang);


    // Directories
    QString defDir =
        QString("%1/A-ToolSuite/AFC").arg(QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).first());

    gDescrDir = QString("%1/descr").arg(defDir);
    QDir(gDescrDir).mkpath(".");
    // logging
    gLogFile = QString("%1/AFC-%2.log").arg(defDir, QDateTime::currentDateTime().toString("yyyyMMdd"));
    QFileInfo info(gLogFile);
    QDir(info.absolutePath()).mkpath(".");
    QString logsFormat = "%{time}{yyyy-MM-dd HH:mm:ss.zzz} %{type:-7}: %{message}";
    qRegisterMetaType<Logger::LogLevel>("LogLevel");
    gFileAppender = new FileAppender;
    gFileAppender->setFileName(gLogFile);
    gFileAppender->setFormat(logsFormat + "\n");
    gFileAppender->setDetailsLevel(Logger::Info);
    cuteLogger->registerAppender(gFileAppender);
    gSignalAppender = new SignalAppender;
    gSignalAppender->setObjectName("logger");
    gSignalAppender->setFormat("%{type:-7}: %{message}");
    gSignalAppender->setDetailsLevel(Logger::Info);
    cuteLogger->registerAppender(gSignalAppender);

    LOG_INFO("+++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    LOG_INFO(QApplication::tr("AFC starting..."));
}

/// @brief Delete global variable pointers.
void finalizeGlobals() {
    LOG_INFO(QApplication::tr("AFC closing..."));
    LOG_INFO("---------------------------------------------------------");
    if (gI18N)
        delete gI18N;
    gI18N = nullptr;
}

/// @brief Set factory default values for application settings.
void setFactoryDefaults(const QString &language) {
    QSettings settings;
    if (QFileInfo::exists(settings.fileName()))
        QFile::remove(settings.fileName());
    // General default settings
    settings.beginGroup("global");
    settings.setValue("language", language.isEmpty() ? QLocale::system().name() : language);
    settings.setValue("font", QVariant::fromValue(QFont("MS Shell Dlg 2", 8)));
    settings.setValue("saveprms", true);
    settings.setValue("precision",12);
    settings.endGroup();
}
