// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "prmchoicewidget.hpp"

/// @brief Constructor.
PrmChoiceWidget::PrmChoiceWidget(QWidget *parent) : QWidget(parent) {
    mOriginal = nullptr;
    mGbxDescriptor = new QGroupBox(this);
    mTableDescriptor = new TableWidget(false, mGbxDescriptor);
    mTableDescriptor->setObjectName("DESCRIPTOR table");
    mTableDescriptor->setColumnCount(2);
    mTableDescriptor->setHorizontalHeaderLabels({tr("Name"), tr("Alias")});
    mTableDescriptor->setDropIndicatorShown(true);
    mTableDescriptor->horizontalHeader()->setSectionsMovable(false);
    mTableDescriptor->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    mTableDescriptor->horizontalHeader()->setStretchLastSection(true);

    QVBoxLayout *layout = new QVBoxLayout(mGbxDescriptor);
    layout->addWidget(mTableDescriptor);

    mGbxOutput = new QGroupBox(this);
    mTableOutput = new TableWidget(true, mGbxOutput);
    mTableOutput->setObjectName("OUTPUT table");
    mTableOutput->setColumnCount(2);
    mTableOutput->setHorizontalHeaderLabels({tr("Name"), tr("Alias")});
    mTableOutput->setDropIndicatorShown(true);
    mTableOutput->horizontalHeader()->setSectionsMovable(false);
    mTableOutput->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
    mTableOutput->horizontalHeader()->setStretchLastSection(true);
    layout = new QVBoxLayout(mGbxOutput);
    layout->addWidget(mTableOutput);

    QHBoxLayout *hLayout = new QHBoxLayout(this);
    hLayout->setContentsMargins(0, 0, 0, 0);
    hLayout->addWidget(mGbxDescriptor);
    hLayout->addWidget(mGbxOutput);

    QObject::connect(mTableDescriptor, &TableWidget::itemMoved, this, &PrmChoiceWidget::itemMoved);
    QObject::connect(mTableOutput, &TableWidget::itemMoved, this, &PrmChoiceWidget::itemMoved);

    translateUI();
}

/// @brief Check whether or not the widget has selected parameters.
bool PrmChoiceWidget::hasSelection() const { return mTableOutput->rowCount(); }

/// @brief Get the selected items.
QVector<PairOfStrings> PrmChoiceWidget::selected() const { return mTableOutput->values(); }

/**
 * @brief Setup the table widgets according to given input.
 * @param input Input XML descriptor.
 */
void PrmChoiceWidget::setupFromXMLDescriptor(XMLDescriptor *input) {
    mTableDescriptor->clearContents();
    mTableDescriptor->setRowCount(0);
    mTableOutput->clearContents();
    mTableOutput->setRowCount(0);
    mOriginal = input;
    if (mOriginal) {
        const QStringList &selection = mOriginal->selection;
        // setup items
        XMLDescriptor left(*mOriginal), right;
        right.instrument = mOriginal->instrument;
        right.version = mOriginal->version;
        for (const QString &n : selection) {
            if (left.hasParameter(n))
                right.parameters.push_back(left.pop(n));
        }

        // "left" contains parameter to show in Descriptor table.
        // "right" contains parameter to show in Output table.
        mTableDescriptor->setPairs(left.parameters);
        mTableOutput->setPairs(right.parameters);
    }
}

/**
 * @brief Update the original XML descriptor from the current content of widget.
 * @internal
 * The original descriptor points a descriptor of the main window descriptor container.
 */
void PrmChoiceWidget::updateOriginal() {
    if (mOriginal) {
        XMLDescriptor left(fromTable(mTableDescriptor));
        XMLDescriptor right(fromTable(mTableOutput));
        for (const PairOfStrings &p : left.parameters)
            mOriginal->setAlias(p.first, p.second);
        // we update from right part only if there is some selected parameter :)
        if (right.parameters.size()) {
            mOriginal->selection.clear();
            for (const PairOfStrings &p : right.parameters) {
                mOriginal->setAlias(p.first, p.second);
                mOriginal->selection.append(p.first);
            }
        }
    }
}

/// @brief Translate UI texts.
void PrmChoiceWidget::translateUI() {
    mGbxDescriptor->setTitle(tr("Descriptor"));
    mGbxOutput->setTitle(tr("Output"));
    QStringList headers = {tr("Name"), tr("Alias")};
    mTableDescriptor->setHorizontalHeaderLabels(headers);
    mTableOutput->setHorizontalHeaderLabels(headers);
}

/**
 * @brief Build an XML descriptor from the content of the given table widget.
 * @param tableWidget TableWidget to process.
 * @return XMLDescriptor instance.
 */
XMLDescriptor PrmChoiceWidget::fromTable(TableWidget *tableWidget) const {
    QVector<PairOfStrings> pairs = tableWidget->values();
    XMLDescriptor out;
    if (mOriginal) {
        out.instrument = mOriginal->instrument;
        out.version = mOriginal->version;
        out.parameters = pairs;
    }
    return out;
}
