// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "tablewidget.hpp"

TableWidget::TableWidget(bool allowInternalMove, QWidget *parent)
    : QTableWidget(parent), mInternalMove(allowInternalMove), mLastDropRow(-1) {

    setDragEnabled(true);
    setAcceptDrops(true);
    setSelectionBehavior(QAbstractItemView::SelectRows);
    setDragDropOverwriteMode(false);
    setColumnCount(2);
}

/**
 * @brief Set new items.
 * @param pairs Pair of parameter name and aliases.
 * @internal
 * The method does not clear the internal model first.
 */
void TableWidget::setPairs(const QVector<PairOfStrings> pairs) {
    for (const PairOfStrings &p : pairs) {
        QTableWidgetItem *n = new QTableWidgetItem(p.first);
        n->setFlags(n->flags() ^ Qt::ItemIsEditable);
        QTableWidgetItem *a = new QTableWidgetItem(p.second);
        int irow = rowCount();
        insertRow(irow);
        setItem(irow, 0, n);
        setItem(irow, 1, a);
    }
}

bool TableWidget::dropMimeData(int row, int, const QMimeData *, Qt::DropAction) {
    mLastDropRow = row;
    return true;
}

void TableWidget::dropEvent(QDropEvent *event) {
    // The QTableWidget from which selected rows will be moved
    TableWidget *sender = qobject_cast<TableWidget *>(event->source());
    bool isMe = sender == this;
    if (!sender || (isMe && !mInternalMove)) {
        event->ignore();
        return;
    }
    // Default dropEvent method fires dropMimeData with appropriate parameters (we're interested in the row index).
    QTableWidget::dropEvent(event);

    // Now we know where to insert selected row(s)
    int dropRow = mLastDropRow;
    QVector<int> selectedRows = sender->getselectedRowsFast();
    int nr = selectedRows.size();
    // Allocate space for transfer
    for (int i = 0; i < nr; i++)
        insertRow(dropRow);

    if (isMe)
        LOG_DEBUG(QString("Moving %1 rows in %2").arg(nr).arg(objectName()));
    else
        LOG_DEBUG(QString("Dropping %1 rows from %2 to %3").arg(nr).arg(sender->objectName()).arg(objectName()));

    QVector<int> selRowsOffsets;
    selRowsOffsets.reserve(nr);
    // if sender == receiver (self), after creating new empty rows selected rows might change their locations
    for (int srow : selectedRows)
        selRowsOffsets.push_back((srow < dropRow || !isMe) ? 0 : nr);

    std::transform(selectedRows.begin(), selectedRows.end(), selRowsOffsets.begin(), selectedRows.begin(), std::plus<int>());

    // copy content of selected rows into empty ones
    int nc = columnCount();
    for (int i = 0; i < nr; i++) {
        int srow = selectedRows[i];
        for (int j = 0; j < nc; j++) {
            QTableWidgetItem *item = sender->item(srow, j);
            if (item) {
                QTableWidgetItem *source = new QTableWidgetItem(*item);
                if (j == 0)
                    source->flags().setFlag(Qt::ItemIsEditable, false);
                setItem(dropRow + i, j, source);
            }
        }
    }
    // delete selected rows !
    std::reverse(selectedRows.begin(), selectedRows.end());
    for (int srow : selectedRows)
        sender->removeRow(srow);
    emit itemMoved();
    event->accept();
}

void TableWidget::dragMoveEvent(QDragMoveEvent *e) {
    if (mInternalMove || e->source() != this)
        e->accept();
    else
        e->ignore();
}

/**
 * @brief Get the items values.
 * @return Vector of string lists.
 */
QVector<PairOfStrings> TableWidget::values() const {
    int nr = rowCount();
    QVector<PairOfStrings> out(nr);
    for (int i = 0; i < nr; i++) {
        PairOfStrings &row = out[i];
        row.first = item(i, 0)->text();
        row.second = item(i, 1)->text();
    }
    return out;
}

/**
 * @brief Compute list of (unique) selected rows.
 * @return Vector with selected rows indexes.
 */
QVector<int> TableWidget::getselectedRowsFast() {
    QVector<int> rows;
    for (auto item : selectedItems()) {
        if (!rows.contains(item->row()))
            rows.push_back(item->row());
    }
    std::sort(rows.begin(), rows.end());
    return rows;
}
