// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024


#include <QtCore>
#include "filefittask.hpp"
#include "fit.hpp"
#include "isp/isp.hpp"
#ifdef WITH_OMP
#include "omp.h"
#endif

typedef QPair<QString, QString> PairOfStrings;

#include <iostream>

int main(int argc, char *argv[]) {
    // int main(int argc, char *argv[]) {
    QCoreApplication app(argc, argv);

    QStringList arguments;
    StringList errors;
    for (int i = 0; i < argc; i++)
        arguments.append(QString(argv[i]));
    QFileInfo appInfo(arguments.first());

    // build command line parser.
    QCommandLineParser parser;
#ifdef WITH_OMP
    parser.setApplicationDescription("\nMulti-threaded AMULSE file fit processing.\nOpenMP version !");
#else
    parser.setApplicationDescription("\nMulti-threaded AMULSE file fit processing.\nQt threadpool version !");
#endif
    const QCommandLineOption hOpt = parser.addHelpOption();
    const QCommandLineOption vOpt({"V", "version"}, "Display version informations.", "", "");
    const QCommandLineOption rOpt("r", "Search recursively in given directory for AMULSE files.", "", "");
    const QCommandLineOption oOpt(
        "o", "Output directory, default to 'results' in current working directory.", "DIR", "");
    const QCommandLineOption lOpt("l", "Write a log file for each processed file.", "", "");
    const QCommandLineOption nOpt("n", "Perform calculations using N threads. Default to 1.", "N", "1");
    const QCommandLineOption sOpt("no-spectras", "Do not write spectras in output files.", "", "");
    const QCommandLineOption csvOpt(
        "csv", "Write output as csv formatted file instead of original results file. Imply --no-spectras.", "", "");
    const QCommandLineOption npOpt("no-preload", "Do not preload spectras before processing input file.", "", "");
    const QCommandLineOption bsOpt("block-step",
        "Process only one block every N block(s). Default to 1 (i.e. process all block in a file).",
        "N",
        "1");

    const QCommandLineOption rsOpt(
        "s", "Reuse sigma-scale coefficients of last successfull fit as first guess.", "", "");
    const QCommandLineOption rbOpt("b", "Reuse base-line coefficients of last successfull fit as first guess.", "", "");
    const QCommandLineOption rcOpt("c", "Reuse concentrations of last successfull fit as first guess.", "", "");

    const QCommandLineOption p512Opt("512",
        "Decimate spectra to 512 points before fit process. --512 and --1024 are mutually exclusive options.",
        "",
        "");
    const QCommandLineOption p1024Opt("1024",
        "Decimate spectra to 1024 points before fit process. --512 and --1024 are mutually exclusive options.",
        "",
        "");

    const QCommandLineOption oldOpt("old", "Use old algorithm version: slow down the process !", "", "");

    parser.addOption(vOpt);
    parser.addOption(rOpt);
    parser.addOption(oOpt);
    parser.addOption(lOpt);
    parser.addOption(nOpt);
    parser.addOption(sOpt);
    parser.addOption(bsOpt);
    parser.addOption(npOpt);
    parser.addOption(rsOpt);
    parser.addOption(rbOpt);
    parser.addOption(rcOpt);
    parser.addOption(csvOpt);
    parser.addOption(p512Opt);
    parser.addOption(p1024Opt);

    parser.addOption(oldOpt);

    parser.addPositionalArgument("CONFIG", "Fit configuration file to use.", "CONFIG");
    parser.addPositionalArgument("INPUT", "Single file or directory with files to process.", "INPUT");

    QTextStream out(stderr);

    // parse arguments.
    if (!parser.parse(arguments)) {
        out << parser.errorText() << Qt::endl << Qt::endl;
        return 1;
    }

    // check for specific options
    // version
    if (parser.isSet(vOpt)) {
        out << QString("%1 %2").arg(appInfo.baseName(), QString(APP_VER)) << Qt::endl;
        return 0;
    }
    // help
    if (parser.isSet(hOpt)) {
        QString help = parser.helpText();
        QString tmp = QDir::toNativeSeparators(appInfo.absolutePath() + QString("/"));
        printString(help.replace(tmp, "").append('\n'));
        return 0;
    }

    // check for mutually exclusive options

    if (parser.isSet(p512Opt) && parser.isSet(p1024Opt)) {
        printString(QString("error: --512 and --1024 options are mutually exclusive: Select one or the other.\n\n"));
        return 1;
    }

    // check for positionals: at least two elements are expected !
    QStringList positionals = parser.positionalArguments();
    if (positionals.size() < 2) {
        printString(QString("error  : Invalid number of arguments.\n"));
        return 1;
    }

    // should we produce csv?
    bool asCSV = parser.isSet(csvOpt);

    // check configuration file.
    QString configPath = positionals.takeFirst();
    FitConfig fitConfig;
    if (!QFileInfo::exists(configPath)) {
        printString(QString("error: %1: no such file.\n").arg(configPath));
        return 1;
    } else if (!fitConfig.parse(configPath.toStdString(), &errors)) {
        printString(QString("error: %1: Invalid configuration file.\n").arg(configPath));
        for (const auto &e : qAsConst(errors)) {
            printString(QString("error: %1\n").arg(QString::fromStdString(e)));
        }
        return 1;
    }
    int cfg_instrum = fitConfig.instrumentNumber();
    int cfg_version = fitConfig.firmwareVersion();

    // build and check list of files.
    bool recurse = parser.isSet(rOpt);
    QString inputDirectory;
    QFileInfo info(positionals.takeFirst());
    QStringList inputFiles;

    if (info.isDir()) {
        inputDirectory = info.absoluteFilePath();
        QDirIterator it(inputDirectory,
            QDir::NoDotAndDotDot | QDir::Files,
            recurse ? QDirIterator::Subdirectories : QDirIterator::NoIteratorFlags);
        while (it.hasNext()) {
            QFileInfo finfo(it.next());
            QString suffix = finfo.suffix();
            if (suffix.compare("bin", Qt::CaseInsensitive) == 0 || suffix.compare("results", Qt::CaseInsensitive) == 0)
                inputFiles << finfo.absoluteFilePath();
        }
    } else {
        inputDirectory = info.absolutePath();
        QString suffix = info.suffix();
        if (suffix.compare("bin", Qt::CaseInsensitive) == 0 || suffix.compare("results", Qt::CaseInsensitive) == 0)
            inputFiles << info.absoluteFilePath();
    }
    QString outputDirectory;
    if (parser.isSet(oOpt)) {
        outputDirectory = QDir(parser.value(oOpt)).absolutePath();
    } else {
        outputDirectory = QString("%1/results").arg(QDir::currentPath());
    }

    QDir topDir(inputDirectory);
    QString dirname = topDir.dirName();
    QVector<PairOfStrings> files;
    int instrum, version;
    QString descr;
    QString ext(asCSV ? "csv" : "results");
    for (const QString &ip : qAsConst(inputFiles)) {
        std::string tdescr;
        auto b = getFileInfos(ip.toStdString(), false, &tdescr);
        instrum = b.instrument();
        version = b.version();
        descr = QString::fromStdString(tdescr);
        if (instrum == -1 && version == -1) {
            printString(QString("warning: %1 is not a valid AMULSE binary file. Skipping file.\n").arg(ip));
            continue;
        } else if (instrum != cfg_instrum || version != cfg_version) {
            if (instrum != cfg_instrum) {
                printString(QString("warning: %1: Inconsistent instrument number: %2 (file) / %3 (cfg).\n")
                                .arg(ip)
                                .arg(instrum)
                                .arg(cfg_instrum));
            }
            if (version != cfg_version) {
                printString(QString("warning: %1: Inconsistent firmware version: %2 (file) / %3 (cfg).\n")
                                .arg(ip)
                                .arg(version)
                                .arg(cfg_version));
            }
            printString(QString("warning: Skipping %1 (inconsistent ins. number/firmware version)...\n").arg(ip));
            continue;
        }
        QFileInfo finfo(QString("%1/%2/%3").arg(outputDirectory, dirname, topDir.relativeFilePath(ip)));
        QString op = QString("%1/%2.%3").arg(finfo.absolutePath(), finfo.completeBaseName(), ext);
        files.push_back(qMakePair(ip, op));
    }
    if (files.isEmpty()) {
        printString(QString("error: no valid files to process.\n"));
        return 1;
    }

    bool logToFile = parser.isSet(lOpt);
    FitParamReuse reusePrms(parser.isSet(rcOpt), parser.isSet(rbOpt), parser.isSet(rsOpt));
    bool skipSpectras = parser.isSet(sOpt);
    int nthreads = parser.isSet(nOpt) ? qBound(1, parser.value(nOpt).toInt(), QThread::idealThreadCount()) : 1;

    Decimation decim = Decimation::No;
    if (parser.isSet(p512Opt))
        decim = Decimation::To_512;
    else if (parser.isSet(p1024Opt))
        decim = Decimation::To_1024;

    int bStep = 1;
    if (parser.isSet(bsOpt))
        bStep = qMax(1, parser.value(bsOpt).toInt());

    bool preLoad = !parser.isSet(npOpt);

    // 4. process !
    int nf = files.size();
    QLatin1Char fc('0');
    nthreads = qMin(nthreads, nf);

    printString(QString("Processing %1 file(s) using %2 thread(s)...\n").arg(nf).arg(nthreads));

    // create directories structure
    for (int i = 0; i < nf; ++i) {
        const PairOfStrings &p = files.at(i);
        QDir dir = QFileInfo(p.second).absoluteDir();
        if (!dir.mkpath(".")) {
            printString(QString("error: Cannot create directory %1\n").arg(dir.absolutePath()));
            return 1;
        }
    }

#ifdef WITH_OMP
    omp_set_num_threads(nthreads);
#pragma omp parallel for
    for (int i = 0; i < nf; ++i) {
        const PairOfStrings &p = files.at(i);
        ExecTimes ttimes;
        bool ret = processFile(p.first,
            p.second,
            fitConfig,
            1000,
            reusePrms,
            bStep,
            decim,
            preLoad,
            logToFile,
            skipSpectras,
            asCSV,
            &ttimes);
        //        int th = omp_get_thread_num();
        //        printString(QString("TH#%1 ; %2 ; %3 ; %4 ; %5")
        //                        .arg(th, 2, 10, fc)
        //                        .arg(QDateTime::currentDateTimeUtc().toString("yyyy.MM.ddThh:mm:ss.zzz"), p.first,
        //                             (ret ? "SUCCESS" : "FAILED"), ttimes.toString(Units::MIN)));
    }
#else
    // Using a Qt thread pool
    QThreadPool *pool = QThreadPool::globalInstance();
    pool->setMaxThreadCount(nthreads);
    for (int i = 0; i < nf; ++i) {
        const PairOfStrings &p = files.at(i);
        FileFitTask *task = new FileFitTask(
            p.first, p.second, &fitConfig, 1000, reusePrms, bStep, decim, preLoad, logToFile, skipSpectras, asCSV);
        pool->start(task);
    }
    pool->waitForDone();
#endif
    return 0;
}
