// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024
/**
 * @file i18n.cpp
 * @brief Internationalization class definitions file.
 * @author J. Burgalat
 * @date 2016
 */

#include "i18n.hpp"
#include <QCoreApplication>
#include <QDir>
#include <QLibraryInfo>
#include <QLocale>
#include <QTranslator>

#include <QDebug>

/**
 * @brief Constructor.
 * @param langPath Directory path of the translation files:
 *   - If path is empty, translation files are assumed to be in the directory given by
 *     QLibraryInfo::location(QLibraryInfo::TranslationsPath)/
 *   - If path is not set as an absolute path, it is assumed to be relative to
 *     the application directory.
 * @param defaultLang Default language to load at initialization. If defaultLang is
 * set to empty string (default value), the constructor attempts to set a language file
 * associated to current system Locale.
 * @param parent Parent QObject.
 */
I18N::I18N(const QString &langPath, QString defaultLang, QObject *parent) : QObject(parent) {
    mTranslator = new QTranslator(this);
    mCurLang = QLocale::C;
    QString tmp = langPath.isEmpty() ? QLibraryInfo::location(QLibraryInfo::TranslationsPath) : langPath;
    if (QDir::isAbsolutePath(tmp))
        mLangPath = tmp;
    else
        mLangPath = QString("%1/%2").arg(QCoreApplication::applicationDirPath()).arg(tmp);
    mLangPath = QDir(mLangPath).absolutePath();
    if (defaultLang.isEmpty())
        defaultLang = I18N::languageCode(QLocale::system().language());
    setLanguage(defaultLang);
}

/// @brief I18N destructor.
I18N::~I18N() {}

/// @brief Get current language.
QLocale::Language I18N::currentLanguage() const { return mCurLang; }

/**
 * @brief Get the list of available languages.
 * @return List of language corresponding to the list of translation files found in
 * I18N::langPath.
 * @details
 * Only languages corresponding to an existing QLocale object are defined in the list: The method builds for each file
 * found, a QLocale object with the suffix part of the file (\<something without underscore\>_\<suffix\>.qm). If the
 * QLocale object fallbacks to "C", it is quietly ignored, otherwise \<suffix\> is appended to the list.
 */
QVector<QLocale::Language> I18N::getLanguageList() const {
    QVector<QLocale::Language> langs;
    langs.push_back(QLocale::C);
    QString appName = QCoreApplication::applicationName();
    QDir dir(mLangPath);
    QStringList fileNames = dir.entryList(QStringList(QString("%1_*.qm").arg(appName)));
    // loop over translation file and retrieve their suffix
    for (QStringList::const_iterator file = fileNames.begin(); file != fileNames.end(); ++file) {
        QString abbr = *file;
        abbr.truncate(abbr.lastIndexOf('.'));
        abbr.remove(0, abbr.indexOf('_') + 1);
        QLocale locale(abbr);
        if (locale.language() != QLocale::C)
            langs.push_back(locale.language());
    }
    return langs;
}

/**
 * @brief Set a new language (and its associated Locale).
 * @param language Name of the language (must follow QLocale::QLocale requirements).
 * @return True on success, false otherwise.
 * @details
 * The method sets a new language by loading its related translation file. The file must be found in I18N::langPath with
 * the following pattern name: \<application name\>_\<language\>.qm.
 *
 * On error, the language is set to QLocale::C.
 */
bool I18N::setLanguage(QLocale::Language language) {
    bool ret = true;
    if (mCurLang != language) {
        mCurLang = language;
        QCoreApplication::removeTranslator(mTranslator);
        if (language == QLocale::C)
            return ret;
        QString langFile = QString("%1/%2_%3");
        QString lang = I18N::languageCode(mCurLang);
        // switch App translator
        ret = mTranslator->load(langFile.arg(mLangPath).arg(QCoreApplication::applicationName()).arg(lang)) &&
              QCoreApplication::installTranslator(mTranslator);
    }
    return ret;
}

/**
 * @brief Set a new language (and its associated Locale).
 * @param language Name of the language (must follow QLocale::QLocale requirements).
 * @return True on success, false otherwise.
 * @details
 * The method sets a new language by loading its related translation file. The file must be found in I18N::langPath with
 * the following pattern name: \<application name\>_\<language\>.qm.
 *
 * On error, the language is set to QLocale::C.
 */
bool I18N::setLanguage(const QString &language) {
    bool ret = true;
    QLocale locale(language);
    if (mCurLang != locale.language())
        return setLanguage(locale.language());
    return ret;
}

///
/**
 * @brief Get the (translated) name of the given language.
 * @note
 * The special locale "C" is translated to "english" (rather than "American English").
 */
QString I18N::languageName(QLocale::Language language){
    return (language == QLocale::C) ? "english" : QLocale(language).nativeLanguageName();
}

/// @brief Get the two-letter ISO 639 language code for the given Language.
QString I18N::languageCode(QLocale::Language language) {
    QString lang = QLocale(language).name();
    int i = lang.indexOf('_');
    lang.remove(i, lang.size() - i);
    if (QLocale::C == language)
        lang = "en";
    return lang;
}

///**
// * @brief Set a new language (and its associated Locale).
// * @param language Name of the language (must follow QLocale::QLocale requirements).
// * @return True on success, false otherwise.
// * @details
// * The method sets a new language by loading its related translation file. The file must be found in I18N::langPath
// with
// * the following pattern name: \<application name\>_\<language\>.qm.
// *
// * On error, the language is set to QLocale::C.
// */
// bool I18N::setLanguage(const QString &language) {
//    bool ret = true;
//    if (mCurLang != language) {
//        mCurLang = language;
//        QString sLang = language;
//        if (sLang.count('_')) {
//            int idx = sLang.lastIndexOf('_');
//            sLang.remove(idx, sLang.size() - idx);
//        }
//        QString langFile = QString("%1/%2_%3");
//        QLocale locale = QLocale(mCurLang);
//        QCoreApplication::removeTranslator(mTranslator);
//        if (locale.language() == QLocale::C)
//            return ret;
//        // switch App translator
//        ret = mTranslator->load(langFile.arg(mLangPath).arg(QCoreApplication::applicationName()).arg(mCurLang)) &&
//              QCoreApplication::installTranslator(mTranslator);
//    }
//    return ret;
//}
