// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024
/**
 * @file i18n.hpp
 * @brief Internationalization class declarations file.
 * @author J. Burgalat
 * @date 2016
 */

#ifndef I18N_HPP
#define I18N_HPP

#include <QObject>
#include <QStringList>
#include <QTranslator>

#include <QLocale>

/**
 * @brief Internationalization and localization handler class
 * @details
 * The class defines a translator for the application translations.
 *
 * Based on a path set at initialization, the class can be used to switch between available languages (through
 * translation files assumed to be set in the path).
 *
 * Otherwise given in the constructor, path is set to languages sub-directory of the application directory.
 *
 * The path is first used to get a list of available languages returned by I18N::getLanguageList. The list can then be
 * used set a new language by calling I18N::setLanguage. When a new language is set, its associated Locale is also set.
 */
class I18N : public QObject {
    Q_OBJECT
  public:
    I18N(const QString &langPath, QString defaultLang = QString(), QObject *parent = nullptr);
    virtual ~I18N();
    QVector<QLocale::Language> getLanguageList() const;
    QLocale::Language currentLanguage() const;
    bool setLanguage(const QString &language);
    bool setLanguage(QLocale::Language language);

    static QString languageName(QLocale::Language language);
    static QString languageCode(QLocale::Language language);

  private:
    QTranslator *mTranslator;   //! Application translator
                                // QString mCurLang;         //! Current language
    QLocale::Language mCurLang; //! Current language
    QString mLangPath;          //! Directory path of language files
};

#endif // I18N_HPP
