// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/**
 * @file validators.cpp
 * @brief floating point value validators classes definitions file.
 * @author J. Burgalat
 * @date 2016-2018
 */

#include "validators.hpp"
#include <QDateTime>
#include <QDebug>

/// @brief Regular expression used to validate string.
const QRegularExpression DoubleValidator::mRe = QRegularExpression("(([+-]?\\d+(\\.\\d*)?|\\.\\d+)([eE][+-]?\\d+)?)");

/**
 * @brief Constructor.
 * @param parent Parent object.
 * @param allowEmpty True to allow empty string. Default to false.
 * @param allowNaN True to allow special word 'nan' for NaN value. Default to true.
 * @details
 * If allowEmpty is set to true, the validator assumes empty string is Acceptable (i.e. is a valid value). Otherwise, an
 * empty string is assumed to be an intermediate result.
 */
DoubleValidator::DoubleValidator(QObject *parent, bool allowEmpty, bool allowNaN)
    : QValidator(parent), mEmptyAllowed(allowEmpty), mNaNAllowed(allowNaN) {}

/// @brief Validate the string as a double value using a regular expression.
bool DoubleValidator::validateDouble(const QString &str) const {
    return (mRe.match(str).captured(0).compare(str) == 0);
}

/// @brief Check whether or not empty string are acceptable during validation.
bool DoubleValidator::emptyAllowed() const { return mEmptyAllowed; }

/// @brief Check whether or not "nan" string is acceptable during validation.
bool DoubleValidator::nanAllowed() const { return mNaNAllowed; }

/// @brief Validate (or not) the input string.
QValidator::State DoubleValidator::validate(QString &input, int &pos) const {
    QString str = input; //.simplified();
    bool empty = str.simplified().isEmpty();
    QValidator::State state = QValidator::Invalid;
    if (empty) { // str.isEmpty()) {
        if (mEmptyAllowed)
            state = QValidator::Acceptable;
        else
            state = QValidator::Intermediate;
    } else if ((mNaNAllowed && (empty || str.simplified().compare("nan", Qt::CaseInsensitive) == 0)) ||
               validateDouble(str)) {
        //} else if ((mNaNAllowed && (str.isEmpty() || str.compare("nan", Qt::CaseInsensitive) == 0)) ||
        //validateDouble(str)) {
        state = QValidator::Acceptable;
    } else {
        // intermediate cases
        if (pos > 0 && (str.at(pos - 1) == 'e' || str.at(pos - 1) == 'E' || str.at(pos - 1) == '.' ||
                        str.at(pos - 1) == '+' || str.at(pos - 1) == '-')) {
            state = QValidator::Intermediate;
        }
    }
    // qDebug() << str << empty << state;
    return state;
}

// original

///// @brief Validate (or not) the input string.
// QValidator::State DoubleValidator::validate(QString &input, int &pos) const {
//    QString str = input; //.simplified();
//    if (str.isEmpty()) {
//        if (mEmptyAllowed)
//            return QValidator::Acceptable;
//        return QValidator::Intermediate;
//    }
//    if ((mNaNAllowed && (str.isEmpty() || str.compare("nan", Qt::CaseInsensitive) == 0)) || validateDouble(str))
//        return QValidator::Acceptable;
//    if ((str.at(pos - 1) == 'e' || str.at(pos - 1) == 'E' || str.at(pos - 1) == '.' || str.at(pos - 1) == '+' ||
//         str.at(pos - 1) == '-'))
//        return QValidator::Intermediate;
//    return QValidator::Invalid;
//}

/// @brief Regular expression used to check for forbidden characters.
const QRegularExpression LogPatternValidator::mReForbidden = QRegularExpression("[\\\\:\\<\\>\"/\\|\\?\\*]");

/// @brief Constructor.
LogPatternValidator::LogPatternValidator(QObject *parent) : QValidator(parent) {}

/// @brief Validate the string as a correct log pattern.
bool LogPatternValidator::validatePattern(const QString &str) const {
    QString tStr = str;
    QRegularExpression regex = QRegularExpression("%\\(([^\\)]+)\\)");
    tStr.replace(QRegExp("%\\(name\\)"), "");
    QRegularExpressionMatch match = regex.match(tStr);
    for (QString fmt : match.capturedTexts()) {
        QString date = QDateTime::currentDateTime().toString(fmt);
        if (date.isEmpty())
            return false;
        tStr.replace(QRegExp(QString("%\\(%1\\)").arg(fmt)), date);
    }
    return true;
}

/// @brief Validate (or not) the input string.
QValidator::State LogPatternValidator::validate(QString &input, int &pos) const {
    Q_UNUSED(pos)
    QString str = input;
    if (str.contains(mReForbidden))
        return QValidator::Invalid;
    if (str.isEmpty())
        return QValidator::Intermediate;
    if (validatePattern(str))
        return QValidator::Acceptable;
    return QValidator::Intermediate;
}
