// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024
/**
 * @file AMDataset.hpp
 * @brief AMDataset class declarations file.
 * @author J. Burgalat
 * @date 2017-2018
 */
#ifndef AMDATASET_HPP
#define AMDATASET_HPP

#include "isp/isp.hpp"

#include <QtCore>

using namespace ISP;

/**
 * @brief Amulse multi-dataset
 * @details
 * The class is intended to amalgamate multiple ADataset.
 *
 * The class is intended to be read-only and does not provide any mechanism to modify its content.
 * User has to manipule the internal datasets.
 * @important
 * All modifications of the internal datasets may break the consistency of the AMDataset.
 */
class AMDataset : public QObject {
    Q_OBJECT
  public:
    AMDataset(const QStringList &paths = QStringList());
    // copy constructor , copy assignment ???
    ~AMDataset();

    // io methods
    bool addFiles(const QStringList &paths, bool sort = true, bool signalNeeded = true);
    bool addFile(const QString &path, bool sort = true, bool signalNeeded = true);
    bool addDataSet(ADataset *set, bool sort = true, bool signalNeeded = true);
    QVector<ADataset *> addDataSets(const QVector<ADataset *> &sets, bool sort = true, bool signalNeeded = true);
    void clear();

    // Time line & indexes
    DoubleVector indexes(bool noGap = true, int step = 1) const;
    DoubleVector timeLine(bool noGap = true, int step = 1, bool *ok = nullptr) const;
    double timeAt(size_t bIndex) const;
    double timeStep() const;
    double startTime();
    double endTime() const;

    // internal data
    const QVector<ADataset *> &datasets() const;
    QVector<ADataset *> &datasets();
    ADataset *dataset(int blockIndex);
    ADataset *datasetAt(long fileIndex);

    ABlock blockAt(int blockIndex) const;
    QVector<ABlock> blocks() const;

    // parameters
    QStringList categories() const;
    QVector<AParam> parameters() const;
    bool hasParameter(const AParam &parameter) const;
    double parameterAt(const AParam &parameter, long bIndex = 0) const;
    DoubleVector parameterAt(const AParam &parameter, const QVector<long> &bIndexes) const;
    DoubleVector parameter(const AParam &parameter, bool noGap = true, int step = 1) const;
    DoubleVector parameter(int n, const AParam &parameter) const;
    // spectras related
    DoubleVector spectraAt(int blockIndex, int spcIndex = 0);
    QStringList spectraLabels() const;
    QString spectraLabel(int lineId) const;
    QStringList spectraMoleculesAt(int lineId, bool noUkn) const;

    // counters
    int blockCount(bool noGap = true) const;
    int parameterCount() const;
    int lineCount() const;
    size_t fileCount() const;
    int spectraSize() const;

    // global infos
    bool isValid() const;
    QStringList filePaths();
    ABlock descriptor() const;
    int instrument() const;
    int version() const;
    const QString &name() const;
    QString friendlyName() const;
    bool alreadyIn(ADataset *set);
    size_t actualBlockIndex(size_t discontinuedIndex) const;
    size_t discontinuedBlockIndex(size_t actualIndex) const;
    long getFileIndex(size_t blockIndex, long *relativeTo = nullptr) const;

    // Modification
    void sortDataSets();
    void setName(const QString &name);

  signals:
    /**
     * @brief Emitted when new file or datasets have been added to the dataset container.
     * @details
     * This signal is emitted by AMDataset::addFiles, AMDataset::addFile, AMDataset::addDataSets and
     * AMDataset::addDataSets only if the relevant method has been called with signalNeeded argument set to true
     * (default).
     */
    void filesAdded();

  private:
    void rebuildDiscontinuities();
    bool discontinued(int fileIndex) const;
    QVector<ADataset *>::iterator sortedIterator(ADataset *set);
    QVector<ADataset *> mFiles;  ///< List of ADataset;
    QVector<bool> mDiscontinued; ///< List of discontinuities
    QString mName;
};

#endif // AMDATASET_HPP
