// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024
#include "fileloader.hpp"
#include "amdataset.hpp"

/// @brief Constructor.
FileLoader::FileLoader(const QStringList &files, QObject *parent) : QObject(parent) {
    mCancel = false;
    mIsRunning = false;
    mFilesToLoad = files;
}

/// @brief Destructor.
FileLoader::~FileLoader() {}

/// @brief Abort operations.
void FileLoader::cancel() {
    QMutexLocker locker(&mDataMutex);
    mCancel = true;
}

/// @brief Check whether or not the worker in the running.
bool FileLoader::isRunning() const {
    QMutexLocker locker(&mDataMutex);
    return mIsRunning;
}

/// @brief Check whether or not the loading process has been canceled.
bool FileLoader::canceled() const {
    QMutexLocker locker(&mDataMutex);
    return mCancel;
}

/// @brief Load files in the dataset.
void FileLoader::process() {
    mDataMutex.lock();
    mCancel = false;
    mIsRunning = true;
    bool cancel = mCancel;
    int numTasks = mFilesToLoad.size();
    QStringList files = mFilesToLoad;
    mDataMutex.unlock();
    emit progressRangeChanged(0, numTasks);
    AMDataset *amdataset = new AMDataset;
    int icount = 0;
    for (const QString &path : files) {
        emit progressChanged(icount++);
        QFileInfo info(path);
        if (!info.size())
            continue;

        ADataset *set = new ADataset(StringList{path.toStdString()});
        if (!set->valid()) {
            emit message(QString("%1: Failed to read file.").arg(path));
            delete set;
        } else if (!amdataset->addDataSet(set, true, false)) {
            emit debugMessage(tr("%1: Failed to register file in the dataset.").arg(path));
            delete set;
        } else
            emit debugMessage(tr("%1: File registered with success in the dataset.").arg(path));
        mDataMutex.lock();
        cancel = mCancel;
        mDataMutex.unlock();
        if (cancel)
            break;
    }
    mDataMutex.lock();
    mIsRunning = false;
    mFilesToLoad.clear();
    mDataMutex.unlock();
    if (cancel) {
        delete amdataset;
        amdataset = nullptr;
        return;
    } else {
        emit progressChanged(numTasks);
        emit datasetLoaded(amdataset);
    }
    emit finished(cancel);
}
