// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024
#include "multifileloader.hpp"
#include "amdataset.hpp"

/// @brief Constructor.
MultiFileLoader::MultiFileLoader(const QVector<QStringList> &files, QObject *parent) : QObject(parent), QRunnable() {
    mCancel = false;
    mIsRunning = false;
    mFilesToLoad = files;
    mNumTasks = 0;
    for (const QStringList &list : mFilesToLoad)
        mNumTasks += list.size();
}

/// @brief Destructor.
MultiFileLoader::~MultiFileLoader() {}

/// @brief Abort operations.
void MultiFileLoader::cancel() {
    QMutexLocker locker(&mDataMutex);
    mCancel = true;
}

/// @brief Check whether or not the worker in the running.
bool MultiFileLoader::isRunning() const {
    QMutexLocker locker(&mDataMutex);
    return mIsRunning;
}

/// @brief Load files in the dataset.
void MultiFileLoader::run() { process(); }

/// @brief Check whether or not the loading process has been canceled.
bool MultiFileLoader::canceled() const {
    QMutexLocker locker(&mDataMutex);
    return mCancel;
}

/// @brief Load files in the dataset.
void MultiFileLoader::process() {
    mDataMutex.lock();
    mCancel = false;
    mIsRunning = true;
    bool cancel = mCancel;
    int numTasks = mNumTasks;
    QVector<QStringList> files = mFilesToLoad;
    mDataMutex.unlock();
    emit progressRangeChanged(0, numTasks);
    int icount = 0;
    emit progressChanged(icount);
    for (const QStringList &paths : files) {
        AMDataset *amdataset = new AMDataset;
        for (const QString &path : paths) {
            QFileInfo info(path);
            if (!info.size()) {
                emit progressChanged(++icount);
                continue;
            }
            ADataset *set = new ADataset(StringList{path.toStdString()});
            if (!set->valid()) {
                emit message(QString("%1: Failed to read file.").arg(path));
                delete set;
            } else if (!amdataset->addDataSet(set, true, false)) {
                emit message(tr("%1: Failed to register file in the dataset.").arg(path));
                delete set;
            } else
                emit message(tr("%1: File registered with success in the dataset.").arg(path));
            mDataMutex.lock();
            cancel = mCancel;
            mDataMutex.unlock();
            if (cancel) {
                emit progressChanged(numTasks);
                mDataMutex.lock();
                mIsRunning = false;
                mFilesToLoad.clear();
                mDataMutex.unlock();
                delete amdataset;
                amdataset = nullptr;
                emit finished(cancel);
                return;
            }
            emit progressChanged(++icount);
        }
        emit datasetLoaded(amdataset);
    }
    emit progressChanged(numTasks);
    mDataMutex.lock();
    mFilesToLoad.clear();
    mIsRunning = false;
    mDataMutex.unlock();
    emit finished(cancel);
}
